/*******************************************************************************
* FILE NAME: ibuffer6.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the IBuffer searching             *
*   functions.                                                                 *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <string.h>
  }

#include <ibuffer.hpp>

#if !defined( _ISTRTEST_ )
  #include <istrtest.hpp>
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBUFFER6_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IBuffer::startSearch                                                         |
|                                                                              |
| This function returns the "adjusted" starting position for a                 |
| search ostensibly starting at position "startPos" for a search               |
| string of length "searchLen."  A return value of 0 indicates                 |
| that the search cannot succeed.                                              |
|                                                                              |
| If the starting position is zero, start the search at the                    |
| beginning of the buffer (at index 1).                                        |
|                                                                              |
| If the starting position plus the length of the search string is             |
| longer than the buffer, then the search cannot succeed, so return 0.         |
| The check for this condition is made carefully to avoid overflow             |
| situations.                                                                  |
------------------------------------------------------------------------------*/
unsigned IBuffer :: startSearch ( unsigned startPos,
                                  unsigned searchLen ) const
  {
  if ( startPos == 0 )
    startPos++;
  if ( startPos > this->length()
       ||
       searchLen > this->length() - startPos + 1 )
    startPos = 0;
  return startPos;
  }

/*------------------------------------------------------------------------------
| IBuffer::startBackwardsSearch                                                |
|                                                                              |
| This function returns the "adjusted" starting position for a                 |
| backward search ostensibly starting at position "startPos" for a             |
| search string of length "searchLen".  A return value of 0 indicates          |
| that the search cannot succeed.                                              |
|                                                                              |
| If the length of the search string exceeds the length of the buffer,         |
| then the search cannot succeed, so return 0.                                 |
|                                                                              |
| If the starting position is zero, or, the specified starting position        |
| is beyond the end of the buffer, then start the search at the end            |
| of the buffer (index equal to buffer length).                                |
|                                                                              |
| Finally, the starting point is moved back so that the search string          |
| starting at that point won't run off the end of the buffer.                  |
------------------------------------------------------------------------------*/
unsigned IBuffer :: startBackwardsSearch ( unsigned startPos,
                                           unsigned searchLen ) const
  {
  if ( searchLen <= this->length() )
    {
    if ( startPos > this->length() )
      startPos = this->length();
    if ( startPos > this->length() - searchLen + 1 )
      startPos = this->length() - searchLen + 1;
    }
  else
    startPos = 0;
  return startPos;
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOf                                                             |
|                                                                              |
| 1. First, check whether the search can succeed (and set the                  |
|    point at which the search is to start) by calling startSearch().          |
| 2. If the search cannot succeed, then return 0.                              |
| 3. Otherwise, compare the buffer contents starting at each index             |
|    to the search string until a match is found (in which case we             |
|    return the index) or the entire buffer is tested (in which case           |
|    0 is returned).                                                           |
|                                                                              |
| Notes:                                                                       |
|   1. If the search string is a single character, then this function is       |
|      implemented via the equivalent invocation of indexOfAnyOf (which is     |
|      optimized for that case).  This saves the additional logic of           |
|      special code to handle that case here, as well.                         |
|   2. If the search string is null, then 0 is returned.                       |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOf ( const char *pSearchString,
                              unsigned    searchLen,
                              unsigned    startPos ) const
  {
  if ( searchLen == 1 )
    return this->indexOfAnyOf( pSearchString, searchLen, startPos );
  startPos = this->startSearch( startPos, searchLen );
  if ( startPos && searchLen )
    {
    startPos--; // Convert to (0-based) array index.
    while ( (startPos + searchLen) <= this->length() )
      if ( memcmp( this->contents() + startPos++,
                   pSearchString,
                   searchLen ) == 0 )
        return startPos;
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOf                                                             |
|                                                                              |
| Simply invokes the functionally equivalent indexOfAnyOf function.            |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOf ( const IStringTest &aTest,
                              unsigned           startPos ) const
  {
  return this->indexOfAnyOf( aTest, startPos );
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOfAnyBut                                                       |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an                  |
|    array (0-based) index.                                                    |
| 4. Search for the first character not in the set of valid characters, based  |
|    on the number of valid characters:                                        |
|      0  - return 1; i.e., the first character of the buffer                  |
|           isn't in the (empty) set of valid characters                       |
|      1  - compare each character of the buffer to the valid                  |
|           character until one doesn't match                                  |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of valid characters;                |
|           return the index of the first that isn't found                     |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOfAnyBut ( const char *pValidChars,
                                    unsigned    numValidChars,
                                    unsigned    startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    startPos--; // Convert to 0-based array index.
    switch ( numValidChars )
      {
      case 0:
        return ++startPos;
      case 1:
        while ( startPos < this->length() )
          if ( this->contents()[ startPos++ ] != *pValidChars )
            return startPos;
        break;
      default:
        while ( startPos < this->length() )
          if ( memchr( pValidChars,
                       this->contents()[ startPos++ ],
                       numValidChars ) == 0 )
            return startPos;
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOfAnyBut                                                       |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a false result.                          |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOfAnyBut ( const IStringTest &aTest,
                                    unsigned           startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    startPos--;
    while ( startPos < this->length() )
      if ( !aTest.test( this->contents()[ startPos++ ] ) )
        return startPos;
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOfAnyOf                                                        |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the                |
|    starting point) by calling startSearch().                                 |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an                  |
|    array (0-based) index.                                                    |
| 4. Search for the first character in the set of search                       |
|    characters, based on the number of search characters:                     |
|      0  - return 0; i.e., no characters of the buffer are in                 |
|           the (empty) set of search characters                               |
|      1  - search for a matching character in the buffer using                |
|           memchr                                                             |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of search characters;               |
|           return the index of the first that is found                        |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOfAnyOf ( const char *pValidChars,
                                   unsigned    numValidChars,
                                   unsigned    startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    startPos--; // Convert to 0-based array index.
    switch ( numValidChars )
      {
      const char *p;
      case 0:
        break;
      case 1:
        p = (const char *) memchr( this->contents() + startPos,
                                   *pValidChars,
                                   this->length() - startPos );
        if ( p )
          return p - this->contents() + 1;
        break;
      default:
        while ( startPos < this->length() )
          if ( memchr( pValidChars,
                       this->contents()[ startPos++ ],
                       numValidChars ) )
            return startPos;
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::indexOfAnyOf                                                        |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a true result.                           |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: indexOfAnyOf ( const IStringTest &aTest,
                                   unsigned           startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    startPos--;
    while ( startPos < this->length() )
      if ( aTest.test( this->contents()[ startPos++ ] ) )
        return startPos;
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::lastIndexOf                                                         |
|                                                                              |
| 1. Adjust starting position and see if the search can succeed by calling     |
|    startBackwardsSearch().                                                   |
| 2. If a match isn't possible, return 0.                                      |
| 3. Otherwise, search (backward) according to length of search string:        |
|      0  - return 0;                                                          |
|      1  - search for a matching character in the buffer by                   |
|           comparing one byte at a time                                       |
|      >1 - call memcmp for each character of the buffer to see                |
|           if the characters starting here match the search                   |
|           string                                                             |
| If no match is found, return 0.                                              |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOf ( const char *pSearchString,
                                  unsigned    searchLen,
                                  unsigned    startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, searchLen );
  if ( startPos )
    {
    switch ( searchLen )
      {
      case 0:
        break;
      case 1:
        while ( startPos-- )
          if ( this->contents()[ startPos ] == *pSearchString )
            return startPos + 1;
        break;
      default:
        while ( startPos-- )
          if ( memcmp( this->contents() + startPos,
                       pSearchString,
                       searchLen ) == 0 )
            return startPos + 1;
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::lastIndexOf                                                         |
|                                                                              |
|  Simply invoke the functionally equivalent lastIndexOfAnyOf.                 |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOf ( const IStringTest &aTest,
                                  unsigned           startPos ) const
  {
  return this->lastIndexOfAnyOf( aTest, startPos );
  }

/*------------------------------------------------------------------------------
| IBuffer::lastIndexOfAnyBut                                                   |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the                |
|    starting point) by calling startBackwardsSearch().                        |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an                  |
|    array (0-based) index.                                                    |
| 4. Search (going backwards) for the first character not in the               |
|    set of valid characters, based on the number of valid                     |
|    characters:                                                               |
|      0  - return the buffer length; i.e., the last character of              |
|           the buffer isn't in the (empty) set of valid characters            |
|      1  - compare each character of the buffer to the valid                  |
|           character until one doesn't match                                  |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of valid characters;                |
|           return the index of the first that isn't found                     |
|                                                                              |
|    If all characters pass the test, return 0.                                |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOfAnyBut ( const char *pValidChars,
                                        unsigned    numValidChars,
                                        unsigned    startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    {
    switch ( numValidChars )
      {
      case 0:
        return startPos;
      case 1:
        while ( startPos-- )
          if ( this->contents()[ startPos ] != *pValidChars )
            return startPos + 1;
        break;
      default:
        while ( startPos-- )
          if ( memchr( pValidChars,
                       this->contents()[ startPos ],
                       numValidChars ) == 0 )
            return startPos + 1;
      }
    }
  return 0;
  }


/*------------------------------------------------------------------------------
| IBuffer::lastIndexOfAnyBut                                                   |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startBackwardsSearch().                                 |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a false result.                          |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOfAnyBut ( const IStringTest &aTest,
                                        unsigned           startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    {
    while ( startPos-- )
      if ( !aTest.test( this->contents()[ startPos ] ) )
        return startPos + 1;
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::lastIndexOfAnyOf                                                    |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the                |
|    starting point) by calling startBackwardsSearch().                        |
| 2. If it can't succeed, return 0.                                            |
| 3. Otherwise, search for the first character in the set of                   |
|    search characters, based on the number of search characters:              |
|      0  - return 0; i.e., no characters of the buffer are in                 |
|           the (empty) set of search characters                               |
|      1  - search for a matching character in the buffer by                   |
|           comparing buffer characters one-by-one (going backwards).          |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of search characters;               |
|           return the index of the first that is found                        |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOfAnyOf ( const char *pValidChars,
                                       unsigned    numValidChars,
                                       unsigned    startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    {
    switch ( numValidChars )
      {
      case 0:
        break;
      case 1:
        while ( startPos-- )
          if ( this->contents()[ startPos ] == *pValidChars )
            return startPos + 1;
        break;
      default:
        while ( startPos-- )
          if ( memchr( pValidChars,
                       this->contents()[ startPos ],
                       numValidChars ) )
            return startPos + 1;
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IBuffer::lastIndexOfAnyOf                                                    |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startBackwardsSearch().                                 |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a true result.                           |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IBuffer :: lastIndexOfAnyOf ( const IStringTest &aTest,
                                       unsigned           startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    {
    while ( startPos-- )
      if ( aTest.test( this->contents()[ startPos ] ) )
        return startPos + 1;
    }
  return 0;
  }
