/*******************************************************************************
* FILE NAME: idbcsb0.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*  This file contains the implementation of the constructors for the           *
*  class IDBCSBuffer (and the implementation of related functions).            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <string.h>
  #include <stdarg.h>
  #include <stdio.h>
  #include <stdlib.h>
  }

#include <idbcsbuf.hpp>
#include <ibuffer.hpp>

#ifndef _IEXCBASE_
  #include <iexcbase.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDBCSB0_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IDBCSBuffer::className                                                       |
------------------------------------------------------------------------------*/
const char *IDBCSBuffer :: className ( ) const
  {
  return "IDBCSBuffer";
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::IDBCSBuffer                                                     |
|                                                                              |
| Construct a buffer of the specified length.  The "data" member               |
| array must actually be of length 1 greater than the argument                 |
| value (this is achieved automatically via use of the overloaded              |
| operator new for class IBuffer).                                             |
|                                                                              |
| The terminating (extra) byte is set to null.                                 |
|                                                                              |
| Notes:                                                                       |
|   1. This method is protected.  IDBCSBufferss must be obtained by using      |
|      IDBCSBuffer::nullBuffer and subsequent newBuffer calls to existing      |
|      IDBCSBuffer objects.  The only non-heap instance of this class is       |
|      the static IDCBSBuffer::nullBuffer object.                              |
------------------------------------------------------------------------------*/
IDBCSBuffer :: IDBCSBuffer ( unsigned len )
  : IBuffer( len ) 
  { 
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::~IDBCSBuffer                                                    |
|                                                                              |
| Empty dtor to prevent generation of static version.                          |
------------------------------------------------------------------------------*/
IDBCSBuffer :: ~IDBCSBuffer ( )
  {
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::allocate                                                        |
|                                                                              |
|  Allocate a new IBuffer of the same type as the receiver                     |
------------------------------------------------------------------------------*/
IBuffer *IDBCSBuffer :: allocate ( unsigned newLen ) const
{
  return new ( newLen ) IDBCSBuffer( newLen );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::rightJustify                                                    |
|                                                                              |
| This function right justifies the receiver in a new buffer of the            |
| given length and returns the address of the new buffer.                      |
|                                                                              |
| The right justification algorithm is as follows:                             |
|                                                                              |
|   1. Calculate how much padding will be needed on the left.                  |
|   2. Calculate how many bytes of the current buffer will be                  |
|      transferred to the right-justified result.                              |
|   3. Determine if the first byte of the new buffer is the 2nd byte           |
|      of a DBCS character.  If so, replace it with the padCharacter.          |
|   4. return a new buffer comprised of the "prefix" pad bytes                 |
|      and the portion of the receiver calculated at step 2.                   |
|                                                                              |
| Notes:                                                                       |
|   1. This function (and likewise all the similar "editing" functions)        |
|      is usually called in the following context:                             |
|         IString::xxxxx ( )                                                   |
|           {                                                                  |
|           ...                                                                |
|           IString old( this->pBuffer );                                      |
|           this->pBuffer = old.pBuffer->rightJustify( ... );                  |
|           ...                                                                |
|           }                                                                  |
|   2. This function appears here (rather than with the other editing          |
|      functions in idbcsb7.cpp) because it is called from the conversion      |
|      functions in istring9/idbcsb9.  Placing it here avoids pulling in       |
|      all the editing functions when only conversion is needed.               |
------------------------------------------------------------------------------*/
IBuffer * IDBCSBuffer :: rightJustify ( unsigned newLength,
                                        char     padCharacter )
{
  IBuffer *result = this;

  if ( newLength != this->length() )
  {
     unsigned prefix       = ( newLength > this->length() )
                               ? newLength - this->length()
                               : 0,
              fromReceiver = ( this->length() < newLength ) ?
                               this->length()
                               :
                               newLength;

     result = this->newBuffer( 0, prefix,
                               this->contents() + this->length() - fromReceiver,
                               fromReceiver,
                               0, 0,
                               padCharacter );

     //-----------------------------------------------------------------------
     // Check to see if the first byte of the returned string is a DBCS2.
     // If so, then it is replaced with the padCharacter.
     //-----------------------------------------------------------------------
     if ( fromReceiver < this->length() )
       for ( unsigned i = 0; i < fromReceiver; i++ )
         if ( this->charType( this->length() - fromReceiver + 1 + i ) >
                IStringEnum::mbcs1 )
           result->contents()[i] = padCharacter;
         else
           break;
  }
  return result;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer :: remove                                                        |
|                                                                              |
| If the buffer will remain unmodified (starting index is beyond               |
| the end or receiver is null), then the receiver is returned                  |
| unmodified.                                                                  |
|                                                                              |
| Else, if the number of characters to delete is zero, or passes               |
| the end of the receiver, then it is defaulted to the rest of the             |
| buffer.  Then, check to see if the result of removing the characters         |
| will split a DBCS character on either end.  If so, replace that              |
| character with a SBCS space. Finally, the new buffer is allocated            |
| and filled in via a call to newBuffer().                                     |
|                                                                              |
| Notes:                                                                       |
|   1. See notes 1. and 2. under IDBCSBuffer::rightJustify.                    |
------------------------------------------------------------------------------*/
IBuffer * IDBCSBuffer :: remove ( unsigned startPos,
                                  unsigned numChars )
{
  IBuffer *result = this;

  if ( startPos <= this->length()
       &&
       this->length() != 0 )
  {
    //------------------------------------------------------------------------
    // Consider startPos 0 to be 1:
    //------------------------------------------------------------------------
    if ( startPos == 0 )
      startPos++;

    //------------------------------------------------------------------------
    // Default numChars to rest of string:
    //------------------------------------------------------------------------
    if ( numChars > this->length() - startPos )
      numChars = this->length() - startPos + 1;

    //------------------------------------------------------------------------
    // Initialize from current contents before/after deleted chars:
    //------------------------------------------------------------------------
    result = this->newBuffer( this->contents(), startPos - 1,
                              this->contents() + startPos + numChars - 1,
                              this->length() - numChars - startPos + 1,
                              0, 0,
                              0 );

    //------------------------------------------------------------------------
    // If the byte at (startPos - 1) is the first byte of a DBCS character,
    // then that byte should be replaced with a SBCS space.
    // If the byte at (startPos + numChars) is the second byte of a DBCS
    // character, then that byte should be replaced with a SBCS space.
    //------------------------------------------------------------------------
    if ( startPos <= this->length () )
    {
      IStringEnum::CharType typeOfFirstRemoved = this->charType ( startPos );
      if ( typeOfFirstRemoved > IStringEnum::mbcs1 )
        for ( unsigned i = 1; i < typeOfFirstRemoved && i < startPos; i++)
          result->contents()[startPos - 1 - i] = ' ';
    }

    for ( unsigned i = 0; startPos + numChars + i <= this->length (); i++ ) {
      if ( this->charType( startPos + numChars + i ) > IStringEnum::mbcs1 )
        result->contents()[startPos - 1 + i] = ' ';
      else
	break;
    }
  }
  return result;
}
