/*******************************************************************************
* FILE NAME: idbcsb1.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*  This file contains the implementation of the IDBCSBuffer testing            *
*  functions.                                                                  *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <ctype.h>
  }

#include <ibuffer.hpp>
#include <istring.hpp>
#include <idbcsbuf.hpp>

#if !defined( _ISTRTEST_ )
  #include <istrtest.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDBCSB1_CPP_
  #include <ipagetun.h>
#endif

size_t IDBCSBuffer :: prevCharLength ( unsigned pos ) const
{
  if ( pos == 1 )
    return 1;
  else
  {
    IStringEnum::CharType ctype =  charType ( pos-1 );
    if ( ctype == IStringEnum::sbcs )
      return 1;
    else
      return (size_t) ctype;
  }
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isDBCS                                                          |
|                                                                              |
| isDBCS returns true if the string consists of only DBCS characters           |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isDBCS( ) const
{
  return isMBCS( );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isMBCS                                                          |
|                                                                              |
| isMBCS returns true if the string consists of only MBCS characters           |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isMBCS( ) const
{
  unsigned    i;
  const char* p = this->contents();

  for ( i = 0; i < this->length(); )
  {
    if ( isSBC ( p ) )
      return false;
    unsigned j = charLength(p);
    p += j;
    i += j;
  }
  return true;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isSBCS                                                          |
|                                                                              |
| isSBCS returns true if the string consists of only SBCS characters           |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isSBCS( ) const
{
  unsigned    i;
  const char* p = this->contents();

  for ( i = 0; i < this->length(); i++ )
  {
    if ( !isSBC ( p ) )
      return false;
    p++;
  }
  return true;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isValidDBCS                                                     |
|                                                                              |
| isValidDBCS checks to see if a null character (\0) appears at the second     |
| byte position of double-byte characters.                                     |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isValidDBCS( ) const
{
  return isValidMBCS( );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isValidMBCS                                                     |
|                                                                              |
| isValidMBCS checks to see if a null character (\0) appears at the second     |
| byte position of multi-byte characters.                                      |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isValidMBCS( ) const
{
  unsigned    i;
  const char* p = this->contents();

  for ( i = 0; i < this->length();)
  {
    unsigned nextChar = i + charLength ( p );
    p++;
    i++;
    if (i < nextChar)
    {
      for ( ; i < nextChar; i++, p++ )
        if ( *p == '\0' )
          return false;
    } else
        return false;
  }
  return true;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::includesDBCS                                                    |
|                                                                              |
| includesDBCS returns true if the string contains any DBCS characters         |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: includesDBCS( ) const
{
  return includesMBCS( );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::includesMBCS                                                    |
|                                                                              |
| includesMBCS returns true if the string contains any MBCS characters         |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: includesMBCS( ) const
{
  return( !isSBCS() );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer :: includesSBCS                                                  |
|                                                                              |
| includesSBCS returns true if the string contains any SBCS characters         |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: includesSBCS( ) const
{
  return( !isDBCS() );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::next                                                            |
|                                                                              |
| next returns a pointer to the next logical character in a string.            |
|                                                                              |
| Notes:                                                                       |
|   This function makes the assumption that the input character string         |
|   (pChars) is either pointing to a SBCS character or the first byte of a     |
|   DBCS character i.e. pChars is NOT pointing to the 2nd byte of a DBCS       |
|   character.                                                                 |
------------------------------------------------------------------------------*/
char* IDBCSBuffer :: next( const char *pChars )
{
  return( (char*)pChars + charLength ( pChars ) );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::next                                                            |
------------------------------------------------------------------------------*/
const char* IDBCSBuffer :: next( const char *pChars ) const
{
  return( (char*)pChars + charLength ( pChars ) );
}

/*------------------------------------------------------------------------------
| IDBCSBuffer :: charType                                                      |
|                                                                              |
| This function evaluates the byte at position index in a string and returns   |
| SBCS if the byte is a single byte character, DBCS1 if the byte is the first  |
| byte of a double byte character, and DBCS2 if the byte is the second byte    |
| of a double byte character.                                                  |
|                                                                              |
| Notes:                                                                       |
|   All invocations of charType from within IDBCSBuffer functions ensure that  |
|   the index is valid (i.e. index > 0 and index <= this->length()) before     |
|   calling charType.                                                          |
------------------------------------------------------------------------------*/
IStringEnum::CharType IDBCSBuffer :: charType( unsigned index ) const
{
  unsigned i = 0;

  for (;;)
  { unsigned next = i + charLength (&contents () [i]);
    if ( next >= index )
    {
      if ( next - i == 1 )
        return IStringEnum::sbcs;
      else
        return IStringEnum::CharType (index - i);
    }
    else
      i = next;
  } /* endfor */
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::isCharValid                                                     |
|                                                                              |
| This function determines if a character, pChar (which can be either a SBCS   |
| or DBCS character) is in the string of valid characters, pValidChars (which  |
| can contain any combination of SBCS and DBCS characters).                    |
------------------------------------------------------------------------------*/
IBase::Boolean IDBCSBuffer :: isCharValid( unsigned    pos,
                                           const char* pValidChars,
                                           unsigned    numValidChars ) const
  {
  char const* c = &contents () [ pos - 1 ];
  size_t lc = charLength ( c );

  while ( numValidChars >= lc )
  { size_t len = charLength ( pValidChars );
    if ( len == lc )
    {
      Boolean match = true;
      for (unsigned i = 0; i < len; i++)
      {
        if ( pValidChars [i] != c [i] )
        {
          match = false;
          break;
        }
      }
      if ( match )
        return true;
    }
    pValidChars += len;
    numValidChars -= len;
  }
  return false;
}
