/*******************************************************************************
* FILE NAME: idbcsb6.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*  This file contains the implementation of the IDBCSBuffer searching          *
*  functions.                                                                  *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <string.h>
  }

#include <ibuffer.hpp>
#include <idbcsbuf.hpp>

#if !defined( _ISTRTEST_ )
  #include <istrtest.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDBCSB6_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IDBCSBuffer::startSearch                                                     |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: startSearch ( unsigned startPos,
                                      unsigned searchLen ) const
  {
  unsigned 
    result = this->IBuffer::startSearch( startPos, searchLen );
  while ( this->charType( result ) > IStringEnum::mbcs1 )
    if ( ++result > this->length() )
    {
      result = 0;
      break;
    }
  return result;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::startBackwardsSearch                                            |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: startBackwardsSearch ( unsigned startPos,
                                               unsigned searchLen ) const
  {
  unsigned 
    result = this->IBuffer::startBackwardsSearch( startPos, searchLen );
  while ( result > 0 && this->charType( result ) > IStringEnum::mbcs1 )
    result--;
  return result;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOf                                                         |
|                                                                              |
| 1. First, check whether the search can succeed (and set the point at which   |
|    the search is to start) by calling startSearch().                         |
| 2. If the search cannot succeed, then return 0.                              |
| 3. Otherwise, compare the buffer contents starting at each index to the      |
|    search string until a match is found (in which case we return the index)  |
|    or the entire buffer is tested (in which case 0 is returned).             |
|                                                                              |
|                                                                              |
| Notes:                                                                       |
|   1. If the search string is a single character, then this function is       |
|      implemented via the equivalent invocation of indexOfAnyOf (which is     |
|      optimized for that case).  This saves the addtional logic of special    |
|      code to handle that case here, as well.                                 |
|   2. If the search string is null, then 0 is returned.                       |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOf ( const char *pSearchString,
                                  unsigned    searchLen,
                                  unsigned    startPos ) const
  {
  if ( searchLen == 1 )
    return this->indexOfAnyOf( pSearchString, searchLen, startPos );

  startPos = this->startSearch( startPos, searchLen );
  if ( startPos && searchLen )
    while ( (startPos + searchLen - 1) <= this->length() )
      if ( memcmp( this->contents() + startPos - 1,
                   pSearchString,
                   searchLen ) == 0 )
        return startPos;
      else 
        startPos += charLength ( startPos );
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOf                                                         |
|                                                                              |
| Simply invokes the functionally equivalent indexOf function.                 |
|                                                                              |
| Notes:                                                                       |
|   1. If the search string is null, then 0 is returned.                       |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOf ( const IStringTest &aTest,
                                  unsigned           startPos ) const
  {
  return this->IBuffer::indexOf( aTest, startPos );
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOfAnyBut                                                   |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an array (0-based)  |
|    index.                                                                    |
| 4. Search for the first character not in the set of valid characters, based  |
|    on the number of valid characters:                                        |
|      0  - return 1; i.e., the first character of the buffer                  |
|           isn't in the (empty) set of valid characters                       |
|      1  - compare each character of the buffer to the valid                  |
|           character until one doesn't match                                  |
|      >1 - call memchr for each character of the buffer to see if the         |
|           character is in the set of valid characters;                       |
|           return the index of the first that isn't found                     |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOfAnyBut ( const char *pValidChars,
                                        unsigned    numValidChars,
                                        unsigned    startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    switch ( numValidChars )
      {
      case 0:
        return startPos;
      case 1:
        // No character can match single DBCS first-byte:
        if ( ! isSBC ( pValidChars ) )
          return startPos;
        // Examine each character of this string...
        while ( startPos <= this->length() )
          // Compare this character.  Note that if current character
          // is DBCS first-byte, this compare will fail.
          if ( this->contents()[ startPos - 1 ] != *pValidChars )
            return startPos;
          else
            startPos++;
        break;
      default:
        while ( startPos <= this->length() )
          {
          if ( !this->isCharValid( startPos, pValidChars, numValidChars ) )
            return startPos;
          else
            startPos += charLength ( startPos );
          }
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOfAnyBut                                                   |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a false result.                          |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOfAnyBut ( const IStringTest &aTest,
                                        unsigned           startPos ) const
{
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
  {
    while ( startPos <= this->length() )
      {
      // Note:  This is broken!  Must somehow pass both bytes of DBCS!
      if ( !aTest.test( this->contents()[ startPos - 1 ] ) )
        return startPos;
      else
        startPos += charLength ( startPos );
      }
  }
  return 0;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOfAnyOf                                                    |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an array (0-based)  |
|    index.                                                                    |
| 4. Search for the first character in the set of search characters, based     |
|    on the number of search characters:                                       |
|      0  - return 0; i.e., no characters of the buffer are in the (empty)     |
|           set of search characters                                           |
|      1  - search for a matching character in the buffer using memchr         |
|      >1 - call memchr for each character of the buffer to see if the         |
|           character is in the set of search characters;                      |
|           return the index of the first that is found                        |
|                                                                              |
|  If all characters fail the test, return 0.                                  |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOfAnyOf ( const char *pValidChars,
                                       unsigned    numValidChars,
                                       unsigned    startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    switch ( numValidChars )
      {
      case 0:
        break;
      case 1:
        while ( startPos <= this->length() )
          {
          if ( this->contents()[ startPos - 1 ] == *pValidChars )
            return startPos;
          startPos += charLength ( startPos );
          }
        break;
      default:
        while ( startPos <= this->length() )
          {
          if ( this->isCharValid( startPos,
                                  pValidChars,
                                  numValidChars ) )
            return startPos;
          else
            startPos += charLength ( startPos );
          }
      }
    }
  return 0;
  }


/*------------------------------------------------------------------------------
| IDBCSBuffer::indexOfAnyOf                                                    |

|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startSearch().                                          |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a true result.                           |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: indexOfAnyOf ( const IStringTest &aTest,
                                       unsigned           startPos ) const
  {
  startPos = this->startSearch( startPos, 1 );
  if ( startPos )
    {
    while ( startPos <= this->length() )
      if ( aTest.test( this->contents()[ startPos - 1 ] ) )
        return startPos;
      else
        startPos++;
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::lastIndexOf                                                     |
|                                                                              |
|  1. Adjust starting position and see if the search can                       |
|     succeed by calling startBackwardsSearch().                               |
|  2. If a match isn't possible, return 0.                                     |
|  3. Otherwise, search (backward) according to length of search               |
|     string:                                                                  |
|       0  - return 0;                                                         |
|       1  - search for a matching character in the buffer by                  |
|            comparing one byte at a time                                      |
|       >1 - call memcmp for each character of the buffer to see               |
|            if the characters starting here match the search string.          |
|                                                                              |
| If no match is found, return 0.                                              |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOf ( const char *pSearchString,
                                      unsigned    searchLen,
                                      unsigned    startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, searchLen );
  if ( startPos )
    {
    switch ( searchLen )
      {
      case 0:
        break;
      case 1:
        while ( startPos )
          if ( this->contents()[ startPos - 1 ] == *pSearchString )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
        break;
      default:
        while ( startPos )
          {
          if ( memcmp( this->contents() + startPos - 1,
                       pSearchString,
                       searchLen ) == 0 )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
          }
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::lastIndexOf                                                     |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOf ( const IStringTest &aTest,
                                      unsigned           startPos ) const
  {
  return this->IBuffer::lastIndexOf( aTest, startPos );
  }


/*------------------------------------------------------------------------------
| IDBCSBuffer::lastIndexOfAnyBut                                               |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the                |
|    starting point) by calling startBackwardsSearch().                        |
| 2. If it can't succeed, return 0.                                            |
| 3. If it can succeed, the starting point is converted to an                  |
|    array (0-based) index.                                                    |
| 4. Search (going backwards) for the first character not in the               |
|    set of valid characters, based on the number of valid                     |
|    characters:                                                               |
|      0  - return the buffer length; i.e., the last character of              |
|           the buffer isn't in the (empty) set of valid characters            |
|      1  - compare each character of the buffer to the valid                  |
|           character until one doesn't match                                  |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of valid characters;                |
|           return the index of the first that isn't found                     |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOfAnyBut ( const char *pValidChars,
                                            unsigned    numValidChars,
                                            unsigned    startPos ) const
{
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    {
    switch ( numValidChars )
      {
      case 0:
        return startPos;
      case 1:
        if ( ! isSBC ( pValidChars ) )
          return 0;
        while ( startPos )
          if ( this->contents()[ startPos - 1 ] != *pValidChars )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
        break;
      default:
        while ( startPos )
          if ( !this->isCharValid( startPos, pValidChars, numValidChars ) )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
      }
    }
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::lastIndexOfAnyBut                                               |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startBackwardsSearch().                                 |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a false result.                          |
|                                                                              |
| If all characters pass the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOfAnyBut ( const IStringTest &aTest,
                                            unsigned   startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  while ( startPos )
    if ( !aTest.test( this->contents()[ startPos - 1 ] ) )
      return startPos;
    else
      startPos -= prevCharLength ( startPos );
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer :: lastIndexOfAnyOf                                              |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the                |
|    starting point) by calling startBackwardsSearch().                        |
| 2. If it can't succeed, return 0.                                            |
| 3. Otherwise, search for the first character in the set of                   |
|    search characters, based on the number of search characters:              |
|      0  - return 0; i.e., no characters of the buffer are in                 |
|           the (empty) set of search characters                               |
|      1  - search for a matching character in the buffer by                   |
|           comparing buffer characters one-by-one (going backwards).          |
|      >1 - call memchr for each character of the buffer to see                |
|           if the character is in the set of search characters                |
|           return the index of the first that is found                        |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOfAnyOf ( const char *pValidChars,
                                           unsigned    numValidChars,
                                           unsigned    startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  if ( startPos )
    switch ( numValidChars )
      {
      case 0:
        break;
      case 1:
        while ( startPos )
          if ( this->contents()[ startPos - 1 ] == *pValidChars )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
        break;
      default:
        while ( startPos )
          if ( this->isCharValid( startPos, pValidChars, numValidChars ) )
            return startPos;
          else
            startPos -= prevCharLength ( startPos );
      }
  return 0;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::lastIndexOfAnyOf                                                |
|                                                                              |
| 1. Checking that the search can succeed (and establishing the starting       |
|    point) by calling startBackwardsSearch().                                 |
| 2. Invokes the IStringTest test method against each character in turn, until |
|    the argument IStringTest returns a true result.                           |
|                                                                              |
| If all characters fail the test, return 0.                                   |
------------------------------------------------------------------------------*/
unsigned IDBCSBuffer :: lastIndexOfAnyOf ( const IStringTest &aTest,
                                          unsigned      startPos ) const
  {
  startPos = this->startBackwardsSearch( startPos, 1 );
  while ( startPos )
    if ( aTest.test( this->contents()[ startPos - 1 ] ) )
      return startPos;
    else
      startPos -= prevCharLength ( startPos );
  return 0;
  }
