/*************************************************************************
  IBM C/C++ Tools Version 3.00 - Collection Class Library
 (C) Copyright IBM Corporation 1992 ,1995, Licensed Program-Property of
 IBM.  All Rights Reserved.  US Government Users Restricted Rights - Use,
 duplication or disclosure restricted by GSA ADP Schedule Contract with
 IBM Corp.
 *************************************************************************/

#include <iiavlkss.h>
#include <ibexcept.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IIAVLKSS_CPP_
  #include <ipagetun.h>
#endif


// assert: avlLeft == - avlRight
typedef short Direction;
static const Direction avlLeft = -1;
static const Direction avlNone = 0;
static const Direction avlRight = 1;

inline Direction avlDirection (long comp)
{
// assume that comp is the result from compareKey... (node, element);
// the result is the direction in which a search for element has to
// proceed from node
  IASSERT (comp != 0);
  return comp < 0 ? avlRight : avlLeft;
}

IAvlKeySortedSetImplNode::
IAvlKeySortedSetImplNode ()
{ ivLeft   = 0;
  ivRight  = 0;
  ivParent = 0;
  ivDeeperSubtree = avlNone;
}

inline IAvlKeySortedSetImplNode*&
IAvlKeySortedSetImplNode::
link (Direction d)
{ IASSERT (d != avlNone);
  return d == avlLeft ? ivLeft : ivRight;
}

inline void IAvlKeySortedSetImplNode::
parent (IAvlKeySortedSetImplNode*& P, Direction& d)
{ P = ivParent;
  if (P != 0)
    d = (P->ivLeft == this ? avlLeft : avlRight);
}

inline void IAvlKeySortedSetImplNode::
connectChild (Direction d, IAvlKeySortedSetImplNode* n)
{ IASSERT (n != 0);
  link (d) = n;
  n->ivParent = this;
}

inline void IAvlKeySortedSetImplNode::
connectChild0 (Direction d, IAvlKeySortedSetImplNode* n)
{ link (d) = n;
  if (n != 0)
    n->ivParent = this;
}

inline void IAvlKeySortedSetImpl::
replace (IAvlKeySortedSetImplNode* oldNode,
         IAvlKeySortedSetImplNode* newNode)
{ IAvlKeySortedSetImplNode *parent = oldNode->ivParent;

  if (newNode != 0)
    newNode->ivParent = parent;

  if (parent == 0)
    ivRoot = newNode;
  else if (oldNode == parent->ivRight)
    parent->ivRight = newNode;
  else
    parent->ivLeft = newNode;
}

IAvlKeySortedSetImpl::
IAvlKeySortedSetImpl ()
{ ivRoot = 0;
  ivNumberOfElements = 0;
}

IAvlKeySortedSetImpl::
IAvlKeySortedSetImpl (IAvlKeySortedSetImpl const&)
{ ivRoot = 0;
  ivNumberOfElements = 0;
}

IAvlKeySortedSetImpl::
~IAvlKeySortedSetImpl ()
{
}

void IAvlKeySortedSetImpl::
operator = (IAvlKeySortedSetImpl const& aTree)
{ if (this != &aTree) {
    removeAll();
    if (aTree.ivRoot != 0) {
      ivRoot = copySubtree(aTree.ivRoot);
      // If Copy does not succeed, next statements are not executed.
      ivNumberOfElements = aTree.ivNumberOfElements;
      ivRoot->ivParent = 0;
    }
  }
}

void IAvlKeySortedSetImpl::
addAllFrom (IAvlKeySortedSetImpl const& aTree)
{ if (aTree.ivRoot != 0)
    unionWithSubtree(aTree.ivRoot);
}

IBoolean IAvlKeySortedSetImpl::
add (void const* element)
{ IAvlKeySortedSetImplNode *dummy;
  return add (element, dummy);
}

IBoolean IAvlKeySortedSetImpl::
add (void const* element, IAvlKeySortedSetImplNode*& node)
{ if (isEmpty ()) {
    node = ivRoot = newNode (element); // Can throw PoolExhausted
    ivNumberOfElements++;
    return True;
  }

  IAvlKeySortedSetImplNode *S = ivRoot;
  IAvlKeySortedSetImplNode *P = ivRoot;
  IAvlKeySortedSetImplNode *Q;
  Direction d;

  for (;;) {
    // P will move down the tree.
    // S will point to the place where rebalancing may be necessary.

    long comp = compareKeyToElement (P, element);

    if (comp == 0) {
      node = P;
      return False;
    }

    d = avlDirection (comp);
    Q = P->link (d);
    if (Q == 0) {
      node = Q = newNode(element); // Can throw PoolExhausted
      P->connectChild (d, Q);
      ivNumberOfElements++;
      break;
    }

    P = Q;

    if (P->ivDeeperSubtree != avlNone)
      S = P;
  } // end for

  // Q is newly inserted node
  // P->link (d) == Q
  // S == min N : (N == ivRoot || ! N is balanced) && N is ancestor of Q
  // where 'min' means 'nearest to Q'

  for (IAvlKeySortedSetImplNode *R = Q;; R = P, R->parent (P, d)) {
    // P->link (d) == R
    // P ancestor of Q && S ancestor of or equal to P.
    // P == S || P is balanced

    if (P == S)
      break;

    // P is balanced
    P->ivDeeperSubtree = d;
  }

  // R->ivParent == S
  // S->link (d) has become higher by 1

  if (S->ivDeeperSubtree == avlNone)

    S->ivDeeperSubtree = d;

  else if (S->ivDeeperSubtree == -d)

    S->ivDeeperSubtree = avlNone;

  else {
    // S->ivDeeperSubtree == d.
    // Rotate

    if (R->ivDeeperSubtree == d) {
      // assume d == avlRight
      // Single rotation: (U, V may be 0)
      //         S(N)                                      R(N)
      //
      //  U(N-2)       R(N)          ----->        S(N-1)        W(N-1)
      //
      //         V(N-2)    W(N-1)            U(N-2)      V(N-2)
      //                         ...
      //                          Q

      replace (S, R);
      S->connectChild0 (d, R->link (-d));
      R->connectChild (-d, S);

      S->ivDeeperSubtree = avlNone;
      R->ivDeeperSubtree = avlNone;
    }
    else {
      // assume d == avlRight
      // Double rotation: (U, W, X, Y may be 0)
      //         S(N)                                P(N)
      //
      //   U(N-2)     R(N)                  S(N-1)           R(M-1)
      //
      //        P(N-1)    W(N-2) -----> U(N-2) X(N-2/3) Y(N-3/2) W(N-2)
      //
      //     X(N-2/3) Y(N-3/2)
      //            ...
      //             Q

      P = R->link (-d);
      replace (S, P);
      R->connectChild0 (-d, P->link (d));
      S->connectChild0 (d, P->link (-d));
      P->connectChild (d, R);
      P->connectChild (-d, S);

      if (P->ivDeeperSubtree == d)
        S->ivDeeperSubtree = -d;
      else
        S->ivDeeperSubtree = avlNone;

      if (P->ivDeeperSubtree == -d)
        R->ivDeeperSubtree = d;
      else
        R->ivDeeperSubtree = avlNone;

      P->ivDeeperSubtree = avlNone;
    } // single of double
  }

  return True;
}

IBoolean IAvlKeySortedSetImpl::
addOrReplaceElementWithKey (void const* element,
                            IAvlKeySortedSetImplNode*& node)
{ IBoolean added = add (element, node);
  if (! added)
    copyFrom(node, element);
  return added;
}

IBoolean IAvlKeySortedSetImpl::
addOrReplaceElementWithKey (void const* element)
{ IAvlKeySortedSetImplNode *dummy;
  return addOrReplaceElementWithKey(element, dummy);
}

IBoolean IAvlKeySortedSetImpl::
locateOrAddElementWithKey (void const* element,
                           IAvlKeySortedSetImplNode*& node)
{ return ! add (element, node);
}

IBoolean IAvlKeySortedSetImpl::
locateOrAddElementWithKey (void const* element)
{ IAvlKeySortedSetImplNode *dummy;
  return locateOrAddElementWithKey(element, dummy);
}

void IAvlKeySortedSetImpl::
removeAt (IAvlKeySortedSetImplNode* node)
{ IBoolean changeDepth;
  IAvlKeySortedSetImplNode* P;
  Direction d, dP;
  node->parent (P, d);

  disconnectAndRemove (node, changeDepth);

  for (node = P; changeDepth && node != 0; node = P, d = dP) {
    node->parent (P, dP);
    balance (node, d, changeDepth);
  }

  ivNumberOfElements--;
}

INumber IAvlKeySortedSetImpl::
removeAll (void* predicate, void* env)
{ IAvlKeySortedSetImplNode *current, *next;
  INumber removed = 0;
  for (setToFirst (current), setToFirst (next);
       current != 0; current = next) {
    setToNext (next);
    if (constantFunctionIteration(predicate, env, current)) {
      removeAt (current);
      removed++;
    }
  }
  return removed;
}

void const* IAvlKeySortedSetImpl::
elementAt (IAvlKeySortedSetImplNode const* node) const
{ return (constElementAtNode (node));
}

void* IAvlKeySortedSetImpl::
elementAt (IAvlKeySortedSetImplNode* node)
{ return elementAtNode (node);
}

void IAvlKeySortedSetImpl::
removeAll ()
{ if (ivRoot != 0) {
    deleteTree(ivRoot);
    ivRoot = 0;
    ivNumberOfElements = 0;
  }
}

IBoolean IAvlKeySortedSetImpl::
isBounded () const
{ return False;
}

IBoolean IAvlKeySortedSetImpl::
isFull () const
{ return False;
}

IBoolean IAvlKeySortedSetImpl::
setToFirst (IAvlKeySortedSetImplNode*& node) const
{ node = ivRoot;
  if (node == 0)
    return False;
  else {
    while (node->ivLeft != 0) {
      node = node->ivLeft;
    }
    return True;
  }
}

IBoolean IAvlKeySortedSetImpl::
setToNext (IAvlKeySortedSetImplNode*& node) const
{ IASSERT (node != 0);
  //
  // if the given node has a right subtree,
  // set the node to the leftmost child of this subtree
  // otherwise find the next ancestor of the given node
  // for which the given node is in its
  // left subtree.
  //

  if (node->ivRight != 0) {
    node = node->ivRight;
    while (node->ivLeft != 0)
      node = node->ivLeft;
  }
  else {
    while (node->ivParent != 0 && node->ivParent->ivLeft != node)
      node = node->ivParent;
    // node->ivParent == 0 || node->ivParent->ivLeft == node
    node = node->ivParent;
  }

  return node != 0;
}

IBoolean IAvlKeySortedSetImpl::
setToLast (IAvlKeySortedSetImplNode*& node) const
{ node = ivRoot;
  if (node == 0)
    return False;
  else {
    while (node->ivRight != 0) {
      node = node->ivRight;
    }
    return True;
  }
}

IBoolean IAvlKeySortedSetImpl::
setToPrevious (IAvlKeySortedSetImplNode*& node) const
{ IASSERT (node != 0);

  // mirror to setToNext

  if (node->ivLeft != 0) {
    node = node->ivLeft;
    while (node->ivRight != 0)
      node = node->ivRight;
  }
  else {
    while (node->ivParent != 0 && node->ivParent->ivRight != node)
      node = node->ivParent;
    // node->ivParent == 0 || node->ivParent->ivRight == node
    node = node->ivParent;
  }

  return node != 0;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDo (void* function, void* env)
{ if (ivRoot != 0)
    return allElementsDoSubtree (function, env, ivRoot);
  else
    return True;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDoSubtree (void* function, void* env,
                      IAvlKeySortedSetImplNode* current)
{ IASSERT (current != 0);

  IBoolean rv = True;

  if (current->ivLeft != 0)
    rv = allElementsDoSubtree (function, env, current->ivLeft);

  rv = rv && functionIteration(function, env, current);

  if (rv && current->ivRight != 0)
    rv = allElementsDoSubtree (function, env, current->ivRight);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDo (void* iterator)
{ if (ivRoot != 0)
    return allElementsDoSubtree(iterator, ivRoot);
  else
    return True;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDoSubtree (void* iterator, IAvlKeySortedSetImplNode* current)
{ IASSERT (current != 0);

  IBoolean rv = True;

  if (current->ivLeft != 0)
    rv = allElementsDoSubtree(iterator, current->ivLeft);

  rv = rv && iteratorIteration(iterator, current);

  if (rv && current->ivRight != 0)
    rv = allElementsDoSubtree(iterator, current->ivRight);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDo (void* function, void* env) const
{ if (ivRoot != 0)
    return allElementsDoSubtree(function, env, ivRoot);
  else
    return True;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDoSubtree (void* function, void* env,
                      IAvlKeySortedSetImplNode const* current) const
{ IASSERT (current != 0);

  IBoolean rv = True;

  if (current->ivLeft != 0)
    rv = allElementsDoSubtree(function, env, current->ivLeft);

  rv = rv && constantFunctionIteration(function, env, current);

  if (rv && current->ivRight != 0)
    rv = allElementsDoSubtree(function, env, current->ivRight);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDo (void* iterator) const
{ if (ivRoot != 0)
    return allElementsDoSubtree(iterator, ivRoot);
  else
    return True;
}

IBoolean IAvlKeySortedSetImpl::
allElementsDoSubtree (void* iterator,
                      IAvlKeySortedSetImplNode const* current) const
{ IASSERT (current != 0);

  IBoolean rv = True;

  if (current->ivLeft != 0)
    rv = allElementsDoSubtree(iterator, current->ivLeft);

  rv = rv && constantIteratorIteration(iterator, current);

  if (rv && current->ivRight != 0)
    rv = allElementsDoSubtree(iterator, current->ivRight);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
containsAllKeysFromSubtree
         (IAvlKeySortedSetImplNode const* current) const
{
  // NOTE: current is not in the 'this' tree!
  IBoolean rv;

  rv = containsElementWithKey(keyAt (current));

  if (rv && current->ivLeft != 0)
    rv = containsAllKeysFromSubtree(current->ivLeft);

  if (rv && current->ivRight != 0)
    rv = containsAllKeysFromSubtree(current->ivRight);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
containsAllKeysFrom (IAvlKeySortedSetImpl const& aTree) const
{ IBoolean rv = True;

  if (aTree.ivRoot != 0)
    rv = containsAllKeysFromSubtree(aTree.ivRoot);
  else
    rv = True;

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
containsElementWithKey (void const* key) const
{ IAvlKeySortedSetImplNode *dummy;
  return locateElementWithKey (key, dummy);
}

IBoolean IAvlKeySortedSetImpl::
replaceElementWithKey (void const* element,
                       IAvlKeySortedSetImplNode*& node)
{ IBoolean rv;

  rv = locateElementWithKeyOfElement (element, node);
  if (rv)
    copyFrom (node, element);

  return rv;
}

IBoolean IAvlKeySortedSetImpl::
replaceElementWithKey (void const* element)
{ IAvlKeySortedSetImplNode *node;
  IBoolean rv;

  rv = locateElementWithKeyOfElement (element, node);
  if (rv)
    copyFrom (node, element);

  return rv;
}

void IAvlKeySortedSetImpl::
replaceAt (IAvlKeySortedSetImplNode* node, void const* element)
{ copyFrom (node, element);
}

IBoolean IAvlKeySortedSetImpl::
locateElementWithKey (void const* key,
                      IAvlKeySortedSetImplNode*& node) const
{ node = ivRoot;

  while (node != 0) {
    long comp = compareKeyToKey (node, key);

    if (comp == 0)
      return True;
    else
      node = node->link (avlDirection (comp));
  }

  return False;
}

IBoolean IAvlKeySortedSetImpl::
locateElementWithKeyOfElement (void const* element,
                               IAvlKeySortedSetImplNode*& node) const
{ node = ivRoot;

  while (node != 0) {
    long comp = compareKeyToElement (node, element);

    if (comp == 0)
      return True;
    else
      node = node->link (avlDirection (comp));
  }

  return False;
}

IBoolean IAvlKeySortedSetImpl::
removeElementWithKey (void const* key)
{ IAvlKeySortedSetImplNode *node;

  if (locateElementWithKey (key, node)) {
    removeAt (node);
    return True;
  }
  else
    return False;
}

void const* IAvlKeySortedSetImpl::
elementWithKey (void const* key) const
{ IAvlKeySortedSetImplNode *node;

  ICHECK (locateElementWithKey(key, node), INotContainsKeyException,
          IKeyNotContainedText);
  return elementAtNode (node);
}

void* IAvlKeySortedSetImpl::
elementWithKey (void const* key)
{ IAvlKeySortedSetImplNode *node;

  ICHECK (locateElementWithKey(key, node), INotContainsKeyException,
          IKeyNotContainedText);
  return elementAtNode (node);
}

void IAvlKeySortedSetImpl::
deleteTree (IAvlKeySortedSetImplNode* current)
{ IASSERT (current != 0);

  // By testing nulliness of the tree to count BEFORE calling this
  // method, about half of the functioncalls is saved.
  // (An optimal tree of depth 'n' contains 2^(n-1) leaves, which
  // would result in 2^n function calls for the leaves (with null
  // child pointers). In our approach, these calls are saved.

  if (current->ivLeft != 0)
    deleteTree (current->ivLeft);

  if (current->ivRight != 0)
    deleteTree (current->ivRight);

  deleteNode (current);
}

void IAvlKeySortedSetImpl::
balance (IAvlKeySortedSetImplNode* P,
         Direction d, IBoolean &changeDepth)
{
  // (re)balance node P for which the depth of the subtree in
  // direction d has decreased by 1;
  // changeDepth is set to False if the depth of n does not change.

  IASSERT (P != 0);

  IAvlKeySortedSetImplNode *Q, *R;

  if (P->ivDeeperSubtree == d) {

    P->ivDeeperSubtree = avlNone;
  }
  else if (P->ivDeeperSubtree == avlNone) {

    P->ivDeeperSubtree = -d;
    changeDepth = False;
  }
  else {
    // P->ivDeeperSubtree == -d
    // Rebalance

    Q = P->link (-d);
    if (Q->ivDeeperSubtree != d) {
      // assume d == avlLeft
      // Single rotation: (U, V, W may be 0)
      //         P(N)                                 Q(N/N-1)
      //
      //  U(N-3)      Q(N-1)         ----->     P(N-1/2)        W(N-2)
      //
      //         V(N-2/3)  W(N-2)          U(N-3)      V(N-2/3)

      replace (P, Q);
      P->connectChild0 (-d, Q->link (d));
      Q->connectChild (d, P);
      if (Q->ivDeeperSubtree == avlNone) {
        P->ivDeeperSubtree = -d;
        Q->ivDeeperSubtree = d;
        changeDepth = False;
      }
      else {
        P->ivDeeperSubtree = avlNone;
        Q->ivDeeperSubtree = avlNone;
      }
    }
    else {
      // assume d == avlLeft
      // Double rotation: (U, W, X, Y may be 0)
      //        P(N)                                R(N-1)
      //
      //   U(N-3)   Q(N-1)                  P(N-2)            Q(M-2)
      //
      //      R(N-2)     W(N-3) ----->  U(N-3) X(N-3/4)  Y(N-4/3) W(N-3)
      //
      //  X(N-3/4) Y(N-4/3)

      R = Q->link (d);

      replace (P, R);
      Q->connectChild0 (d, R->link (-d));
      P->connectChild0 (-d, R->link (d));
      R->connectChild (-d, Q);
      R->connectChild (d, P);

      if (R->ivDeeperSubtree == -d)
          P->ivDeeperSubtree = d;
      else
          P->ivDeeperSubtree = avlNone;

      if (R->ivDeeperSubtree == d)
          Q->ivDeeperSubtree = -d;
      else
          Q->ivDeeperSubtree = avlNone;

      R->ivDeeperSubtree = avlNone;
    } // single or double rotate
  } // P->ivDeeperSubtree == -d
}

void IAvlKeySortedSetImpl::
disconnectAndRemove (IAvlKeySortedSetImplNode* P, IBoolean& changeDepth)
{
  // remove P and rebalance subtrees
  // changeDepth tells whether the depth of the replaced tree has
  // changed

  if (P->ivRight == 0) {

    replace (P, P->ivLeft);
    changeDepth = True;
  }
  else if (P->ivLeft == 0) {

    replace (P, P->ivRight);
    changeDepth = True;
  }
  else {

    // Find the node R which is in-order just to the
    // left of the node P which is to be removed.
    // That is, go left and then right until 0 encountered.
    //
    // A, ..., B, W may be 0
    //
    //          P                                  R
    //
    //    A         V                        A           V
    //     .                                  .
    //      .                                  .
    //   ..  B              ----->          ..  B
    //
    //    ..   R                             ..  W
    //
    //        W
    //

    IAvlKeySortedSetImplNode* R = P->ivLeft;
    while (R->ivRight != 0)
      R = R->ivRight;

    IAvlKeySortedSetImplNode* B;
    if (R != P->ivLeft) {
      B = R->ivParent;
      B->connectChild0 (avlRight, R->ivLeft);
      R->connectChild (avlLeft, P->ivLeft);
    }
    else {
    // there is no A, ..., B, and W remains the left child of R
      B = 0;
    }

    replace (P, R);
    R->connectChild (avlRight, P->ivRight);
    R->ivDeeperSubtree = P->ivDeeperSubtree;
    // possibly diminished depth of R->ivLeft is considered later

    // rebalance A, ..., B
    changeDepth = True;
    IAvlKeySortedSetImplNode* BParent;
    for (changeDepth = True;
         changeDepth && B != 0 && B != R;
         B = BParent) {
      BParent = B->ivParent;
      balance (B, avlRight, changeDepth);
    }

    if (changeDepth)
      balance (R, avlLeft, changeDepth);
  }

  deleteNode (P);
}

IBoolean IAvlKeySortedSetImpl::
checkNode (IAvlKeySortedSetImplNode const* node) const
{ if (node == 0)
    return True;
  else if (ivRoot != 0)
    return checkNode (node, ivRoot);
  else
    return False;
}

IBoolean IAvlKeySortedSetImpl::
checkNode (IAvlKeySortedSetImplNode const* node,
           IAvlKeySortedSetImplNode* current) const
{ IBoolean found = False;

  found = (node == current);

  if (! found && current->ivLeft != 0)
    found = checkNode (node, current->ivLeft);

  if (! found && current->ivRight != 0)
    found = checkNode (node,current->ivRight);

  return found;
}

IAvlKeySortedSetImplNode* IAvlKeySortedSetImpl::
copySubtree (IAvlKeySortedSetImplNode const* current)
{ IASSERT (current != 0);

  IAvlKeySortedSetImplNode* newNode = copyNode(current);
  // Will throw if OutOfMemory
  newNode->ivDeeperSubtree = current->ivDeeperSubtree;

  ITRY {
    if (current->ivLeft != 0) {
      newNode->ivLeft = copySubtree(current->ivLeft);
      newNode->ivLeft->ivParent = newNode;
    }
    if (current->ivRight != 0) {
      newNode->ivRight = copySubtree(current->ivRight);
      newNode->ivRight->ivParent = newNode;
    }
  }
  IENDTRY
  ICATCH (IOutOfMemory exc) {
    deleteTree(newNode);
    IRETHROW (exc);
  }
  IENDCATCH

  return newNode;
}

void IAvlKeySortedSetImpl::
removeFirst ()
{ IAvlKeySortedSetImplNode *node;

  setToFirst(node);
  removeAt(node);
}

void IAvlKeySortedSetImpl::
removeLast ()
{ IAvlKeySortedSetImplNode *node;

  setToLast(node);
  removeAt(node);
}

void IAvlKeySortedSetImpl::
removeAtPosition (IPosition position)
{ IAvlKeySortedSetImplNode *node;

  setToPosition(position, node);
  removeAt(node);
}

void const* IAvlKeySortedSetImpl::
firstElement () const
{ IAvlKeySortedSetImplNode *node;

  setToFirst(node);
  return elementAtNode (node);
}

void const* IAvlKeySortedSetImpl::
lastElement () const
{ IAvlKeySortedSetImplNode *node;

  setToLast(node);
  return elementAtNode (node);
}

void const* IAvlKeySortedSetImpl::
elementAtPosition (IPosition position) const
{ IAvlKeySortedSetImplNode *node;

  setToPosition(position, node);
  return elementAtNode (node);
}

void IAvlKeySortedSetImpl::
setToPosition (IPosition position,
               IAvlKeySortedSetImplNode*& node) const
{ // When the position is past the half of the tree, the node
  // is moved backwards, otherwise it is moved forward.

  INumber num = numberOfElements ();

  IASSERT (position >= 1);
  IASSERT (position <= num);

  if (position < (num / 2))      // count forwards
  { setToFirst(node);
    while (--position > 0)
      setToNext(node);
  }
  else
  { position = num - position + 1;    // count backwards
    setToLast(node);
    while (--position > 0)
      setToPrevious(node);
  }
}

#ifndef __BIN_COMPAT_BASE_43__
IPosition IAvlKeySortedSetImpl::
position (IAvlKeySortedSetImplNode const* node) const
{ IAvlKeySortedSetImplNode *tmpnode;
  IPosition pos = 1;

  setToFirst(tmpnode);
  while (tmpnode != node) {
    setToNext (tmpnode);
    pos++;
  }
  return pos;
}
#endif

IBoolean IAvlKeySortedSetImpl::
isFirst (IAvlKeySortedSetImplNode const* node) const
{ IAvlKeySortedSetImplNode *node1;
  setToFirst (node1);
  return node == node1;
}

IBoolean IAvlKeySortedSetImpl::
isLast (IAvlKeySortedSetImplNode const* node) const
{ IAvlKeySortedSetImplNode *node1;
  setToLast (node1);
  return node == node1;
}

void IAvlKeySortedSetImpl::
unionWithSubtree (IAvlKeySortedSetImplNode const* current)
{ add (constElementAtNode (current));

  if (current->ivLeft != 0)
    unionWithSubtree(current->ivLeft);

  if (current->ivRight != 0)
    unionWithSubtree(current->ivRight);

}

long IAvlKeySortedSetImpl::
compare (IAvlKeySortedSetImpl const& aTree, void *function) const
{ IAvlKeySortedSetImplNode *node1, *node2;

  IBoolean next1 = setToFirst(node1);
  IBoolean next2 = aTree.setToFirst(node2);

  long compare = 0;
  while (next1 && next2 && compare == 0) {
    compare = functionComparison(function, node1, node2);
    if (compare == 0) {
      next1 = setToNext(node1);
      next2 = setToNext(node2);
    }
  }

  if (compare == 0 && (next1 || next2)) {
    //
    // we have to return a result based on the size of the collection
    //
    compare = numberOfElements() - aTree.numberOfElements();
  }

  return compare;
}

IBoolean IAvlKeySortedSetImpl::
isConsistent () const
{ if (ivRoot != 0 && ivRoot->ivParent != 0) {
    return False;
  }
  int dummy, nrOfElm;
  if (! isConsistent (ivRoot, dummy, nrOfElm))
    return False;
  if (nrOfElm != ivNumberOfElements) {
    return False;
  }
  return True;
}

IBoolean IAvlKeySortedSetImpl::
isConsistent (IAvlKeySortedSetImplNode const* n,
              int& depth, int& nrOfElm) const
{ if (n == 0) {
    depth = 0;
    nrOfElm = 0;
    return True;
  }

  if (n->ivLeft != 0 && n->ivLeft->ivParent != n)
    return False;
  if (n->ivRight != 0 && n->ivRight->ivParent != n)
    return False;
  if (! isConsistent (n->ivLeft, depth, nrOfElm))
    return False;

  int d, nr;
  if (! isConsistent (n->ivRight, d, nr))
    return False;
  if (n->ivDeeperSubtree != d - depth)
    return False;

  nrOfElm += nr + 1;

  if (d > depth)
    depth = d;
  depth++;

  return True;
}
