/*******************************************************************************
* FILE NAME: iobslist.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iobslist.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_DOSPROCESS

#include <iwindefs.h>
#include <iobslist.hpp>
#include <iobservr.hpp>
#include <inotifev.hpp>
#include <iobsptr.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IOBSLIST_CPP_
  #include <ipagetun.h>
#endif

#include <iseq.h>

// Define the Collection of observer data
// Note: We implement a collection of IObserverPtr objects
//       since we must store the observer unique data as
//       well as the observer pointer.
//       Most of the functions in this class are a simple
//       passthru to their ISequence counterpart.
class IObserverDataList : public ISequence<IObserverPtr>
{
public:
  IObserverDataList  ( );
 ~IObserverDataList  ( );
};

/*------------------------------------------------------------------------------
| IObserverDataList::IObserverDataList                                         |
------------------------------------------------------------------------------*/
IObserverDataList::IObserverDataList( )
  : ISequence<IObserverPtr>(50)
{ }

/*------------------------------------------------------------------------------
| IObserverDataList::~IObserverDataList                                        |
------------------------------------------------------------------------------*/
IObserverDataList::~IObserverDataList( )
{ }

/*------------------------------------------------------------------------------
| static Data.                                                                 |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IString IObserverList::obsLockKeyPrefix =
  "CPPOOB3\\OBSLOCK\\";
#pragma data_seg()

/*------------------------------------------------------------------------------
| IObserverList::observerKey                                                   |
------------------------------------------------------------------------------*/
const char* IObserverList::observerKey()
{
  return fObserverKey;
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::Cursor                                                |
------------------------------------------------------------------------------*/
IObserverList::Cursor::Cursor ( IObserverList& observerList)
                :    fCursor(new IObserverDataList::Cursor(
                          *(observerList.fObserverDataList)))
{ }

/*------------------------------------------------------------------------------
| IObserverList::Cursor::~Cursor                                               |
------------------------------------------------------------------------------*/
IObserverList::Cursor::~Cursor ( )
{
  delete ((IObserverDataList::Cursor*)fCursor);
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::setToFirst                                            |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::Cursor::setToFirst ( )
{
  return ((IObserverDataList::Cursor*)fCursor)->setToFirst();
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::setToNext                                             |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::Cursor::setToNext ( )
{
  return ((IObserverDataList::Cursor*)fCursor)->setToNext();
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::setToLast                                             |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::Cursor::setToLast ( )
{
  return ((IObserverDataList::Cursor*)fCursor)->setToLast();
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::setToPrevious                                         |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::Cursor::setToPrevious ( )
{
  return ((IObserverDataList::Cursor*)fCursor)->setToPrevious();
}

/*------------------------------------------------------------------------------
| IObserverList::Cursor::isValid                                               |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::Cursor::isValid    ( ) const
{
  return ((IObserverDataList::Cursor*)fCursor)->isValid();
}


/*------------------------------------------------------------------------------
| IObserverList::Cursor::invalidate                                            |
------------------------------------------------------------------------------*/
void IObserverList::Cursor::invalidate  ( )
{
  ((IObserverDataList::Cursor*)fCursor)->invalidate();
}


/*------------------------------------------------------------------------------
| IObserverList::IObserverList                                                 |
------------------------------------------------------------------------------*/
IObserverList::IObserverList ()
          : fObserverDataList(new IObserverDataList()),
            fNumObservers(0), fpObserverArray(0),
            fObserverKey( obsLockKeyPrefix + IString( (unsigned long)this ).d2x() +
                          "\\" + IString(IObserverList::myPID()))
{}

/*------------------------------------------------------------------------------
| IObserverList::myPID                                                         |
------------------------------------------------------------------------------*/
unsigned long IObserverList::myPID()
{
  PTIB    ptib;  //Thread Information Block
  PPIB    ppib;  //Process Information Block

  DosGetInfoBlocks(&ptib, &ppib);
  return ppib->pib_ulpid;
}

/*------------------------------------------------------------------------------
| IObserverList::~IObserverList                                                |
------------------------------------------------------------------------------*/
IObserverList::~IObserverList ()
{
  delete fObserverDataList;
}

/*------------------------------------------------------------------------------
| IObserverList::notifyObservers                                               |
|                                                                              |
| This code is convoluted because of the fact that when we call                |
| dispatchNotificationEvent, the application has control and the ability to    |
| add or delete items from the collection we are using.  Therefore, when we    |
| come back from the dispatchNotificationEvent call, our cursor which we are   |
| using to run the list of observers may be undefined.  However, we don't      |
| have a way of knowing it is undefined.                                       |
|                                                                              |
| The solution to this is to make a copy of the collection of observers.       |
| Then we iterate this copy and do our notifications from it.  In addition,    |
| the code that removes observers, will update this copy if it removes one     |
| of its members.  This will put the performance hit into the removeAt() and   |
| remove() functions, and not in this function which needs to be as fast       |
| as possible.                                                                 |
|                                                                              |
| NOTE:  Any observers added during a call to dispatchNotificationEvent will   |
|        not be notified.  Only observers that existed at the time the         |
|        notification was generated are a candidate for notification.          |
------------------------------------------------------------------------------*/
IObserverList& IObserverList::notifyObservers(const INotificationEvent& anEvent)
{
  /********************************************************************/
  /* Only notify observers when there are observers to notify         */
  /********************************************************************/
  if ( this->numberOfElements() > 0 )
  {
    /********************************************************************/
    /* Grab the semaphore first to be sure no other thread is accessing */
    /* the collection of observers. Use ISharedResource with a name     */
    /* unique to this IObserverList instance so we only lock out truely */
    /* dangerous access (i.e. altering THIS observer array or list).    */
    /********************************************************************/
    ISharedResource obsLock( observerKey() );
    obsLock.lock();

    IObserverPtr *fpObserverArray1 = 0;
    if (fpObserverArray)
    {
      fpObserverArray1 = fpObserverArray;
    }

    /**********************************************************/
    /* Allocate an array to hold the collection of observers. */
    /**********************************************************/
    fNumObservers = fObserverDataList->numberOfElements();
    fpObserverArray = new IObserverPtr[fNumObservers];

    /********************************************************************/
    /* Loop through the list of observers, and add each one to our      */
    /* temporary array.  We will use the temporary list to notify from. */
    /********************************************************************/
    unsigned long i = 0;
    IObserverDataList::Cursor cursor(*fObserverDataList);
    for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
    {
      fpObserverArray[i] = fObserverDataList->elementAt(cursor);
      i++;
    }

    /*******************************************************************/
    /* Adjust the number of observers in the very rare case that the   */
    /* cursor got messed up and could not complete its walk of all the */
    /* observers.                                                      */
    /*******************************************************************/
    fNumObservers = i;

    for (unsigned long j = 0; j < fNumObservers; j++)
    {
      /************************************************************************/
      /* Check to see if our temporary array is still pointing to an observer */
      /* object.  If it isn't, it means that it was removed during a call to  */
      /* dispatchNotificationEvent().  See remove() and removeAt() which will */
      /* update our temporary array and zero-out the observer pointer when    */
      /* it is removed.  If the client calls removeAll() during processing    */
      /* of dispatchNotificationEvent, the removeAll() function will assign   */
      /* fNumObservers to 0 which will stop this loop dead in it's tracks.    */
      /************************************************************************/
      if (fpObserverArray[j].obsrvr)
      {
        IObserverPtr obsptr = fpObserverArray[j];
        ((INotificationEvent&)anEvent).setObserverData(obsptr.observerData);
        obsptr.obsrvr->dispatchNotificationEvent(anEvent);

        /******************************************************************/
        /* On the call to dispatchNotificationEvent, observers could have */
        /* been added and notifyObservers called after they were added.   */
        /* If this happened, fNumObservers will be wrong.  We need to     */
        /* first check to see if it is not 0.  If not, reassign it to be  */
        /* equal to our local variable of the number of observers.  This  */
        /* protects the integrity of our loop.  If it is 0, it means that */
        /* all the observers have been removed in which case we want this */
        /* loop to stop.  The case were observers get removed is already  */
        /* handled.                                                       */
        /******************************************************************/
        if (fNumObservers)
        {
          fNumObservers = i;
        }
      }
    }

    delete [] fpObserverArray;
    fpObserverArray = fpObserverArray1;
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IObserverList::add                                                           |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::add(IObserver& anObserver,
                                  void* observerData)
{
  /********************************************************************/
  /* Grab the semaphore first to be sure no other thread is accessing */
  /* the collection of observers.                                     */
  /********************************************************************/
  ISharedResource obsLock( observerKey() );
  obsLock.lock();
  return fObserverDataList->add(IObserverPtr(&anObserver, observerData));
}

/*------------------------------------------------------------------------------
| IObserverList::elementAt                                                     |
------------------------------------------------------------------------------*/
IObserver& IObserverList::elementAt(const Cursor& aCursor) const
{
  IObserverPtr& obsptr = fObserverDataList->elementAt
                       (*(IObserverDataList::Cursor*)(aCursor.fCursor));
  return obsptr.observer();
}

/*------------------------------------------------------------------------------
| IObserverList::removeAt                                                      |
------------------------------------------------------------------------------*/
IObserverList& IObserverList::removeAt(const Cursor& cursor)
{
  /********************************************************************/
  /* Grab the semaphore first to be sure no other thread is accessing */
  /* the collection of observers.                                     */
  /********************************************************************/
  ISharedResource obsLock( observerKey() );
  obsLock.lock();
  fObserverDataList->removeAt(*((IObserverDataList::Cursor*)cursor.fCursor));

  /********************************************************************/
  /* Remove this observer from our tempory array of observers that we */
  /* may be currently iterating.                                      */
  /********************************************************************/
  removeFromObsArray(*((fObserverDataList->elementAt(
             *((IObserverDataList::Cursor*)cursor.fCursor))).obsrvr));
  return *this;
}

/*------------------------------------------------------------------------------
| IObserverList::remove                                                        |
------------------------------------------------------------------------------*/
IObserverList& IObserverList::remove(const IObserver& observer)
{
  /********************************************************************/
  /* Grab the semaphore first to be sure no other thread is accessing */
  /* the collection of observers.                                     */
  /********************************************************************/
  ISharedResource obsLock( observerKey() );
  obsLock.lock();
  IObserverDataList::Cursor cursor(*fObserverDataList);
  for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
  {
    IObserverPtr obsptr = fObserverDataList->elementAt(cursor);
    if (&(obsptr.observer()) == &observer)
    {
      /********************************************************************/
      /* Remove this observer from our tempory array of observers that we */
      /* may be currently iterating.                                      */
      /********************************************************************/
      removeFromObsArray(observer);
      fObserverDataList->removeAt(cursor);
      break;
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IObserverList::removeAll                                                     |
------------------------------------------------------------------------------*/
IObserverList& IObserverList::removeAll( )
{
  /********************************************************************/
  /* Grab the semaphore first to be sure no other thread is accessing */
  /* the collection of observers.                                     */
  /********************************************************************/
  ISharedResource obsLock( observerKey() );
  obsLock.lock();
  fObserverDataList->removeAll();

  /********************************************************************/
  /* Set the number of observers to 0 since we just removed them all. */
  /* This will stop the dispatchNotificationEvent loop in the         */
  /* notifyObservers() function in case we are in the process of      */
  /* notifying.                                                       */
  /********************************************************************/
  fNumObservers = 0;
  return *this;
}

/*------------------------------------------------------------------------------
| IObserverList::numberOfElements                                              |
------------------------------------------------------------------------------*/
unsigned long IObserverList::numberOfElements( ) const
{
  return fObserverDataList->numberOfElements();
}

/*------------------------------------------------------------------------------
| IObserverList::isEmpty                                                       |
------------------------------------------------------------------------------*/
IBase::Boolean IObserverList::isEmpty( ) const
{
  return fObserverDataList->isEmpty();
}

/*------------------------------------------------------------------------------
| IObserverList::removeFromObsArray                                            |
------------------------------------------------------------------------------*/
void IObserverList::removeFromObsArray(const IObserver& observer)
{
  if (fpObserverArray)
  {
    for (unsigned long i = 0; i < fNumObservers; i++)
    {
      if (fpObserverArray[i].obsrvr == &observer)
      {
        fpObserverArray[i].obsrvr = 0;
        break;
      }
    }
  }
}

/*------------------------------------------------------------------------------
| IObserverList::removeFromObsArray                                            |
------------------------------------------------------------------------------*/
void IObserverList::removeFromObsArray(IObserver& observer)
{
  if (fpObserverArray)
  {
    for (unsigned long i = 0; i < fNumObservers; i++)
    {
      if (fpObserverArray[i].obsrvr == &observer)
      {
        fpObserverArray[i].obsrvr = 0;
        break;
      }
    }
  }
}

