/*******************************************************************************
* FILE NAME: ireslock.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IResource                                                                *
*     IPrivateResource                                                         *
*     ISharedResource                                                          *
*     IResourceLock                                                            *
*                                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#ifdef IC_REF_COUNT_SEMS
  #pragma options("/Oi+")   // Ensure optimization is on
#endif
extern "C" {
  #define INCL_DOSSEMAPHORES
  #define INCL_DOSERRORS
  #define INCL_WINMESSAGEMGR
  #include <iwindefs.h>
}

#include "ireslock.hpp"
#include <istring.hpp>
#include <iexcept.hpp>
#include <itrace.hpp>
#include <stdio.h>
#include <icconst.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IRESLOCK_CPP_
  #include <ipagetun.h>
#endif


/***************************************************************/
/* FUNCTION:  IPrivateResource :: handle                       */
/*                                                             */
/* PURPOSE:   returns handle to the resource                   */
/*            Cannot be inlined due to being virtual           */
/***************************************************************/
ISemaphoreHandle& IPrivateResource :: handle()
{ return resourceHandleCl;   }


/***************************************************************/
/* FUNCTION:  ISharedResource :: handle                        */
/*                                                             */
/* PURPOSE:   returns handle to the resource                   */
/*            Cannot be inlined due to being virtual           */
/***************************************************************/
ISemaphoreHandle& ISharedResource :: handle()
{ return resourceHandleCl;   }


/***************************************************************/
/* FUNCTION:  IResource :: IResource                           */
/*                                                             */
/* PURPOSE:   Constructor for base resource object             */
/***************************************************************/
IResource :: IResource()
{
  lClReferenceCount=0;
  reslockGate=0;
}


/***************************************************************/
/* FUNCTION:  IResource :: ~IResource                          */
/*                                                             */
/* PURPOSE:   Destructor for base resource object              */
/***************************************************************/
IResource :: ~IResource()
{
  IMODTRACE_ALL("IRes::~IRes");
  if (reslockGate)
    delete reslockGate;
  reslockGate=0;
  ITRACE_ALL(IString("Use Count is: ")+IString(lClReferenceCount));
  lClReferenceCount = 0;
}


/***************************************************************/
/* FUNCTION:  IPrivateResource :: IPrivateResource             */
/*                                                             */
/* PURPOSE:   Constructor the key for a private resource.      */
/***************************************************************/
IPrivateResource :: IPrivateResource()
                        : resourceHandleCl(0)
{
  IMODTRACE_ALL("IPrivRes::IPrivRes");
  unsigned long ulRC = 0;

  ulRC = DosCreateMutexSem(0, (PHMTX)&resourceHandleCl, 0, false);
  if (ulRC!=0)
  {
     ITHROWSYSTEMERROR( ulRC,
                        "DosCreateMutexSem",
                        IErrorInfo::accessError,
                        IException::recoverable);
  }
}


/***************************************************************/
/* FUNCTION:  IPrivateResource :: ~IPrivateResource            */
/*                                                             */
/* PURPOSE:   Destructor.                                      */
/***************************************************************/
IPrivateResource :: ~IPrivateResource()
{
  IMODTRACE_ALL("IPrivRes::~IPrivRes");
  unsigned long ulRC=0;

  // Copied from IResource::~IResource
  if (reslockGate) delete reslockGate;
  reslockGate=0;

  if (resourceHandleCl!=0)
     ulRC = DosCloseMutexSem((HMTX)resourceHandleCl);
  if (ulRC!=0)
  {
     ITHROWSYSTEMERROR( ulRC,
                        "DosCloseMutexSem",
                        IErrorInfo::accessError,
                        IException::recoverable);
  }
  ITRACE_ALL(IString("Handle is: ")+IString(handle().asUnsigned()).d2x());
  resourceHandleCl=0;
}


/***************************************************************/
/* FUNCTION:  ISharedResource :: ISharedResource               */
/*                                                             */
/* PURPOSE:   Construct the key for a named semaphore.         */
/***************************************************************/
ISharedResource :: ISharedResource(const char* pszKeyName)
                     : keyNameCl(""), resourceHandleCl(0)
{
  IMODTRACE_ALL("ISharedRes::ISharedRes");
  unsigned long ulRC=0;

  keyNameCl += "\\SEM32\\";
  keyNameCl += pszKeyName;
  ulRC = DosCreateMutexSem((PSZ)(char*)keyNameCl,
                      (PHMTX)&resourceHandleCl, 0, false);
  if ( ulRC == ERROR_DUPLICATE_NAME )
  {
    ulRC = DosOpenMutexSem((PSZ)(char*)keyNameCl,
                           (PHMTX)&resourceHandleCl);
    if (ulRC!=0)
    {
       ITHROWSYSTEMERROR( ulRC,
                          "DosOpenMutexSem",
                          IErrorInfo::accessError,
                          IException::recoverable);
    }
  }
  else if (ulRC!=0)
  {
     ITHROWSYSTEMERROR( ulRC,
                        "DosCreateMutexSem",
                        IErrorInfo::accessError,
                        IException::recoverable);
  }
  ITRACE_ALL(IString("Handle is:")+IString(handle().asUnsigned()).d2x());
}


/***************************************************************/
/* FUNCTION:  ISharedResource :: ~ISharedResource              */
/*                                                             */
/* PURPOSE:   Destructor.                                      */
/***************************************************************/
ISharedResource :: ~ISharedResource()
{
  IMODTRACE_ALL("ISharedRes::~ISharedRes");
  unsigned long ulRC=0;

  // Copied from IResource::~IResource
  if (reslockGate)
    delete reslockGate;
  reslockGate=0;

  if (resourceHandleCl!=0)
  {
     ulRC = DosCloseMutexSem((HMTX)resourceHandleCl);
     ITRACE_ALL(IString("Handle is:")+IString(resourceHandleCl.asUnsigned()).d2x());
  }

  if (ulRC!=0)
  {
     ITHROWSYSTEMERROR( ulRC,
                        "DosCloseMutexSem",
                        IErrorInfo::accessError,
                        IException::recoverable);
  }
  ITRACE_ALL(IString("Handle is: ")+IString(handle().asUnsigned()).d2x());
  resourceHandleCl=0;
}


/***************************************************************/
/* FUNCTION: IResource :: lock                                 */
/*                                                             */
/* PURPOSE: provide ability to explicitly acquire a resouce    */
/*          lock intead of requiring x = new IResourceLock(yy) */
/***************************************************************/
IResource& IResource :: lock(long lTimeOut)
{
  IMODTRACE_ALL("IResource::lock");
  if (reslockGate == 0) {
    reslockGate = new IResourceLock(*this,lTimeOut);
  }
  else {
    reslockGate->setLock(lTimeOut);
  }
  ITRACE_ALL(IString("Use Count is: ")+IString(lClReferenceCount));
  return *this;
}


/***************************************************************/
/* FUNCTION: IResource :: unlock                               */
/*                                                             */
/* PURPOSE: provide ability to explicitly release a resouce    */
/*          lock intead of requiring delete x                  */
/***************************************************************/
IResource& IResource :: unlock()
{
  IMODTRACE_ALL("IResource::unlock");
  if (reslockGate != 0)
    reslockGate->clearLock();
  ITRACE_ALL(IString("Use Count is: ")+IString(lClReferenceCount));
  return *this;
}


/***************************************************************/
/* FUNCTION: IResourceLock :: setLock                          */
/*                                                             */
/* PURPOSE: sets (requests) the semaphore                      */
/***************************************************************/
IResourceLock& IResourceLock::setLock(long lTimeOut)
{
  IMODTRACE_ALL("IResourceLock::setLock");
  unsigned long ulRC = 0;
  if (presCl->handle())
  {
    ITRACE_ALL( IString("Acquiring Handle: ")+
                IString(presCl->handle().asUnsigned()).d2x());
    ulRC = WinRequestMutexSem((HMTX)presCl->handle(), lTimeOut);
  }
  if (ulRC==ERROR_TIMEOUT)
    ITHROWLIBRARYERROR( IC_SEMAPHORES_EXHAUSTED,
                        IErrorInfo::outOfSystemResource,
                        IException::recoverable);
  else if (ulRC!=0)
    ITHROWSYSTEMERROR( ulRC,
                       "WinRequestMutexSem",
                       IErrorInfo::accessError,
                       IException::recoverable);
  // Only bump ref count if semaphore request was successful
  ++presCl->lClReferenceCount;
  ITRACE_ALL( IString("Use Count is: ")+
              IString(presCl->lClReferenceCount));
  return *this;

}


/***************************************************************/
/* FUNCTION:  IResourceLock :: IResourceLock                   */
/*                                                             */
/* PURPOSE:   Construct an instance of a resource lock.        */
/***************************************************************/
IResourceLock :: IResourceLock(IResource& res, long lTimeOut)
{
  this->presCl = &res;
  setLock(lTimeOut);
}


/***************************************************************/
/* FUNCTION:  IResourceLock :: ~IResourceLock                  */
/*                                                             */
/* PURPOSE:   Free the lock on the resource.                   */
/***************************************************************/
IResourceLock :: ~IResourceLock()
{
  if ( this->presCl && this->presCl->lClReferenceCount )
    clearLock();
}


/***************************************************************/
/* FUNCTION: IResourceLock :: clearLock                        */
/*                                                             */
/* PURPOSE: clears the associated semaphore                    */
/***************************************************************/
IResourceLock& IResourceLock::clearLock()
{
  IMODTRACE_ALL("IResourceLock::clearLock");
  unsigned long ulRC = 0;

  if (presCl->handle())
  {
    ITRACE_ALL( IString("Releasing Handle:")+
                IString(presCl->handle().asUnsigned()).d2x());
    ulRC = DosReleaseMutexSem((HMTX)presCl->handle());
  }
  if (ulRC!=0)
    ITHROWSYSTEMERROR( ulRC,
                       "DosReleaseMutexSem",
                       IErrorInfo::accessError,
                       IException::recoverable);
  // Only drop ref count if semaphore release was successful
  presCl->lClReferenceCount--;
  ITRACE_ALL( IString("Use Count is: ")+
              IString(presCl->lClReferenceCount));
  return *this;
}

