/*******************************************************************************
* FILE NAME: istdntfy.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in istdntfy.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <istdntfy.hpp>
#include <iobslist.hpp>
#include <iobservr.hpp>
#include <inotifev.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTDNTFY_CPP_
  #include <ipagetun.h>
#endif

#pragma data_seg(ICLStaticConst)
INotificationId const IStandardNotifier::deleteId = "IStandardNotifier::delete";
#pragma data_seg()

/*------------------------------------------------------------------------------
| INotifier::INotifier                                                         |
------------------------------------------------------------------------------*/
INotifier::INotifier( )
{ }

/*------------------------------------------------------------------------------
| INotifier::~INotifier                                                        |
------------------------------------------------------------------------------*/
INotifier::~INotifier( )
{ }

/*------------------------------------------------------------------------------
| IStandardNotifier::IStandardNotifier                                         |
|                                                                              |
| Default IStandardNotifier constructor. By default, IStandardNotifier objects |
| are created in a disabled state and therefore require a call to              |
| enableNotification before they will pass notifications on to observer        |
| objects.                                                                     |
------------------------------------------------------------------------------*/
IStandardNotifier::IStandardNotifier ( )
    : observers(0),
      enabled(false)
{ }

/*------------------------------------------------------------------------------
| IStandardNotifier::IStandardNotifier                                         |
|                                                                              |
| IStandardNotifier copy constructor.                                          |
| NOTE:  The ObserverList will be initialized to 0 and will not be copied.     |
------------------------------------------------------------------------------*/
IStandardNotifier::IStandardNotifier (const IStandardNotifier& copy)
{
  this->enabled = copy.enabled;
  this->observers = 0;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::operator=                                                 |
|                                                                              |
| IStandardNotifier assignment operator.                                       |
| NOTE:  The ObserverList will be initialized to 0 and will not be copied.     |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::operator= (const IStandardNotifier& stdN)
{
  /* Make sure we are not assigning to ourselves. */
  if (this != &stdN)
  {
    this->enabled = stdN.enabled;

    if (this->observers)
    {
      delete(this->observers);
      this->observers = 0;
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::~IStandardNotifier                                        |
|                                                                              |
| IStandardNotifier destructor.  Deletes the observer list.                    |
------------------------------------------------------------------------------*/
IStandardNotifier::~IStandardNotifier ( )
{
  notifyObservers(INotificationEvent(IStandardNotifier::deleteId, *this));
  delete this->observers;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::enableNotification                                        |
|                                                                              |
| Flags this notifier as being ready to accept notifications.  A Notifier,     |
| like IStandardNotifier, does not send notifications to observer objects      |
| until it is enabled.                                                         |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::enableNotification( Boolean enabled )
{
  this->enabled = enabled;
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::disableNotification                                       |
|                                                                              |
| Flags this notifier as being not ready to accept notifications.  A Notifier, |
| like IStandardNotifier, does not send notifications to observer objects      |
| unless it is  enabled                                                        |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::disableNotification( )
{
  this->enabled = false;
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::isEnabledForNotification                                  |
|                                                                              |
| Returns true if this notifier is enabled for notification.                   |
------------------------------------------------------------------------------*/
IBase::Boolean IStandardNotifier::isEnabledForNotification( ) const
{
  return this->enabled;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::addObserver                                               |
|                                                                              |
| This is the implementation of the INotifier::addObserver protocol.  We just  |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::addObserver(IObserver& observer,
                                                  const IEventData& userData)
{
  observerList().add(observer, userData);
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::removeObserver                                            |
|                                                                              |
| This is the implementation of the INotifier::removeObserver protocol. We     |
| just pass requests on to the observer list.                                  |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::removeObserver( IObserver& observer )
{
  observerList().remove(observer);
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::removeObserver                                            |
|                                                                              |
| This is the implementation of the INotifier::removeAllObservers protocol. We |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::removeAllObservers ( )
{
  observerList().removeAll();
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::notifyObservers                                           |
|                                                                              |
| This is the implementation of the INotifier::notifyObservers protocol. We    |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::notifyObservers(
                                         const INotificationEvent& anEvent)
{
  if(isEnabledForNotification())
    observerList().notifyObservers(anEvent);
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::notifyObservers                                           |
|                                                                              |
| This is the implementation of the INotifier::notifyObservers protocol. We    |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IStandardNotifier& IStandardNotifier::notifyObservers(
                                         const INotificationId& nId)
{
  if (isEnabledForNotification())
  {
    INotificationEvent anEvent(nId, *this);
    observerList().notifyObservers(anEvent);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IStandardNotifier::observers                                                 |
|                                                                              |
| Returns a reference to the observer list. The creation of the observer       |
| list is delayed until this function is called for the first time -- all      |
| access to the observer list MUST be obtained through this function.          |
------------------------------------------------------------------------------*/
IObserverList& IStandardNotifier::observerList( ) const
{
  if(this->observers==0)
    ((IStandardNotifier*)this)->observers = new IObserverList();
  return *(this->observers);
}
