/*******************************************************************************
* FILE NAME: istring0.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the constructors for the          *
*   class IString (and the implementation of related functions).               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <stdio.h>
  #include <float.h>
  #include <ctype.h>
  }

#include <istream.hpp>

#include <istring.hpp>

#if !defined( _IBUFFER_ )
  #include <ibuffer.hpp>
#endif

#ifndef _IEXCBASE_
  #include <iexcbase.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING0_CPP_
  #include <ipagetun.h>
#endif

#define QUOTE(x) #x
#define STRING(x) QUOTE(x)

/*------------------------------------------------------------------------------
|Member Name: IString :: null                                                  |
|             IString :: zero                                                  |
|             IString :: maxLong                                               |
|             IString :: hexDigits                                             |
|                                                                              |
| These are static IString and char* objects used by the ISring                |
| implementation:                                                              |
|                                                                              |
|   null      - Null (zero length) IString                                     |
|   zero      - IString with value == "0"                                      |
|   maxLong   - IString with ascii representation of LONG_MAX                  |
|   hexDigits - pointer to array of hex digits, in order                       |
------------------------------------------------------------------------------*/

#pragma data_seg(ICLNonConst)
const char
 *IString::null = "",
 *IString::zero = "0",
 *IString::maxLong = STRING(LONG_MAX);

char
 *IString::nullBuffer(0);

#pragma data_seg()

/*------------------------------------------------------------------------------
| IString::IString                                                             |
|                                                                              |
| Initialize the string with as accurrate a representation of the input        |
| argument as possible.                                                        |
|                                                                              |
| All constructors (except as noted below) *must* initialize the pBuffer data  |
| member to point to an appropriate null IBuffer object.  This will cause the  |
| resulting IString to be DBCS-enabled if need be.  This is accomplished via   |
| the static member function IString::defaultBuffer.  It will return a pointer |
| to a null Buffer of the appropriate type.                                    |
|                                                                              |
| The IString is actually "constructed" (that is, storage buffer allocated and |
| initialized) via application of the initBuffer function to the just          |
| initialized IBuffer.                                                         |
|                                                                              |
| The copy constructor simply sets the pBuffer data member to point to the     |
| buffer of the argument IString and bumps that buffer's use count.            |
|                                                                              |
| The special protected constructor that accepts an IBuffer pointer is another |
| special case.  This constructor is used to pass control of an IBuffer to     |
| another IString.  It is typically used within non-const IString member       |
| functions to delay disposal of an IString's IBuffer (in cases where that     |
| IBuffer may be needed during construction of the new IString contents).      |
|                                                                              |
| Notes:                                                                       |
|   1. The use count for the null buffer returned by                           |
|      IString::defaultBuffer is *not* incremented.  This is because           |
|      that buffer (in almost all instances) will be "discarded" (via          |
|      the call to IString::initBuffer).  The exception is the default         |
|      constructor, which will retain its reference to the null buffer         |
|      and therefore must increment the buffer's use count.                    |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| operator<<                                                                   |
|                                                                              |
| Display the IString contents on the output stream.                           |
------------------------------------------------------------------------------*/
ostream& operator << ( ostream& aStream, const IString& aString )
  {
  return aStream << aString.data();
  }


/*------------------------------------------------------------------------------
| operator >>                                                                  |
|                                                                              |
| Read characters one at a time until a whitespace is reached.                 |
------------------------------------------------------------------------------*/
istream &operator >> ( istream &aStream, IString &aString )
  {
  if ( aString.length() == 0 )
    aString = IString( 0, 128 );

  unsigned
    bytes = 0;

  char
    c;

  for( aStream.get( c ); !isspace(c) && !aStream.fail(); aStream.get(c) )
    {
    if ( bytes == aString.length() )
      aString += IString( 0, bytes );
    aString[ ++bytes ] = c;
    }

  // Remove trailing stuff...
  aString.remove( bytes+1 );

  return aStream;
  }

/*------------------------------------------------------------------------------
| IString::lineFrom                                                            |
|                                                                              |
| Read characters up to a delimiter.                                           |
------------------------------------------------------------------------------*/
IString IString :: lineFrom ( istream &aStream, char delim )
  {
  IString
    result( 0, 128 );

  unsigned
    bytes = 0;

  char
    c;

  for( aStream.get( c ); c != delim && !aStream.fail(); aStream.get(c) )
    {
    if ( bytes == result.length() )
      result += IString( 0, bytes );
    result[ ++bytes ] = c;
    }

  // Remove trailing stuff...
  result.remove( bytes+1 );

  return result;
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
|                                                                              |
| The "base" version of this function simply invokes the corresponding         |
| IBuffer::newBuffer function against the pBuffer member.                      |
|                                                                              |
| The long/unsigned long/double functions convert their arguments to strings   |
| and invoke the "base" version.                                               |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( const void *p1,
                                 unsigned    len1,
                                 const void *p2,
                                 unsigned    len2,
                                 const void *p3,
                                 unsigned    len3,
                                 char        padChar )
  {
  return this->setBuffer( this->buffer() -> newBuffer( p1, len1,
                                                     p2, len2,
                                                     p3, len3,
                                                     padChar ) );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( long n )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%ld", n );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( unsigned long n )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%lu", n );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( double d )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%.*g", DBL_DIG, d );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::defaultBuffer                                                       |
|                                                                              |
| If DBCS environment is active, return &IDBCSBuffer::nullBuffer.              |
| Otherwise, return &IBuffer::nullBuffer.                                      |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::data                                                                |
|                                                                              |
| Simply return a pointer to the start of the associated IBuffer's buffer.     |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::lengthOf                                                            |
|                                                                              |
| If the argument char* pointer is zero, return 0;  Otherwise, return the      |
| length as calculated by strlen.                                              |
------------------------------------------------------------------------------*/
// See istring.inl


/*------------------------------------------------------------------------------
| IString::prepareToChange                                                     |
|                                                                              |
| If the reference count for the referenced IBuffer is greater than one        |
| (i.e., if another IString is pointing to it), then make a copy.              |
------------------------------------------------------------------------------*/
IString &IString :: prepareToChange ( )
  {
  if ( this->buffer()->useCount() > 1 )
    {
    IString old( this->buffer() );
    this->initBuffer( this->data(), this->length(),
                      0, 0,
                      0, 0,
                      0 );
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::rightJustify                                                        |
|                                                                              |
| If the new length differs from the current length, delegate the call to the  |
| IBuffer contents.  The receiver's contents will be replaced with the new     |
| buffer returned.                                                             |
|                                                                              |
| Notes:                                                                       |
|   1. This function is implemented here rather than in istring7.cpp (with     |
|      the other "editing" functions) because it is invoked from istring9.cpp  |
|      and we don't want to cause all the editing code to be linked in if      |
|      only conversion functions are required.                                 |
------------------------------------------------------------------------------*/
IString & IString :: rightJustify ( unsigned newLength,
                                    char     padCharacter )
  {
  if ( this->length() != newLength )
    {
    IString old( this->buffer() );
    this->setBuffer( this->buffer() -> rightJustify( newLength,
                                                     padCharacter ));
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::remove                                                              |
|                                                                              |
| If the number of characters to be removed is non-zero, then delegate this    |
| function via the IBuffer contents.  The receiver's contents are replaced     |
| with the IBuffer returned from that call.                                    |
|                                                                              |
| Notes:                                                                       |
|   1. This function is implemented here rather than in istring7.cpp (with     |
|      the other "editing" functions) because it is invoked from istring9.cpp  |
|      and we don't want to cause all the editing code to be linked in if      |
|      only conversion functions are required.                                 |
------------------------------------------------------------------------------*/
IString & IString :: remove ( unsigned startPos )
  {
  return this->remove( startPos, UINT_MAX );
  }

/*------------------------------------------------------------------------------
| IString::remove                                                              |
------------------------------------------------------------------------------*/
IString & IString :: remove ( unsigned startPos,
                              unsigned numChars )
  {
  if ( numChars > 0
       &&
       startPos <= this->length()
       &&
       this->length() > 0 )
    {
    IString old( this->buffer() );
    this->setBuffer( this->buffer() -> remove( startPos,
                                               numChars ));
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::remove                                                              |
------------------------------------------------------------------------------*/
IString IString :: remove ( const IString &aString,
                            unsigned       startPos,
                            unsigned       numChars )
  {
  return IString(aString).remove( startPos, numChars );
  }

/*------------------------------------------------------------------------------
| IString :: setBuffer                                                         |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::asDebugInfo                                                         |
------------------------------------------------------------------------------*/
IString IString :: asDebugInfo ( ) const
  {
  IString result( "IString(@" );
  result += IString( (unsigned long) this );
  result += ",pBuffer->";
  result += this->buffer()->asDebugInfo();
  result += ")";
  return result;
  }

/*------------------------------------------------------------------------------
| IString::operator =                                                          |
------------------------------------------------------------------------------*/
IString &IString :: operator = ( const IString &aString )
  {
  aString.buffer()->addRef();
  this->buffer()->removeRef();
  this->setBuffer( aString.buffer() );
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer,
                     unsigned    len,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer,
                    len,
                    0, 0, 0, 0,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer1,
                     unsigned    len1,
                     const void *pBuffer2,
                     unsigned    len2,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer1,
                    len1,
                    pBuffer2,
                    len2,
                    0, 0,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer1,
                     unsigned    len1,
                     const void *pBuffer2,
                     unsigned    len2,
                     const void *pBuffer3,
                     unsigned    len3,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer1,
                    len1,
                    pBuffer2,
                    len2,
                    pBuffer3,
                    len3,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( int n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (long) n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( long n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( short n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (long) n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (unsigned long) n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned long n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned short n )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (unsigned long) n );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( double aDouble )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( aDouble );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( signed char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const char *p )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) p,
                    lengthOf( p ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const unsigned char *p )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) p,
                    lengthOf( (const char *) p ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const signed char *p )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) p,
                    lengthOf( (const char *) p ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const IString &aString )
  {
  aString.buffer() -> addRef();
  this->setBuffer( aString.buffer() );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( )
  : pBuffer( IString::defaultBuffer() )
  {
  this->buffer() -> addRef();
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( IBuffer * oldBuffer )
  {
  this->setBuffer( oldBuffer );
  }

/*------------------------------------------------------------------------------
| IString::~IString                                                            |
------------------------------------------------------------------------------*/
IString :: ~IString ( )
  {
  this->buffer()->removeRef();
  this->pBuffer = 0;
  }
