/*******************************************************************************
* FILE NAME: istring5.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the "accessing" functions for     *
*   the class IString.                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <istring.hpp>

#ifndef _ISTREAM_
  #include <istream.hpp>
#endif

#ifndef _IBUFFER_
  #include <ibuffer.hpp>
#endif

#ifndef _IEXCBASE_
  #include <iexcbase.hpp>
#endif

#ifndef _ICCONSTA_
  #include <icconsta.h>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING5_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IString::length                                                              |
|                                                                              |
| Simply return the length field from the associated IBuffer.                  |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::size                                                                |
|                                                                              |
| This function is simply a synonym for IString::length.                       |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::subString                                                           |
|                                                                              |
| The result string is constructed using the IBuffer obtained via the          |
| function IBuffer::subString applied to the receiver's IBuffer.               |
|                                                                              |
| Notes:                                                                       |
|   1. The result of this function varies for DBCS vs. SBCS strings.           |
|      In the case of DBCS, substringing is accomplished as follows:           |
|        a. First, the specified bytes are (logically) extracted               |
|           from the string.                                                   |
|        b. If the first byte of this substring is the second byte             |
|           of a DBCS character, then the preceding byte is                    |
|           prepended to the result and the last byte is removed.              |
|        c. If the last byte is the substring is now the first byte            |
|           of a DBCS character, then it is converted to the                   |
|           padCharacter.                                                      |
|      This ensures that the result has length specified by the                |
|      input length argument.                                                  |
------------------------------------------------------------------------------*/
IString IString :: subString ( unsigned startPos ) const
  {
  if ( startPos <= this->length() )
    return this->subString( startPos, this->length() - startPos + 1 );
  else
    return IString::null;
  }

/*------------------------------------------------------------------------------
| IString::subString                                                           |
------------------------------------------------------------------------------*/
IString IString :: subString ( unsigned startPos,
                               unsigned len,
                               char     padCharacter ) const
  {
  IString result( this->buffer()->subString( startPos,
                                            len,
                                            padCharacter ) );
  return result;
  }

/*------------------------------------------------------------------------------
| IString::operator []                                                         |
|                                                                              |
| The const and non-const version of these functions differ in the way         |
| indexing off the end of the string is handled.  For the const case, an       |
| Invalid request exception is thrown. In the non-const case, the string is    |
| extended with blanks (and a reference to the last blank added is returned).  |
|                                                                              |
| Otherwise, return the specified element of the receiver's buffer.            |
|                                                                              |
| Notes:                                                                       |
|   1. The indexing is 1-based.  For convenience, index zero                   |
|      is treated like index 1 (this is consistent with the treatment          |
|      of index zero by other member functions; e.g., for searching).          |
|   2. This function presents some problems for the integrity of IString       |
|      contents (as does operator char*).  While some precautions are          |
|      taken (ensuring the receiver has the only reference to its              |
|      buffer) subsequent assignment of the IString and manipulation via       |
|      the char& returned by this function may cause unexpected result.        |
------------------------------------------------------------------------------*/
const char& IString :: operator [] ( unsigned index ) const
   {
   static char overFlow = 0;
   index -= ( index > 0 ); // Get 0-based index.
   if ( index >= this->length() )
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR), IC_ISTRING_INDEX_ERROR, IException::recoverable );
      ITHROW( exception );
      }
   return this->data()[ index ];
   }

/*------------------------------------------------------------------------------
| IString::operator []                                                         |
------------------------------------------------------------------------------*/
char& IString :: operator [] ( unsigned index )
   {
   IString old( this->buffer() );
   index -= ( index > 0 ); // Get 0-based index.
   // Test whether buffer has to grow:
   if ( index >= old.length() )
     { // It does,
     this->initBuffer( old.data(), old.length(),
                      0, index - old.length() + 1,
                      0, 0,
                      ' ' );
     }
   else
     { // It doesn't, but make sure we've got our own copy.
     if ( this->buffer()->useCount() == 1 )
       { // Keep it, instead.
       this->buffer() -> addRef();
       }
     else
       { // Need to make copy.
       this->initBuffer( old.data(), old.length(),
                         0, 0,
                         0, 0,
                         0 );
       }
     }
   return this->data()[ index ];
   }
const char& IString :: operator [] ( signed index ) const
   {
   if (index < 0)
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }

   index -= ( index > 0 ); // Get 0-based index.
   if ( index >= this->length() )
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }
   return this->data()[ index ];
   }

char& IString :: operator [] ( signed index )
   {
   if (index < 0)
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }

   IString old( this->buffer() );
   index -= ( index > 0 ); // Get 0-based index.
   // Test whether buffer has to grow:
   if ( index >= old.length() )
     { // It does,
     this->initBuffer( old.data(), old.length(),
                      0, index - old.length() + 1,
                      0, 0,
                      ' ' );
     }
   else
     { // It doesn't, but make sure we've got our own copy.
     if ( this->buffer()->useCount() == 1 )
       { // Keep it, instead.
       this->buffer() -> addRef();
       }
     else
       { // Need to make copy.
       this->initBuffer( old.data(), old.length(),
                         0, 0,
                         0, 0,
                         0 );
       }
     }
   return this->data()[ index ];
   }

