/*******************************************************************************
* FILE NAME: itrace.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in itrace.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_DOSSEMAPHORES
#define INCL_DOSINFOSEG
#define INCL_DOSPROCESS

#include <iwindefs.h>
#include <itrace.hpp>
#include <istring.hpp>
#include <istream.hpp>


extern "C"
{
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <imprintf.h>
}


#define IC_MAX_THREADS 255


#pragma data_seg(ICLNonConst)
int ITrace::iClState = ITrace::uninitialized;

ITrace::Destination ITrace::iClTraceLocation = ITrace::queue;
static unsigned ITrace__uClIndent[IC_MAX_THREADS] = {0};
static unsigned long ITrace__ulClProcessId =0;
static Boolean ITrace__fCheckMemory = false;
static Boolean ITrace__fCheckStack = false;
static const unsigned INDENT_LEVEL(2);
#pragma data_seg()

class ITraceSetup
{
  public:

/*------------------------------------------------------------------------------
| ITraceSetup::traceOn                                                         |
|                                                                              |
| Returns true if tracing is on                                                |
------------------------------------------------------------------------------*/
static Boolean traceOn()
     {
#ifdef PTR7989
       if (ITrace::iClState == ITrace::uninitialized)
         initialize();
#else
       if (ITrace::iClState & ITrace::uninitialized)
         initialize();
#endif
       return (ITrace::iClState & ITrace::on);
     }
/*------------------------------------------------------------------------------
| ITraceSetup::initialize                                                      |
------------------------------------------------------------------------------*/
static void initialize();
};

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ITRACE_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| ITrace::ITrace                                                               |
|                                                                              |
| Default constructor for tracing.                                             |
------------------------------------------------------------------------------*/
ITrace :: ITrace(const char* pszTraceName,
                 long lLineNo)
#ifndef PTR7989
        : pszClTraceName(0)
#endif
{
   if(ITraceSetup::traceOn())
   {
      if (pszTraceName!=0)
      {
        pszClTraceName = (char*)pszTraceName;
        IString str(pszTraceName);
        if(isWriteLineNumberEnabled() && lLineNo>0)
        {
           str+=IString("(");
           str+=IString(lLineNo);
           str+=IString(")");
        }
        writeFormattedString(str, "+");
      }
      else
         pszClTraceName = 0;
   }
}


/*------------------------------------------------------------------------------
| ITrace::~ITrace                                                              |
|                                                                              |
| Write exit trace and flush the buffer.                                       |
------------------------------------------------------------------------------*/
ITrace :: ~ITrace()
{
   if(ITraceSetup::traceOn())
   {
      if (pszClTraceName!=0)
      {
        writeFormattedString(IString(pszClTraceName),"-");
      }
   }
}


/*------------------------------------------------------------------------------
| ITrace::write                                                                |
|                                                                              |
| Write an IString using writeFormattedString after adding the appropriate     |
| padding for indentation.                                                     |
------------------------------------------------------------------------------*/
void ITrace :: write(const IString& strString)
{
   if(ITraceSetup::traceOn())
   {
       writeFormattedString(strString, ">");
   }
#ifdef __DEBUG_ALLOC__
   if(ITrace__fCheckMemory)
     _heap_check();
#endif

}

/*------------------------------------------------------------------------------
| ITrace::write                                                                |
|                                                                              |
| Write a char array after adding the appropriate padding for indentation.     |
------------------------------------------------------------------------------*/
void ITrace::write(const char* pszText)
{
   write(IString(pszText));
}

/*------------------------------------------------------------------------------
| ITrace::writeFormattedString                                                 |
|                                                                              |
| Write a formatted IString using writeString.                                 |
------------------------------------------------------------------------------*/
void  ITrace :: writeFormattedString(const IString& strString,
                                     char* pszMarker)
{
   static unsigned long ulSequence = 0;

   unsigned long ulThreadId = threadId();
   if(*pszMarker=='-' && ITrace__uClIndent[ulThreadId] >= INDENT_LEVEL)
      ITrace__uClIndent[ulThreadId]-=INDENT_LEVEL;


   IString strOut("");
   IString strPrefix("");
   strPrefix.insert(pszMarker, ITrace__uClIndent[ulThreadId], ' ');

   if(isWritePrefixEnabled())
   {
      IString strThreadId  = IString(ulThreadId).rightJustify(2, '0');
      IString strProcId  = IString(ITrace__ulClProcessId).rightJustify(6, '0');
      IString strLineNo  = IString(ulSequence).rightJustify(8, '0');
      if(ITrace__fCheckStack)
      {
        IString strStack =
                        IString(ITrace::remainingStack()).rightJustify(7, '0');
        strPrefix = IString(" ") + strStack + " " + strPrefix;
      }
      strPrefix = strLineNo + " " + strProcId + ":" + strThreadId + " " +
          strPrefix;
   }


   // copy the string and add the prefix info where a newline found
   strOut = IString::change(strString, '\n', "\n" + strPrefix);
   strOut = strPrefix+strOut+"\n";

   writeString((char*)strOut);

   ulSequence++;
   if(*pszMarker=='+')
      ITrace__uClIndent[ulThreadId]+=INDENT_LEVEL;

}

/*------------------------------------------------------------------------------
| ITrace::threadId                                                             |
|                                                                              |
| Determine the threadId of the current thread                                 |
------------------------------------------------------------------------------*/
unsigned long ITrace::threadId()
{
   unsigned long traceid;

   PTIB    ptib;      /*  Thread Information Block  */
   PPIB    ppib;      /*  Process Information Block */

   DosGetInfoBlocks(&ptib, &ppib);

   traceid = ptib->tib_ptib2->tib2_ultid;
   ITrace__ulClProcessId = ppib->pib_ulpid;
   return (traceid);
}

/*------------------------------------------------------------------------------
| ITrace::remainingStack                                                       |
|                                                                              |
| Determine the size of the remaining stack of the current thread.             |
------------------------------------------------------------------------------*/
unsigned long ITrace::remainingStack()
{

   PTIB    ptib;      /*  Thread Information Block  */
   PPIB    ppib;      /*  Process Information Block */
   DosGetInfoBlocks(&ptib, &ppib);
   void
     *base = ptib->tib_pstack,
     *cur  = &ptib;
   return (char*)cur - (char*)base;
}

/*------------------------------------------------------------------------------
| ITrace::writeString                                                          |
|                                                                              |
| Write a text string to file.                                                 |
|                                                                              |
| Notes:                                                                       |
|                                                                              |
|   1) We cannot use cout or cerr. since these are C++ objects that may not    |
|      be initialized when this routine is called.                             |
------------------------------------------------------------------------------*/
void  ITrace :: writeString(char* pszString)
{
  switch(ITrace::iClTraceLocation)
  {
    case ITrace::standardError:
    {
      fprintf(stderr,"%s", pszString);
      fflush(stderr);
      break;
    }

    case ITrace::standardOutput:
    {
      printf("%s",pszString);
      fflush(stdout);
      break;
    }
    case ITrace::queue:
    {
      imprintf("%0.*s", IC_PRINTFMAXLEN, pszString);
      break;
    }
    case ITrace::file:
    {
      // check to see whether the environment variable is defined
      char *envptr = getenv("ICLUI TRACEFILE");
      if (!envptr)
        envptr = getenv("ICLUI_TRACEFILE");

      // If either version of the environment variable is defined
      if (envptr)
      {
        // Strip leading/trailing blanks from value and uppercase it.
        IString strTraceOutputFile(envptr);
        strTraceOutputFile.strip().upperCase();

        // Use converted value as filename and open file for appending.
        FILE *fp = fopen(strTraceOutputFile, "a");

        // If file opened successfully, append string to file and close.
        if (fp)
        {
          fprintf(fp, "%s", pszString);
          fclose(fp);
        }
      }
      break;
    }
  }
}


/*------------------------------------------------------------------------------
| ITraceSetup::initialize                                                      |
|                                                                              |
| Initialize uses the following environment variables to set up the tracing    |
| options.                                                                     |
|                                                                              |
| Environment variables and possible values:                                   |
|   ICLUI TRACE                                                                |
|      ON                - turns tracing on                                    |
|      NOPREFIX          - no prefixes                                         |
|   ICLUI TRACETO                                                              |
|      STDERR | ERR      - standard error for tracing output                   |
|      STDOUT | OUT      - standard output for tracing output                  |
|      QUEUE  | PMPRINTF - tracing output to queue                             |
|      FILE              - tracing output to file                              |
|   ICLUI CHECKMEMORY                                                          |
|      ON | TRUE         - check memory                                        |
|   ICLUI CHECKSTACK                                                           |
|      ON | TRUE         - check stack                                         |
------------------------------------------------------------------------------*/
void ITraceSetup::initialize()
{
   char *envptr;
   /******************************************************************/
   /* Reset uninitialized bit and setup trace defaults.  OR values   */
   /* in to prevent resetting other trace bits that may have been    */
   /* set PRIOR to initialization being called.                      */
   /******************************************************************/
   ITrace::iClState &= ~ITrace::uninitialized;
   ITrace::iClState |= ITrace::writeLineNumber | ITrace::writePrefix;

   /* Decide whether tracing is on */
   if (!(envptr = getenv("ICLUI TRACE")))
      envptr = getenv("ICLUI_TRACE");
   IString strTrace(envptr);
   strTrace.strip().upperCase();
   if(strTrace== "ON")
      ITrace::iClState |=ITrace::on;
   else if(strTrace== "NOPREFIX")
   {
      ITrace::iClState |=ITrace::on;
      ITrace::iClState &= ~ITrace::writePrefix;
   }

   /* decide where to send the output */
   if (!(envptr = getenv("ICLUI TRACETO")))
      envptr = getenv("ICLUI_TRACETO");
   IString strTraceTo(envptr);
   strTraceTo.strip().upperCase();

   if(strTraceTo.length() > 2)
   {
      /* Turn tracing on */
      ITrace::iClState |=ITrace::on;

      /* Setup output location; default is PM Queue  */
      if(strTraceTo=="STDERR" || strTraceTo=="ERR")
        ITrace::iClTraceLocation =ITrace::standardError;
      else if(strTraceTo=="STDOUT" || strTraceTo=="OUT")
        ITrace::iClTraceLocation=ITrace::standardOutput;
      else if(strTraceTo=="FILE")
        ITrace::iClTraceLocation=ITrace::file;
      else if(strTraceTo=="QUEUE" || strTraceTo=="PMPRINTF")
        ITrace::iClTraceLocation=ITrace::queue;
      else
        ITrace::iClTraceLocation=ITrace::queue;
   }
#ifdef __DEBUG_ALLOC__
   /* decide whether to check memory */
   if (!(envptr = getenv("ICLUI CHECKMEMORY")))
      envptr = getenv("ICLUI_CHECKMEMORY");
   IString strCheckMemory(envptr);
   strTraceTo.strip().upperCase();
   if(strCheckMemory.length() > 1)
   {
      if(strCheckMemory=="ON" | strCheckMemory=="TRUE")
        ITrace__fCheckMemory = true;
   }
#endif

   /* decide whether to check stack usage */
   if (!(envptr = getenv("ICLUI CHECKSTACK")))
      envptr = getenv("ICLUI_CHECKSTACK");
   IString strCheckStack(envptr);
   strCheckStack.strip().upperCase();
   if(strCheckStack.length() > 1)
   {
      if(strCheckStack=="ON" | strCheckStack=="TRUE")
        ITrace__fCheckStack = true;
   }
}

