/*******************************************************************************
* FILE NAME: iddeevt.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Definition of the class(es):                                               *
*     IDDEEvent                   - Base DDE event class.                      *
*     IDDEAcknowledgeEvent        - Base DDE acknowledgement class.            *
*     IDDESetAcknowledgeInfoEvent - Base class for all events that need to set *
*                                   acknowledgement information.               *
*     IDDEClientAcknowledgeEvent  - Information for acknowledgements from a    *
*                                   server to requestData(), beginHotLink()    *
*                                   and endHotLink() requests.                 *
*     IDDEServerAcknowledgeEvent  - Information for acknowledgements from a    *
*                                   client.                                    *
*     IDDEAcknowledgePokeEvent    - Information for acknowledgements from a    *
*                                   server to pokeData() requests.             *
*     IDDEAcknowledgeExecuteEvent - Information for acknowledgements from a    *
*                                   server to executeCommands() requests.      *
*     IDDEDataEvent               - Information and data from a server in      *
*                                   response to requestData() requests and as  *
*                                   a result of beginHotLink() requests.       *
*     IDDEClientHotLinkEvent      - Information from a server as a result of   *
*                                   beginHotLink() requests.                   *
*     IDDERequestDataEvent        - Information for requestData() requests     *
*                                   from a client.                             *
*     IDDEPokeEvent               - Information for pokeData() requests from   *
*                                   a client.                                  *
*     IDDEServerHotLinkEvent      - Information for beginHotLink() requests    *
*                                   from a client.                             *
*     IDDEExecuteEvent            - Information for executeCommands() requests *
*                                   from a client.                             *
*     IDDEEndEvent                - Base DDE end conversation event class.     *
*     IDDEClientEndEvent          - Information for a client about the ending  *
*                                   of a conversation.                         *
*     IDDEBeginEvent              - Information for a server about             *
*                                   beginConversation() requests.              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
   #define INCL_WINDDE
   #include <os2.h>
}

#include <iddeevt.hpp>
#include <iddeinfo.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDDEEVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IDDEEvent::IDDEEvent                                                         |
------------------------------------------------------------------------------*/
IDDEEvent :: IDDEEvent ( const IEvent& ddeEvent )
                : IEvent(ddeEvent)
{
   IMODTRACE_DEVELOP("IDDEEvent::ctor");
   PDDESTRUCT pddes = (_DDESTRUCT*)(void*)(ddeEvent.parameter2());
   usClStatus = pddes->fsStatus;
   strClItem = (char*)DDES_PSZITEMNAME(pddes);
   if (pddes->usFormat != 0)
      strClFormat = IDDEInfo__stringFromAtom(pddes->usFormat);
   else
      strClFormat = "";
   if (pddes->cbData != 0)
      strClData = IString( (void *)DDES_PABDATA(pddes), pddes->cbData );
   else
      strClData = "";
}

/*------------------------------------------------------------------------------
| IDDEEvent::IDDEEvent                                                         |
------------------------------------------------------------------------------*/
IDDEEvent :: IDDEEvent  ( const IDDEEvent& ddeEvent )
                : IEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEEvent::~IDDEEvent                                                        |
------------------------------------------------------------------------------*/
IDDEEvent :: ~IDDEEvent ( )
{
   ;
}


/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::IDDEAcknowledgeEvent                                   |
------------------------------------------------------------------------------*/
IDDEAcknowledgeEvent :: IDDEAcknowledgeEvent ( const IEvent& ddeEvent )
                                    : IDDEEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::~IDDEAcknowledgeEvent                                  |
------------------------------------------------------------------------------*/
IDDEAcknowledgeEvent :: ~IDDEAcknowledgeEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::isApplicationBusy                                      |
------------------------------------------------------------------------------*/
Boolean IDDEAcknowledgeEvent :: isApplicationBusy ( ) const
{
   if (status() & DDE_FBUSY)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::isAckPositive                                          |
------------------------------------------------------------------------------*/
Boolean IDDEAcknowledgeEvent :: isAckPositive ( ) const
{
   if (status() & DDE_FACK)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::isMessageUnderstood                                    |
------------------------------------------------------------------------------*/
Boolean IDDEAcknowledgeEvent :: isMessageUnderstood ( ) const
{
   if (status() & DDE_NOTPROCESSED)
      return false;
   return true;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeEvent::applicationSpecificData                                |
------------------------------------------------------------------------------*/
unsigned char IDDEAcknowledgeEvent :: applicationSpecificData ( ) const
{
   unsigned short usTemp = status();
   usTemp = usTemp & DDE_FAPPSTATUS;
   usTemp = usTemp >> 8;
   unsigned char ucTemp = usTemp;
   return ucTemp;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::IDDESetAcknowledgeInfoEvent                     |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent :: IDDESetAcknowledgeInfoEvent (
                                  const IEvent& ddeEvent )
   : IDDEEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::IDDESetAcknowledgeInfoEvent                     |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent :: IDDESetAcknowledgeInfoEvent (
                                  const IDDESetAcknowledgeInfoEvent& ddeEvent )
   : IDDEEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::~IDDESetAcknowledgeInfoEvent                    |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent :: ~IDDESetAcknowledgeInfoEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::setApplicationSpecificData                      |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent&
   IDDESetAcknowledgeInfoEvent :: setApplicationSpecificData (
                                     unsigned char appData)
{
   IMODTRACE_DEVELOP("IDDESetAckInfoEvent::setAppSpecificData");
   unsigned short usTemp = appData;
   unsigned short usTemp1 = status();
   usTemp = usTemp << 8;
   usTemp1 = usTemp1 & 0x00ff;
   usTemp = usTemp1 | usTemp;
   setStatus(usTemp);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::setApplicationBusy                              |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent&
   IDDESetAcknowledgeInfoEvent :: setApplicationBusy ( )
{
   setStatus( status() | DDE_FBUSY);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDESetAcknowledgeInfoEvent::setMessageNotUnderstood                         |
------------------------------------------------------------------------------*/
IDDESetAcknowledgeInfoEvent&
   IDDESetAcknowledgeInfoEvent :: setMessageNotUnderstood ( )
{
   setStatus( status() | DDE_NOTPROCESSED);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDEClientAcknowledgeEvent::IDDEClientAcknowledgeEvent                       |
------------------------------------------------------------------------------*/
IDDEClientAcknowledgeEvent :: IDDEClientAcknowledgeEvent (
                                 const IEvent& ddeEvent )
   : IDDEAcknowledgeEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientAcknowledgeEvent::~IDDEClientAcknowledgeEvent                      |
------------------------------------------------------------------------------*/
IDDEClientAcknowledgeEvent :: ~IDDEClientAcknowledgeEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientAcknowledgeEvent::isAckToRequestData                               |
------------------------------------------------------------------------------*/
Boolean IDDEClientAcknowledgeEvent :: isAckToRequestData ( ) const
{
   if (transactionType() == WM_DDE_REQUEST)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEClientAcknowledgeEvent::isAckToBeginHotLink                              |
------------------------------------------------------------------------------*/
Boolean IDDEClientAcknowledgeEvent :: isAckToBeginHotLink ( ) const
{
   if (transactionType() == WM_DDE_ADVISE)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEClientAcknowledgeEvent::isAckToEndHotLink                                |
------------------------------------------------------------------------------*/
Boolean IDDEClientAcknowledgeEvent :: isAckToEndHotLink ( ) const
{
   if (transactionType() == WM_DDE_UNADVISE)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEServerAcknowledgeEvent::IDDEServerAcknowledgeEvent                       |
------------------------------------------------------------------------------*/
IDDEServerAcknowledgeEvent :: IDDEServerAcknowledgeEvent (
                                 const IEvent& ddeEvent)
   : IDDEAcknowledgeEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEServerAcknowledgeEvent::~IDDEServerAcknowledgeEvent                      |
------------------------------------------------------------------------------*/
IDDEServerAcknowledgeEvent :: ~IDDEServerAcknowledgeEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEServerAcknowledgeEvent::isAckToHotLinkUpdate                             |
------------------------------------------------------------------------------*/
Boolean IDDEServerAcknowledgeEvent :: isAckToHotLinkUpdate ( ) const
{
   if (transactionType() == WM_DDE_ADVISE)
      return true;
   return false;
}


/*------------------------------------------------------------------------------
| IDDEAcknowledgePokeEvent::IDDEAcknowledgePokeEvent                           |
------------------------------------------------------------------------------*/
IDDEAcknowledgePokeEvent :: IDDEAcknowledgePokeEvent ( const IEvent& ddeEvent )
                              : IDDEAcknowledgeEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgePokeEvent::~IDDEAcknowledgePokeEvent                          |
------------------------------------------------------------------------------*/
IDDEAcknowledgePokeEvent :: ~IDDEAcknowledgePokeEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeExecuteEvent::IDDEAcknowledgeExecuteEvent                     |
------------------------------------------------------------------------------*/
IDDEAcknowledgeExecuteEvent :: IDDEAcknowledgeExecuteEvent (
                                  const IEvent& ddeEvent )
                      : IDDEAcknowledgeEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEAcknowledgeExecuteEvent::~IDDEAcknowledgeExecuteEvent                    |
------------------------------------------------------------------------------*/
IDDEAcknowledgeExecuteEvent :: ~IDDEAcknowledgeExecuteEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEDataEvent::IDDEDataEvent                                                 |
------------------------------------------------------------------------------*/
IDDEDataEvent :: IDDEDataEvent ( const IEvent& ddeEvent )
                         : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEDataEvent::~IDDEDataEvent                                                |
------------------------------------------------------------------------------*/
IDDEDataEvent :: ~IDDEDataEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEDataEvent::isAckRequested                                                |
------------------------------------------------------------------------------*/
Boolean IDDEDataEvent :: isAckRequested ( ) const
{
   if (status() & DDE_FACKREQ)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEDataEvent::isDataFromHotLink                                             |
------------------------------------------------------------------------------*/
Boolean IDDEDataEvent :: isDataFromHotLink ( ) const
{
   if (status() & DDE_FRESPONSE)
      return false;
   return true;
}

/*------------------------------------------------------------------------------
| IDDEClientHotLinkEvent::IDDEClientHotLinkEvent                               |
------------------------------------------------------------------------------*/
IDDEClientHotLinkEvent :: IDDEClientHotLinkEvent ( const IEvent& ddeEvent )
                              : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientHotLinkEvent::IDDEClientHotLinkEvent                               |
------------------------------------------------------------------------------*/
IDDEClientHotLinkEvent :: IDDEClientHotLinkEvent (
                                    const IDDEClientHotLinkEvent& ddeEvent )
                              : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientHotLinkEvent::~IDDEClientHotLinkEvent                              |
------------------------------------------------------------------------------*/
IDDEClientHotLinkEvent :: ~IDDEClientHotLinkEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientHotLinkEvent::isAckRequested                                       |
------------------------------------------------------------------------------*/
Boolean IDDEClientHotLinkEvent :: isAckRequested ( ) const
{
   if (status() & DDE_FACKREQ)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEClientHotLinkEvent::isDataRequested                                      |
------------------------------------------------------------------------------*/
Boolean IDDEClientHotLinkEvent :: isDataRequested ( ) const
{
   if (status() & DDE_FNODATA)
      return false;
   return true;
}

/*------------------------------------------------------------------------------
| IDDERequestDataEvent::IDDERequestDataEvent                                   |
------------------------------------------------------------------------------*/
IDDERequestDataEvent :: IDDERequestDataEvent  ( const IEvent& ddeEvent )
                             : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDERequestDataEvent::~IDDERequestDataEvent                                  |
------------------------------------------------------------------------------*/
IDDERequestDataEvent :: ~IDDERequestDataEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDERequestDataEvent::setData                                                |
------------------------------------------------------------------------------*/
IDDERequestDataEvent& IDDERequestDataEvent :: setData ( const char* data )
{
   IString strTemp(data);
   setBuffer(strTemp);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDERequestDataEvent::setData                                                |
------------------------------------------------------------------------------*/
IDDERequestDataEvent& IDDERequestDataEvent :: setData ( const void* data,
                                                      unsigned long dataLength )
{
   IString strTemp(data, dataLength);
   setBuffer(strTemp);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDERequestDataEvent::requestAck                                             |
------------------------------------------------------------------------------*/
IDDERequestDataEvent& IDDERequestDataEvent :: requestAck ( )
{
   setStatus( status() | DDE_FACKREQ);
   return *this;
}

/*------------------------------------------------------------------------------
| IDDEPokeEvent::IDDEPokeEvent                                                 |
------------------------------------------------------------------------------*/
IDDEPokeEvent :: IDDEPokeEvent  ( const IEvent& ddeEvent )
                        : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEPokeEvent::~IDDEPokeEvent                                                |
------------------------------------------------------------------------------*/
IDDEPokeEvent :: ~IDDEPokeEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEExecuteEvent::IDDEExecuteEvent                                           |
------------------------------------------------------------------------------*/
IDDEExecuteEvent :: IDDEExecuteEvent  ( const IEvent& ddeEvent )
                             : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEExecuteEvent::~IDDEExecuteEvent                                          |
------------------------------------------------------------------------------*/
IDDEExecuteEvent :: ~IDDEExecuteEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEServerHotLinkEvent::IDDEServerHotLinkEvent                               |
------------------------------------------------------------------------------*/
IDDEServerHotLinkEvent :: IDDEServerHotLinkEvent  ( const IEvent& ddeEvent )
                                  : IDDESetAcknowledgeInfoEvent(ddeEvent)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEServerHotLinkEvent::~IDDEServerHotLinkEvent                              |
------------------------------------------------------------------------------*/
IDDEServerHotLinkEvent :: ~IDDEServerHotLinkEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEServerHotLinkEvent::isPacingRequested                                    |
------------------------------------------------------------------------------*/
Boolean IDDEServerHotLinkEvent :: isPacingRequested ( ) const
{
   if (status() & DDE_FACKREQ)
      return true;
   return false;
}

/*------------------------------------------------------------------------------
| IDDEServerHotLinkEvent::isDataRequested                                      |
------------------------------------------------------------------------------*/
Boolean IDDEServerHotLinkEvent :: isDataRequested ( ) const
{
   if (status() & DDE_FNODATA)
      return false;
   return true;
}

/*------------------------------------------------------------------------------
| IDDEEndEvent::IDDEEndEvent                                                   |
------------------------------------------------------------------------------*/
IDDEEndEvent :: IDDEEndEvent ( const IEvent& ddeEvent,
                               Source        endSource )
                     : IEvent(ddeEvent), sourceCl(endSource)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEEndEvent::~IDDEEndEvent                                                  |
------------------------------------------------------------------------------*/
IDDEEndEvent :: ~IDDEEndEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientEndEvent::IDDEClientEndEvent                                       |
------------------------------------------------------------------------------*/
IDDEClientEndEvent :: IDDEClientEndEvent ( const IEvent&  ddeEvent,
                                           Source         endSource,
                                           IString        application,
                                           IString topic )
                      : IDDEEndEvent(ddeEvent,endSource),
                        strClApp(application), strClTopic(topic)
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEClientEndEvent::~IDDEClientEndEvent                                      |
------------------------------------------------------------------------------*/
IDDEClientEndEvent :: ~IDDEClientEndEvent ( )
{
   ;
}

/*------------------------------------------------------------------------------
| IDDEBeginEvent::IDDEBeginEvent                                               |
------------------------------------------------------------------------------*/
IDDEBeginEvent :: IDDEBeginEvent ( const IEvent& ddeBeginEvent )
                        : IEvent(ddeBeginEvent)
{
   PDDEINIT pddei = (_DDEINIT*)(void*)(ddeBeginEvent.parameter2());
   strClApp = pddei->pszAppName;
   strClTopic = pddei->pszTopic;
   unsigned long ulCaseSens = DDEI_PCONVCONTEXT(pddei)->fsContext;
   if (ulCaseSens == DDECTXT_CASESENSITIVE)
      bClCaseSens =  true;
   else bClCaseSens =  false;
}

/*------------------------------------------------------------------------------
| IDDEBeginEvent::~IDDEBeginEvent                                              |
------------------------------------------------------------------------------*/
IDDEBeginEvent :: ~IDDEBeginEvent ( )
{
   ;
}
