/*******************************************************************************
* FILE NAME: immabuf.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immabuf.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_OS2MM
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#ifndef _ISTRING_
   #include <istring.hpp>
#endif

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif

#include <immabuf.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMABUF_CPP_
  #include <ipagetun.h>
#endif


// Initialize static objects
#pragma data_seg(ICLStaticConst)
const unsigned long
  IMMAudioBuffer::audio  = MMIO_MEDIATYPE_AUDIO;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMMAudioBufferData                                                           |
| This class is reserved for future use (to store                              |
| private instance variables).                                                 |
------------------------------------------------------------------------------*/
class IMMAudioBufferData : public IBase
{
public:
  IMMAudioBufferData()
     { }
};


/*------------------------------------------------------------------------------
| IMMAudioBuffer::IMMAudioBuffer                                               |
| This Constructor creates an audio buffer object, containing the data         |
| passed in.                                                                   |
------------------------------------------------------------------------------*/
IMMAudioBuffer::IMMAudioBuffer( void *audioData,
                                unsigned long len)
              : fBuffer(new IString(audioData,len)),
                fLength(len)
{
  fHeader = new MMAUDIOHEADER;
  memset(fHeader, 0, sizeof(MMAUDIOHEADER));
  PMMAUDIOHEADER(fHeader)->ulHeaderLength = sizeof(MMAUDIOHEADER);
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::IMMAudioBuffer                                               |
| This Constructor creates an empty audio buffer object.                       |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer::IMMAudioBuffer(unsigned long len)
              : fBuffer(new IString(0,len)),
                fLength(len)
{
  fHeader = new MMAUDIOHEADER;
  memset(fHeader, 0, sizeof(MMAUDIOHEADER));
  PMMAUDIOHEADER(fHeader)->ulHeaderLength = sizeof(MMAUDIOHEADER);
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::IMMAudioBuffer                                               |
| This Constructor creates an audio buffer object from an existing audio       |
| buffer object                                                                |
------------------------------------------------------------------------------*/
IMMAudioBuffer::IMMAudioBuffer(const IMMAudioBuffer &audioBuffer)
              : fBuffer(audioBuffer.fBuffer),
                fLength(audioBuffer.fLength)
{
  fHeader = new MMAUDIOHEADER;
  memcpy(fHeader, audioBuffer.fHeader, sizeof(MMAUDIOHEADER));
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::operator =                                                   |
| Copies the audio buffer information from the passed in audio buffer.         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::operator = (const IMMAudioBuffer &audioBuffer)
{
  memcpy(fHeader, audioBuffer.fHeader, sizeof(MMAUDIOHEADER));
  if (fBuffer)
     delete fBuffer;
  fBuffer = new IString(audioBuffer.fBuffer,audioBuffer.fLength);
  fLength = audioBuffer.fLength;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::~IMMAudioBuffer                                              |
| Deletes the buffer and associated header information.                        |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer::~IMMAudioBuffer()
{
  if (fHeader)
     delete fHeader;
  if (fBuffer)
     delete fBuffer;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::headerData                                                   |
| Returns the operating system specific header information.                    |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMAudioBuffer::headerData() const
{
  return IString(fHeader,sizeof(MMAUDIOHEADER));
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::data                                                         |
! Returns a copy of the data in the audio buffer.                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMAudioBuffer::data() const
{
  return IString(*fBuffer);
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setData                                                      |
| Sets the contents and size of the audio buffer to that of the passed in      |
| buffer.                                                                      |
------------------------------------------------------------------------------*/
IMMAudioBuffer& IMMAudioBuffer::setData( const void* data,
                                         unsigned long dataLength)
{
  if (fBuffer)
     delete fBuffer;
  fBuffer = new IString(data,dataLength);
  fLength = dataLength;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::length                                                       |
| Returns the size of the buffer.                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::length() const
{
  return fLength;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::bytesPerSecond                                               |
| Returns the average number of bytes-per-second played or recorded (the       |
| rate that the data is transferred to the hardware).                          |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::bytesPerSecond() const
{
  return PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.ulAvgBytesPerSec;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::bitsPerSample                                                |
| Returns the number of bits-per-sample used to represent each sample of       |
| each channel.  The standard values are 4, 8, and 16.                         |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::bitsPerSample() const
{
  return PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usBitsPerSample;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::blockAlignment                                               |
| Returns the block alignment of the data in bytes.                            |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::blockAlignment() const
{
  return PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usBlockAlign;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::channels                                                     |
| Returns the number of audio channels set.  For example:                      |
| mono returns 1; stereo returns 2.                                            |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::channels() const
{
  return PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usChannels;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::contentType                                                  |
| Returns the audio content type.                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer::ContentType IMMAudioBuffer::contentType() const
{
  return IMMAudioBuffer::ContentType(PMMAUDIOHEADER(fHeader)->ulContentType);
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::format                                                       |
| Returns the audio format being used.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer::Format IMMAudioBuffer::format() const
{
  return IMMAudioBuffer::Format(PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usFormatTag);
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::mediaType                                                    |
| Returns the type of multimedia data in the buffer.                           |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::mediaType() const
{
  return PMMAUDIOHEADER(fHeader)->ulMediaType;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::samplesPerSecond                                             |
| Returns the number of samples-per-second played or recorded.                 |
| Standard values are 11025, 22050, and 44100.                                 |
------------------------------------------------------------------------------*/
unsigned long IMMAudioBuffer::samplesPerSecond() const
{
  return PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.ulSamplesPerSec;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setBytesPerSecond                                            |
| Sets the average bytes per second to be played or recorded.                  |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setBytesPerSecond(unsigned long avgBytes)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.ulAvgBytesPerSec = avgBytes;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setBitsPerSample                                             |
| Sets the bits-per-sample to be played or recorded.                           |
| The standard values are 4, 8, and 16.                                        |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setBitsPerSample(unsigned long bitsPer)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usBitsPerSample = USHORT(bitsPer);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setBlockAlignment                                            |
| Sets the block alignment of the data in bytes.                               |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setBlockAlignment(unsigned long blockAlign)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usBlockAlign =
                                                        USHORT(blockAlign);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setChannels                                                  |
| Sets the number of audio channels.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setChannels(unsigned long channels)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usChannels = (unsigned short)channels;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setContentType                                               |
| Sets the audio content type.                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setContentType(IMMAudioBuffer::ContentType contentType)
{
  PMMAUDIOHEADER(fHeader)->ulContentType = ULONG(contentType);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setFormat                                                    |
| Sets the audio format to be used.                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setFormat(IMMAudioBuffer::Format tag)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.usFormatTag = USHORT(tag);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setMediaType                                                 |
| Sets the type of multimedia data in the buffer.                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setMediaType(unsigned long mediaType)
{
  PMMAUDIOHEADER(fHeader)->ulMediaType = mediaType;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAudioBuffer::setSamplesPerSecond                                          |
| Sets the samples rate for playing and recording.                             |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAudioBuffer &IMMAudioBuffer::setSamplesPerSecond(unsigned long samps)
{
  PMMAUDIOHEADER(fHeader)->mmXWAVHeader.WAVEHeader.ulSamplesPerSec = samps;
  return *this;
}

