/*******************************************************************************
* FILE NAME: immamix.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immamix.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_MCIOS2
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif
#include <immowin.hpp>
#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif

#include <immamix.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMAMIX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMMAmpMixer::IMMAmpMixer                                                     |
| This constructor creates a new device object.                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer::IMMAmpMixer( Boolean openNow,
                          unsigned long instance,
                          Boolean openShareable)
           : IMMDevice("AMPMIX")
{
  if (openNow)
  {
     IString fullName("AMPMIX");

     if (instance)
        fullName += IString(instance).rightJustify(2, '0');

     openOnThread(fullName, openShareable);
  }
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::IMMAmpMixer                                                     |
| This constructor wrappers an existing device, and gives a derived class the  |
| ability to specify a desired alias (instead of ICLUI generating it).         |
------------------------------------------------------------------------------*/
IMMAmpMixer::IMMAmpMixer(unsigned long newDeviceId,
                         const IString &newAlias)
           : IMMDevice(newDeviceId, newAlias)
{ }

/*------------------------------------------------------------------------------
| IMMAmpMixer::~IMMAmpMixer                                                    |
| Virtual destructor - nothing special needed as parent does cleanup.          |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer::~IMMAmpMixer()
{ }

/*------------------------------------------------------------------------------
| IMMAmpMixer::balance                                                         |
| Returns the balance, where 0 is defined as full left balance and             |
| 100 is defined as full right balance.                                        |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::balance(CallType call) const
{
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_BALANCE, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"itemStatus");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::bass                                                            |
| Returns the bass magnitude. The value returned is interpreted as a           |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::bass(CallType call) const
{
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_BASS, 0, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::gain                                                            |
| Returns the gain. The value returned is interpreted as a                     |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::gain(CallType call) const
{
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_GAIN, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::isMonitoringEnabled                                             |
| Returns true if the amp-mixer is able to listen to the audio signal from one |
| device while the audio signal from a second device is being recorded.        |
------------------------------------------------------------------------------*/
Boolean IMMAmpMixer::isMonitoringEnabled(CallType call) const
{
  IASSERTSTATE(isOpen());
  Boolean rc = itemStatus(MCI_AMP_STATUS_MONITOR, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::pitch                                                           |
| Returns the pitch. The value returned is interpreted as a                    |
| percentage: 0 is the least amount; 100 is the maximum.                       |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::pitch(CallType call) const
{
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_PITCH, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::treble                                                          |
| Returns the treble. The value returned is interpreted as a                   |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::treble(CallType call) const
{
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_TREBLE, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setBalance                                                      |
| Sets the balance, where 0 is defined as full left balance and                |
| 100 is defined as full right balance.                                        |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setBalance(unsigned long balance, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)balance;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_BALANCE;
  parms.ulValue          = (unsigned long)balance;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_BALANCE | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setBass                                                         |
| Sets the bass magnitude. The value passed in is interpreted as a             |
| percentage: 0% is the least amount; 100 is the maximum                       |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setBass(unsigned long bass, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)bass;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_BASS;
  parms.ulValue          = (unsigned long)bass;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_BASS | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setGain                                                         |
| Sets the gain. The value passed in is interpreted as a                       |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setGain(unsigned long gain, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)gain;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_GAIN;
  parms.ulValue          = (unsigned long)gain;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_GAIN | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::enableMonitoring                                                |
| Enable the amp-mixer to be able to listen to the audio signal from one       |
| device while the audio signal from a second device is being recorded.        |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::enableMonitoring(Boolean enable, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_MONITOR;
  sendCommand(MCI_SET,
              MCI_SET_ITEM | (enable ? MCI_SET_ON : MCI_SET_OFF) | call,
              &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::disableMonitoring                                               |
| Disable the monitoring function.                                             |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::disableMonitoring(CallType call)
{
  enableMonitoring(false,call);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setPitch                                                        |
| Sets the pitch. The value passed in is interpreted as a                      |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setPitch(unsigned long pitch, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)pitch;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_PITCH;
  parms.ulValue          = (unsigned long)pitch;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_PITCH | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setTreble                                                       |
| Sets the treble. The value passed in is interpreted as a                     |
| percentage: 0 is the least amount; 100 is the maximum.                       |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setTreble(unsigned long treble, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)treble;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_TREBLE;
  parms.ulValue          = (unsigned long)treble;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_TREBLE | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::isOpenStringValid                                               |
| Returns true if the string passed in is valid for this device.               |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMAmpMixer::isOpenStringValid(const IString &deviceName) const
{
  IString tester(deviceName);
  return (tester.upperCase().indexOf("AMPMIX") == 1 ? true : false);
}
