/*******************************************************************************
* FILE NAME: immdev1.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immdevnh.hpp and some in immdev.hpp.                                    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define INCL_WINSYS

extern "C" {
  #define IC_MULTIMEDIA
  #include <iwindefs.h>
}

#include <iwindow.hpp>
#include <immowin.hpp>
#include <immdev.hpp>
#include <immdevnh.hpp>
#include <ievent.hpp>
#include <inotifev.hpp>
#include <iexcept.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMDEV1_CPP_
  #include <ipagetun.h>
#endif

// Class to clean up static objects
class IMMDeviceNotifyStatics
{
public:
  ~IMMDeviceNotifyStatics ()
  {
    delete deviceNotifyHandler;
    deviceNotifyHandler = 0;
  }

static IMMDeviceNotifyHandler
  &notificationHandler ( );

static IMMDeviceNotifyHandler
  *deviceNotifyHandler;
};

// Initialize static objects and notification strings.
#pragma data_seg(ICLStaticConst)
static IMMDeviceNotifyStatics deviceStatics;
INotificationId const IMMDevice::commandNotifyId           = "IMMDevice::commandNotify";
INotificationId const IMMDevice::passDeviceId              = "IMMDevice::passDevice";
INotificationId const IMMDevice::cuePointId                = "IMMDevice::cuePoint";
INotificationId const IMMDevice::positionChangeId          = "IMMDevice::positionChange";
INotificationId const IMMDevice::deviceEventId             = "IMMDevice::deviceEvent";
#pragma data_seg()

#pragma data_seg(ICLNonConst)
IMMDeviceNotifyHandler*
  IMMDeviceNotifyStatics::deviceNotifyHandler = 0;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMMDeviceNotifyStatics::notificationHandler                                  |
|                                                                              |
| Static function to create and return a notification handler for IMMDevice.   |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler& IMMDeviceNotifyStatics::notificationHandler ()
{
  if (!(IMMDeviceNotifyStatics::deviceNotifyHandler))
  {
    IMMDeviceNotifyStatics::deviceNotifyHandler =
                                   new IMMDeviceNotifyHandler();
  }
  return( *IMMDeviceNotifyStatics::deviceNotifyHandler );
}


/*------------------------------------------------------------------------------
| IMMDevice::enableNotification                                                |
|                                                                              |
| Flags this part as being ready to accept notifications.  A Notifier, like    |
| IMMDevice, does not send notifications to observer objects until it          |
| is enabled.                                                                  |
|                                                                              |
| In this class, we override the INotifier function to allocate the correct    |
| notification handler. We use our msg window to manage the handler & boolean. |
------------------------------------------------------------------------------*/
IMMDevice& IMMDevice::enableNotification (Boolean enabled)
{
  if (!notificationHandler())
  {
    setNotificationHandler(
               &(IMMDeviceNotifyStatics::notificationHandler()));
  }
  deviceWindow().enableNotification(enabled);    // Enable our proxy window, so our notification
                                                 // handler gets called (also enables IWindow notifIds e.g., destroyId)
  Inherited::enableNotification(enabled);                   // Enable this device also, so notifyObservers works
  return *this;
}


/*------------------------------------------------------------------------------
| IMMDevice::notificationHandler                                               |
|                                                                              |
| Returns a pointer to the notification handler                                |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler* IMMDevice::notificationHandler    ( ) const
{
  IMMDeviceNotifyHandler* notifyHandler =
           (IMMDeviceNotifyHandler*)(deviceWindow().notificationHandler());
  return notifyHandler;
}

/*------------------------------------------------------------------------------
| IMMDevice::setNotificationHandler                                            |
|                                                                              |
| Sets the notification handler to the passed value.                           |
------------------------------------------------------------------------------*/
IMMDevice& IMMDevice::setNotificationHandler ( 
                                     IMMDeviceNotifyHandler* notifyHandler )
{
  IASSERTSTATE(notifyHandler!=0);
  deviceWindow().setNotificationHandler(notifyHandler);
  return *this;
}


/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::IMMDeviceNotifyHandler                               |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler::IMMDeviceNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::~IMMDeviceNotifyHandler                              |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler::~IMMDeviceNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::dispatchHandlerEvent                                 |
|                                                                              |
| This notifies IMMDevice Windows of the following events:                     |
|                                                                              |
| mediaLoaded(Id) - either loaded (true) or unloaded (false)                   |
------------------------------------------------------------------------------*/

Boolean IMMDeviceNotifyHandler::dispatchHandlerEvent (IEvent& anEvent)
{
  switch (anEvent.eventId())
  {
    case MM_MCINOTIFY:
      {
      IMMNotifyEvent notifyEvent(anEvent);
      IMMDevice* device = notifyEvent.device();
      device->notifyObservers(
                     INotificationEvent(IMMDevice::commandNotifyId,
                                        *device,true, &notifyEvent));
      break;
      }

    case MM_MCIPASSDEVICE:
      {
      IMMPassDeviceEvent passEvent(anEvent);
      IMMDevice* device = passEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::passDeviceId,
                                     *device,true, &passEvent));
      break;
      }

    case MM_MCIPOSITIONCHANGE:
      {
      IMMPositionChangeEvent posEvent(anEvent);
      IMMDevice* device = posEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::positionChangeId,
                                     *device,true, &posEvent));
      break;
      }

    case MM_MCICUEPOINT:
      {
      IMMCuePointEvent cueEvent(anEvent);
      IMMDevice* device = cueEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::cuePointId,
                                     *device,true, &cueEvent));
      break;
      }

    case MM_MCIEVENT:
      {
      IMMDeviceEvent devEvent(anEvent);
      IMMDevice* device = devEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::deviceEventId,
                                     *device,true, &devEvent));
      break;
      }

      default: 
         break;
  }  /* end switch */

   return (Inherited::dispatchHandlerEvent(anEvent));
}

