/*******************************************************************************
* FILE NAME: immfilem.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immfilem.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_MCIOS2
   #define INCL_WINATOM
   #define INCL_WINMESSAGEMGR
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#include <immfilem.hpp>

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif

#ifndef _ITHREAD_
   #include <ithread.hpp>
#endif

#ifndef _IEXCEPT_
   #include <iexcept.hpp>
#endif

#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif
#include <immowin.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMFILEM_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMMFileMedia::IMMFileMedia                                                   |
| This constructor creates a new device object.                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMFileMedia::IMMFileMedia(const IString &deviceName,
                           Boolean openNow,
                           unsigned long instance,
                           Boolean shareable)
            : IMMPlayableDevice(deviceName, openNow, instance, shareable),
              fWriteable(true)
{ }

/*------------------------------------------------------------------------------
| IMMFileMedia::IMMFileMedia                                                   |
| This constructor wrappers an existing device, and gives a derived class the  |
| ability to specify a desired alias (instead of ICLUI generating it).         |
------------------------------------------------------------------------------*/
IMMFileMedia::IMMFileMedia(unsigned long newDeviceId,
                           const IString &newAlias)
             : IMMPlayableDevice(newDeviceId, newAlias),
               fWriteable(true)
{ }

/*------------------------------------------------------------------------------
| IMMFileMedia::~IMMFileMedia                                                  |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMFileMedia::~IMMFileMedia()
{ }

/*------------------------------------------------------------------------------
| IMMFileMedia::load                                                           |
| Loads the given file into memory.  If the file does not exist, a temporary   |
| file is created.                                                             |
------------------------------------------------------------------------------*/
IMMFileMedia& IMMFileMedia::load(const IString &filename,
                                 Boolean readOnly, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_LOAD_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback   = HWND(deviceWindow().handle());
  parms.pszElementName = (PSZ)filename;
  sendCommand(MCI_LOAD,
              MCI_OPEN_ELEMENT | call | (readOnly ? MCI_READONLY : 0),
              &parms);
  setReadOnlyFlag(!readOnly);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMFileMedia::isWriteable                                                    |
| Returns false if the file was loaded as read-only.                           |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMFileMedia::isWriteable() const
{
  return fWriteable;
}

/*------------------------------------------------------------------------------
| IMMFileMedia::isReadOnly                                                     |
| Returns true if the file was loaded as read-only.                            |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMFileMedia::isReadOnly() const
{
  return !isWriteable();
}

/*------------------------------------------------------------------------------
| IMMFileMedia::setReadOnlyFlag                                                |
| Allows the device to track read-only status of the loaded file.              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMFileMedia& IMMFileMedia::setReadOnlyFlag(Boolean rdOnly)
{
  enableDataUpdate(!rdOnly);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMFileMedia::enableDataUpdate                                               |
| Allows the device to track read-only status of the loaded file.              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMFileMedia& IMMFileMedia::enableDataUpdate  ( Boolean update)
{
  fWriteable = update;
  return *this;
}


/*------------------------------------------------------------------------------
| IMMFileMedia::loadOnThread                                                   |
| Loads the given file into memory by creating a thread to do the actual       |
| loading.                                                                     |
------------------------------------------------------------------------------*/
IMMFileMedia& IMMFileMedia::loadOnThread(const IString& filename,
                                         Boolean readOnly)
{
   QMSG qmsg;

   class LoadSyncFn : public IThreadFn
   {
      public:
         LoadSyncFn(IMMFileMedia& dev, const IString& file, Boolean readOnly)
            : device(dev),
              file(file),
              ro(readOnly)
            {;}
         void run() { device.load(file, ro, IMMDevice::nowait); }

      private:
         IMMFileMedia &device;
         const IString file;
         Boolean ro;
   };

   // Create our thread function object then construct an IThread
   // with the function object and auto initializing PM
   LoadSyncFn *loadSync = new LoadSyncFn(*this, filename, readOnly);
   IThread thread(loadSync, true);

   while (WinGetMsg(IThread::current().anchorBlock(),
                    &qmsg,
                    (HWND)0,
                    0,
                    0))
   {
      WinDispatchMsg(IThread::current().anchorBlock(), &qmsg);
      if ((HWND)deviceWindow().handle() == qmsg.hwnd)
      {
         IEvent evt(qmsg.hwnd, qmsg.msg, qmsg.mp1, qmsg.mp2);
         IMMNotifyEvent nEvt(evt);
         if (IMMNotifyEvent::load == nEvt.command())
         {
            setLastError(nEvt.errorId());
            break;
         }
      }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IMMFileMedia::filename                                                       |
| Returns the currently loaded file name.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMFileMedia::filename(CallType call) const
{
  IASSERTSTATE(isOpen());
  MCI_INFO_PARMS parms;
  IString        result;
  char           fileName[CCHMAXPATH + 1];

  memset(&parms, 0, sizeof(parms));
  memset(fileName, 0, sizeof(fileName));
  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.pszReturn    = (PSZ)fileName;
  parms.ulRetSize    = sizeof(fileName);
  ((IMMFileMedia*)this)->sendCommand(MCI_INFO, MCI_INFO_FILE |
                                     call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  else
     result = fileName;
  return result;
}
