/*******************************************************************************
* FILE NAME: immplayd.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immplayd.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_MCIOS2
   #define INCL_WINMESSAGEMGR
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#include <immplayd.hpp>

#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif

#ifndef _ITHREAD_
   #include <ithread.hpp>
#endif

#include <immowin.hpp>

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMPLAYD_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMMPlayableDevice::~IMMPlayableDevice                                        |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice::~IMMPlayableDevice()
{ }

/*------------------------------------------------------------------------------
| IMMPlayableDevice::cueForPlayback                                            |
| Cues the device for playback. Not required, but for some devices it          |
| might reduce the delay associated with the play function.                    |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::cueForPlayback(CallType call)
{
   IASSERTSTATE(isOpen());
   MCI_GENERIC_PARMS parms;

   memset(&parms, 0, sizeof(parms));
   parms.hwndCallback = HWND( deviceWindow().handle());
   sendCommand(MCI_CUE, MCI_CUE_OUTPUT | call, &parms);
   return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::play                                                      |
| Starts playing the device from the passed in start position                  |
| to the passed in end position.                                               |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::play(const IMMTime &from,
                                           const IMMTime &to,
                                           Boolean resumeIfPaused,
                                           CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_PLAY_PARMS parms;
  unsigned long flags = 0;

  if (resumeIfPaused && IMMDevice::paused == mode())
     resume(call);
  else
  {
     memset(&parms, 0, sizeof(parms));
     parms.hwndCallback = HWND(deviceWindow().handle());
     if (from.isValid())
     {
        parms.ulFrom = (unsigned long)from;
        flags |= MCI_FROM;
     }
     if (to.isValid())
     {
        parms.ulTo = (unsigned long)to;
        flags |= MCI_TO;
     }
     sendCommand(MCI_PLAY,
                 flags | call,
                 &parms);
  }
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::length                                                    |
| Returns the total length in the current time format.                         |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMPlayableDevice::length(CallType call) const
{
  unsigned long result = itemStatus(MCI_STATUS_LENGTH, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"itemStatus");
  return result;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::pause                                                     |
| Pauses the device if it is playing.                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::pause( CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_GENERIC_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_PAUSE, call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::position                                                  |
| Returns the current position in the current time format.                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMPlayableDevice::position(CallType call) const
{
  unsigned long result = itemStatus(MCI_STATUS_POSITION, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"itemStatus");
  return result;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::resume                                                    |
| Resumes playback of the device from a paused state.                          |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::resume(Boolean resume ,
                                             CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_GENERIC_PARMS parms;

  if (!(!resume && IMMDevice::paused == mode()))
  {
     parms.hwndCallback = HWND(deviceWindow().handle());
     sendCommand(MCI_RESUME, call, &parms);
     if (lastError())
        ITHROWMMERROR(lastError(),"sendCommand");
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::removeCuePoint                                            |
| Removes the given cue point. A cue point is a location or time position in a |
| media device.                                                                |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::removeCuePoint(const IMMTime &time, CallType call)
{
  setCuePoint(time, false, call);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::addCuePoint                                               |
| Sets a cue point at the specified location.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice&  IMMPlayableDevice::addCuePoint(const IMMTime &time, CallType call)
{
  setCuePoint(time, true, call);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::setCuePoint                                               |
| Sets or removes a cue point.                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
void IMMPlayableDevice::setCuePoint(const IMMTime &time, Boolean on,
                                    CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_CUEPOINT_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.ulCuepoint   = (unsigned long)time;
  parms.usUserParm   = userParameter();
  sendCommand(MCI_SET_CUEPOINT,
              (on ? MCI_SET_CUEPOINT_ON : MCI_SET_CUEPOINT_OFF) | call,
              &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::stopPositionTracking                                      |
| Stops position tracking. Position tracking causes a position change event to |
| be generated every time the media position changes a given time amount.      |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::stopPositionTracking(CallType call)
{
  setPositionAdvise(IMMTime(), false, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::startPositionTracking                                     |
| Start position tracking - generate a change event to every time the media    |
| position changes by the time amount passed in.                               |
------------------------------------------------------------------------------*/
IMMPlayableDevice&  IMMPlayableDevice::startPositionTracking(const IMMTime &time,
                                                             CallType call)
{
  setPositionAdvise(time, true, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::setPositionAdvise                                         |
| Start or stop position tracking.                                             |
|                                                                              |
------------------------------------------------------------------------------*/
void IMMPlayableDevice::setPositionAdvise(const IMMTime &time,
                                          Boolean on,
                                          CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_POSITION_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  if (on)
     parms.ulUnits   = (unsigned long)time;
  parms.usUserParm   = userParameter();
  sendCommand(MCI_SET_POSITION_ADVISE,
              (on ? MCI_SET_POSITION_ADVISE_ON :
              MCI_SET_POSITION_ADVISE_OFF) |
              call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::stepFrame                                                 |
| Steps the play one or more time units forward or backward.                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::stepFrame(unsigned long frames,
                                                Boolean forward,
                                                CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_STEP_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.ulStep       = frames;
  sendCommand(MCI_STEP, MCI_STEP_FRAMES |
              (forward ? 0 : MCI_STEP_REVERSE) | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::stop                                                      |
| Stops playback of the device.                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::stop(CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_GENERIC_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_STOP, call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::seek                                                      |
| Sets the current position of the device to the passed in position.           |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::seek(const IMMTime &to, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_SEEK_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.ulTo         = (to.isValid() ? (unsigned long)to : 0);
  sendCommand(MCI_SEEK, MCI_TO | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::seekToEnd                                                 |
| Moves the current position to the end of the device's data.                  |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::seekToEnd(CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_SEEK_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.ulTo         = 0;
  sendCommand(MCI_SEEK, MCI_TO_END | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::seekToStart                                               |
| Moves the current position to the beginning of the device's data.            |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice& IMMPlayableDevice::seekToStart(CallType call)
{
  IASSERTSTATE(isOpen());

  MCI_SEEK_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.ulTo         = 0;
  sendCommand(MCI_SEEK, MCI_TO_START | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::IMMPlayableDevice                                         |
| This constructor creates a new device object.                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMPlayableDevice::IMMPlayableDevice(const IString &deviceOrFileName,
                                     Boolean openNow,
                                     unsigned long instance,
                                     Boolean shareable)
                 : IMMDevice(deviceOrFileName)
{
  if (openNow)
  {
     IString fullName(deviceOrFileName);
     if (instance)
        fullName += IString(instance).rightJustify(2, '0');
     openOnThread(fullName, shareable);
  }
}

/*------------------------------------------------------------------------------
| IMMPlayableDevice::IMMPlayableDevice                                         |
| This constructor wrappers an existing device, and gives a derived class the  |
| ability to specify a desired alias (instead of ICLUI generating it).         |
------------------------------------------------------------------------------*/
IMMPlayableDevice::IMMPlayableDevice(unsigned long newDeviceId,
                                     const IString &newAlias)
                 : IMMDevice(newDeviceId, newAlias)
{ }

