/*******************************************************************************
* FILE NAME: immrecrd.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immrecrd.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_MCIOS2
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif
#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif
#include <immowin.hpp>
#include <immrecrd.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMRECRD_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMMRecordable::IMMRecordable                                                 |
| This constructor creates a new device object.                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable::IMMRecordable(const IString &baseDeviceName,
                             Boolean openNow,
                             unsigned long instance,
                             Boolean shareable)
             : IMMFileMedia(baseDeviceName, openNow, instance, shareable)
{ }

/*------------------------------------------------------------------------------
| IMMRecordable::IMMRecordable                                                 |
| This constructor wrappers an existing device, and gives a derived class the  |
| ability to specify a desired alias (instead of ICLUI generating it).         |
------------------------------------------------------------------------------*/
IMMRecordable::IMMRecordable(unsigned long newDeviceId,
                             const IString &newAlias)
             : IMMFileMedia(newDeviceId, newAlias)
{ }

/*------------------------------------------------------------------------------
| IMMRecordable::~IMMRecordable                                                |
| Destructor                                                                   |          |
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable::~IMMRecordable()
{ }

/*------------------------------------------------------------------------------
| IMMRecordable::canRedo                                                       |
| Returns true if the device has any actions to be redone.                     |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMRecordable::canRedo() const
{
  IASSERTSTATE(isOpen());
  Boolean rc = itemStatus(MCI_STATUS_CAN_REDO);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMRecordable::canUndo                                                       |
| Returns true if the device has any actions that can be undone.               |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMRecordable::canUndo() const
{
  IASSERTSTATE(isOpen());
  Boolean rc = itemStatus(MCI_STATUS_CAN_UNDO);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
}

/*------------------------------------------------------------------------------
| IMMRecordable::copy                                                          |
| Copies data from the passed in start position to the passed in end position  |
| into the clipboard.                                                          |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::copy(const IMMTime &from,
                                   const IMMTime &to,
                                   CallType call) const
{
  IASSERTSTATE(isOpen());
  MCI_EDIT_PARMS parms;
  unsigned long flags = call;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  if (from.isValid())
  {
     flags |= MCI_FROM;
     parms.ulFrom = (unsigned long)from;
  }
  if (to.isValid())
  {
     flags |= MCI_TO;
     parms.ulTo = (unsigned long)to;
  }
  ((IMMRecordable*)this)->sendCommand(MCI_COPY, flags, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *((IMMRecordable*)this);
}

/*------------------------------------------------------------------------------
| IMMRecordable::cut                                                           |
| Cuts data from the passed in start position to the passed in end position    |
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::cut(const IMMTime &from,
                                  const IMMTime &to,
                                  CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_EDIT_PARMS parms;
  unsigned long flags = call;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  if (from.isValid())
  {
     flags |= MCI_FROM;
     parms.ulFrom = (unsigned long)from;
  }
  if (to.isValid())
  {
     flags |= MCI_TO;
     parms.ulTo = (unsigned long)to;
  }
  sendCommand(MCI_CUT, flags, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::deleteSelection                                               |
| Deletes data from the passed in start position to the passed in end position.|
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::deleteSelection(const IMMTime &from,
                                              const IMMTime &to,
                                              CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_EDIT_PARMS parms;
  unsigned long flags = call;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  if (from.isValid())
  {
     flags |= MCI_FROM;
     parms.ulFrom = (unsigned long)from;
  }
  if (to.isValid())
  {
     flags |= MCI_TO;
     parms.ulTo = (unsigned long)to;
  }
  sendCommand(MCI_DELETE, flags, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::paste                                                         |
| Replaces the data from the passed in start position to the passed in end     |
| position with data from the clipboard.                                       |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::paste(const IMMTime &from,
                                    const IMMTime &to,
                                    Boolean convert,
                                    CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_EDIT_PARMS parms;
  unsigned long flags = call;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());

  if (from.isValid())
  {
     flags |= MCI_FROM;
     parms.ulFrom = (unsigned long)from;
  }
  if (to.isValid())
  {
     flags |= MCI_TO;
     parms.ulTo = (unsigned long)to;
  }

  if (convert)
     flags |= MCI_CONVERT_FORMAT;
  sendCommand(MCI_PASTE, flags, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::redo                                                          |
| Re-does the record, cut, paste, or delete operation most recently undone by  |
| an undo.  The position is at the beginning of the file after a redo.         |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::redo(CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_GENERIC_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_REDO, call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::undo                                                          |
| Un-does the record, cut, paste, or delete operation most recently undone by  |
| an undo.  The position is at the beginning of the file after a redo.         |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::undo(CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_GENERIC_PARMS parms;

  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_UNDO, call, &parms);
  if (lastError() && lastError()!=MCIERR_CANNOT_UNDO)
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::cueForRecording                                               |
| Cues the device for recording. Not required, but may reduce the delay        |
| associated with recording for some devices.                                  |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::cueForRecording(CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_GENERIC_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_CUE, MCI_CUE_INPUT | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::record                                                        |
| Starts recording at the begin location and continues until it reaches the    |
| end location.                                                                |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::record(Boolean insert,
                                     const IMMTime &from,
                                     const IMMTime &to,
                                     Boolean resumeIfPaused,
                                     CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_RECORD_PARMS parms;
  unsigned long    flags = 0;

  if (resumeIfPaused && IMMDevice::paused == mode())
     resume(call);
  else
  {
     memset(&parms, 0, sizeof(parms));
     parms.hwndCallback = HWND(deviceWindow().handle());
     if (from.isValid())
     {
        parms.ulFrom = (unsigned long)from;
        flags       |= MCI_FROM;
     }
     if (to.isValid())
     {
        parms.ulTo = (unsigned long)to;
        flags     |= MCI_TO;
     }
     flags |= (insert ? MCI_RECORD_INSERT : MCI_RECORD_OVERWRITE) | call;
     sendCommand(MCI_RECORD, flags, &parms);
  }
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::save                                                          |
| Saves the current file.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::save(CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_SAVE_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  sendCommand(MCI_SAVE, call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

/*------------------------------------------------------------------------------
| IMMRecordable::saveAs                                                        |
| Saves the current file with a new name, specified by the IString passed in.  |
|                                                                              |
------------------------------------------------------------------------------*/
IMMRecordable& IMMRecordable::saveAs(const IString &filename, CallType call)
{
  IASSERTSTATE(isOpen());
  MCI_SAVE_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback = HWND(deviceWindow().handle());
  parms.pszFileName  = (PSZ)filename;
  sendCommand(MCI_SAVE, MCI_SAVE_FILE | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return *this;
}

