/*******************************************************************************
* FILE NAME: immtime.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immtime.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #define INCL_MCIOS2
   #define IC_MULTIMEDIA
   #include <iwindefs.h>
   }

#ifndef _ITRACE_
   #include <itrace.hpp>
#endif

#ifndef _IMMEXCPT_
   #include <immexcpt.hpp>
#endif

#include <immtime.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMTIME_CPP_
  #include <ipagetun.h>
#endif

#pragma data_seg(ICLStaticConst)
const unsigned long
  IMMTime::defaultTime = 0xFFFFFFFFL;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMMTime::~IMMTime                                                            |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::~IMMTime()
{ }

/*------------------------------------------------------------------------------
| IMMTime::IMMTime                                                             |
| Construct an instance with the value of the passed in time                   |
| (in IMMTime units).                                                          |
------------------------------------------------------------------------------*/
IMMTime::IMMTime(unsigned long time)
       : fValue(time),
         fValid(true)
{
  if (time == defaultTime)
  {
     fValue = 0;
     fValid = false;
  }
}

/*------------------------------------------------------------------------------
| IMMTime::IMMTime                                                             |
| Construct an instance with the value of the passed in IMMTime object         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::IMMTime(const IMMTime &time)
       : fValue(time.fValue),
         fValid(time.fValid)
{ }

/*------------------------------------------------------------------------------
| IMMTime::asMMTime                                                            |
| Returns this time value as an unsigned long. Derived classes                 |
| should override this function to provide a pertinent conversion.             |
------------------------------------------------------------------------------*/
unsigned long IMMTime::asMMTime() const
{
  return fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::isValid                                                             |
| Returns true if the time value is valid.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::isValid() const
{
  return fValid;
}

/*------------------------------------------------------------------------------
| IMMTime::setMMTime                                                           |
| Sets the object's time to the passed in value.                               |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::setMMTime(unsigned long newTime)
{
  fValue = newTime;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::setValid                                                            |
| Sets the validity of the current time value.  Derived classes                |
| should override where appropriate.                                           |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::setValid(Boolean isValid)
{
  fValid = isValid;
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::asString                                                            |
| Returns the time value as a string formatted as HH:MM:SS.                    |
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMTime::asString() const
{
   unsigned long msec(MSECFROMMM(fValue));
   unsigned long hours   = 0;
   unsigned long minutes = 0;
   unsigned long seconds = 0;

   if (msec)
   {
      hours = msec / (1000 * 60 * 60);
      msec %= 1000 * 60 * 60;
   }
   IString hoursStr(hours);
   if (msec)
   {
      minutes = msec / (1000 * 60);
      msec   %= 1000 * 60;
   }
   IString minutesStr(minutes);
   if (msec)
   {
      seconds = msec / 1000;
      msec   %= 1000;
   }
   IString secondsStr(seconds);
   IString msecStr(msec);
   minutesStr.rightJustify(2, '0');
   secondsStr.rightJustify(2, '0');
   msecStr.rightJustify(3, '0');
   return  IString(hoursStr   + ":" + minutesStr + ":" +
                   secondsStr + "." + msecStr);
}

/*------------------------------------------------------------------------------
| IMMTime::hours                                                               |
| Returns the hours component of the time.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::hours() const
{
  unsigned long msec(MSECFROMMM(fValue));
  return msec / (1000 * 60 * 60);
}

/*------------------------------------------------------------------------------
| IMMTime::minutes                                                             |
| Returns the minutes component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::minutes() const
{
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  return msec / (1000 * 60);
}

/*------------------------------------------------------------------------------
| IMMTime::seconds                                                             |
| Returns the seconds component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::seconds() const
{
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  msec %= 1000 * 60;         // strip minutes component
  return msec / 1000;
}

/*------------------------------------------------------------------------------
| IMMTime::hundredths                                                          |
| Returns the millisecond component of the time rounded to the nearest         |
| hundredth.                                                                   |
------------------------------------------------------------------------------*/
unsigned long IMMTime::hundredths() const
{
  unsigned long temp(thousandths());
  Boolean roundUp(temp % 10 > 4);
  temp /= 10;
  if (roundUp)
     ++temp;
  return temp;
}

/*------------------------------------------------------------------------------
| IMMTime::thousandths                                                         |
| Returns the millisecond component of the time rounded to the nearest         |
| thousandth.                                                                  |
------------------------------------------------------------------------------*/
unsigned long IMMTime::thousandths() const
{
  unsigned long msec(MSECFROMMM(fValue));
  msec %= 1000 * 60 * 60;    // strip hours component
  msec %= 1000 * 60;         // strip minutes component
  msec %= 1000;              // strip seconds component
  return msec;
}

/*------------------------------------------------------------------------------
| IMMTime::operator unsigned long                                              |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime::operator unsigned long () const
{
  return fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator =                                                          |
| Sets the time to be equal to the passed in time.                             |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator = (const IMMTime &time)
{
  if (this != &time)   // check for assignment to self
  {
     fValue = time.asMMTime();
     fValid = time.isValid();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator +                                                          |
| Returns the sum of this time and the passed in time.                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime IMMTime::operator + (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return IMMTime(fValue + time.asMMTime());
}

/*------------------------------------------------------------------------------
| IMMTime::operator -                                                          |
| Returns the result of this time minus the passed in time.                    |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime IMMTime::operator - (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return IMMTime(fValue - time.asMMTime());
}

/*------------------------------------------------------------------------------
| IMMTime::operator +=                                                         |
| Adds the time passed in from this time object                                |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator += (const IMMTime &time)
{
  IASSERTSTATE(isValid() && time.isValid());
  fValue += time.asMMTime();
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator -=                                                         |
| Subtracts the time passed in from this time object                           |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime &IMMTime::operator -= (const IMMTime &time)
{
  IASSERTSTATE(isValid() && time.isValid());
  fValue -= time.asMMTime();
  return *this;
}

/*------------------------------------------------------------------------------
| IMMTime::operator <                                                          |
| Returns true if this time is less than the passed in time.                   |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator < (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue < time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator >                                                          |
| Returns true if this time is greater than the passed in time.                |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator > (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue > time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator ==                                                         |
| Returns true if this time is equal to the passed in time.                    |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator == (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue == time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator !=                                                         |
| Returns true if this time is not equal to the passed in time.                |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator != (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue != time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator <=                                                         |
| Returns true if this time is less than or equal to the passed in time.       |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator <= (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue <= time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::operator >=                                                         |
| Returns true if this time is greater than or equal to the passed in time.    |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMTime::operator >= (const IMMTime &time) const
{
  IASSERTSTATE(isValid() && time.isValid());
  return fValue >= time.fValue;
}

/*------------------------------------------------------------------------------
| IMMTime::ordinal                                                             |
| Returns an ordinal number in milliseconds.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMTime::ordinal() const
{
  return MSECFROMMM(fValue);
}

/*------------------------------------------------------------------------------
| IMMTime::setTimeToOrdinal                                                    |
| Sets the time object to the value represented by the ordinal number          |
| in milliseconds.                                                             |
------------------------------------------------------------------------------*/
IMMTime& IMMTime::setTimeToOrdinal(unsigned long ordinal)
{
  setMMTime(MSECTOMM(ordinal));
  return *this;
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
| Construct an object representing the millsecond                              |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(const IMMMillisecondTime &time)
                  : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
| Construct an object representing the millsecond                              |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(const IMMTime &time)
                  : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMMillisecondTime::IMMMillisecondTime                                       |
| Construct an object representing the millsecond                              |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMillisecondTime::IMMMillisecondTime(unsigned long time)
                  : IMMTime(MSECTOMM(time))
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::operator unsigned long                                   |
| Returns the time as an unsigned long where each time unit is equal to        |
| one-thousandth of a second.                                                  |
------------------------------------------------------------------------------*/
IMMMillisecondTime::operator unsigned long () const
{
  return MSECFROMMM(asMMTime());
}

/*------------------------------------------------------------------------------
| IMMMillisecondTime::~IMMMillisecondTime                                      |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMillisecondTime::~IMMMillisecondTime()
{ }

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::IMMHourMinSecFrameTime                               |
| Construct an object representing the hours-minutes-seconds-frames (HHMMSSFF) |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(unsigned long val, unsigned long fps)
                      : IMMTime(HourMinSecFrameToMM(val, fps)),
                        fFramesPerSec(fps),
                        fOriginal(val)
{
  if (val == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
     fOriginal = 0;
  }
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::IMMHourMinSecFrameTime                               |
| Construct an object representing the hours-minutes-seconds-frames (HHMMSSFF) |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::IMMHourMinSecFrameTime(const IMMHourMinSecFrameTime& time)
                      : IMMTime(time),
                        fFramesPerSec(time.fFramesPerSec),
                        fOriginal(time.fOriginal)
{ }

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::~IMMHourMinSecFrameTime                              |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::~IMMHourMinSecFrameTime()
{ }

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::asString                                             |
| Returns the time value as a string formatted for this object's representation|
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMHourMinSecFrameTime::asString() const
{
  return IString(hours()) + ":" +
         IString(minutes()).rightJustify(2, '0') + ":" +
         IString(seconds()).rightJustify(2, '0') + "." +
         IString(frames()).rightJustify(2, '0');
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::framesPerSecond                                      |
| Returns the number of frames per second.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::framesPerSecond() const
{
  return fFramesPerSec;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::operator unsigned long                               |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecFrameTime::operator unsigned long() const
{
  return ((IMMHourMinSecFrameTime*)this)->HourMinSecFrameFromMM(asMMTime(),fFramesPerSec);
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::ordinal                                              |
| Returns an ordinal number in frames                                          |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::ordinal() const
{
   // change internal HourMinSecF value to frames
   // this will allow us to convert back to HourMinSecFrame from ordinal
  return (seconds() + minutes() * 60 + hours() * 60 * 60) *
         framesPerSecond() + frames();
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::setTimeToOrdinal                                     |
| Sets the time object to the value represented by the ordinal number.         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime& IMMHourMinSecFrameTime::setTimeToOrdinal(unsigned long ordinal)
{
  unsigned long h, m, s, f;

  // first get odd frames (not on a second boundary)
  f = ordinal % framesPerSecond();
  ordinal -= f;
  ordinal /= framesPerSecond();

  // now ordinal is in seconds
  h = ordinal / (60 * 60);
  ordinal %= 60 * 60;

  m = ordinal / 60;
  s = ordinal % 60;

  fOriginal = h | (m << 8) | (s << 16) | (f << 24);
  setMMTime(HourMinSecFrameToMM(fOriginal, framesPerSecond()));
  setValid(true);
  return *this;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::frames                                               |
| Returns the frames component of the time.                                    |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::frames() const
{
  return (((unsigned long) (*this)) >> 24) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::hours                                                |
| Returns the hours component of the time.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::hours() const
{
  return ((unsigned long) (*this)) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::minutes                                              |
| Returns the minutes component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::minutes() const
{
  return (((unsigned long) (*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecFrameTime::seconds                                              |
| Returns the seconds component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecFrameTime::seconds() const
{
  return (((unsigned long) (*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::IMM24FramesPerSecondTime                           |
| Construct an object representing the 24 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 24)
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::IMM24FramesPerSecondTime                           |
| Construct an object representing the 24 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::IMM24FramesPerSecondTime(const IMM24FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{ }

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::operator unsigned long                             |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::operator unsigned long () const
{
  return (unsigned long)(FPS24FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM24FramesPerSecondTime::~IMM24FramesPerSecondTime                          |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMM24FramesPerSecondTime::~IMM24FramesPerSecondTime()
{ }


/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::IMM25FramesPerSecondTime                           |
| Construct an object representing the 25 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(const IMM25FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{ }

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::IMM25FramesPerSecondTime                           |
| Construct an object representing the 25 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::IMM25FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 25)
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::operator unsigned long                             |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::operator unsigned long () const
{
  return (unsigned long)(FPS25FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM25FramesPerSecondTime::~IMM25FramesPerSecondTime                          |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMM25FramesPerSecondTime::~IMM25FramesPerSecondTime()
{ }

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::IMM30FramesPerSecondTime                           |
| Construct an object representing the 30 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(const IMM30FramesPerSecondTime &time)
                        : IMMHourMinSecFrameTime(time)
{ }

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::IMM30FramesPerSecondTime                           |
| Construct an object representing the 30 frames per second                    |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::IMM30FramesPerSecondTime(unsigned long time)
                        : IMMHourMinSecFrameTime(time, 30)
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::operator unsigned long                             |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::operator unsigned long () const
{
  return (unsigned long)(FPS30FROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMM30FramesPerSecondTime::~IMM30FramesPerSecondTime                          |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMM30FramesPerSecondTime::~IMM30FramesPerSecondTime()
{ }

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
| Construct an object representing the hours-minutes-seconds-frames            |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(unsigned long time)
                  : IMMTime(REDBOOKTOMM(time))
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
| Construct an object representing the hours-minutes-seconds-frames            |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(const IMMMinSecFrameTime &time)
                  : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::IMMMinSecFrameTime                                       |
| Construct an object representing the hours-minutes-seconds-frames            |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::IMMMinSecFrameTime(const IMMTime &time)
                  : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::asString                                                 |
| Returns the time value as a string formatted for this object's representation|
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMMinSecFrameTime::asString() const
{
  return IString(minutes()) + ":" +
         IString(seconds()).rightJustify(2, '0') + "." +
         IString(frames()).rightJustify(3, '0');
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::operator unsigned long                                   |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::operator unsigned long () const
{
  return REDBOOKFROMMM(asMMTime());
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::~IMMMinSecFrameTime                                      |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMMinSecFrameTime::~IMMMinSecFrameTime()
{ }

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::frames                                                   |
| Returns the frames component of the time.                                    |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::frames() const
{
  return (((unsigned long) (*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::minutes                                                  |
| Returns the minutes component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::minutes() const
{
  return (((unsigned long) (*this))) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMMinSecFrameTime::seconds                                                  |
| Returns the seconds component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMMinSecFrameTime::seconds() const
{
  return (((unsigned long) (*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
| Construct an object representing the hours-minutes-seconds                   |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(unsigned long time)
                 : IMMTime(HMSTOMM(time))
{
  if (time == defaultTime)
  {
     setMMTime (0);
     setValid  (false);
  }
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
| Construct an object representing the hours-minutes-seconds                   |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(const IMMHourMinSecTime &time)
                 : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::IMMHourMinSecTime                                         |
| Construct an object representing the hours-minutes-seconds                   |
| time format.                                                                 |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::IMMHourMinSecTime(const IMMTime &time)
                 : IMMTime(time)
{ }

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::hours                                                     |
| Returns the hours component of the time.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::hours() const
{
  return ((unsigned long)(*this)) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::minutes                                                   |
| Returns the minutes component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::minutes() const
{
  return (((unsigned long)(*this)) >> 16) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::seconds                                                   |
| Returns the seconds component of the time.                                   |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::seconds() const
{
  return (((unsigned long)(*this)) >> 8) & 0xFF;
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::asString                                                  |
| Returns the time value as a string formatted for this object's representation|
|                                                                              |
------------------------------------------------------------------------------*/
IString IMMHourMinSecTime::asString() const
{
  return IString(hours()) + ":" +
         IString(minutes()).rightJustify(2, '0') + ":" +
         IString(seconds()).rightJustify(2, '0');
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::operator unsigned long                                    |
| Returns the time as a unsigned long.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::operator unsigned long () const
{
  return (unsigned long)(HMSFROMMM(asMMTime()));
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::~IMMHourMinSecTime                                        |
| Destructor                                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IMMHourMinSecTime::~IMMHourMinSecTime()
   {
   }

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::ordinal                                                   |
| Returns an ordinal number in seconds.                                        |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMHourMinSecTime::ordinal() const
{
  // return total seconds
  return seconds() + (minutes() * 60) + (hours() * 60 * 60);
}

/*------------------------------------------------------------------------------
| IMMHourMinSecTime::setTimeToOrdinal                                          |
| Sets the time object to the value represented by the ordinal number          |
|                                                                              |
------------------------------------------------------------------------------*/
IMMTime& IMMHourMinSecTime::setTimeToOrdinal(unsigned long ordinal)
{
  setMMTime(MSECTOMM(ordinal * 1000));
  return *this;
}

