/*******************************************************************************
* FILE NAME: i3statbx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in i3statbx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C"
{
  #define INCL_WINBUTTONS
  #define INCL_WINERRORS
  #define INCL_GPIBITMAPS
  #define INCL_GPIPRIMITIVES
  #define INCL_WINPOINTERS
  #define INCL_WINSYS
  #include <iwindefs.h>
}

#ifndef _I3STATBX_
  #include <i3statbx.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFONT_
  #include <ifont.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _I3STATBX_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public three-state check box styles.                                         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const I3StateCheckBox::Style
  I3StateCheckBox::autoSelect        ( 0, IBS_AUTO3STATE ),
  I3StateCheckBox::classDefaultStyle ( WS_VISIBLE, IBS_AUTO3STATE );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new three-state check box objects (initial value).         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  I3StateCheckBox::Style
         I3StateCheckBox::currentDefaultStyle ( WS_VISIBLE, IBS_AUTO3STATE );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Class:  I3StateCheckBoxData                                                  |
|                                                                              |
| Store private data used by I3StateCheckBox.                                  |
------------------------------------------------------------------------------*/
class I3StateCheckBoxData : public IBase {        
public:                                           
static ISize                                      
  gBitmapSize;                                    
}; // I3StateCheckBoxData                         

#pragma data_seg(ICLNonConst)                     
  ISize I3StateCheckBoxData::gBitmapSize ( 0, 0 );
#pragma data_seg()                                


/*------------------------------------------------------------------------------
| I3StateCheckBox::I3StateCheckBox                                             |
|                                                                              |
| Constructor to create a three-state check box on a standard window.          |
------------------------------------------------------------------------------*/
I3StateCheckBox :: I3StateCheckBox(unsigned long ulId,
                                   IWindow* pwndParent,
                                   IWindow* pwndOwner,
                                   const IRectangle& rectInit,
                                   const Style& s3bsStyle)
                                 : f3StateCheckBoxData( 0 )
{
  IASSERTPARM(pwndParent!=0);          //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | s3bsStyle.asExtendedUnsignedLong() );

  //Create three-state box control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle wh3StateBox =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( s3bsStyle ),
                      WC_BUTTON,
                      pwndParent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0 );

  startHandlingEventsFor(wh3StateBox);
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::I3StateCheckBox                                             |
|                                                                              |
| Constructor to instantiate an object for a a dialog template.                |
------------------------------------------------------------------------------*/
I3StateCheckBox :: I3StateCheckBox(unsigned long ulId,
                                   IWindow* pdlgwndParent)
                                 : f3StateCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::I3StateCheckBox                                             |
|                                                                              |
| Constructor to instantiate from an existing three-state check box control.   |
------------------------------------------------------------------------------*/
I3StateCheckBox :: I3StateCheckBox(const IWindowHandle& wh)
                                 : f3StateCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(wh);
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::~I3StateCheckBox                                            |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
I3StateCheckBox :: ~I3StateCheckBox()
{
  // delete f3StateCheckBoxData;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::defaultStyle                                                |
|                                                                              |
| Return the default style for new three-state check box objects.              |
------------------------------------------------------------------------------*/
I3StateCheckBox::Style  I3StateCheckBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::setDefaultStyle                                             |
|                                                                              |
| Replace the default style for new three-state check box objects.             |
------------------------------------------------------------------------------*/
void  I3StateCheckBox :: setDefaultStyle(const Style& s3bsStyle)
{
  currentDefaultStyle = s3bsStyle;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::convertToGUIStyle                                           |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long I3StateCheckBox :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                   Boolean bExtOnly) const
{
  // Obtain the style from the class (ISettingButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // BS_ styles in the lower 4 bits of the style overlap.  Therefore, map
    // our extended style bits, back to these values.
    unsigned long ulTmp = extendedStyle() & IBS_EXT3STATE;

    // Handle overlapping styles
    if (ulTmp & IBS_AUTO3STATE)
      ulStyle |= BS_AUTO3STATE;
    else
      ulStyle |= BS_3STATE;

  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::enableAutoSelect                                            |
|                                                                              |
| Sets or removes the autoSelect style for the three-state check box control.  |
------------------------------------------------------------------------------*/
I3StateCheckBox&  I3StateCheckBox :: enableAutoSelect ( Boolean enable )
{
  unsigned long ulOldStyle = this->style();
  unsigned long ulStyle = ( ulOldStyle & (unsigned long)(~BS_PRIMARYSTYLES) );
                                       // Use masking style from pmwin.h.

  if ( enable )
  {
     ulStyle |= BS_AUTO3STATE;
  }
  else
  {
     ulStyle |= BS_3STATE;
  }

  if ( ulStyle != ulOldStyle )
  {
     this->setStyle( ulStyle );
     this->refresh();                  // Force repaint.
  }
  return *this;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::disableAutoSelect                                           |
------------------------------------------------------------------------------*/
I3StateCheckBox&  I3StateCheckBox :: disableAutoSelect()
{
  enableAutoSelect(false);
  return *this;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::isAutoSelect                                                |
------------------------------------------------------------------------------*/
Boolean  I3StateCheckBox :: isAutoSelect() const
{
  return( (style() & BS_AUTO3STATE & (unsigned long)(~BS_3STATE)) ? true : false );
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::selectHalftone                                              |
|                                                                              |
| Selects a halftone three-state check box.  Halftone is equivalent to         |
| undetermined state.                                                          |
------------------------------------------------------------------------------*/
I3StateCheckBox&  I3StateCheckBox :: selectHalftone()
{
  if (!(this->isHalftone()))
  {
    handle().sendEvent(BM_SETCHECK, IEventParameter1(2),
                                    IEventParameter2(0));
  }
  return *this;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::isHalftone                                                  |
------------------------------------------------------------------------------*/
Boolean  I3StateCheckBox :: isHalftone() const
{
  IEventResult evt = handle().sendEvent(BM_QUERYCHECK,
                                        IEventParameter1(0),
                                        IEventParameter2(0));
  if (evt.asUnsignedLong() == 2)
     return true;
  else
     return false;
}


/*------------------------------------------------------------------------------
| I3StateCheckBox::calcMinimumSize                                             |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The minimum        |
| size is based on the maximum height and width of the font of the             |
| three-state check box text (plus the size of the check box itself).          |
------------------------------------------------------------------------------*/
ISize  I3StateCheckBox :: calcMinimumSize() const
{
   long bmpWidth = I3StateCheckBoxData::gBitmapSize.width();
   if ( bmpWidth == 0 )
   {            // Find size of check box bitmap.
      BITMAPINFOHEADER pbmpData;
      pbmpData.cbFix=sizeof(BITMAPINFOHEADER);
      HBITMAP hbit = (HBITMAP) IGETSYSBITMAP(SBMP_CHECKBOXES);
      IQUERYBITMAPPARAMETERS(hbit, &pbmpData);
      IDELETEBITMAP(hbit);

      bmpWidth = pbmpData.cx;                                   
      I3StateCheckBoxData::gBitmapSize.setWidth( bmpWidth );    
      I3StateCheckBoxData::gBitmapSize.setHeight( pbmpData.cy );
   }

   ISize size = displaySize(text());
   unsigned long ulWidth = size.width();
   unsigned long i=textLength();

   if (i==0)
     ulWidth += ((bmpWidth/4) + 2 + (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   else
     ulWidth += (ulWidth/i + (bmpWidth/4) + 2 +
                                    (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   unsigned long ulHeight = size.height() +
                                     (2 * IQUERYSYSVALUE(SV_CYBORDER));
   return ISize(ulWidth, ulHeight);
}
