/*******************************************************************************
* FILE NAME: iaccelky.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iaccelky.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINACCELERATORS          // For struct _ACCEL.
  #define INCL_WINFRAMEMGR              // For SC_xxx.
  #define INCL_WININPUT                 // For VK_xxx.
  #include <iwindefs.h>
  }

// Flags used to map to/from an ACCEL structure.

#include <iaccelky.hpp>
#include <iexcept.hpp>
#include <iframe.hpp>
#include <ihandle.hpp>
#include <ireslib.hpp>
#include <istrenum.hpp>
#include <istring.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IACCELKY_CPP_
  #include <ipagetun.h>
#endif

// Private implementation classes.
/*------------------------------------------------------------------------------
| IAcceleratorKeyData                                                          |
------------------------------------------------------------------------------*/
class IAcceleratorKeyData : public IBase {
typedef IBase
  Inherited;
public:
  IAcceleratorKeyData ( );
  IAcceleratorKeyData ( const ACCEL&               accel );
  IAcceleratorKeyData ( const IAcceleratorKeyData& keyData );
IAcceleratorKeyData
 &operator=           ( const IAcceleratorKeyData& keyData );
Boolean
  operator==          ( const IAcceleratorKeyData& keyData ) const;

unsigned short
  fFlags;
unsigned short
  fKey;
ICommand::CommandId
  fCmd;
Boolean
  fFromACCEL;
unsigned long
  fUniqueKey;
};  // IAcceleratorKeyData

IAcceleratorKeyData::IAcceleratorKeyData ( )
  : IAcceleratorKeyData::Inherited(),
    fFlags( 0 ), fKey( 0 ), fCmd( 0 ),
    fFromACCEL( false ),
    fUniqueKey( 0 )
{ }

IAcceleratorKeyData::IAcceleratorKeyData ( const ACCEL& accel )
  : IAcceleratorKeyData::Inherited(),
    fFlags( accel.fs ),
    fKey( accel.key ),
    fCmd( accel.cmd ),
    fFromACCEL( true ),
    fUniqueKey( 0 )          // Set by IAcceleratorKey.
{
}

IAcceleratorKeyData::IAcceleratorKeyData ( const IAcceleratorKeyData& keyData )
  : IAcceleratorKeyData::Inherited(),
    fFlags( keyData.fFlags ), fKey( keyData.fKey ), fCmd( keyData.fCmd ),
    fFromACCEL( keyData.fFromACCEL ),
    fUniqueKey( keyData.fUniqueKey )
{ }

IAcceleratorKeyData&
  IAcceleratorKeyData::operator= ( const IAcceleratorKeyData& keyData )
{
  if ( this != &keyData )
  {           // Not copying self.
     this->fFlags = keyData.fFlags;
     this->fKey = keyData.fKey;
     this->fCmd = keyData.fCmd;
     this->fFromACCEL = keyData.fFromACCEL;
     this->fUniqueKey = keyData.fUniqueKey;
  }
  return *this;
}

IBase::Boolean
  IAcceleratorKeyData::operator== ( const IAcceleratorKeyData& keyData ) const
{
  return ( this->fFlags == keyData.fFlags  &&
           this->fKey == keyData.fKey  &&
           this->fCmd == keyData.fCmd );
}

/*------------------------------------------------------------------------------
| ::key                                                                        |
|                                                                              |
| Global function for key set of IAcceleratorKey objects.  Allow only one      |
| accelerator table entry for a given key and modifier.                        |
------------------------------------------------------------------------------*/
unsigned long const& key ( IAcceleratorKey const& item )
{
  return item.fData->fUniqueKey;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::IAcceleratorKey                                             |
|                                                                              |
| Constructor for virtual key accelerator.                                     |
------------------------------------------------------------------------------*/
IAcceleratorKey::IAcceleratorKey ( IKey::VirtualKey         virtualKey,
                                   const IKey::KeyModifier& modifier,
                                   ICommand::ActionType     actionType,
                                   ICommand::CommandId      commandId )
  : IAcceleratorKey::Inherited(),
    fData( new IAcceleratorKeyData() )
{
  (*this)
   .setKey( virtualKey, modifier );

  switch ( actionType )
  {
     case (ICommand::applicationCommand):
        this->setCommand( commandId );
        break;
     case (ICommand::systemCommand):
        this->setSystemCommand( commandId );
        break;
     case (ICommand::help):
        this->setHelpKey();
        break;
     default:
        ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                            IBaseErrorInfo::invalidParameter,
                            IException::recoverable );
        break;
  }
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::IAcceleratorKey                                             |
|                                                                              |
| Constructor for character key accelerator.                                   |
------------------------------------------------------------------------------*/
IAcceleratorKey::IAcceleratorKey ( const IString&           commandKey,
                                   const IKey::KeyModifier& modifier,
                                   ICommand::ActionType     actionType,
                                   ICommand::CommandId      commandId )
  : IAcceleratorKey::Inherited(),
    fData( new IAcceleratorKeyData() )
{
  (*this)
   .setKey( commandKey, modifier );

  switch ( actionType )
  {
     case (ICommand::applicationCommand):
        this->setCommand( commandId );
        break;
     case (ICommand::systemCommand):
        this->setSystemCommand( commandId );
        break;
     case (ICommand::help):
        this->setHelpKey();
        break;
     default:
        ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                            IBaseErrorInfo::invalidParameter,
                            IException::recoverable );
        break;
  }
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::IAcceleratorKey                                             |
|                                                                              |
| Constructor for ACCEL structure.                                             |
| Note: For Windows, ACCEL has no way to hold command type information.        |
------------------------------------------------------------------------------*/
IAcceleratorKey::IAcceleratorKey ( const ACCEL& accelerator )
  : IAcceleratorKey::Inherited(),
    fData( new IAcceleratorKeyData( accelerator ))
{
  this->fData->fUniqueKey = IAcceleratorKey::uniqueKeyFor( *this );
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::IAcceleratorKey                                             |
|                                                                              |
| Copy constructor.                                                            |
------------------------------------------------------------------------------*/
IAcceleratorKey::IAcceleratorKey ( const IAcceleratorKey& key )
  : IAcceleratorKey::Inherited(),
    fData( new IAcceleratorKeyData( *( key.fData ) ))
{ }

/*------------------------------------------------------------------------------
| IAcceleratorKey::operator=                                                   |
|                                                                              |
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::operator= ( const IAcceleratorKey& key )
{
  if ( this != &key )
  {           // Not copying self.
     *(this->fData) = *(key.fData);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::~IAcceleratorKey                                            |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IAcceleratorKey::~IAcceleratorKey ( )
{
  delete this->fData;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::operator==                                                  |
|                                                                              |
| Equality check.                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IAcceleratorKey::operator== ( const IAcceleratorKey& key ) const
{
  return ( *(this->fData) == *(key.fData) );
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::isLike                                                      |
|                                                                              |
| Check for same key and modifier.                                             |
------------------------------------------------------------------------------*/
IBase::Boolean IAcceleratorKey::isLike ( const IAcceleratorKey& key ) const
{
  return ( ::key( *this ) == ::key( key ) );
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setKey                                                      |
|                                                                              |
| Assign character accelerator key.                                            |
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::setKey ( const IString&           accelKey,
                                           const IKey::KeyModifier& modifier )
{
  char accelChar[ 2 ];
  accelChar[ 0 ] = accelKey[ 1 ];
  if ( accelKey.charType( 1 ) == IStringEnum::sbcs )
  {
     accelChar[ 1 ] = 0;
  }
  else
  {
     accelChar[ 1 ] = accelKey[ 2 ];
  }
  this->fData->fKey = *(unsigned short *)accelChar;

  unsigned long flags = (unsigned long)this->fData->fFlags;
  flags &= (unsigned long)~(AF_VIRTUALKEY | AF_SCANCODE);
  flags |= AF_CHAR;
  this->fData->fFlags = (unsigned short)flags;

  this->setKeyModifier( modifier );
  this->fData->fFromACCEL = false;
  this->fData->fUniqueKey = IAcceleratorKey::uniqueKeyFor( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setKey                                                      |
|                                                                              |
| Assign virtual key accelerator key.                                          |
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::setKey ( IKey::VirtualKey         virtualKey,
                                           const IKey::KeyModifier& modifier )
{
  IASSERTPARM( virtualKey != IKey::kNoKey );

  this->fData->fKey = (unsigned short)virtualKey;

  unsigned long flags = (unsigned long)this->fData->fFlags;
  flags &= (unsigned long)~(AF_CHAR | AF_SCANCODE);
  flags |= AF_VIRTUALKEY;
  this->fData->fFlags = (unsigned short)flags;

  this->setKeyModifier( modifier );
  this->fData->fFromACCEL = false;
  this->fData->fUniqueKey = IAcceleratorKey::uniqueKeyFor( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::character                                                   |
|                                                                              |
| Return the character key accelerator key.                                    |
------------------------------------------------------------------------------*/
IString IAcceleratorKey::character ( ) const
{
  char accelChar[ 3 ];
  *(unsigned short *)accelChar = this->fData->fKey;
  accelChar[ 2 ] = 0;
  return ( this->fData->fFlags & AF_CHAR ) ? IString( accelChar ) : IString();
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::virtualKey                                                  |
|                                                                              |
| Return the virtual key accelerator key.                                      |
------------------------------------------------------------------------------*/
IKey::VirtualKey IAcceleratorKey::virtualKey ( ) const
{
  return ( this->fData->fFlags & AF_VIRTUALKEY ) ?
            this->fData->fKey :
            IKey::kNoKey;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::keyModifier                                                 |
|                                                                              |
| Return the keyboard modifier for the accelerator key.                        |
------------------------------------------------------------------------------*/
IKey::KeyModifier IAcceleratorKey::keyModifier ( ) const
{
  IKey::KeyModifier modifier = IKey::noModifier;
  unsigned long flags = (unsigned long)this->fData->fFlags;

  if ( flags & AF_CONTROL )
  {
     modifier |= IKey::ctrl;
  }
  if ( flags & AF_ALT )
  {
     modifier |= IKey::alt;
  }
  if ( flags & AF_SHIFT )
  {
     modifier |= IKey::shift;
  }

  // Remove the default value if some modifier is being used.
  if ( modifier != IKey::noModifier )
  {
     modifier &= ~IKey::noModifier;
  }

  return modifier;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setCommand                                                  |
|                                                                              |
| Assign the command that the accelerator key will run.                        |
#ifdef IC_WIN
| Note: The command ID should be below SC_SIZE.                                |
#endif
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::setCommand ( unsigned long commandId )
{
  // Set the flags to show this key will run an application command.
  unsigned long flags = this->fData->fFlags;
  flags &= (unsigned long)~(AF_SYSCOMMAND | AF_HELP);
  this->fData->fFlags = (unsigned short)flags;

  // Now store the ID of the application command.
  this->fData->fCmd = commandId;

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setSystemCommand                                            |
|                                                                              |
| Assign the system command that the accelerator key will run.                 |
#ifdef IC_WIN
| Note: The command ID should be SC_SIZE or greater in value.                  |
#endif
------------------------------------------------------------------------------*/
IAcceleratorKey&
  IAcceleratorKey::setSystemCommand ( unsigned long systemCommand )
{
  // Set the flags to show this key will run a system command.
  unsigned long flags = this->fData->fFlags;
  flags &= (unsigned long)~AF_HELP;
  flags |= AF_SYSCOMMAND;
  this->fData->fFlags = (unsigned short)flags;

  // Now store the ID of the system command.
  this->fData->fCmd = systemCommand;

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setHelpKey                                                  |
|                                                                              |
| Specify that this key will invoke a help request.                            |
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::setHelpKey ( )
{
  // Set the flags to show this key will invoke help.
  unsigned long flags = this->fData->fFlags;
  flags &= (unsigned long)~AF_SYSCOMMAND;
  flags |= AF_HELP;
  this->fData->fFlags = (unsigned short)flags;

  this->fData->fCmd = 0;

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::actionType                                                  |
|                                                                              |
| Return the type of event that the accelerator key will run.                  |
------------------------------------------------------------------------------*/
ICommand::ActionType IAcceleratorKey::actionType ( ) const
{
  ICommand::ActionType action = ICommand::applicationCommand;
  unsigned long flags = (unsigned long)this->fData->fFlags;

  if ( flags & AF_HELP )
  {
     action = ICommand::help;
  }
  else if ( flags & AF_SYSCOMMAND )
  {
     action = ICommand::systemCommand;
  }

  return action;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::commandId                                                   |
|                                                                              |
| Return the identifier of the command that the accelerator key will run.      |
------------------------------------------------------------------------------*/
unsigned long IAcceleratorKey::commandId ( ) const
{
  unsigned long cmdId = this->fData->fCmd;
  if ( this->actionType() == ICommand::help )
  {
     cmdId = 0;
  }
  return cmdId;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::asACCEL                                                     |
|                                                                              |
| Convert to ACCEL structure.                                                  |
------------------------------------------------------------------------------*/
ACCEL IAcceleratorKey::asACCEL ( ) const
{
  ACCEL accel;
  accel.fs = this->fData->fFlags;
  accel.key = this->fData->fKey;
  accel.cmd = (unsigned short)this->fData->fCmd;


  return accel;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::uniqueKeyFor                                                |
|                                                                              |
| Convert accelerator key object to a unique key value for a keyed collection. |
------------------------------------------------------------------------------*/
unsigned long
  IAcceleratorKey::uniqueKeyFor ( const IAcceleratorKey& key )
{
  ACCEL accelData = key.asACCEL();
  unsigned long
    value = accelData.key,
    flags = ( accelData.fs &
               ( AF_CHAR | AF_VIRTUALKEY | AF_SCANCODE |
                 AF_SHIFT | AF_CONTROL | AF_ALT | AF_LONEKEY ) );
  value |= ( flags << 16 );
  return value;
}

/*------------------------------------------------------------------------------
| IAcceleratorKey::setKeyModifier                                              |
|                                                                              |
| Assign a key modifier for the accelerator.                                   |
------------------------------------------------------------------------------*/
IAcceleratorKey& IAcceleratorKey::setKeyModifier ( const IKey::KeyModifier& modifier )
{
  unsigned long flags = this->fData->fFlags;
  flags &= (unsigned long)~( AF_LONEKEY | AF_CONTROL | AF_ALT | AF_SHIFT );

  if ( modifier & IKey::ctrl )
  {
     flags |= AF_CONTROL;
  }
  if ( modifier & IKey::alt )
  {
     flags |= AF_ALT;
  }
  if ( modifier & IKey::shift )
  {
     flags |= AF_SHIFT;
  }

  this->fData->fFlags = (unsigned short)flags;

  return *this;
}
