/*******************************************************************************
* FILE NAME: iacceltb.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iacceltb.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINACCELERATORS          // For WinQueryAccelTable, etc.
  #define INCL_WINCOUNTRY               // For WinQueryCp.
  #include <iwindefs.h>
  }

#include <iacceltb.hpp>
#include <iaccelky.hpp>
#include <ihandle.hpp>
#include <ikeyset.h>
#include <ireslib.hpp>
#include <ithread.hpp>
#include <iwindow.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IACCELTB_CPP_
  #include <ipagetun.h>
#endif

// Platform-specific macro definitions.
  #define IQUERYACCELTABLE(hwnd) \
      (WinQueryAccelTable( IThread::current().anchorBlock(), hwnd ))
  #define ICOPYACCELTABLE(haccel,pbuf,size) \
      (WinCopyAccelTable( haccel, pbuf, size ))
  #define ICREATEACCELTABLE(pbuf,count) \
      (WinCreateAccelTable( IThread::current().anchorBlock(), pbuf ))

// Private implementation classes.

typedef IKeySet< IAcceleratorKey, unsigned long >
  IAcceleratorKeySet;

class IAcceleratorTableData : public IBase {
typedef IBase
  Inherited;
public:
  IAcceleratorTableData ( );
 ~IAcceleratorTableData ( );
  // Compiler-generated copy ctor and assignment operators are ok.
IWindowHandle
  fAccelWindow;
IAcceleratorKeySet
  fKeySet;
};  // IAcceleratorTableData

IAcceleratorTableData::IAcceleratorTableData ( )
  : IAcceleratorTableData::Inherited(),
    fAccelWindow( 0 ),
    fKeySet( 30 )
{ }

IAcceleratorTableData::~IAcceleratorTableData ( )
{ }

class IAcceleratorTableCursorData : public IBase {
typedef IBase
  Inherited;
public:
  IAcceleratorTableCursorData ( const IAcceleratorKeySet* keySet );
 ~IAcceleratorTableCursorData ( );
IAcceleratorKeySet::Cursor
  fCursor;
};  // IAcceleratorTableCursorData

IAcceleratorTableCursorData::IAcceleratorTableCursorData
                                 ( const IAcceleratorKeySet* keySet )
  : IAcceleratorTableCursorData::Inherited(), fCursor( *keySet )
{ }

IAcceleratorTableCursorData::~IAcceleratorTableCursorData ( )
{ }

/*------------------------------------------------------------------------------
| IAcceleratorTable::IAcceleratorTable                                         |
|                                                                              |
| Constructor to create an empty accelerator table, or one that wrappers the   |
| accelerator table for the system queue.                                      |
------------------------------------------------------------------------------*/
IAcceleratorTable::IAcceleratorTable ( Boolean systemQueueKeys )
  : IAcceleratorTable::Inherited(),
    fData( new IAcceleratorTableData() )
{
  if ( systemQueueKeys )
  {
     IThread::current().initializeGUI();
     IAccelTblHandle accelTblHandle = IQUERYACCELTABLE( 0 );
     this->initialize( accelTblHandle );
  }
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::IAcceleratorTable                                         |
|                                                                              |
| Constructor to create an accelerator table object from a frame window's      |
| accelerator table.                                                           |
------------------------------------------------------------------------------*/
IAcceleratorTable::IAcceleratorTable ( const IWindow* window )
  : IAcceleratorTable::Inherited(),
    fData( new IAcceleratorTableData() )
{
  if ( window )
  {
     IAccelTblHandle accelTbl = window->acceleratorHandle();
     this->initialize( accelTbl );
     fData->fAccelWindow = window->handle();
  }
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::IAcceleratorTable                                         |
|                                                                              |
| Constructor to create an accelerator table object from the handle of an      |
| existing accelerator table.                                                  |
------------------------------------------------------------------------------*/
IAcceleratorTable::IAcceleratorTable ( const IAccelTblHandle& accelTblHandle )
  : IAcceleratorTable::Inherited(),
    fData( new IAcceleratorTableData() )
{
  this->initialize( accelTblHandle );
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::IAcceleratorTable                                         |
|                                                                              |
| Constructor to create an accelerator table object from an accelerator table  |
| loaded from a resource library.                                              |
------------------------------------------------------------------------------*/
IAcceleratorTable::IAcceleratorTable ( const IResourceId& resId )
  : IAcceleratorTable::Inherited(),
    fData( new IAcceleratorTableData() )
{
  IAccelTblHandle accelTbl =
     resId.resourceLibrary().loadAccelTable( resId.id() );
  this->initialize( accelTbl );
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::IAcceleratorTable                                         |
|                                                                              |
| Copy constructor to create an accelerator table object from an existing      |
| accelerator table object.                                                    |
------------------------------------------------------------------------------*/
IAcceleratorTable::IAcceleratorTable ( const IAcceleratorTable& accelTbl )
  : IAcceleratorTable::Inherited(),
    fData( new IAcceleratorTableData() )
{
  *(this->fData) = *(accelTbl.fData);
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::operator=                                                 |
|                                                                              |
| Assignment operator to replace the contents of an accelerator table object   |
| with another.                                                                |
------------------------------------------------------------------------------*/
IAcceleratorTable&
  IAcceleratorTable::operator= ( const IAcceleratorTable& accelTbl )
{
  if ( this != &accelTbl )
  {           // Not copying self.
     *(this->fData) = *(accelTbl.fData);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::~IAcceleratorTable                                        |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IAcceleratorTable::~IAcceleratorTable ( )
{
  delete this->fData;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::keyAt                                                     |
|                                                                              |
| Return the accelerator key object at the cursor position.                    |
------------------------------------------------------------------------------*/
IAcceleratorKey IAcceleratorTable::keyAt ( const Cursor& cursor )
{
  return cursor.element();
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::removeKeyAt                                               |
|                                                                              |
| Removes the accelerator key object at the cursor position from the table's   |
| list of accelerator keys.                                                    |
------------------------------------------------------------------------------*/
IAcceleratorTable&
  IAcceleratorTable::removeKeyAt ( IAcceleratorTable::Cursor& cursor,
                                   Boolean updateWindow )
{
  this->fData->fKeySet.removeAt( cursor.fData->fCursor );

  if ( updateWindow )
  {
     this->updateWindow();
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::removeAllKeys                                             |
|                                                                              |
| Empties the list of keys.                                                    |
------------------------------------------------------------------------------*/
IAcceleratorTable&
  IAcceleratorTable::removeAllKeys ( Boolean updateWindow )
{
  this->fData->fKeySet.removeAll();

  if ( updateWindow )
  {
     this->updateWindow();
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::replaceKeyAt                                              |
|                                                                              |
| Replaces the accelerator key object at the cursor position with the          |
| specified accelerator key.                                                   |
------------------------------------------------------------------------------*/
IAcceleratorTable&
  IAcceleratorTable::replaceKeyAt ( const IAcceleratorTable::Cursor& cursor,
                                    const IAcceleratorKey& newKey,
                                    Boolean updateWindow )
{
  this->fData->fKeySet.replaceAt( cursor.fData->fCursor, newKey );

  if ( updateWindow )
  {
     this->updateWindow();
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::addKey                                                    |
|                                                                              |
| Adds an accelerator key object to the collection of keys.                    |
------------------------------------------------------------------------------*/
IBase::Boolean IAcceleratorTable::addKey ( const IAcceleratorKey& newKey,
                                           Boolean updateWindow )
{
  Boolean added = this->fData->fKeySet.add( newKey );

  if ( updateWindow )
  {           // Update whether or not key was added.
     this->updateWindow();
  }

  return added;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::addOrReplaceKey                                           |
|                                                                              |
| Adds an accelerator key object to the collection of keys, if the collection  |
| does not already contain an entry with the same key and modifier.  Otherwise |
| replaces the similar entry.                                                  |
------------------------------------------------------------------------------*/
IBase::Boolean
  IAcceleratorTable::addOrReplaceKey ( const IAcceleratorKey& newKey,
                                       Boolean updateWindow )
{
  Boolean added = this->fData->fKeySet.addOrReplaceElementWithKey( newKey );

  if ( updateWindow )
  {           // Update whether key was added or replaced.
     this->updateWindow();
  }

  return added;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::removeKeyLike                                             |
|                                                                              |
| Removes the accelerator key object with the same key and modifier values as  |
| the specified key (the entry in the collection with the same unique key).    |
------------------------------------------------------------------------------*/
IBase::Boolean
  IAcceleratorTable::removeKeyLike( const IAcceleratorKey& key,
                                    Boolean updateWindow )
{
  Boolean removed = this->fData->fKeySet.removeElementWithKey( ::key( key ) );

  if ( updateWindow )
  {           // Update whether or not the key was removed.
     this->updateWindow();
  }

  return removed;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::containsKeyLike                                           |
|                                                                              |
| Returns if the collection contains an accelerator key object with the same   |
| key and modifier data.                                                       |
------------------------------------------------------------------------------*/
IBase::Boolean
  IAcceleratorTable::containsKeyLike ( const IAcceleratorKey& key ) const
{
  return this->fData->fKeySet.containsElementWithKey( ::key( key ) );
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::locateKeyLike                                             |
|                                                                              |
| Sets the cursor to the accelerator key object with the same key and modifier |
| data in the collection.                                                      |
------------------------------------------------------------------------------*/
Boolean
  IAcceleratorTable::locateKeyLike ( const IAcceleratorKey& key,
                                     IAcceleratorTable::Cursor& cursor ) const
{
  return this->fData->fKeySet.locateElementWithKey( ::key( key ),
                                                    cursor.fData->fCursor );
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::keyCount                                                  |
|                                                                              |
| Returns the number of accelerator keys in the table.                         |
------------------------------------------------------------------------------*/
unsigned long IAcceleratorTable::keyCount ( ) const
{
  return this->fData->fKeySet.numberOfElements();
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::initialize                                                |
|                                                                              |
| Sets the object with the accelerators in the specified accelerator table.    |
------------------------------------------------------------------------------*/
IAcceleratorTable&
  IAcceleratorTable::initialize ( const IAccelTblHandle& accelTblHandle )
{
  if ( ! accelTblHandle )
  {                     // Empty table.
     this->fData->fKeySet.removeAll();  // Empty out any existing entries.
  }
  else
  {                     // Non-empty table.
     unsigned long tableSize = ICOPYACCELTABLE( accelTblHandle, 0, 0 );
     if ( tableSize == 0 )
     {                  // Assume error, rather than no accelerators.
        ITHROWGUIERROR( "ICOPYACCELTABLE" );
     }

     // Get the entries from the existing accelerator table.
     // Note: In PM, tableSize is the size of the table in bytes.
     //       In Windows, it is the number of entries in the table.
     PACCELTABLE accelTableBuf = (PACCELTABLE) new char[ tableSize ];
     unsigned long copied =
        ICOPYACCELTABLE( accelTblHandle, accelTableBuf, tableSize );
     if ( ! copied )
     {                  // Error occurred.
        ITHROWGUIERROR( "ICOPYACCELTABLE" );
     }

     // Copy the entries into the collection of IAcceleratorKey objects.
     this->fData->fKeySet.removeAll();   // First empty original entries.
                // (No need to delete entries, since removeAll deletes them.)
     unsigned long itemCount = accelTableBuf->cAccel;
     ACCEL *accelItem = &accelTableBuf->aaccel[ 0 ];
     while ( itemCount )
     {
        IAcceleratorKey key( *accelItem );
        this->fData->fKeySet.add( key );
        accelItem++;
        itemCount--;
     }

  delete [] (char*)accelTableBuf;
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::handle                                                    |
|                                                                              |
| Create and return the accelerator table handle for the accelerator keys.     |
| The caller is responsible for destroying the handle.                         |
------------------------------------------------------------------------------*/
IAccelTblHandle IAcceleratorTable::handle ( ) const
{
  IAccelTblHandle accelTblHandle( 0 );
  unsigned long
    keyCount = this->keyCount();
  if ( keyCount )
  {                // Table has some entries in it.
     // Fill in the first part of the table structure.
     unsigned long
       tableSize = sizeof( ACCELTABLE ) +
                   ( sizeof(ACCEL) * ( keyCount ? keyCount - 1 : 0 ));
     PACCELTABLE accelTableBuf = (PACCELTABLE) new char[ tableSize ];
     accelTableBuf->cAccel = (unsigned short)keyCount;
     accelTableBuf->codepage =
        (unsigned short)WinQueryCp( IThread::current().messageQueue() );
     ACCEL *accelItem = accelTableBuf->aaccel;

     // Copy in the array of key structures.
     IAcceleratorTable::Cursor cursor( *(IAcceleratorTable*)this );
     for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
     {
         IAcceleratorKey key = IAcceleratorTable::keyAt( cursor );
         *accelItem = key.asACCEL();
         accelItem++;
     }

     accelTblHandle = ICREATEACCELTABLE( accelTableBuf, keyCount );
     // The caller is assumed to destroy this accelerator table handle.

     // Clean-up.
     delete [] (char*)accelTableBuf;
  }

  return accelTblHandle;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::destroyHandle                                             |
|                                                                              |
| Destroy the specified accelerator table handle.  Presumably, but not         |
| necessarily, this handle was returned by IAcceleratorTable::handle.          |
------------------------------------------------------------------------------*/
void IAcceleratorTable::destroyHandle ( IAccelTblHandle& accelTblHandle )
{
  // Ignore any errors, since the application could have already
  // found and destroyed this using system APIs, and we are trying
  // to do some behind-the-scenes cleanup as well.
    IDESTROYACCELTABLE( accelTblHandle );
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::updateWindow                                              |
|                                                                              |
| Update the associated window with the current contents of this object.       |
------------------------------------------------------------------------------*/
IBase::Boolean
  IAcceleratorTable::updateWindow ( )
{
  Boolean didUpdate = false;
  IWindowHandle accelHwnd( fData->fAccelWindow );

  if ( accelHwnd.isValid() )
  {
     IWindow* accelWindow = IWindow::windowWithHandle( accelHwnd );
     if ( accelWindow )
     {        // Have a window to update changes to.
        accelWindow->setAcceleratorTable( this );
        didUpdate = true;
     }
  }
  else
  {           // Clear out the invalid window handle.
     fData->fAccelWindow = 0;
  }

  return didUpdate;
}

/*------------------------------------------------------------------------------
| IAcceleratorTable::Cursor functions                                          |
------------------------------------------------------------------------------*/

IAcceleratorTable::Cursor :: Cursor ( IAcceleratorTable& accelTbl )
  : IAcceleratorTable::Cursor::Inherited(),
    fData( new IAcceleratorTableCursorData( &(accelTbl.fData->fKeySet) ))
{ }

IAcceleratorTable::Cursor :: ~Cursor ( )
{
  delete this->fData;
}

IBase::Boolean IAcceleratorTable::Cursor :: setToFirst ( )
{
  return this->fData->fCursor.setToFirst();
}

IBase::Boolean IAcceleratorTable::Cursor :: setToNext ( )
{
  return this->fData->fCursor.setToNext();
}

IBase::Boolean IAcceleratorTable::Cursor :: isValid ( ) const
{
  return this->fData->fCursor.isValid();
}

void IAcceleratorTable::Cursor :: invalidate ( )
{
  this->fData->fCursor.invalidate();
}

IAcceleratorKey IAcceleratorTable::Cursor :: element ( ) const
{
  IAcceleratorKey accelKey = this->fData->fCursor.element();
  return accelKey;
}
