/*******************************************************************************
* FILE NAME: icheckbx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icheckbx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#ifndef _IBASE_
  #include <ibase.hpp>
#endif
#define INCL_WINBUTTONS
#define INCL_WINERRORS
#define INCL_WINPOINTERS
#define INCL_GPIBITMAPS
#define INCL_WINSYS
#define INCL_GPIPRIMITIVES
#include <iwindefs.h>

#ifndef _ICHECKBX_
  #include <icheckbx.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFONT_
  #include <ifont.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICHECKBX_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public check box styles                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ICheckBox::Style
  ICheckBox::autoSelect        ( 0, IBS_AUTOCHECKBOX ),
  ICheckBox::classDefaultStyle ( WS_VISIBLE, IBS_AUTOCHECKBOX );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  ICheckBox::Style
        ICheckBox::currentDefaultStyle ( WS_VISIBLE, IBS_AUTOCHECKBOX );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Class:  ICheckBoxData                                                        |
|                                                                              |
| Store private data used by ICheckBox.                                        |
------------------------------------------------------------------------------*/
class ICheckBoxData : public IBase {        
public:                                     
static ISize                                
  gBitmapSize;                              
}; // ICheckBoxData                         

#pragma data_seg(ICLNonConst)               
  ISize ICheckBoxData::gBitmapSize ( 0, 0 );
#pragma data_seg()                          


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to create a check box on a standard window.                      |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(unsigned long ulId,
                       IWindow* pwndParent,
                       IWindow* pwndOwner,
                       const IRectangle& rectInit,
                       const Style& chbsStyle)
                     : fCheckBoxData( 0 )
{

  IASSERTPARM(pwndParent!=0);          //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | chbsStyle.asExtendedUnsignedLong() );

  //Create the check box control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle whCheckBox =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( chbsStyle ),
                      WC_BUTTON,
                      pwndParent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0 );
  startHandlingEventsFor(whCheckBox);
}


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(unsigned long ulId,
                       IWindow* pdlgwndParent)
                     : fCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to instantiate an object from an existing check box control.     |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(const IWindowHandle& wh)
                     : fCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(wh);
}


/*------------------------------------------------------------------------------
| ICheckBox::~ICheckBox                                                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
ICheckBox :: ~ICheckBox()
{
  // delete fCheckBoxData;
}


/*------------------------------------------------------------------------------
| ICheckBox::defaultStyle                                                      |
|                                                                              |
| Return the default style for new check box objects.                          |
------------------------------------------------------------------------------*/
ICheckBox::Style  ICheckBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| ICheckBox::setDefaultStyle                                                   |
------------------------------------------------------------------------------*/
void  ICheckBox :: setDefaultStyle(const Style& chbsStyle)
{
  currentDefaultStyle = chbsStyle;
}


/*------------------------------------------------------------------------------
| ICheckBox::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ICheckBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // Obtain the style from the class (ISettingButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // BS_ styles in the lower 4 bits of the style overlap.  Therefore, map
    // our extended style bits, back to these values.
    unsigned long ulTmp = extendedStyle() & IBS_EXTCHECKBX;

    // Handle overlapping styles
    if (ulTmp & IBS_AUTOCHECKBOX)
      ulStyle |= BS_AUTOCHECKBOX;
    else
      ulStyle |= BS_CHECKBOX;

  }


  return( ulStyle );
}


/*------------------------------------------------------------------------------
| ICheckBox::enableAutoSelect                                                  |
|                                                                              |
| Sets or removes the autoSelect style for the check box control.              |
------------------------------------------------------------------------------*/
ICheckBox&  ICheckBox :: enableAutoSelect ( Boolean enable )
{
  unsigned long ulOldStyle = this->style();
  unsigned long ulStyle = ( ulOldStyle & (unsigned long)(~BS_PRIMARYSTYLES) );
                                       // Use masking style from pmwin.h.

  if ( enable )
  {
     ulStyle |= BS_AUTOCHECKBOX;
  }
  else
  {
     ulStyle |= BS_CHECKBOX;
  }

  if ( ulStyle != ulOldStyle )
  {
     this->setStyle( ulStyle );
     this->refresh();                  // Force repaint.
  }
  return *this;
}


/*------------------------------------------------------------------------------
| ICheckBox::disableAutoSelect                                                 |
------------------------------------------------------------------------------*/
ICheckBox&  ICheckBox :: disableAutoSelect()
{
  enableAutoSelect(false);
  return *this;
}


/*------------------------------------------------------------------------------
| ICheckBox::isAutoSelect                                                      |
------------------------------------------------------------------------------*/
Boolean  ICheckBox :: isAutoSelect() const
{
  return( (style() & BS_AUTOCHECKBOX & (unsigned long)(~BS_CHECKBOX)) ? true : false );
}


/*------------------------------------------------------------------------------
| ICheckBox::calcMinimumSize                                                   |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The minimum size   |
| is based on the maximum height of the font and the font width of the         |
| check box text (plus the size of the check box itself).                      |
------------------------------------------------------------------------------*/
ISize  ICheckBox :: calcMinimumSize() const
{
   long bmpWidth = ICheckBoxData::gBitmapSize.width();
   if ( bmpWidth == 0 )                               
   {            // Find size of check box bitmap.     
      BITMAPINFOHEADER pbmpData;
      pbmpData.cbFix=sizeof(BITMAPINFOHEADER);
      HBITMAP hbit = (HBITMAP) IGETSYSBITMAP(SBMP_CHECKBOXES);
      IQUERYBITMAPPARAMETERS(hbit, &pbmpData);    //Retrieve size of check
      IDELETEBITMAP(hbit);                        //box bitmap

      bmpWidth = pbmpData.cx;                                   
      ICheckBoxData::gBitmapSize.setWidth( bmpWidth );          
      ICheckBoxData::gBitmapSize.setHeight( pbmpData.cy );      
   }                                                         

   ISize size = displaySize(text());
   unsigned long ulWidth = size.width();
   unsigned long i=textLength();

   if (i==0)
     ulWidth += ((bmpWidth/4) + 2 + (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   else
     ulWidth += (ulWidth/i + (bmpWidth/4) + 2 +
                                    (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   unsigned long ulHeight = size.height() +
                                     (2 * IQUERYSYSVALUE(SV_CYBORDER));
   return ISize(ulWidth, ulHeight);
}
