/*******************************************************************************
* FILE NAME: iclipbrd.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iclipbrd.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINATOM
#define INCL_DOSMEMMGR
#define INCL_WINCLIPBOARD
#define INCL_GPI
extern "C" {
   #include <iwindefs.h>
     #include <bsememf.h>
}
#include <string.h>
#include <iclipbrd.hpp>
#include <ihandle.hpp>
#include <iexcept.hpp>
#include <ithread.hpp>
#include <istring.hpp>
#include <igbitmap.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICLIPBRD_CPP_
  #include <ipagetun.h>
#endif

/////////////////// Known Problems
// 1)  For some reason there is a problem with delayed rendering within
// a single process.  It looks like, the setData() of the delayed renderer
// gives the data to the shell and never gets it back on the data() side.


/*------------------------------------------------------------------------------
| Macros for Win32 functions                                                   |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| Macros for PM functions                                                      |
------------------------------------------------------------------------------*/
  /* The following functions not directly supported:         */
  /*    WinSetClipbrdData, WinQueryClipbrdFormat             */
  #define IOPENCLIPBOARD(hab, hwnd)           (WinOpenClipbrd(hab))
  #define ICLOSECLIPBOARD(hab)                (WinCloseClipbrd(hab))
  #define IREGISTERCLIPBOARDFORMAT(psz)       (WinAddAtom(WinQuerySystemAtomTable(), psz))
  #define IFINDCLIPBOARDFORMAT(psz)           (WinFindAtom(WinQuerySystemAtomTable(), psz))
  #define IGETCLIPBOARDFORMATTEXT(fmt, psz, len) (WinQueryAtomName(WinQuerySystemAtomTable(),fmt,psz, len))
  #define IEMPTYCLIPBOARD(hab)                (WinEmptyClipbrd(hab))
  #define IGETCLIPBOARDOWNER(hab)             (WinQueryClipbrdOwner(hab))
  #define ISETCLIPBOARDVIEWER(hab, hwnd)      (WinSetClipbrdViewer(hab, hwnd))
  #define IGETCLIPBOARDVIEWER(hab)            (WinQueryClipbrdViewer(hab))
  #define IENUMCLIPBOARDFORMATS(hab, ulFormat) (WinEnumClipbrdFmts(hab, ulFormat))
  #define IGETCLIPBOARDDATA(hab, ulFormat )   (WinQueryClipbrdData(hab, ulFormat))



/*------------------------------------------------------------------------------
| Clipboard Data class                                                         |
------------------------------------------------------------------------------*/
class IClipboardData
{
public:
IClipboardData (const IWindowHandle& owner  = 0)
  : ownerHandle(owner),
    opened(false),
    autoOpened(false)
{ }
friend class IClipboard;

IWindowHandle
 owner ( )  { return ownerHandle; }

Boolean
 isOpen ( )  { return opened; }

// AutoOpen means that we opened the clipboard because we needed
// it open.  This is turned off when the user callse open.
Boolean
 isAutoOpen ( )  { return autoOpened; }

void
 setOpen  ( Boolean open = true)
 {
 opened = open;
 }

void
 setAutoOpen  ( Boolean open = true)
 {
 autoOpened = open;
 }

void
 setOwner ( const IWindowHandle& owner) { ownerHandle = owner; }


private:
IWindowHandle
  ownerHandle;
Boolean
  opened,
  autoOpened;
};



/*------------------------------------------------------------------------------
| Constant Data                                                                |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)

const char
          * const IClipboard::textFormat(SZFMT_TEXT),
          * const IClipboard::displayTextFormat(SZFMT_DSPTEXT),
          * const IClipboard::bitmapFormat(SZFMT_BITMAP),
          * const IClipboard::displayBitmapFormat(SZFMT_DSPBITMAP),
          * const IClipboard::metafileFormat(SZFMT_METAFILE),
          * const IClipboard::displayMetafileFormat(SZFMT_DSPMETAFILE);


#pragma data_seg()

/*------------------------------------------------------------------------------
| Static Data                                                                  |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
//  IWindowHandle IClipboardData::ownerHandle(0);
// Boolean       IClipboardData::opened(false);
#pragma data_seg()



/*------------------------------------------------------------------------------
| IClipboard::IClipboard                                                       |
|                                                                              |
| Private constructor - used for Enumeration clipboard formats.                |
------------------------------------------------------------------------------*/
IClipboard::IClipboard ( )
       : fClipboardData(new IClipboardData(0))
{
  IMODTRACE_DEVELOP("Clipboard::Clipboard(clipboardWindow)");
}


/*------------------------------------------------------------------------------
| IClipboard::IClipboard                                                       |
|                                                                              |
| Constructor to create a clipboard object with an owner.                      |
------------------------------------------------------------------------------*/
IClipboard::IClipboard ( const IWindowHandle& clipboardWindow)
       : fClipboardData(new IClipboardData(clipboardWindow))
{
  IMODTRACE_DEVELOP("Clipboard::Clipboard(clipboardWindow)");
  IASSERTPARM(clipboardWindow!=0 && clipboardWindow.isValid());
}

/*------------------------------------------------------------------------------
| IClipboard::~IClipboard                                                      |
|                                                                              |
| Delete a clipboard object.                                                   |
------------------------------------------------------------------------------*/
IClipboard::~IClipboard ( )
{
  IMODTRACE_DEVELOP("Clipboard::~Clipboard");
  if(fClipboardData->isOpen())
    close();
  delete fClipboardData;

}

/*------------------------------------------------------------------------------
| IClipboard::open                                                             |
|                                                                              |
| Open the clipboard object.                                                   |
| A window handle is required for Windows but not for OS/2.  For consistency,  |
| we have moved the requirement for specifying a window handle up to the       |
| constructor.  Its this handle that is used as the owner if an owner is       |
| necessary(empty).  Its used here to open the clipboard (on Windows).         |
------------------------------------------------------------------------------*/
IClipboard &IClipboard::open( )
{
  IMODTRACE_DEVELOP("Clipboard::open");

  if (!fClipboardData->isOpen())
  {
    Boolean successful = IOPENCLIPBOARD(
                             IThread::current().anchorBlock(),
                             fClipboardData->owner() );
    if (!successful)
       ITHROWGUIERROR("OpenClipboard");
    fClipboardData->setAutoOpen(false);
    fClipboardData->setOpen(true);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::autoOpen                                                         |
|                                                                              |
| Private function that opens the clipboard and turns the auto open bit on.    |
------------------------------------------------------------------------------*/
IClipboard &IClipboard::autoOpen( )
{
  IMODTRACE_DEVELOP("Clipboard::autoOpen");
  if (!fClipboardData->isOpen())
  {
    open( );
    fClipboardData->setAutoOpen(true);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::close                                                            |
|                                                                              |
| Close the clipboard object.                                                  |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::close( )
{
  IMODTRACE_DEVELOP("Clipboard::close");

  if (fClipboardData->isOpen())
  {
    Boolean successful = ICLOSECLIPBOARD(IThread::current().anchorBlock());
    if (!successful)
      ITHROWGUIERROR("CloseClipboard");
  }
  fClipboardData->setOpen(false);
  fClipboardData->setAutoOpen(false);
  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::empty                                                            |
|                                                                              |
| Empty the contents of the clipboard and establish a new clipboard owner. We  |
| use the owner that was supplied on the open call.                            |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::empty( )
{
  IMODTRACE_DEVELOP("Clipboard::empty");

   // Ensure the clipboard is open
   if (!fClipboardData->isOpen())
      autoOpen();

  Boolean successful = IEMPTYCLIPBOARD(IThread::current().anchorBlock());

  if(fClipboardData->isAutoOpen())
    close();

  if (!successful)
     ITHROWGUIERROR("EmptyClipboard");

  // Note: in Win32, the owner is automatically estabished on empty.  To keep
  // things the same, we also set the owner in PM
     successful = WinSetClipbrdOwner(
                            IThread::current().anchorBlock(),
                            fClipboardData->owner());
     if (!successful)
        ITHROWGUIERROR("WinSetClipboardOwner");

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::isOpen                                                           |
|                                                                              |
| Return true if the clipboard is open.                                        |
------------------------------------------------------------------------------*/
Boolean IClipboard::isOpen( ) const
{
  IMODTRACE_DEVELOP("Clipboard::isOpen");
  return fClipboardData->isOpen();
}


/*------------------------------------------------------------------------------
| IClipboard::setText                                                          |
|                                                                              |
| Make a copy of the passed text and place it on the clipboard by calling      |
| setData.                                                                     |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setText( const char* text)
{
  IMODTRACE_DEVELOP("Clipboard::setText");
  IASSERTPARM(text!=0);
  return setData(IClipboard::textFormat, text, strlen(text));
}

/*------------------------------------------------------------------------------
| IClipboard::setBitmap                                                        |
|                                                                              |
| Make a copy of the passed bitmap and place it on the clipboard by calling    |
| setHandle.                                                                   |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setBitmap( const IBitmapHandle& bitmap)
{
  IMODTRACE_DEVELOP("Clipboard::setBitmap");
  return setHandle(IClipboard::bitmapFormat, IGBitmap::copy(bitmap));
}




/*------------------------------------------------------------------------------
| IClipboard::hasText                                                          |
|                                                                              |
| Returns true if the clipboard has text in it (IClipboard::textFormat).       |
------------------------------------------------------------------------------*/
Boolean IClipboard::hasText( ) const
{
  IMODTRACE_DEVELOP("Clipboard::hasText");
  return hasData(IClipboard::textFormat);
}

/*------------------------------------------------------------------------------
| IClipboard::hasText                                                          |
|                                                                              |
| Returns true if the clipboard has a bitmap in it (IClipboard::bitmapFormat). |
------------------------------------------------------------------------------*/
Boolean IClipboard::hasBitmap( ) const
{
  IMODTRACE_DEVELOP("Clipboard::hasBitmap");
  return hasData(IClipboard::bitmapFormat);
}


/*------------------------------------------------------------------------------
| IClipboard::text                                                             |
|                                                                              |
| Returns text from the clipboard by calling the function data with the format |
| IClipboard::textFormat.  The function data will throw an exception if the    |
| clipboard has no text format data.                                           |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
IString IClipboard::text( )
{
  IMODTRACE_DEVELOP("Clipboard::text");

  // Note: We cannot close the clipboard until we make
  //       a copy of the text.
  void* textData = data(textFormat);
  IString str((char*)textData);

  if (fClipboardData->isAutoOpen())
     close();
  return str;

}

/*------------------------------------------------------------------------------
| IClipboard::bitmap                                                           |
|                                                                              |
| Returns a copy of a bitmap on the clipboard by calling the function data with|
| the format IClipboard::bitmapFormat.  The function data will throw an        |
| exception if the clipboard has no bitmap format data.                        |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
IBitmapHandle IClipboard::bitmap( )
{
  IMODTRACE_DEVELOP("Clipboard::bitmap");

  // Note: We cannot close the clipboard until we make
  //       a copy of the bitmap.
  IBitmapHandle bitmap = (unsigned long)data(IClipboard::bitmapFormat);

  IBitmapHandle bitmapCopy = IGBitmap::copy(bitmap);

  if (fClipboardData->isAutoOpen())
      close();

  return bitmapCopy;
}


/*------------------------------------------------------------------------------
| IClipboard::data                                                             |
|                                                                              |
| Returns data of all pointer types from the clipboard.  Throws an             |
| IInvalidRequest exception if the clipboard doesn't have the requested        |
| format of data.                                                              |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
void* IClipboard::data( const char* aFormat )
{
  IMODTRACE_DEVELOP("Clipboard::data");
  ITRACE_DEVELOP(IString("Format=")+aFormat);

  FormatHandle formatHandle = IClipboard::formatAsHandle(aFormat);

  // Get the data from the clipboard
  void* pData = (void*)IGETCLIPBOARDDATA(
                        IThread::current().anchorBlock(),
                        formatHandle);
   // If we didn't succeed and the clipboard hasn't been opened,
   // open it and try again
   if (!pData && !fClipboardData->isOpen())
   {
      ITRACE_DEVELOP("Opening");
      autoOpen();
      pData = (void*)IGETCLIPBOARDDATA(
                            IThread::current().anchorBlock(),
                            formatHandle);
   }

 if (!pData)
    ITHROWLIBRARYERROR(IC_NOCLIPBOARDDATA,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);


 return pData;
}

/*------------------------------------------------------------------------------
| IClipboard::primaryFormat                                                    |
|                                                                              |
| Returns the primary Format of data on the clipboard.  The primary format     |
| is the first format of data entered after the clipboard is emptied.          |
------------------------------------------------------------------------------*/
IString IClipboard::primaryFormat( )
{
  IMODTRACE_DEVELOP("Clipboard::primaryFormat");
  IClipboard clipboard;
  IClipboard::Cursor cursor(clipboard);
  cursor.setToFirst();
  return IClipboard::format(cursor);
}

/*------------------------------------------------------------------------------
| IClipboard::format                                                           |
|                                                                              |
| Returns a format at the cursor location.                                     |
------------------------------------------------------------------------------*/
IString IClipboard::format ( const IClipboard::Cursor& cursor)
{
  IASSERTPARM(cursor.isValid());
  char buffer[256];
  int count = IGETCLIPBOARDFORMATTEXT( cursor.formatHandle, buffer, 256);
  if(!count)
    ITHROWLIBRARYERROR(IC_CLIPBOARDFORMATNOTFOUND,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);
  return IString(buffer);
}

/*------------------------------------------------------------------------------
| IClipboard::format                                                           |
|                                                                              |
| Returns a format for the passed format handle.                               |
------------------------------------------------------------------------------*/
IString IClipboard::format ( const IClipboard::FormatHandle& handle)
{
  char buffer[256];
  int count = IGETCLIPBOARDFORMATTEXT( handle, buffer, 256);
  if(!count)
    ITHROWLIBRARYERROR(IC_CLIPBOARDFORMATNOTFOUND,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);
  return IString(buffer);
}

/*------------------------------------------------------------------------------
| IClipboard::registerFormat                                                   |
|                                                                              |
| Registers a new Format of clipboard data.  The handle of the format is       |
| returned.  In OS/2, the format is added to the system atom table.  Windows   |
| has a special clipboard registration function.                               |
------------------------------------------------------------------------------*/
IClipboard::FormatHandle IClipboard::registerFormat ( const char* privateFormat)
{
  IMODTRACE_DEVELOP("Clipboard::registerFormat");
  ITRACE_DEVELOP(IString("Format =")+privateFormat);
  unsigned long format = IREGISTERCLIPBOARDFORMAT( privateFormat);
  if (!format)
    ITHROWLIBRARYERROR(IC_REGISTERCLIPBOARDFORMATFAILED,
                       IErrorInfo::accessError,
                       IException::recoverable);
  return (FormatHandle)format;
}

/*------------------------------------------------------------------------------
| IClipboard::formatAsHandle                                                   |
|                                                                              |
| Return the handle of a clipboard format.                                     |
------------------------------------------------------------------------------*/
IClipboard::FormatHandle IClipboard::formatAsHandle ( const char* aFormat)
{
  FormatHandle formatHandle;

   formatHandle = IFINDCLIPBOARDFORMAT((const char*)aFormat);
   if(!formatHandle)
     ITHROWLIBRARYERROR(IC_CLIPBOARDFORMATNOTFOUND,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);

return formatHandle;
}


/*------------------------------------------------------------------------------
| IClipboard::formatCount                                                      |
|                                                                              |
| Returns the number of clipboard formats.                                     |
------------------------------------------------------------------------------*/
unsigned long IClipboard::formatCount ( ) const
{
  unsigned long formatHandle = 0;
  unsigned long count = 0;

  do {

   formatHandle = IENUMCLIPBOARDFORMATS(
                    IThread::current().anchorBlock(),
                    formatHandle);
   if (formatHandle)
      count++;
  }
  while (formatHandle!=0);

return count;
}



/*------------------------------------------------------------------------------
| IClipboard::owner                                                            |
|                                                                              |
| Returns the current clipboard owner.                                         |
|                                                                              |
| Note: The clipboard owner is provided on the open function but ownership     |
|       is not established until the call to empty.  This function returns     |
|       the real clipboard owner (e.g. if you call this function after calling |
|       open but before calling empty, you will receive the prior owner ( or   |
|       perhaps none).                                                         |
------------------------------------------------------------------------------*/
IWindowHandle IClipboard::owner( ) const
{
  IMODTRACE_DEVELOP("Clipboard::owner");
  unsigned long owner = IGETCLIPBOARDOWNER(
                            IThread::current().anchorBlock());
  return IWindowHandle(owner);
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::Cursor                                                   |
|                                                                              |
| The constructor simply saves the clipboard object and initializes the cursor |
| state to an "invalid" state.                                                 |
------------------------------------------------------------------------------*/
IClipboard::Cursor::Cursor ( IClipboard& clipboardIn)
  : clipboard ( clipboardIn ),
    formatHandle( 0 )
{
}

/*------------------------------------------------------------------------------
| IClipboard::~ChildCursor                                                     |
|                                                                              |
| Currently empty.                                                             |
------------------------------------------------------------------------------*/
IClipboard::Cursor::~Cursor ( )
{
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::setToFirst                                               |
|                                                                              |
| Set to the first format in the clipboard.                                    |
------------------------------------------------------------------------------*/
Boolean IClipboard::Cursor::setToFirst ( )
{
  this->formatHandle = IENUMCLIPBOARDFORMATS(
                    IThread::current().anchorBlock(), 0);



  return this->formatHandle!=0;
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::setToNext                                                |
|                                                                              |
| Move the cursor to the next format.                                          |
------------------------------------------------------------------------------*/
Boolean IClipboard::Cursor::setToNext ( )
{
  this->formatHandle = IENUMCLIPBOARDFORMATS(
                    IThread::current().anchorBlock(),
                    this->formatHandle);
  return this->formatHandle != 0;
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::isValid                                                |
|                                                                              |
| Return true if the saved window handle isn't null.                           |
------------------------------------------------------------------------------*/
Boolean IClipboard::Cursor::isValid ( ) const
{
  return this->formatHandle != 0;
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::invalidate                                             |
|                                                                              |
| If enum handle is valid, end the enumeration.                                |
------------------------------------------------------------------------------*/
void IClipboard::Cursor::invalidate ( )
{
  this->formatHandle = 0;
}





/*------------------------------------------------------------------------------
| OS/2 Only functions                                                          |
|                                                                              |
| The following functions are OS/2 unique.  Their Win32 counterparts can       |
| be found later in this file.                                                 |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| IClipboard::setOwner                                                         |
|                                                                              |
| Identifies a new owner for the clipboard.                                    |
|                                                                              |
| Note:  OS/2 only function.                                                   |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setOwner ( )
{
  IMODTRACE_DEVELOP("Clipboard::setOwner");
  Boolean successful = WinSetClipbrdOwner(
                          IThread::current().anchorBlock(),
                          fClipboardData->owner());
  if(!successful)
     ITHROWGUIERROR("SetOwner");
  return *this;
}


/*------------------------------------------------------------------------------
| IClipboard::setData                                                          |
|                                                                              |
| OS/2 function to place memory data into the clipboard. This function         |
| allocates the data in shared memory, copies in the new data, and sets it     |
| into the clipboard.                                                          |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setData( const char*        aFormat,
                                 const void*        data,
                                 unsigned long      dataLength)
{
   IMODTRACE_DEVELOP("Clipboard::setData");
   ITRACE_DEVELOP(IString("Format=")+aFormat);

   // Allocate shared memory for the data
   char* memory=0;
   if(data !=0)
   {
     IASSERTPARM(dataLength!=0);
     unsigned long rc = DosAllocSharedMem(
                            (void**)(&memory), 0,
                            dataLength+1,
                            PAG_WRITE | PAG_COMMIT | OBJ_GIVEABLE);
     if (rc)
     {
        ITHROWSYSTEMERROR(rc, "DosAllocSharedMem",
                          IErrorInfo::accessError,
                          IException::recoverable);
     }

     // Copy to shared memory and ensure '0' termination
     memcpy(memory, data, dataLength);
     memory[dataLength] = '\0';
   }


   // Note there is a special case of getting data from the clipboard
   // during the processing of the WM_RENDERFMT or WM_RENDERALLFMTS
   // messages.  We don't want the clipboard open since it might already
   // be open by the process requesting the data.  We work around this
   // by first setting the data without opening the clipboard.  If
   // this fails, we open and try again.
   // Try to set the clipboard data
   IClipboard::FormatHandle formatHandle = IClipboard::formatAsHandle(aFormat);
   Boolean successful = WinSetClipbrdData(
                             IThread::current().anchorBlock(),
                             (unsigned long)memory,
                             formatHandle,
                             CFI_POINTER);

   // If we didn't succeed and the clipboard hasn't been opened,
   // open it and try again
   if (!successful && !fClipboardData->isOpen())
   {
      ITRACE_DEVELOP("Opening");
      autoOpen();
      successful = WinSetClipbrdData(
                     IThread::current().anchorBlock(),
                     (unsigned long)memory,
                     formatHandle,
                     CFI_POINTER);
   }

   if (fClipboardData->isAutoOpen())
     close();

   if(!successful)
     ITHROWGUIERROR("SetClipboardData");
   return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::setHandle                                                        |
|                                                                              |
| OS/2 function to place a handle into the clipboard. It is the callers        |
| responsibility to copy the data that the handle references into shared       |
| memory so that the paste process can find it.                                |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setHandle( const char*     aFormat,
                                   unsigned long   data)
{
  IMODTRACE_DEVELOP("Clipboard::setHandle");


   // Note there is a special case of getting data from the clipboard
   // during the processing of the WM_RENDERFMT or WM_RENDERALLFMTS
   // messages.  We don't want the clipboard open since it might already
   // be open by the process requesting the data.  We work around this
   // by first writting the data without opening the clipboard.  If
   // this fails, we open and try again.
   Boolean clipboardOpened = false;

   // Try to set the clipboard data
   IClipboard::FormatHandle formatHandle = IClipboard::formatAsHandle(aFormat);
   Boolean successful = WinSetClipbrdData(
                             IThread::current().anchorBlock(),
                             data,
                             formatHandle,
                             CFI_HANDLE);

   // If we didn't succeed and the clipboard hasn't been opened,
   // open it and try again
   if (!successful && !fClipboardData->isOpen())
   {
      clipboardOpened = true;
      autoOpen();
      successful = WinSetClipbrdData(
                             IThread::current().anchorBlock(),
                             data,
                             formatHandle,
                             CFI_HANDLE);
   }
   if (fClipboardData->isAutoOpen() && clipboardOpened)
     close();
   if(!successful)
     ITHROWGUIERROR("WinSetClipbrdData");
   return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::hasData                                                          |
|                                                                              |
| OS/2 function to return true if the requested format of data is on the       |
| clipboard.                                                                   |
------------------------------------------------------------------------------*/
Boolean IClipboard::hasData( const char* aFormat) const
{
  IMODTRACE_DEVELOP("Clipboard::hasData");
  unsigned long usageFlags;

  IClipboard::FormatHandle formatHandle = IClipboard::formatAsHandle(aFormat);
  Boolean fExists = WinQueryClipbrdFmtInfo(
                             IThread::current().anchorBlock(),
                             formatHandle,
                             &usageFlags);
  return fExists;
}


/*------------------------------------------------------------------------------
| Win32 Only functions                                                         |
|                                                                              |
| The following functions are Win32 unique.  Their OS/2 counterparts can       |
| be found earlier in this file.                                               |
------------------------------------------------------------------------------*/
