/*******************************************************************************
* FILE NAME: icnrctl2.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the IContainerControl class       *
*   declared in icnrctl.hpp.                                                   *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WIN
#define INCL_WINSTDCNR
#define INCL_WINWINDOWMGR
#define INCL_GPI

#include <icnrrec.hpp>  // Must be first for OS flags

#include <itrace.hpp>
#include <iexcept.hpp>
#include <irect.hpp>
#include <icnrctl.hpp>
#include <icnrcol.hpp>
#include <ifont.hpp>
#include <icoordsy.hpp>
#include <ithread.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICNRCTL2_CPP_
  #include <ipagetun.h>
#endif

/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: viewPortOnWorkspace

 Implementation: Query the rectangle bounding the client area
            relative to the origin of the containers workspace.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: viewPortOnWorkspace() const
{
   IRectangle rect = viewport(CMA_WORKSPACE, false);
   return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: viewPortOnWindow

 Implementation: Query the rectangle bounding the client area
            in container window coordinates.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: viewPortOnWindow() const
{
   IRectangle rect = viewport(CMA_WINDOW, false);
   return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: detailsViewPortOnWindow

 Implementation: Query the rectangle bounding the client area
   in container window coordinates of Details View.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: detailsViewPortOnWindow(Boolean fRight) const
{
  IRectangle rect = viewport(CMA_WINDOW, fRight);

  return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: detailsViewPortOnWorkspace

 Implementation: Query the rectangle bounding the client area
  relative to the origin of the containers workspace
  in the details view.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: detailsViewPortOnWorkspace(Boolean fRight) const
{
  IRectangle rect = viewport(CMA_WORKSPACE, fRight);


  return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: viewport

 Implementation: Query the rectangle bounding the client area
   relative to the passed coordinate space.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: viewport(unsigned long ulCoordinateSpace,
                                         Boolean fRightSplitWindow) const
{
  RECTL rectl;
  Boolean fSuccess = (Boolean)sendEvent(CM_QUERYVIEWPORTRECT,
                                   MPFROMP(&rectl),
                                   MPFROM2SHORT(ulCoordinateSpace,
                                                fRightSplitWindow));
  if(!fSuccess)
  {
     ITHROWGUIERROR("CM_QUERYVIEWPORT");
  }
  IRectangle rect(rectl.xLeft, rectl.yBottom, rectl.xRight, rectl.yTop);
  // convert to application coordinates if needed
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     rect.moveTo( IPoint( rect.minX(),
                          size().height() - rect.maxY() ) );
     }
  return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scrollVertically

 Implementation: Scroll the container up and down.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scrollVertically(long lPixels)
{
  scroll(lPixels, 0, false);
  return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scrollHorizontally

 Implementation: Scroll the container left and right.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scrollHorizontally(long lPixels,
                                                           Boolean fRight)//ppb1
{
   scroll(0, lPixels, fRight);
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scroll

 Implementation: Scroll the container up/down and left/right.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scroll(long lVerticalPixels,
                                 long lHorizontalPixels, Boolean fRight)
{
   Boolean fSuccess;

   //We need to make sure any posted CM_CLOSEEDIT is processed before we
   //scroll the window (to not lose changes).  Remove any from the queue
   //and call CM_CLOSEEDIT now.  We post this message in closeEdit().
   QMSG  qmsg;
   Boolean closeFound = false;
   while ( IPEEKMSG( IThread::current().anchorBlock(), &qmsg, handle(),
                     CM_CLOSEEDIT, CM_CLOSEEDIT, PM_REMOVE ) )
   {
     closeFound = true;
   }
   if ( closeFound )
   {
     sendEvent(CM_CLOSEEDIT, 0, 0);
   }

   if(lVerticalPixels)
   {
     fSuccess = (Boolean)sendEvent(CM_SCROLLWINDOW,
                                    MPFROMLONG(CMA_VERTICAL),
                                    MPFROMLONG(lVerticalPixels));
     if(!fSuccess)
     {
        ITHROWGUIERROR("CM_SCROLLWINDOW");
     }
   }

   if(lHorizontalPixels)
   {
     if ( isDetailsView() && detailsViewSplit() )
     {
       scrollDetailsHorizontally(lHorizontalPixels, fRight);
     }
     else
     {
       fSuccess = (Boolean)sendEvent( CM_SCROLLWINDOW,
                                      MPFROMLONG(CMA_HORIZONTAL),
                                      MPFROMLONG(lHorizontalPixels));
       if(!fSuccess)
       {
          ITHROWGUIERROR("CM_SCROLLWINDOW");
       }
     }
   }
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scrollDetailsHorizontally

 Implementation: Scroll the details view left and right.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scrollDetailsHorizontally(long lHorizontalPixels,
                                                    Boolean fRight)
{
   unsigned long flSide;
   if (fRight)
     flSide = CMA_RIGHT;
   else
     flSide = CMA_LEFT;

   if (detailsViewSplit())
   {
     Boolean fSuccess = (Boolean)sendEvent(CM_HORZSCROLLSPLITWINDOW,
                                    MPFROMLONG(flSide),
                                    MPFROMLONG(lHorizontalPixels));
       if(!fSuccess)
       {
          ITHROWGUIERROR("CM_SCROLLSPLITWINDOW");
       }
   }
   else
     scroll( 0, lHorizontalPixels, false);

   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scrollToObject

 Implementation: Scroll the Window to include an object.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scrollToObject(const IContainerObject* pcnrobj)
{
IRectangle rectViewport = viewPortOnWindow();
IRectangle rectObject   = iconRectangle(pcnrobj, true);
long lMargin = 4;

  // always want to use the top of the object and viewport as the
  // reference points.
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     scroll(rectViewport.minY() - rectObject.minY() - lMargin,
            rectObject.minX() - lMargin);
     }
  else
     {
     scroll(rectViewport.maxY() - rectObject.maxY() - lMargin,
            rectObject.minX() - lMargin);
     }
return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: scrollToObject

 Implementation: Scroll the Window to include an object in a column
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: scrollToObject(
                                           const IContainerObject* pcnrobj,
                                           const IContainerColumn* pcnrcol,
                                           Boolean bLeftJustify)
{
  IASSERTPARM(pcnrobj!=0);
  long        lMargin = 4;
  long        horizontalDistance=0;
  long        verticalDistance  =0;
  Boolean     rightSide=false;
  IRectangle  rectViewport;
  IRectangle  rectObject;

  if (pcnrcol)
  {
    rightSide = isColumnRight(pcnrcol);
    rectViewport = detailsViewPortOnWindow(rightSide);
    rectObject   = detailsObjectRectangle(pcnrobj, pcnrcol);
  }
  else
  {
    rectViewport = detailsViewPortOnWindow(false);
    rectObject   = detailsObjectRectangle(pcnrobj, false);
  }

  // convert coordinates so that reference points are correct in rectangles
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     long   cnrheight = size().height();
     rectViewport.moveTo( IPoint( rectViewport.minX(),
                          cnrheight - rectViewport.maxY() ) );
     rectObject.moveTo( IPoint( rectObject.minX(),
                        cnrheight - rectObject.maxY() ) );
     }

  if (bLeftJustify)  //scroll to top-left corner
  {
    if (pcnrcol)     //only scroll horizontally if given a column
    {
      horizontalDistance = rectObject.left() - lMargin;
    }
    verticalDistance = rectViewport.top() - rectObject.top() - lMargin;
  }
  else             //scroll object just into view at closest border
  {
    if (pcnrcol)   //only scroll horizontally if given a column
    {
      if ( (rectObject.left() < rectViewport.left()) ||
           ( (rectObject.right()-rectObject.left()) >
             (rectViewport.right()-rectViewport.left()) ) )
      {
        horizontalDistance = rectObject.left() - rectViewport.left();
      }
      else if (rectObject.right() > rectViewport.right() )
      {
        horizontalDistance = rectObject.right() - rectViewport.right();
      }
    }

    if ( (rectObject.top() > rectViewport.top()) ||
         ( (rectObject.top()-rectObject.bottom()) >
           (rectViewport.top()-rectViewport.bottom()) ) )
    {
      verticalDistance = rectViewport.top() - rectObject.top();
    }
    else if (rectObject.bottom() < 0)
    {
      verticalDistance = rectViewport.bottom() - rectObject.bottom();
    }
  }
  scroll( verticalDistance, horizontalDistance, rightSide );

  return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: setIconSize

 Implementation: Set the size of icons.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: setIconSize(const ISize& sizeIcon)
{
  ICnrInfo cnrInfo;
  cnrInfo.slBitmapOrIcon.cx = sizeIcon.width();
  cnrInfo.slBitmapOrIcon.cy = sizeIcon.height();
  setContainerInfo(&cnrInfo, CMA_SLBITMAPORICON);
  return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: iconSize

 Implementation: Query the size of icons.
------------------------------------------------------------------------------*/
ISize IContainerControl :: iconSize() const
{
  ICnrInfo cnrInfo;
  containerInfo(&cnrInfo);
  return ISize(cnrInfo.slBitmapOrIcon.cx, cnrInfo.slBitmapOrIcon.cy);
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: moveIconTo

 Implementation: Move an icon in the Icon View.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: moveIconTo(IContainerObject* object,
                                     const IPoint& pt)
{
  IMODTRACE_DEVELOP("cnrctl::moveIconTo");
  IASSERTPARM(object!=0);
  Boolean fContainsObject = containsObject(object);
  IMiniCnrRecord* pcnrrec = IRecFromObj(object);

  if(fContainsObject)
  {
    IASSERTSTATE(parentObject(object)==0);
    /* remove record from previous location */
    if (!(Boolean)sendEvent(CM_ERASERECORD,
                        MPFROMP(pcnrrec),0))
    {
       ITHROWGUIERROR("CM_ERASERECORD");
    }
  }

  /* set new position */
  pcnrrec->ptlIcon.x = pt.x();
  pcnrrec->ptlIcon.y = pt.y();
  // convert to application coordinates if needed
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     pcnrrec->ptlIcon.y = size().height() - pt.y();
     }

  if(fContainsObject)
  {
    unsigned short fsFlags = CMA_REPOSITION;

    if(!(Boolean)sendEvent(CM_INVALIDATERECORD,
           MPFROMP(&pcnrrec),MPFROM2SHORT(1,fsFlags)))
    {
       ITHROWGUIERROR("CM_INVALIDATERECORD");
    }
  }

  ITRACE_ALL(IString("New Icon location is:")+ iconRectangle(object, false).asString());
  ITRACE_ALL(IString("New Icon & Text location is:")+ iconRectangle(object, true).asString());

   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: iconRectangle

 Implementation: Query an icon's rectangle.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: iconRectangle(const IContainerObject* object,
                                             Boolean fIncludeText) const
{
   IASSERTSTATE(object!=0 && containsObject(object));
   RECTL rectl;
   QUERYRECORDRECT  qryrecrect;

   /* Avoid error when calling CM_QUERYRECORDRECT by not setting */
   /* CMA_ICON in either the text view or the tree-text view.    */
   if ( isTextView() )
     qryrecrect.fsExtent = 0;
   else
     qryrecrect.fsExtent = CMA_ICON;

   if(fIncludeText!=false)
      qryrecrect.fsExtent |= CMA_TEXT;
   qryrecrect.pRecord = (PRECORDCORE)IRecFromObj(object);
   qryrecrect.cb = sizeof(QUERYRECORDRECT);
   Boolean fSuccess = (Boolean)sendEvent( CM_QUERYRECORDRECT,
                               MPFROMP(&rectl),
                               MPFROMLONG(&qryrecrect));
   if(fSuccess)
   {
     IRectangle rect(rectl.xLeft, rectl.yBottom, rectl.xRight, rectl.yTop);
     // convert to application coordinates if needed
     if (ICoordinateSystem::applicationOrientation() ==
         ICoordinateSystem::originUpperLeft )
        {
        rect.moveTo( IPoint( rect.minX(),
                             size().height() - rect.maxY() ) );
        }
     return rect;
   }
   else
     return IRectangle();

}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: textRectangle

 Implementation: Query an object's text rectangle.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: textRectangle(const IContainerObject* object) const
{
   IASSERTSTATE(object!=0 && containsObject(object));
   RECTL rectl;
   QUERYRECORDRECT  qryrecrect;
   qryrecrect.fsExtent = CMA_TEXT;
   qryrecrect.pRecord = (PRECORDCORE)IRecFromObj(object);
   qryrecrect.cb = sizeof(QUERYRECORDRECT);
   Boolean fSuccess = (Boolean)sendEvent( CM_QUERYRECORDRECT,
                               MPFROMP(&rectl),
                               MPFROMLONG(&qryrecrect));
   if(fSuccess)
   {
     IRectangle rect(rectl.xLeft, rectl.yBottom, rectl.xRight, rectl.yTop);
     // convert to application coordinates if needed
     if (ICoordinateSystem::applicationOrientation() ==
         ICoordinateSystem::originUpperLeft )
        {
        rect.moveTo( IPoint( rect.minX(),
                             size().height() - rect.maxY() ) );
        }
     return rect;
   }
   else
     return IRectangle();
}

/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: detailsObjectRectangle

 Implementation: Query an object's details view rectangle.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: detailsObjectRectangle
                                       (const IContainerObject* object,
                                        Boolean fRightWindow ) const
{
   IASSERTSTATE(object!=0 && containsObject(object));
   RECTL rectl;
   QUERYRECORDRECT  qryrecrect;
   qryrecrect.fsExtent = 0;
   qryrecrect.pRecord = (PRECORDCORE)IRecFromObj(object);
   qryrecrect.cb = sizeof(QUERYRECORDRECT);
   qryrecrect.fRightSplitWindow = fRightWindow;
   Boolean fSuccess = (Boolean)sendEvent( CM_QUERYRECORDRECT,
                               MPFROMP(&rectl),
                               MPFROMLONG(&qryrecrect));
   if(fSuccess)
   {
     IRectangle rect(rectl.xLeft, rectl.yBottom, rectl.xRight, rectl.yTop);
     // convert to application coordinates if needed
     if (ICoordinateSystem::applicationOrientation() ==
         ICoordinateSystem::originUpperLeft )
        {
        rect.moveTo( IPoint( rect.minX(),
                             size().height() - rect.maxY() ));
        }
     return rect;
   }
   else
     return IRectangle();
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: detailsObjectRectangle

 Implementation: Query an object's column rectangle (the cell).
                 Returns rectangle in window coordinates.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: detailsObjectRectangle
                                       (const IContainerObject* object,
                                        const IContainerColumn* column) const
{
   IASSERTPARM(column!=0 && object!=0);
   IASSERTSTATE(containsObject(object) && column->pfieldinfoCl);

   // get the complete details rectangle for this object
   IRectangle lRectRecord = detailsObjectRectangle(object, false);

   /*-----------------------------------------------------------
   if the column is not visible, return a rectangle with a height but no width
  ------------------------------------------------------------*/
   if (!column->isVisible())
   {
     IRectangle rect(0,
                     lRectRecord.bottom(),
                     0,
                     lRectRecord.top());
     return rect;
   }

   /*-----------------------------------------------------------
  else if it is visible, continue calculating the cell size
  ------------------------------------------------------------*/
  // get font metrics
   long        LeftSide=0, RightSide=0;
   long        LeftMargin, RightMargin;
   long        lAveCharWidth;
   HPS         hps;
   FONTMETRICS FontMetrics;
   hps = WinGetPS( handle() );
   GpiQueryFontMetrics( hps, sizeof(FONTMETRICS), &FontMetrics );
   WinReleasePS( hps );
   lAveCharWidth = FontMetrics.lAveCharWidth;
   LeftMargin  = (3 * lAveCharWidth) / 2;
   RightMargin = (3 * lAveCharWidth) - LeftMargin;

   // determine which side the column is on
   IContainerControl::ColumnCursor visiblecolcur(*this, true);
   if( isColumnRight(column) )
   {
     visiblecolcur.setCurrent(detailsViewSplit());
     visiblecolcur.setToNext();
   }
   else
     visiblecolcur.setToFirst();

   // set up a second column cursor
   IContainerControl::ColumnCursor lastvisiblecur(*this, true);
   if( !isColumnRight(column)  && detailsViewSplit() )
   {
     lastvisiblecur.setCurrent(detailsViewSplit());
     if ( !(detailsViewSplit()->isVisible()) )
       lastvisiblecur.setToPrevious();
   }
   else
   {
     lastvisiblecur.setToLast();
   }

   //  determine the right & left sides side of the cell
   RightSide = columnAt(visiblecolcur)->displayWidth();
   RightSide += lAveCharWidth + RightMargin;

   while (visiblecolcur.isValid() && columnAt(visiblecolcur) != column )
   {
     LeftSide = RightSide;
     visiblecolcur.setToNext();
     RightSide += columnAt(visiblecolcur)->displayWidth();
     RightSide += LeftMargin + RightMargin;
   }

   if ( columnAt(visiblecolcur) == columnAt(lastvisiblecur) )
     RightSide += lAveCharWidth - RightMargin;

   /*-----------------------------------------------------------
    Convert the columns workspace values to window coordinates.
  ------------------------------------------------------------*/
   IRectangle lRectViewportWork;
   lRectViewportWork = detailsViewPortOnWorkspace(isColumnRight(column));
   IRectangle rect(LeftSide - lRectViewportWork.left(),
                   lRectRecord.bottom(),
                   RightSide - lRectViewportWork.left(),
                   lRectRecord.top());
   if (isColumnRight(column))
     {
     IRectangle lRectViewportWindow;

     lRectViewportWindow = detailsViewPortOnWindow(true);
     rect = IRectangle(LeftSide - lRectViewportWork.left()
                              + lRectViewportWindow.left(),
                     lRectRecord.bottom(),
                     RightSide - lRectViewportWork.left()
                              + lRectViewportWindow.left(),
                     lRectRecord.top());
     }
   return rect;
// Stub for now on Motif until IContainerColumn is ported
// Not sure if we can ever get this function to work though...
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: disableDrawItem

 Implementation: Turn Owner draw off.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: disableDrawItem()
{
   setContainerAttributes(CA_OWNERDRAW,0);
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl::enableDrawBackground

 Implementation: Enable the container to paint the background.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: enableDrawBackground(Boolean fEnable)
{
   if(fEnable)
     setContainerAttributes(0,CA_OWNERPAINTBACKGROUND);
   else
     setContainerAttributes(CA_OWNERPAINTBACKGROUND, 0);
   return *this;

}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: isDrawBackgroundEnabled

 Implementation: Query Background drawing state.
------------------------------------------------------------------------------*/
IBase::Boolean IContainerControl :: isDrawBackgroundEnabled() const
{
  return ( (containerAttributes() & CA_OWNERPAINTBACKGROUND) ? true : false);
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl::disableDrawBackground

 Implementation: Don't allow the container to paint the background.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: disableDrawBackground()
{
   setContainerAttributes(CA_OWNERPAINTBACKGROUND,0);
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: convertToWorkspace

 Implementation: Convert the given rectangle in window coordinates to
                 container workspace coordinates.  This conversion is valid
                 only for the current view as we need to query the current
                 viewport for the conversion.
------------------------------------------------------------------------------*/
IRectangle IContainerControl :: convertToWorkspace
                                       (const IRectangle& windowRectangle,
                                        Boolean fRightWindow ) const
{
  IRectangle lRectViewportWindow;
  IRectangle rect = windowRectangle;
  IRectangle lRectViewportWork = detailsViewPortOnWorkspace(fRightWindow);
  ISize  adjustSize;
  // convert from application coordinates if needed
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     adjustSize = size();
     lRectViewportWork.moveTo( IPoint( lRectViewportWork.minX(),
                                       adjustSize.height() -
                                       lRectViewportWork.maxY()));
     rect.moveTo( IPoint( rect.minX(),
                          adjustSize.height() -
                          rect.maxY()));
     }

  if (fRightWindow)
  {
    lRectViewportWindow = detailsViewPortOnWindow(true);
    rect = IRectangle(rect.left()   + lRectViewportWork.left()
                         - lRectViewportWindow.left(),
                      rect.bottom() + lRectViewportWork.bottom(),
                      rect.right()  + lRectViewportWork.left()
                         - lRectViewportWindow.left(),
                      rect.top()    + lRectViewportWork.bottom());
  }
  else
  {
    rect = IRectangle(rect.left()   + lRectViewportWork.left(),
                      rect.bottom() + lRectViewportWork.bottom(),
                      rect.right()  + lRectViewportWork.left(),
                      rect.top()    + lRectViewportWork.bottom());
  }
  // convert to application coordinates if needed
  if (ICoordinateSystem::applicationOrientation() ==
      ICoordinateSystem::originUpperLeft )
     {
     rect.moveTo( IPoint( rect.minX(),
                          adjustSize.height() -
                          rect.maxY()));
     }
  return rect;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: columnUnderPoint

 Implementation: Determine the column under a point in window
                 coordinates.
------------------------------------------------------------------------------*/
IContainerColumn* IContainerControl :: columnUnderPoint(const IPoint& point) const
{
  IMODTRACE_DEVELOP("CnrCtl :: columnUnderPoint");
  IASSERTSTATE( isDetailsView() );
  IRectangle lRectLeftViewportWindow;
  IRectangle lRectRightViewportWindow;
  IContainerControl::ColumnCursor visiblecolcur(*this, true);
  IContainerControl::ColumnCursor lastcolcur(*this, true);

  /*****************************************************************/
  /* Make sure x value of point is within the confines of the      */
  /* current details viewport.  Also, set up the starting column   */
  /* cursor and the last column cursor.                            */
  /*****************************************************************/
  lRectLeftViewportWindow = detailsViewPortOnWindow(false);
  if (detailsViewSplit())
  {
    lRectRightViewportWindow = detailsViewPortOnWindow(true);

    IASSERTPARM( ((point.x() >= lRectLeftViewportWindow.left()) &&
                  (point.x() <= lRectRightViewportWindow.right())) );
    /*******************************************************/
    /* Set up the start and end columns.                   */
    /* Count the split bar area as the left D.V, window.   */
    /*******************************************************/
    if ( (point.x() >= lRectLeftViewportWindow.left()) &&
         (point.x() <= lRectRightViewportWindow.left()) )
    {
      visiblecolcur.setToFirst();
      lastcolcur.setCurrent(detailsViewSplit());
      if ( !(detailsViewSplit()->isVisible()) )
        lastcolcur.setToPrevious();
    }
    else
    {
      visiblecolcur.setCurrent(detailsViewSplit());
      visiblecolcur.setToNext();
      lastcolcur.setToLast();
    }
  }
  else
  {
    IASSERTPARM( (point.x() >= lRectLeftViewportWindow.left()) &&
                 (point.x() <= lRectLeftViewportWindow.right()) );
    visiblecolcur.setToFirst();
    lastcolcur.setToLast();
  }

  /****************************************************************/
  /* Query the containers font to determine the column margins    */
  /****************************************************************/
  long        LeftSide=0, RightSide=0;
  long        LeftMargin, RightMargin;
  long        lAveCharWidth;
  HPS         hps;
  FONTMETRICS FontMetrics;
  hps = WinGetPS( handle() );
  GpiQueryFontMetrics( hps, sizeof(FONTMETRICS), &FontMetrics );
  WinReleasePS( hps );
  lAveCharWidth = FontMetrics.lAveCharWidth;
  LeftMargin  = (3 * lAveCharWidth) / 2;
  RightMargin = (3 * lAveCharWidth) - LeftMargin;
  /*****************************************************************/
  /* Adjust the starting point to container window coordinates     */
  /*****************************************************************/
  IRectangle lRectViewportWork;
  lRectViewportWork = detailsViewPortOnWorkspace(isColumnRight
                           (columnAt(visiblecolcur)) );
  LeftSide -= lRectViewportWork.left();
  RightSide -= lRectViewportWork.left();
  if (isColumnRight(columnAt(visiblecolcur)))
  {
    LeftSide += lRectRightViewportWindow.left();
    RightSide += lRectRightViewportWindow.left();
  }

  /******************************************************************/
  /* Compute the first visible column, if it is it, return now.     */
  /******************************************************************/
  RightSide += columnAt(visiblecolcur)->displayWidth();
  RightSide += lAveCharWidth + RightMargin;
  if ( point.x() <= RightSide )
    return (columnAt(visiblecolcur));

  /******************************************************************/
  /* Loop thru all visible columns and return if point matches      */
  /******************************************************************/
  while (visiblecolcur.isValid() && (columnAt(visiblecolcur) !=
                                     columnAt(lastcolcur)) )
  {
    LeftSide = RightSide;
    visiblecolcur.setToNext();
    RightSide += columnAt(visiblecolcur)->displayWidth();
    RightSide += LeftMargin + RightMargin;
    if ( point.x() <= RightSide )
      return (columnAt(visiblecolcur));
  }

  return(columnAt(lastcolcur));
}

