/*******************************************************************************
* FILE NAME: icnrctl8.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the sorting and filtering         *
*   functions for class IContainerControl, declared in icnrctl.hpp.            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WINSTDCNR
#define INCL_WINCOUNTRY
#define INCL_DOSNLS

#include <icnrrec.hpp>  // Must be first for OS flags

#include <itrace.hpp>
#include <iexcept.hpp>
#include <icnrctl.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICNRCTL8_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
   Imbedded classes: definition and implementation
------------------------------------------------------------------------------*/
class ICnrSortPackage
{
public:
const IContainerControl::CompareFn
 *pCompareFn;

IContainerControl
 *pcnrctl;
};

class ICnrFilterPackage
{
public:
const IContainerControl::FilterFn
 *pFilterFn;

IContainerControl
 *pcnrctl;
};

// Keep the compiler from generating the following statically
IContainerControl::FilterFn  ::  FilterFn ( ) {}
IContainerControl::FilterFn  :: ~FilterFn ( ) {}
IContainerControl::CompareFn ::  CompareFn( ) {}
IContainerControl::CompareFn :: ~CompareFn( ) {}

class ICnrSortByIcon : public IContainerControl::CompareFn
{
public:
   ICnrSortByIcon (Boolean ascending);
  ~ICnrSortByIcon ();
 Boolean
   isEqual( IContainerObject*  pcnrobj1,
            IContainerObject*  pcnrobj2,
            IContainerControl* pcnrctl) const;
private:
Boolean
  fAscending;
};

/* Functions to sort by icon */
ICnrSortByIcon :: ICnrSortByIcon (Boolean ascending) : fAscending(ascending) {}
ICnrSortByIcon :: ~ICnrSortByIcon ( ) {}
ICnrSortByIcon :: isEqual( IContainerObject*  pcnrobj1,
            IContainerObject*  pcnrobj2,
            IContainerControl* pcnrctl) const
{

 if(fAscending)
    return pcnrctl->nlsCompare( pcnrobj1->iconText(),
                                pcnrobj2->iconText());
 else
    return pcnrctl->nlsCompare( pcnrobj2->iconText(),
                                pcnrobj1->iconText());
}

class ICnrFilterNone : public IContainerControl::FilterFn
{
public:
  virtual Boolean
    isMemberOf( IContainerObject* object,
                IContainerControl* container) const;
   ICnrFilterNone ( );
  ~ICnrFilterNone ( );
};

IBase::Boolean ICnrFilterNone :: isMemberOf( IContainerObject* object,
                IContainerControl* container) const
{ return true; }

ICnrFilterNone ::   ICnrFilterNone ( ) {}
ICnrFilterNone ::  ~ICnrFilterNone ( ) {}


/*-----------------------------------------------------------------------------
 Function Name:  IContainerControl__compare

 Implementation: Non-member function to compare two objects text.
------------------------------------------------------------------------------*/
long  _System  IContainerControl__compare(const IMiniCnrRecord* pcnrrec1,
                                          const IMiniCnrRecord* pcnrrec2,
                                          ICnrSortPackage*      pSortPackage)
{
   long ulReturn;
   if(pSortPackage)
       ulReturn = pSortPackage->pCompareFn->isEqual( IObjFromRec(pcnrrec1),
                                          IObjFromRec(pcnrrec2),
                                          pSortPackage->pcnrctl);
   else
      ulReturn = IContainerControl::nlsCompare(pcnrrec1->strIcon, pcnrrec2->strIcon);

   return ulReturn;
}



/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: sort

 Implementation: Sort by the User Defined Sort function.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: sort(const IContainerControl::CompareFn& sortFunction)
{
   ICnrSortPackage cnrsort = {&sortFunction, this};

   Boolean fSuccess = (Boolean)sendEvent( CM_SORTRECORD,
                                 MPFROMP(IContainerControl__compare),
                                 MPFROMP(&cnrsort));
   if(!fSuccess)
   {
      ITHROWGUIERROR("CM_SORTRECORD");
   }
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl__filter

 Implementation: Non-memeber filter function.
------------------------------------------------------------------------------*/
IBase::Boolean _System IContainerControl__filterFn( const IMiniCnrRecord* pcnrrec,
                                           ICnrFilterPackage* pFilterPackage)
{
   return pFilterPackage->pFilterFn->isMemberOf(IObjFromRec(pcnrrec), pFilterPackage->pcnrctl);
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: filter

 Implementation: Filter the container with User Defined Filter.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: filter(const IContainerControl::FilterFn& filterFn)
{
   ICnrFilterPackage cnrfilter = {&filterFn, this};

   Boolean fSuccess = true;
   // Only filter the container if there are objects as the WC_CONTAINER
   // views this as an error
   ICnrInfo cnrInfo;
   containerInfo(&cnrInfo);
   if (cnrInfo.cRecords > 0)
      fSuccess = (Boolean)sendEvent( CM_FILTER,
                                     MPFROMP(IContainerControl__filterFn),
                                     &cnrfilter);
   if(!fSuccess)
      ITHROWGUIERROR("CM_FILTER");
   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: filter

 Implementation: Restore filtered item in the container.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: filter()
{
   ICnrFilterNone filterNone;
   ICnrFilterPackage cnrfilter = {&filterNone, this};

   Boolean fSuccess = true;
   // Only filter the container if there are objects as the WC_CONTAINER
   // views this as an error
   ICnrInfo cnrInfo;
   containerInfo(&cnrInfo);
   if (cnrInfo.cRecords > 0)
      fSuccess = (Boolean)sendEvent( CM_FILTER,
                                     MPFROMP(IContainerControl__filterFn),
                                     &cnrfilter);

   if(!fSuccess)
      ITHROWGUIERROR("CM_FILTER");

   return *this;
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: sortByIconText

 Implementation: Sort the container by the icons text.
------------------------------------------------------------------------------*/
IContainerControl& IContainerControl :: sortByIconText(Boolean ascending)
{
   ICnrSortByIcon iconSort(ascending);
   return sort(iconSort);
}


/*-----------------------------------------------------------------------------
 Function Name: IContainerControl :: nlsCompare

 Implementation: Compare two strings with National Language Support.
------------------------------------------------------------------------------*/
long IContainerControl :: nlsCompare(const char* pszString1,
                                      const char* pszString2)
{
  IASSERT(pszString1!=0 && pszString2!=0);

  unsigned long  ulLength = 0;
  unsigned long  ulResult = 0;
  long      lCompare = 1;

  COUNTRYCODE cntry;
  COUNTRYINFO cntryinfo;
  cntry.country = 0;
  cntry.codepage = 0;

  ulResult = DosGetCtryInfo(sizeof(cntryinfo),
                        &cntry,
                        &cntryinfo,
                        &ulLength);

  if (ulResult==0)
  {
     ulResult = WinCompareStrings ((HAB)0, cntry.codepage,
                                    cntry.country,
                                    (PSZ)pszString1,
                                    (PSZ)pszString2,
                                    0);
     if (ulResult == WCS_EQ)
        lCompare = 0;
     else if (ulResult == WCS_LT)
        lCompare = -1;
     else   /* greater then by default */
        lCompare = 1;
  }

  return lCompare;

}
