/*******************************************************************************
* FILE NAME: icnrobj0.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icnrobj.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINSTDDRAG
#define INCL_WINSTDCNR
extern "C" {
  #include <string.h>
  #include <stdio.h>
}

#include <icnrrec.hpp>  // Must be first for OS flags

#include <ireslib.hpp>
#include <itrace.hpp>
#include <iexcept.hpp>
#include <istring.hpp>
#include <icnrobj.hpp>
#include <icnrctl.hpp>
#include <icconst.h>
#include <icnrfilt.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICNROBJ0_CPP_
  #include <ipagetun.h>
#endif



/*------------------------------------------------------------------------------
| IContainerObject::initialize                                                 |
|                                                                              |
| Common constructor initialization                                            |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject::initialize()
{

   this->pcnrobjClParent = 0;
   this->pbase = new(this) IMiniCnrRecord();

   // Check to make sure that we were created with operator new
   if (this->pbase->cb != sizeof(IMiniCnrRecord))
   {
      ITHROWLIBRARYERROR(IC_CNROBJ_INVALID_CREATE,
                           IErrorInfo::invalidRequest,
                           IException::recoverable);
   }

   this->pbase->flRecordAttr = CRA_RECORDREADONLY | CRA_DROPONABLE;

// Ewald claimed have to set flClState this way under Motif to prevent
// setIcon() from being invisible until call to setRefreshOn()
   this->flClState |= ok;

   this->ulClUseCount = 0;
   this->delayFree = false;
   return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Default constructor with no arguments.                                       |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject()
{
   IMODTRACE_ALL("CnrObj:CnrObj()");
   initialize();

   return;
}

/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Create an container object using an IString for text and an IPointerHandle   |
| for the icon.                                                                |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject(const IString& strIconName,
                                     const IPointerHandle& ptrhIconInit)
{
   IMODTRACE_ALL("CnrObj:CnrObj(str,ptrh)");
   initialize();

   this->pbase->hptrIcon = ptrhIconInit;
   this->pbase->strIcon = strIconName;

#ifdef IC_NOBINARYCOMP
   this->pbase->cRecordText = (char*)(this->pbase->strIcon);
   this->pbase->pRecordIcon = (IHandle::Value)(this->pbase->hptrIcon);
#endif

   return;
}


/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Create an container object by providing an IString for text and an           |
| IResourceId for the icon.                                                    |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject(const IString& strIconName,
                                     const IResourceId& residIcon)
{
   IMODTRACE_ALL("CnrObj:CnrObj(str,resid)");
   initialize();

   this->pbase->strIcon = strIconName;

#ifdef IC_NOBINARYCOMP
   this->pbase->cRecordText = (char*)(this->pbase->strIcon);
#endif

   setRefreshOff();
   if(residIcon)
     setIcon(residIcon);
   setRefreshOn();

   return;
}

/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Create a container object by providing an IString for text and an unsigned   |
| long that is used as a resource ID.                                          |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject(const IString& strIconName,
                                     unsigned long ResidIcon)
{
   IMODTRACE_ALL("CnrObj:CnrObj(str,ulResid)");
   initialize();

   this->pbase->strIcon = strIconName;

#ifdef IC_NOBINARYCOMP
   this->pbase->cRecordText = (char*)(this->pbase->strIcon);
#endif

   setRefreshOff();
   if(ResidIcon)
     setIcon(ResidIcon);
   setRefreshOn();

   return;
}

/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Create a container object by providing resource IDs for both the text and    |
| the icon.                                                                    |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject(const IResourceId& residIconName,
                                     const IResourceId& residIcon)
{
   IMODTRACE_ALL("CnrObj:CnrObj(resid,resid)");
   initialize();

   setRefreshOff();
   if(residIcon)
     setIcon(residIcon);

   if(residIconName)
     setIconText(residIconName);
   setRefreshOn();

   return;
}


/*------------------------------------------------------------------------------
| IContainerObject::IContainerObject                                           |
|                                                                              |
| Copy Constructor.                                                            |
------------------------------------------------------------------------------*/
IContainerObject :: IContainerObject (const IContainerObject& cnrobjOld)
{
   IMODTRACE_ALL("CnrObj:CnrObj(&)");
   initialize();

   IMiniCnrRecord* precOld = IRecFromObj(&cnrobjOld);
   this->pbase->flRecordAttr = precOld->flRecordAttr;
   this->pbase->flRecordAttr &=
                  (unsigned long)(~(CRA_CURSORED | CRA_INUSE | CRA_SELECTED));
   this->pbase->ptlIcon = precOld->ptlIcon;

   setRefreshOff();
   setIcon(precOld->hptrIcon);
   setIconText((char*)precOld->strIcon);
   setRefreshOn();
}

/*------------------------------------------------------------------------------
| IContainerObject::newRecord                                                  |
|                                                                              |
| Implementation of operator new.                                              |
------------------------------------------------------------------------------*/
void* IContainerObject :: newRecord(size_t size)
{
  IMiniCnrRecord* pcnrrec;
  IContainerObject* pcnrobj;

  pcnrrec = (IMiniCnrRecord*)(void*)
                IContainerControl::hwndAllocation().sendEvent(
                                                           CM_ALLOCRECORD,
 #ifdef IC_NOBINARYCOMP
                                                           MPFROMLONG(size+48),  //DAB hack
 #else
                                                           MPFROMLONG(size),
 #endif
                                                           MPFROMLONG(1L));


  if(pcnrrec==0)
  {
     ITHROWGUIERROR2("CM_ALLOCRECORD",
                          IErrorInfo::outOfMemory,
                          IException::recoverable);
  }
  pcnrrec->cb = sizeof(IMiniCnrRecord);
  pcnrobj = IObjFromRec(pcnrrec);

  if(size > sizeof(IContainerObject))
     pcnrobj->flClState = subclassed;

  return(pcnrobj);
}

/*------------------------------------------------------------------------------
| IContainerObject::deleteRecord                                               |
|                                                                              |
| Implementation of operator delete.                                           |
------------------------------------------------------------------------------*/
void IContainerObject :: deleteRecord(void* pcnrobj)
{
if (!(((IContainerObject*)pcnrobj)->isDelayedFree()))
{
  /********************************************************************/
  /* Call Container control to free the MINIRECORDCORE and associated */
  /* memory that was allocated for this object via CM_ALLOCRECORD     */
  /* during the construction of this object.                          */
  /********************************************************************/
  void* pcnrrec = IRecFromObj((IContainerObject*)pcnrobj);
  int fSuccess = (int) IContainerControl::hwndAllocation().sendEvent(
                                   CM_FREERECORD,
                                   MPFROMP(&pcnrrec),
                                   MPFROMSHORT(1));

  if (fSuccess == -1)
  {
    ITHROWGUIERROR("CM_FREERECORD");
  }
}

}

/*------------------------------------------------------------------------------
| IContainerObject::setIcon                                                    |
|                                                                              |
| Sets an object's icon.                                                       |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::  setIcon(const IPointerHandle& ptrh)
{
   this->pbase->hptrIcon = ptrh;
#ifdef IC_NOBINARYCOMP
  this->pbase->pRecordIcon = (IHandle::Value)(this->pbase->hptrIcon);
#endif

   if(isRefreshOn())
     refresh();

   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setIcon                                                    |
|                                                                              |
| Sets an object's icon via an resource id.                                    |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::  setIcon(const IResourceId& resid)
{
   setIcon(resid.resourceLibrary().loadIcon(resid));
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setIcon                                                    |
|                                                                              |
| Sets an object's icon via an resource id.                                    |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::  setIcon(unsigned long ulId)
{
   IResourceId resid(ulId);
   if(ulId)
     setIcon(resid.resourceLibrary().loadIcon(resid));
   else
   {
     this->pbase->hptrIcon = 0;
#ifdef IC_NOBINARYCOMP
  this->pbase->pRecordIcon = (IHandle::Value)(this->pbase->hptrIcon);
#endif
   }
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setIconText                                                |
|                                                                              |
| Indicate the text for Icon View.  Refresh if necessary.                      |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setIconText(const IString& strIconNew)
{
  this->pbase->strIcon = strIconNew;
#ifdef IC_NOBINARYCOMP
  this->pbase->cRecordText = (char*)(this->pbase->strIcon);
#endif
  if(isRefreshOn())
    refresh();
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setIconText                                                |
|                                                                              |
| Indicate the text for Icon View.                                             |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setIconText(const char* pszIconNew)
{
   IString strIconNew(pszIconNew);
   setIconText(strIconNew);
   return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::setIconText                                                |
|                                                                              |
| Indicate the text for Icon View from resource id.                            |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setIconText(const IResourceId& resid)
{
   IString strText = resid.resourceLibrary().loadString(resid);
   setIconText(strText);
   return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::iconText                                                   |
|                                                                              |
| Query an object's icon text.                                                 |
------------------------------------------------------------------------------*/
IString IContainerObject :: iconText() const
{
   return this->pbase->strIcon;
}

/*------------------------------------------------------------------------------
| IContainerObject::icon                                                       |
|                                                                              |
| Query an object's icon .                                                     |
------------------------------------------------------------------------------*/
IPointerHandle IContainerObject :: icon() const
{
  return this->pbase->hptrIcon;
}


/*------------------------------------------------------------------------------
| IContainerObject::setInUse                                                   |
|                                                                              |
| Set the object's in-use emphasis                                             |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setInUse(Boolean fInUse,
                                               IContainerControl* pcnrctl )
{
   setEmphasis(CRA_INUSE, fInUse, pcnrctl);
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::removeInUse                                                |
|                                                                              |
| Remove the object's in-use emphasis.                                         |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: removeInUse(
                                               IContainerControl* pcnrctl )
{
   setEmphasis(CRA_INUSE, false, pcnrctl);
   return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::enableDataUpdate                                           |
|                                                                              |
| Make an object editable.                                                     |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: enableDataUpdate(
                                               Boolean fEnable,
                                               IContainerControl* pcnrctl)
{
   if(fEnable)
     setAttributes(CRA_RECORDREADONLY, 0, pcnrctl);
   else
     setAttributes(0, CRA_RECORDREADONLY, pcnrctl);
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::disableDataUpdate                                          |
|                                                                              |
| Make an object non-editable.                                                 |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: disableDataUpdate(
                                                 IContainerControl* pcnrctl)
{
   setAttributes(0, CRA_RECORDREADONLY, pcnrctl);
   return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::enableDrop                                                 |
|                                                                              |
| Allow other objects to be dropped on this object.                            |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: enableDrop(Boolean fEnable,
                                                 IContainerControl* pcnrctl)
{
   if(fEnable)
     setAttributes(0, CRA_DROPONABLE, pcnrctl);
   else
     setAttributes(CRA_DROPONABLE, 0, pcnrctl);
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::disableDrop                                                |
|                                                                              |
| Prevent any object from being dropped on this object.                        |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: disableDrop( IContainerControl* pcnrctl)
{
   setAttributes(CRA_DROPONABLE, 0, pcnrctl);
   return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::isDropOnAble                                               |
|                                                                              |
| Returns true if this object can accept a drop.                               |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject :: isDropOnAble(IContainerControl* pcnrctl) const
{
   return isAttribute(CRA_DROPONABLE, pcnrctl);
}


/*------------------------------------------------------------------------------
| IContainerObject::isInUse                                                    |
|                                                                              |
| Query if this object has in-use emphasis.                                    |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject :: isInUse(IContainerControl* pcnrctl) const
{
   return isAttribute(CRA_INUSE, pcnrctl);
}


/*------------------------------------------------------------------------------
| IContainerObject::isWriteable                                                |
|                                                                              |
| Query if this object can be edited.                                          |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject :: isWriteable(IContainerControl* pcnrctl) const
{
   return ( !isAttribute(CRA_RECORDREADONLY, pcnrctl) );
}

/*------------------------------------------------------------------------------
| IContainerObject::hide                                                       |
|                                                                              |
| Makes this object invisible by filtering it out of the collection.           |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject::hide(IContainerControl* pcnrctl)
{
  return show(false, pcnrctl);
}

/*------------------------------------------------------------------------------
| IContainerObject::isVisible                                                  |
|                                                                              |
| Determine if an object is filtered out of a container.                       |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject::isVisible(IContainerControl* pcnrctl) const
{
  return !isAttribute(CRA_FILTERED, pcnrctl);
}


/*------------------------------------------------------------------------------
| IContainerHideFilter::IContainerHideFilter                                   |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
IContainerHideFilter ::  IContainerHideFilter ( ) {}

/*------------------------------------------------------------------------------
| IContainerHideFilter::~IContainerHideFilter                                  |
|                                                                              |
| Default destructor here for page tuning.                                     |
------------------------------------------------------------------------------*/
IContainerHideFilter ::  ~IContainerHideFilter ( ) {}

/*------------------------------------------------------------------------------
| IContainerHideFilter::isMemberOf                                             |
|                                                                              |
| Determine if an object is filtered out of a container.                       |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerHideFilter :: isMemberOf( IContainerObject* object,
                                                   IContainerControl* ) const
{
  IMiniCnrRecord* pcnrrec = IRecFromObj(object);
  if (pcnrrec->flRecordAttr & CRA_FILTERED)
     return false;
  else
     return true;
}


/*------------------------------------------------------------------------------
| IContainerObject::incrementUseCount                                          |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::  incrementUseCount()
{
  ((IContainerObject*)this)->ulClUseCount++;
  return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::decrementUseCount                                          |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::  decrementUseCount()
{
  ((IContainerObject*)this)->ulClUseCount--;
  return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setBase                                                    |
|                                                                              |
| Stores the address of the real base of this object.                          |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setBase( const IMiniCnrRecord* pcnrrec)
{
  this->pbase = (IMiniCnrRecord*)pcnrrec;
  return *this;
}


/*------------------------------------------------------------------------------
| IContainerObject::setRefreshOff                                              |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setRefreshOff()
{
  flClState |=  noRefresh;
  return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setRefreshOn                                               |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject :: setRefreshOn(IBase::Boolean fOn)
{
  if ( fOn )
     flClState &= ~noRefresh;
  else
     flClState |= noRefresh;
  return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::isRefreshOn                                                |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject :: isRefreshOn() const
{
  return ((flClState&noRefresh) ? false : true);
}

/*------------------------------------------------------------------------------
| IContainerObject::setClosed                                                  |
|                                                                              |
| Removes the open emphasis in all containers.                                 |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::setClosed()
{
  flClState &= ~opened;
  removeInUse();
  return *this;
}

/*------------------------------------------------------------------------------
| IContainerObject::setOpen                                                    |
|                                                                              |
| Sets the open emphasis.                                                      |
------------------------------------------------------------------------------*/
IContainerObject& IContainerObject ::setOpen(IBase::Boolean fOpen)
{
  if(fOpen)
  {
    flClState |= opened;
    setInUse();
    return *this;
  }
  return setClosed();
}

/*------------------------------------------------------------------------------
| IContainerObject::handleOpen                                                 |
|                                                                              |
| Called when Enter is pressed on an object or a mouse double-click occurs on  |
| an object and the IContainerControl containing the object has been provided  |
| with an ICnrHandler.                                                         |
------------------------------------------------------------------------------*/
void IContainerObject :: handleOpen ( IContainerControl* )
{
  setOpen();
}

/*------------------------------------------------------------------------------
| IContainerObject::isOpen                                                     |
|                                                                              |
| Returns true if this object is open.                                         |
------------------------------------------------------------------------------*/
IBase::Boolean IContainerObject :: isOpen() const
{
  return isInUse();
}

/*------------------------------------------------------------------------------
| IContainerObject::handleCursoredChange                                       |
|                                                                              |
| Called when cursored emphasis changes on an object.                          |
------------------------------------------------------------------------------*/
void IContainerObject :: handleCursoredChange ( IContainerControl*,
                                                IBase::Boolean )
{  }

/*------------------------------------------------------------------------------
| IContainerObject::handleSelectedChange                                       |
|                                                                              |
| Called when selection emphasis changes on an object and the IContainerControl|
| containing the object has been provided with an ICnrHandler.                 |
------------------------------------------------------------------------------*/
void IContainerObject :: handleSelectedChange ( IContainerControl*,
                                                IBase::Boolean )
{  }

/*------------------------------------------------------------------------------
| IContainerObject::handleInuseChange                                          |
|                                                                              |
| Called when in-use emphasis changes on an object and the IContainerControl   |
| containing the object has been provided with an ICnrHandler.                 |
------------------------------------------------------------------------------*/
void IContainerObject :: handleInuseChange ( IContainerControl*,
                                             IBase::Boolean )
{  }

/*------------------------------------------------------------------------------
| IContainerObject::handleTreeCollapse                                         |
|                                                                              |
| Called when this object is a parent in the tree view and the descendent      |
| objects have been collapsed.                                                 |
------------------------------------------------------------------------------*/
void IContainerObject :: handleTreeCollapse ( IContainerControl* )
{  }

/*------------------------------------------------------------------------------
| IContainerObject::handleTreeExpand                                           |
|                                                                              |
| Called when this object is a parent in the tree view and the descendent      |
| objects have been expanded.                                                  |
------------------------------------------------------------------------------*/
void IContainerObject :: handleTreeExpand ( IContainerControl* )
{  }

/*------------------------------------------------------------------------------
| IContainerObject::helpId                                                     |
|                                                                              |
| Called when an ICnrHelpEvent is dispatched to ICnrHandler.                   |
------------------------------------------------------------------------------*/
unsigned long IContainerObject :: helpId() const
{
  return 0;
}

/*------------------------------------------------------------------------------
| ICnrAllocator::ICnrAllocator                                                 |
|                                                                              |
| ICnrAllocator constructor.                                                   |
------------------------------------------------------------------------------*/
ICnrAllocator::ICnrAllocator(unsigned long num, unsigned long size)
{
  IMiniCnrRecord* pcnrRec;
  if (num)
  {
    pcnrRec = (IMiniCnrRecord*)(void*)
                  IContainerControl::hwndAllocation().sendEvent(
                                                  CM_ALLOCRECORD,
                                                  MPFROMLONG(size),
                                                  MPFROMLONG(num));
    if (!pcnrRec)
    {
      ITHROWGUIERROR2("CM_ALLOCRECORD",
                      IErrorInfo::outOfMemory,
                      IException::recoverable);
    }
  }
  else
  {
    ITHROWLIBRARYERROR(IC_CNRALLOCATOR_INVALIDNUMVALUE,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);
  }
  this->pFirst     = pcnrRec;
  this->pAvailable = pcnrRec;
  this->numInitialized = 0;
}

/*------------------------------------------------------------------------------
| ICnrAllocator::~ICnrAllocator                                                |
|                                                                              |
| ICnrAllocator destructor.                                                    |
------------------------------------------------------------------------------*/
ICnrAllocator::~ICnrAllocator()
{
  void* pFreeRec = this->nextAvailable();

  while (pFreeRec)
  {
    int fSuccess = (int)IContainerControl::hwndAllocation().sendEvent(
                                           CM_FREERECORD,
                                           MPFROMP(&pFreeRec),
                                           MPFROMSHORT(1));
    if (fSuccess == -1)
    {
      ITHROWGUIERROR("CM_FREERECORD");
    }
    pFreeRec = this->nextAvailable();
  }
}

/*------------------------------------------------------------------------------
| ICnrAllocator::nextAvailable                                                 |
|                                                                              |
| Returns the next available item that has not been initialized.               |
------------------------------------------------------------------------------*/
void* ICnrAllocator::nextAvailable()
{
  void* pAvail = pAvailable;
  if (pAvailable)
  {
    pAvailable = ((IMiniCnrRecord*)pAvailable)->preccNextRecord;
  }
  return pAvail;
}

/*------------------------------------------------------------------------------
| ICnrAllocator::next                                                          |
|                                                                              |
| Returns the next item of the item passed in.                                 |
------------------------------------------------------------------------------*/
void* ICnrAllocator::next(void* pRec)
{
  return ((IMiniCnrRecord*)pRec)->preccNextRecord;
}

/*------------------------------------------------------------------------------
| ICnrAllocator::updateForInsert                                               |
|  Updates the state of an ICnrAllocator object after addObjects or            |
|  addObjectsAfter is called.                                                  |
------------------------------------------------------------------------------*/
ICnrAllocator& ICnrAllocator::updateForInsert()
{
  this->pFirst = this->pAvailable;
  this->numInitialized = 0;
  return *this;
}
