/*******************************************************************************
* FILE NAME: icombobs.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icombobs.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

/**********************************************************************/
/* Need IC_USE_CB defined before iwindefs since Windows macros are    */
/* different if processing listboxes that are part of comboboxes      */
/**********************************************************************/
extern "C" {
  #define IC_USE_CB
  #define INCL_WINENTRYFIELDS       // ES_xxx, CBS_xxx, CBID_EDIT.
  #define INCL_WINERRORS            // PMERR_HEAP_MAX_SIZE_REACHED
  #define INCL_WININPUT
  #define INCL_WINLISTBOXES         // LS_HORZSCROLL
  #define INCL_WINMESSAGEMGR        // For WM_PRESPARAMCHANGED.
  #define INCL_NLS                  // ES_ANY, ES_SBCS, ES_DBCS, ES_MIXED
  #define INCL_WINWINDOWMGR         // WinQueryClassName
  #define INCL_WINSYS               // SV_CYBORDER, PP_FONTNAMESIZE, etc.
  #define INCL_GPICONTROL           // From ILISTBX2.HPP
  #define INCL_GPILCIDS             //
  #define INCL_GPILOGCOLORTABLE     //
  #define INCL_GPIPRIMITIVES        //

  #include <iwindefs.h>
  #include <stdlib.h>            // atoi
  #include <string.h>
}

#include <inotifev.hpp>
#include <icombobs.hpp>

#include <iapp.hpp>
#include <icconst.h>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <ihandle.hpp>
#include <ilistbox.hpp>
#include <imphdr.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iwcname.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICOMBOBS_CPP_
  #include <ipagetun.h>
#endif

/******************************************************************************/
/* Public combobox styles.                                                    */
/******************************************************************************/
#pragma data_seg(ICLStaticConst)
const IBaseComboBox::Style
  IBaseComboBox::simpleType           = CBS_SIMPLE,
  IBaseComboBox::dropDownType         = CBS_DROPDOWN,
  IBaseComboBox::horizontalScroll     = LS_HORZSCROLL,
  IBaseComboBox::border3D             ( 0, IWS_BORDER3D ),
  IBaseComboBox::readOnlyDropDownType = CBS_DROPDOWNLIST,
  IBaseComboBox::anyData              ( 0, IES_ANYDATA ),
  IBaseComboBox::sbcsData             = ES_SBCS,
  IBaseComboBox::dbcsData             = ES_DBCS,
  IBaseComboBox::mixedData            ( 0, IES_MIXEDDATA ),
  IBaseComboBox::autoScroll           = ES_AUTOSCROLL,
  IBaseComboBox::classDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE), IES_ANYDATA );

const unsigned long    IBaseComboBox::notFound = (unsigned long)ULONG_MAX;
const unsigned long    IBaseComboBox::first    = (unsigned long)ULONG_MAX;
#pragma data_seg()

/******************************************************************************/
/* Default style for new objects (initial value)                              */
/******************************************************************************/
#pragma data_seg(ICLNonConst)
IBaseComboBox::Style
  IBaseComboBox::currentDefaultStyle ( (WS_VISIBLE | CBS_SIMPLE), IES_ANYDATA );
#pragma data_seg()


class IBaseComboBoxData : public IBase {
/*******************************************************************************
* The IBaseComboBoxData class encapsulates private data and functions used by the  *
* IBaseComboBox class.  An object of this class is created in the IBaseComboBox        *
* constructors.                                                                *
*******************************************************************************/
public:
  IBaseComboBoxData() : minimumRows (4),
                    minimumCharacters (25)
                    {;}

 ~IBaseComboBoxData() {;}

  unsigned long    minimumRows,
                   minimumCharacters;

//  Returns the handle of the entry field part of a combo box
static IWindowHandle
  handleOfEntryField(const IWindowHandle& combobox );

IWinProc
 *fdefaultEntryProc; // Default procedure for Entry field part of combo
IWindowHandle
  fentryHandle;      // handle of the entry field

private:
  IBaseComboBoxData::IBaseComboBoxData (const IBaseComboBoxData&);
  IBaseComboBoxData& operator= (const IBaseComboBoxData&);
};

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                         |
|                                                                              |
| Construct an IBaseComboBox on an IWindow.                                        |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox(unsigned long id,
                     IWindow* pParent,
                     IWindow* pOwner,
                     const IRectangle& rectInit,
                     const Style& style)
                   : fBaseComboBoxData( 0 )
{
  // assertions on input parms
  IASSERTPARM(pParent!=0);
  IASSERTPARM((style & (simpleType | dropDownType | readOnlyDropDownType))
               !=0);
  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  // Create a private data object.
  fBaseComboBoxData = new IBaseComboBoxData;
  IWindowHandle owner = (pOwner == 0) ? IWindowHandle(0) : pOwner->handle();

  IWindowHandle whComboBox =
      this -> create( id,
                      0,
                      convertToGUIStyle( style ),
                      WC_COMBOBOX,
                      pParent->handle(),
                      owner,
                      rectInit,
                      0,
                      0 );

  startHandlingEventsFor(whComboBox);
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

  // Subclass the entry field in the combo box so that we can get
  // navigation keys.
  fBaseComboBoxData->fentryHandle =
     IBaseComboBoxData::handleOfEntryField( whComboBox );
  if (fBaseComboBoxData->fentryHandle)
  {
    // setup our procedure in the window.
    fBaseComboBoxData->fdefaultEntryProc = (IWinProc*)
        ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                        icomboboxEntryWinProc);
    if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
    {
      // throw now rather than trap later.
      ITHROWGUIERROR( "ISUBCLASSWINDOW" );
    }
  }
}

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                         |
|                                                                              |
| Construct an IBaseComboBox on a dialog.                                          |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox( unsigned long id, IWindow* parent)
                            : fBaseComboBoxData( 0 )
{
   // Create a private data object.
   fBaseComboBoxData = new IBaseComboBoxData;

   setAutoDestroyWindow(false);
   startHandlingEventsFor(id, parent);
   IMousePointerHandler::defaultHandler()->handleEventsFor(this);

   // Subclass the entry field in the combo box so that we can get
   // navigation keys.
   fBaseComboBoxData->fentryHandle =
      IBaseComboBoxData::handleOfEntryField( this->handle() );
   if (fBaseComboBoxData->fentryHandle)
   {
     // setup our procedure in the window.
     fBaseComboBoxData->fdefaultEntryProc = (IWinProc*)
         ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                         icomboboxEntryWinProc);
     if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
     {
       // throw now rather than trap later.
       ITHROWGUIERROR( "ISUBCLASSWINDOW" );
     }
   }
}

/*------------------------------------------------------------------------------
| IBaseComboBox::IBaseComboBox                                                         |
|                                                                              |
| Construct an IBaseComboBox from an existing IBaseComboBox.                           |
------------------------------------------------------------------------------*/
IBaseComboBox::IBaseComboBox( const IWindowHandle& handle )
                            : fBaseComboBoxData( 0 )
{
   // Create a private data object.
   fBaseComboBoxData = new IBaseComboBoxData;

   setAutoDestroyWindow(false);
   startHandlingEventsFor(handle);
   IMousePointerHandler::defaultHandler()->handleEventsFor(this);

   // Subclass the entry field in the combo box so that we can get
   // navigation keys.
   fBaseComboBoxData->fentryHandle =
      IBaseComboBoxData::handleOfEntryField( handle );
   if (fBaseComboBoxData->fentryHandle)
   {
     // setup our procedure in the window.
     fBaseComboBoxData->fdefaultEntryProc = (IWinProc*)
        ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                        icomboboxEntryWinProc);
     if ( fBaseComboBoxData->fdefaultEntryProc == 0 )
     {
       // throw now rather than trap later.
       ITHROWGUIERROR( "ISUBCLASSWINDOW" );
     }
   }
}

/*------------------------------------------------------------------------------
| IBaseComboBox::~IBaseComboBox                                                        |
|                                                                              |
| Reomve the default mouse pointer handler.                                    |
------------------------------------------------------------------------------*/
IBaseComboBox :: ~IBaseComboBox()
{
  // Remove the subclass of the entry field
  if ( (fBaseComboBoxData->fentryHandle) &&
        IISWINDOW(0, fBaseComboBoxData->fentryHandle) )
  {
    ISUBCLASSWINDOW(fBaseComboBoxData->fentryHandle,
                    fBaseComboBoxData->fdefaultEntryProc);
  }

  IMousePointerHandler::defaultHandler()->stopHandlingEventsFor(this);
  if ( fBaseComboBoxData != 0 )
    delete fBaseComboBoxData;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::type                                                              |
------------------------------------------------------------------------------*/
IBaseComboBox::ControlType  IBaseComboBox::type() const
{
  unsigned long ulType =
           style() & (simpleType.asUnsignedLong()  |
                      dropDownType.asUnsignedLong() |
                      readOnlyDropDownType.asUnsignedLong());
  ControlType eReturn = simple;


  switch (ulType)
  {
     case CBS_SIMPLE:
        break;
     case CBS_DROPDOWN:
        eReturn = dropDown;
        break;
     case CBS_DROPDOWNLIST:
        eReturn = dropDownList;
        break;
     default:
        break;
  }

  return eReturn;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::defaultStyle                                                      |
------------------------------------------------------------------------------*/
IBaseComboBox::Style  IBaseComboBox::defaultStyle()
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::setDefaultStyle                                                   |
------------------------------------------------------------------------------*/
void IBaseComboBox::setDefaultStyle(const Style& style)
{
  currentDefaultStyle = style;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IBaseComboBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // Obtain the style from the class (IEntryField) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean CBS_ styles, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & ICBS_MASK;

  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| IBaseComboBox::isListShowing                                                     |
------------------------------------------------------------------------------*/
Boolean   IBaseComboBox::isListShowing() const
{
  if (type() != simple)
  {
     IEventResult evt = handle().sendEvent(CBM_ISLISTSHOWING,
                                IEventParameter1(0),
                                IEventParameter2(0));
     if (evt.asUnsignedLong())
        return true;
     else
        return false;
  }
  else
     return true;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::showList                                                          |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox::showList(Boolean show)
{
  Boolean stateChanged = false;

  if (type() != simple)
  {
    if (show)
    {
      if (!isListShowing())
      {
        stateChanged = true;
      }
    }
    else if (isListShowing())
    {
      stateChanged = true;
    }

    if (stateChanged)
    {
      handle().sendEvent(CBM_SHOWLIST, IEventParameter1(show),
                                       IEventParameter2(0));
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::hideList                                                          |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox::hideList()
{
  showList(false);
  return *this;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::hasFocus                                                      |
| Cannot add override because not virtual at IWindow level - future change     |
------------------------------------------------------------------------------*/
//IBase::Boolean IBaseComboBox::hasFocus() const
//{
//   if ( style() & readOnlyDropDownType.asUnsignedLong() )
//     return Inherited::hasFocus();      // static text type combobox
//   else
//   {                                    // has an Edit window
//     HWND hwndCombo = handle();
//     HWND hwndEdit = this->fBaseComboBoxData->fentryHandle;
//     HWND hwndFocus = IQUERYFOCUS( HWND_DESKTOP );
//     // Check both combobox and entryfield.  If either has focus, then true.
//     return (( hwndCombo == hwndFocus ) || ( hwndEdit == hwndFocus ));
//   }
//}

/*------------------------------------------------------------------------------
| IBaseComboBox::calcMinimumSize                                                   |
|                                                                              |
| Calculate the minimum size needed by the control.  The height is based on    |
| the font height and the width is based on the average font width multiplied  |
| by the text limit.                                                           |
------------------------------------------------------------------------------*/
ISize  IBaseComboBox::calcMinimumSize() const
{
  ISize sizMin = Inherited::calcMinimumSize();
  IBaseComboBox::ControlType comboType = type();
  // If we aren't a drop-down, add room for minimum rows
  if ( comboType != dropDown && comboType != dropDownList)
  {
    sizMin.setHeight ( sizMin.height() +
                       minimumRows() * IFont(this).maxCharHeight() +
                       IQUERYSYSVALUE(SV_CYBORDER));
    // Note: Only the bottom border is added to the height since the
    // listbox top border overlaps the entry field bottom border
  }

  // Add size for the entry field border (doesn't seem to happen
  // via entry field calculation)
  sizMin += ISize((3 * IQUERYSYSVALUE(SV_CXBORDER) + 2),
                     (3 * IQUERYSYSVALUE(SV_CYBORDER) + 2));
  return sizMin;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::layoutAdjustment                                                  |
|                                                                              |
| Calculate the size and distance a dropDown Combination box should be         |
| adjusted after layout.                                                       |
------------------------------------------------------------------------------*/
IRectangle  IBaseComboBox::layoutAdjustment() const
{
  IBaseComboBox::ControlType comboType = type();
  if ( comboType == dropDown || comboType == dropDownList)
  {
      // Adjust position so that minimum rows of text will fit in the
      // drop down listbox
      IRectangle rect( 0, -(minimumRows() * IFont(this).maxCharHeight() +
                               IQUERYSYSVALUE(SV_CYBORDER)),
                       0, 0);
      // lowerLeft to upperLeft coordinates.  Rectangle should represent
      // area of the drop down list relative to the entry field
      if ( ICoordinateSystem::isConversionNeeded() )
         rect.moveBy( IPair(0, rect.height()) );
      return rect;
  }
  else
     return Inherited::layoutAdjustment();
}

/*------------------------------------------------------------------------------
| IBaseComboBox::visibleRectangle                                              |
|                                                                              |
| Calculate the actual displayed (painted) ares of the combo box.              |
------------------------------------------------------------------------------*/
IRectangle  IBaseComboBox::visibleRectangle() const
{
  IRectangle    trueRect = nativeRect();
  SWP           swp;

  IBaseComboBox::ControlType comboType = type();
  if ( comboType == dropDown || comboType == dropDownList)
  {
    IQUERYWINDOWPOS(IWINDOWFROMID(handle(),CBID_EDIT),&swp);
    unsigned long cyBorder = IQUERYSYSVALUE(SV_CYBORDER);
    trueRect = IRectangle ( IPoint(trueRect.left(),
                                   trueRect.bottom()+swp.y-(2*cyBorder)),
                            trueRect.topRight() );
  }
  else
  {
    IQUERYWINDOWPOS(IWINDOWFROMID(handle(),CBID_LIST),&swp);
    trueRect = IRectangle ( IPoint(trueRect.left(),
                            trueRect.bottom()+swp.y),
                            trueRect.topRight() );
  }
  return ICoordinateSystem::convertToApplication( trueRect, parentSize() );
}

/*------------------------------------------------------------------------------
| IBaseComboBox::setMinimumRows                                                |
|                                                                              |
| Sets the number of rows in a minimum size combobox.                          |
------------------------------------------------------------------------------*/
IBaseComboBox&  IBaseComboBox :: setMinimumRows (unsigned long minimumRows)
{
  fBaseComboBoxData->minimumRows = minimumRows;
  setLayoutDistorted (minimumSizeChanged, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| IBaseComboBox::minimumRows                                                   |
|                                                                              |
| Returns the number of rows in a minimum size combobox.                       |
------------------------------------------------------------------------------*/
unsigned long  IBaseComboBox :: minimumRows () const
{
  return fBaseComboBoxData->minimumRows;
}


/*------------------------------------------------------------------------------
| IBaseComboBox::setAlignment                                                  |
|                                                                              |
| Replaces the text alignment style of the entry field                         |
------------------------------------------------------------------------------*/
IBaseComboBox& IBaseComboBox::setAlignment(IEntryField::Alignment alignment)
{
   ITHROWLIBRARYERROR1(IC_MEMBER_ACCESS_ERROR,
                       IErrorInfo::invalidRequest,
                       IException::recoverable,
                       "IBaseComboBox");
   return *this;
}


/*------------------------------------------------------------------------------
| IBaseComboBoxData::handleOfEntryField                                        |
------------------------------------------------------------------------------*/
IWindowHandle IBaseComboBoxData::handleOfEntryField(
                                         const IWindowHandle& combobox )
{
   // Retrieve the edit control child of the combobox.
   IWindowHandle hwndChild = IWINDOWFROMID( combobox, CBID_EDIT );
   return hwndChild;
}

/*------------------------------------------------------------------------------
| icomboboxEntryWinProc                                                        |
| Window procedure for the entryfield part of a combo box.  This procedure     |
| intercepts the messages used for keyboard navigation and sends them to       |
| the combobox owner.  Other messages are sent on to the original procedure    |
| for the combobox entryfield.                                                 |
------------------------------------------------------------------------------*/
void*   _System icomboboxEntryWinProc( unsigned long
                                                     hwnd,
                                       unsigned long msg,
                                       void* mp1, void* mp2 )
{
   IWindowHandle  parent   = IPARENTOF(hwnd);
   IBaseComboBox* combobox = (IBaseComboBox*)
      IWindow::windowWithHandle( parent, false );
   if (combobox)
   {
     if ( msg == WM_PRESPARAMCHANGED )
     {
       char fontName[120];   // Use same size as in IFont ctor
       unsigned long fontNameLen =
           WinQueryPresParam( hwnd, PP_FONTNAMESIZE, 0, 0,
                              sizeof( fontName ), &fontName,
                              QPF_NOINHERIT );
       if ( fontNameLen )
       {
         if (( ((long)mp1) == PP_FONTNAMESIZE) ||
                                      (((long)mp1) == PP_FONTHANDLE))
         {
           WinSetPresParam( combobox->handle(), PP_FONTNAMESIZE,
                            fontNameLen+1, (char*)fontName );
           WinRemovePresParam( hwnd, PP_FONTNAMESIZE );
         }
       }
     }
     return combobox->fBaseComboBoxData->fdefaultEntryProc(
                                                  hwnd, msg, mp1, mp2 );
   }
   else
   {
      // We are basically busted if we can't find our parent IWindow.  We do
      // what is probably the most reasonable thing and call the
      // default procedure.
      return IDEFWINDOWPROC( hwnd, msg, mp1, mp2 );
   }
}

#include <ilistbx3.hpp>

