/*******************************************************************************
* FILE NAME: idmefit.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMEFItem                                                                *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_WINPOINTERS
  #define INCL_DOSPROCESS
  #include <os2.h>
}

// tracing functions
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

// Exception handling
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _IDMEFIT_
  #include <idmefit.hpp>
#endif

#ifndef _IDMEVENT_
  #include <idmevent.hpp>
#endif

#ifndef _IDMSRCOP_
  #include <idmsrcop.hpp>
#endif

#ifndef _IENTRYFD_
  #include <ientryfd.hpp>
#endif

#ifndef _IDMIMAGE_
  #include <idmimage.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDMEFIT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IDMEFItem::IDMEFItem                                                         |
|                                                                              |
| Constructor with source item arguments.                                      |
------------------------------------------------------------------------------*/
IDMEFItem::IDMEFItem(IDMSourceOperation *pDMSrcOp)  :
                     IDMItem(pDMSrcOp,
                             IDM::text,
                             (IDMItem::moveable | IDMItem::copyable),
                             none)
{
  IMODTRACE_DEVELOP( "IDMEFItem::ctor (source)" );

  /********************************************************************/
  /* Obtain pointer to entry field control from the source window     */
  /********************************************************************/
  IEntryField *pEF = (IEntryField *)pDMSrcOp->sourceWindow();

  /********************************************************************/
  /* Get the actual text, selected characters only if selection       */
  /* in use.                                                          */
  /********************************************************************/
  setContents(((pEF->hasSelectedText()) ? pEF->selectedText()  :
                                          pEF->text()));

  /********************************************************************/
  /* Set the native renderering mechanism and format for the item.    */
  /********************************************************************/
  IString tempStr = generateSourceName();
  if (tempStr.size())
  {
    setSourceName(tempStr);
    setRMFs(rmfsFrom(IDM::rmLibrary,
                     rfForThisProcess() + "," + IDM::rfText));
  }
  else
  {
    setRMFs(rmfsFrom(IDM::rmLibrary,
                     rfForThisProcess() + "," + IDM::rfSharedMem));

    /******************************************************************/
    /* Use render prepare event to determine size of transfer buffer  */
    /******************************************************************/
    setRequiresPreparation(true);
  }

  /********************************************************************/
  /* Add a drag image for the item.                                   */
  /********************************************************************/
  IDMImage image = IDMImage(ISystemPointerHandle(
                            ISystemPointerHandle::text));
  setImage(image);
}


/*------------------------------------------------------------------------------
| IDMEFItem::IDMEFItem                                                         |
|                                                                              |
| Constructor with target item arguments.                                      |
------------------------------------------------------------------------------*/
IDMEFItem::IDMEFItem(const IDMItem::Handle &pEFItem)  :
                     IDMItem(pEFItem)
{
  IMODTRACE_DEVELOP( "IDMEFItem::ctor (target)" );

  /********************************************************************/
  /* Verify that one of these rmfs is supported by the source.        */
  /********************************************************************/
  IString defaultRMFs = rmfsFrom(IDM::rmLibrary,
                                      rfForThisProcess() + "," +
                                      IDM::rfText + "," +
                                      IDM::rfSharedMem);
  IString rmf = matchingRMFs (rmfsFrom( rmfs().change(' ', "") ),
                              defaultRMFs);

  /******************************************************************/
  /* The RMFs that IDMEFItem understands are                        */
  /* (IDM::rmLibrary)x                                              */
  /*           (IDM::rfProcess=pid,IDM::rfText,IDM::sharedMem)      */
  /******************************************************************/
  setRMFs(rmf);
}

/*------------------------------------------------------------------------------
| IDMEFItem::~IDMEFItem                                                        |
|                                                                              |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
IDMEFItem :: ~IDMEFItem()
{
  IMODTRACE_DEVELOP( "IDMEFItem::dtor" );
}

/*------------------------------------------------------------------------------
| IDMEFItem::generateSourceItems                                               |
|                                                                              |
| Gives IDMEFItem an opportunity to attach new IDMEFItem's to the argument     |
| IDMSourceOperation object.  This function is called when a drag operation    |
| commences over an IEntryField object.  Returns true if items were provided,  |
| false otherwise.                                                             |
------------------------------------------------------------------------------*/
Boolean IDMEFItem ::
        generateSourceItems(IDMSourceOperation *pDMSrcOp)
{
  IMODTRACE_DEVELOP( "IDMEFItem::generateSourceItems" );

  /********************************************************************/
  /* Obtain pointer to entry field control from the source window     */
  /********************************************************************/
  IEntryField *pEF = (IEntryField *)pDMSrcOp->sourceWindow();

  /********************************************************************/
  /* Do not attempt drag if the entry field is empty                  */
  /********************************************************************/
  if (pEF->isEmpty())
    return(false);

  /********************************************************************/
  /* Create the entry field drag item                                 */
  /********************************************************************/
  IDMEFItem *pEFItem = new IDMEFItem(pDMSrcOp);

  /********************************************************************/
  /* Add entry field drag item to the source operation.               */
  /********************************************************************/
  pDMSrcOp->addItem(pEFItem);

  /********************************************************************/
  /* Set the drag image style.                                        */
  /********************************************************************/
  pDMSrcOp->setImageStyle(IDM::allStacked);

  /********************************************************************/
  /* Indicate that the entry field drag item was successfully created.*/
  /********************************************************************/
  return(true);
}

/*------------------------------------------------------------------------------
| IDMEFItem::object                                                            |
|                                                                              |
| Returns a pointer to the item text.                                          |
------------------------------------------------------------------------------*/
void* IDMEFItem :: object() const
{
  return((void *)&(this->strContents));
}

/*------------------------------------------------------------------------------
| IDMEFItem::targetDrop                                                        |
|                                                                              |
| Writes text into the target entry field.                                     |
------------------------------------------------------------------------------*/
Boolean IDMEFItem :: targetDrop(IDMTargetDropEvent &)
{
  IMODTRACE_DEVELOP( "IDMEFItem::targetDrop" );

  /********************************************************************/
  /* Cast IWindow pointer to entry field                              */
  /********************************************************************/
  IEntryField *pEF = (IEntryField *)this->targetOperation()->targetWindow();

  /********************************************************************/
  /* Add the text to the entry field control                          */
  /********************************************************************/
  pEF->setText(contents());

  return(true);
}

