/*******************************************************************************
* FILE NAME: idmmleit.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMMLEItem                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  // define what subsections of PM headers to include
  #define INCL_WINPOINTERS
  #define INCL_DOSPROCESS
  #include <os2.h>
}

// tracing functions
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

// Exception handling
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

// include our classes's stuff
#ifndef _IDMMLEIT_
  #include <idmmleit.hpp>
#endif

#ifndef _IDMEVENT_
  #include <idmevent.hpp>
#endif

#ifndef _IDMSRCOP_
  #include <idmsrcop.hpp>
#endif

#ifndef _IMLE_
  #include <imle.hpp>
#endif

#ifndef _IDMIMAGE_
  #include <idmimage.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDMMLEIT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IDMMLEItem::IDMMLEItem                                                       |
|                                                                              |
| Constructor using source item arguments.                                     |
------------------------------------------------------------------------------*/
IDMMLEItem::IDMMLEItem(IDMSourceOperation *pDMSrcOp)  :
                       IDMItem(pDMSrcOp,
                               IDM::text,
                               (IDMItem::moveable | IDMItem::copyable),
                               none)
{
  IMODTRACE_DEVELOP( "IDMMLEItem::ctor (source)" );

  /********************************************************************/
  /* Obtain pointer to entry field control from the source window     */
  /********************************************************************/
  IMultiLineEdit *pMLE = (IMultiLineEdit *)pDMSrcOp->sourceWindow();

  /********************************************************************/
  /* Get the actual text, selected characters only if selection       */
  /* in use.                                                          */
  /********************************************************************/
  setContents(((pMLE->hasSelectedText()) ? pMLE->selectedText()  :
                                           pMLE->text()));

  /********************************************************************/
  /* Set the native renderering mechanism and format for the item.    */
  /********************************************************************/
  IString tempStr = generateSourceName();
  if (tempStr.size())
  {
    setSourceName(tempStr);
    setRMFs(rmfsFrom(IDM::rmLibrary,
                     rfForThisProcess() + "," + IDM::rfText));
  }
  else
  {
    setRMFs(rmfsFrom(IDM::rmLibrary,
                     rfForThisProcess() + "," + IDM::rfSharedMem));

    /******************************************************************/
    /* Use render prepare event to determine size of transfer buffer  */
    /******************************************************************/
    setRequiresPreparation(true);
  }

  /********************************************************************/
  /* Add a drag image for the item.                                   */
  /********************************************************************/
  IDMImage image = IDMImage(ISystemPointerHandle(
                            ISystemPointerHandle::text));
  setImage(image);
}


/*------------------------------------------------------------------------------
| IDMMLEItem::IDMMLEItem                                                       |
|                                                                              |
| Constructor using target item arguments.                                     |
------------------------------------------------------------------------------*/
IDMMLEItem::IDMMLEItem(const IDMItem::Handle &pMLEItem)  :
                       IDMItem(pMLEItem)
{
  IMODTRACE_DEVELOP( "IDMMLEItem::ctor (target)" );

  /********************************************************************/
  /* Verify that one of these rmfs is supported by the source.        */
  /********************************************************************/
  IString defaultRMFs = rmfsFrom(IDM::rmLibrary,
                                      rfForThisProcess() + "," +
                                      IDM::rfText + "," +
                                      IDM::rfSharedMem);
  IString rmf = matchingRMFs (rmfsFrom( rmfs().change(' ', "") ),
                              defaultRMFs);

  /******************************************************************/
  /* The RMFs that IDMMLEItem understands are                       */
  /* (IDM::rmLibrary)x                                              */
  /*           (IDM::rfProcess=pid,IDM::rfText,IDM::sharedMem)      */
  /******************************************************************/
  setRMFs(rmf);
}

/*------------------------------------------------------------------------------
| IDMMLEItem::~IDMMLEItem                                                      |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IDMMLEItem :: ~IDMMLEItem()
{
  IMODTRACE_DEVELOP( "IDMMLEItem::dtor" );
}

/*------------------------------------------------------------------------------
| IDMMLEItem::generateSourceItems                                              |
|                                                                              |
| Called to give IDMMLEItem opportunity to attach new IDMMLEItems to the       |
| argument IDMSourceOperation object.  This function is called when a drag     |
| operation commences over an IMLE object.  Returns true if items were         |
| provided, false otherwise.                                                   |
------------------------------------------------------------------------------*/
Boolean IDMMLEItem ::
        generateSourceItems(IDMSourceOperation *pDMSrcOp)
{
  IMODTRACE_DEVELOP( "IDMMLEItem::generateSourceItems" );

  /********************************************************************/
  /* Obtain pointer to MLE control from the source window             */
  /********************************************************************/
  IMultiLineEdit *pMLE = (IMultiLineEdit *)pDMSrcOp->sourceWindow();

  /********************************************************************/
  /* Do not attempt drag if the MLE is empty                          */
  /********************************************************************/
  if (!(pMLE->textLength()))
    return(false);

  /********************************************************************/
  /* Create the MLE drag item                                         */
  /********************************************************************/
  IDMMLEItem *pMLEItem = new IDMMLEItem(pDMSrcOp);

  /********************************************************************/
  /* Add MLE drag item to the source operation.                       */
  /********************************************************************/
  pDMSrcOp->addItem(pMLEItem);

  /********************************************************************/
  /* Set the drag image style.                                        */
  /********************************************************************/
  pDMSrcOp->setImageStyle(IDM::allStacked);

  /********************************************************************/
  /* Indicate that the MLE drag item was successfully created.        */
  /********************************************************************/
  return(true);
}

/*------------------------------------------------------------------------------
| IDMMLEItem::object                                                           |
|                                                                              |
| Returns a pointer to the item text.                                          |
------------------------------------------------------------------------------*/
void* IDMMLEItem :: object() const
{
  return((void *)&(this->strContents));
}

/*------------------------------------------------------------------------------
| IDMMLEItem::targetDrop                                                       |
|                                                                              |
| Write text into entry field from drag item.                                  |
------------------------------------------------------------------------------*/
Boolean IDMMLEItem :: targetDrop(IDMTargetDropEvent &)
{
  IMODTRACE_DEVELOP( "IDMMLEItem::targetDrop" );

  /********************************************************************/
  /* Cast IWindow pointer to MLE                                      */
  /********************************************************************/
  IMultiLineEdit *pMLE =
                 (IMultiLineEdit *)this->targetOperation()->targetWindow();

  /********************************************************************/
  /* Add the text to the MLE control                                  */
  /********************************************************************/
  pMLE->add(contents(), contentsSize());

  return(true);
}

