/*******************************************************************************
* FILE NAME: idmrendr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMRenderer                                                              *
*       IDMSourceRenderer                                                      *
*       IDMTargetRenderer                                                      *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
// include PM headers
  #define INCL_WINSTDDRAG       // Standard drag stuff
  #include <os2.h>
}

#ifndef _IDMITEM_
  #include <idmitem.hpp>
#endif

#ifndef _IDMRENDR_
  #include <idmrendr.hpp>
#endif

#ifndef _IDMSRCRN_
  #include <idmsrcrn.hpp>
#endif

#ifndef _IDMTGTRN_
  #include <idmtgtrn.hpp>
#endif

#ifndef _IDMOPER_
  #include <idmoper.hpp>
#endif

#ifndef _IDMEVENT_
  #include <idmevent.hpp>
#endif

#ifndef _ICTLEVT_
  #include <ictlevt.hpp>
#endif

#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDMRENDR_CPP_
  #include <ipagetun.h>
#endif

typedef IDM::DropIndicator DropIndicator;

#pragma data_seg(ICLNonConst)
/*------------------------------------------------------------------------------
| Static objects to ensure default renderers get deleted at shutdown.          |
------------------------------------------------------------------------------*/
static struct IDMRenderer_Static {
static IDMSourceRenderer
 *pSrcRenderer;
static IDMTargetRenderer
 *pTgtRenderer;

  ~IDMRenderer_Static ( )
    {
    IMODTRACE_DEVELOP( "IDMRenderer_Static dtor" );
    delete pSrcRenderer;
    delete pTgtRenderer;
    }

} statics;

IDMSourceRenderer
 *IDMRenderer_Static::pSrcRenderer = 0;
IDMTargetRenderer
 *IDMRenderer_Static::pTgtRenderer = 0;

// IDMRenderer static members:
IDMSourceRenderer
 *IDMRenderer::pSource = 0;
IDMTargetRenderer
 *IDMRenderer::pTarget = 0;

#pragma data_seg()

/*------------------------------------------------------------------------------
| IDMTargetRendererData class definition                                       |
------------------------------------------------------------------------------*/
class IDMTargetRendererData : public IBase
{
public:
  IDMTargetRendererData  ( );

  ~IDMTargetRendererData ( );

unsigned long
  ulDragTransferCount;

_DRAGTRANSFER
 *pCurrentDragTransfer;

};

/*------------------------------------------------------------------------------
| IDMTargetRendererData::IDMTargetRendererData                                 |
|                                                                              |
| This class contains the target renderer data pointers.                       |
------------------------------------------------------------------------------*/
IDMTargetRendererData::IDMTargetRendererData()
  : pCurrentDragTransfer (0),
    ulDragTransferCount (0)
{ }

/*------------------------------------------------------------------------------
| IDMTargetRendererData::~IDMTargetRendererData                                |
|                                                                              |
| Destructor here for page tuning.                                             |
------------------------------------------------------------------------------*/
IDMTargetRendererData::~IDMTargetRendererData()
{ }

/*-------------------------------- BASE RENDERER ----------------------------*/

/*------------------------------------------------------------------------------
| IDMRenderer::IDMRenderer                                                     |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMRenderer :: IDMRenderer(const char *pRMFs, const char *pTypes)
  : IVBase(),
    strRMFs( IDMItem::rmfsFrom( pRMFs ) ),
    strTypes(pTypes)
{
}

/*------------------------------------------------------------------------------
| IDMRenderer::IDMRenderer                                                     |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMRenderer :: IDMRenderer()
  : IVBase(),
    strTypes(0),
    strRMFs(0)
{
  setSupportedTypes(IDM::any);
}

/*------------------------------------------------------------------------------
| IDMRenderer::~IDMRenderer                                                    |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IDMRenderer :: ~IDMRenderer()
{
}

/*------------------------------------------------------------------------------
| IDMRenderer::supportedTypes                                                  |
|                                                                              |
| Returns the types of items supported by the renderer.                        |
------------------------------------------------------------------------------*/
IString IDMRenderer :: supportedTypes() const
{
  return(this->strTypes);
}

/*------------------------------------------------------------------------------
| IDMRenderer::supportedRMFs                                                   |
|                                                                              |
| Returns RMFs types of items supported by the renderer.                       |
------------------------------------------------------------------------------*/
IString IDMRenderer :: supportedRMFs() const
{
  return(this->strRMFs);
}

/*------------------------------------------------------------------------------
| IDMRenderer::setSupportedTypes                                               |
|                                                                              |
| Sets the types of items supported by the renderer.                           |
------------------------------------------------------------------------------*/
IDMRenderer& IDMRenderer :: setSupportedTypes(const char *pTypes)
{
  IMODTRACE_DEVELOP("IDMRenderer::setSupportedTypes");

  this->strTypes = pTypes;
  return(*this);
}

/*------------------------------------------------------------------------------
| IDMRenderer::setSupportedRMFs                                                |
|                                                                              |
| Sets the RMFs supported by this renderer.                                    |
------------------------------------------------------------------------------*/
IDMRenderer& IDMRenderer :: setSupportedRMFs(const char *pRMFs)
{
  IMODTRACE_DEVELOP("IDMRenderer::setSupportedRMFs");

  this->strRMFs = IDMItem::rmfsFrom( pRMFs );
  return(*this);
}

/*------------------------------------------------------------------------------
| IDMRenderer::defaultSourceRenderer                                           |
|                                                                              |
| Return default source renderer, initializing if this is the first time.      |
------------------------------------------------------------------------------*/
IDMSourceRenderer &IDMRenderer :: defaultSourceRenderer ( )
  {
  if ( !IDMRenderer::pSource )
    {
    if ( !statics.pSrcRenderer )
      statics.pSrcRenderer = new IDMSourceRenderer;
    IDMRenderer::pSource = statics.pSrcRenderer;
    }
  return *IDMRenderer::pSource;
  }

/*------------------------------------------------------------------------------
| IDMRenderer::defaultTargetRenderer                                           |
|                                                                              |
| Return default target renderer, initializing if this is the first time.      |
------------------------------------------------------------------------------*/
IDMTargetRenderer &IDMRenderer :: defaultTargetRenderer ( )
  {
  if ( !IDMRenderer::pTarget )
    {
    if ( !statics.pTgtRenderer )
      statics.pTgtRenderer = new IDMTargetRenderer;
    IDMRenderer::pTarget = statics.pTgtRenderer;
    }
  return *IDMRenderer::pTarget;
  }

/*------------------------------------------------------------------------------
| IDMRenderer::setDefaultSourceRenderer                                        |
|                                                                              |
| Set default source renderer pointer to point to argument renderer.           |
------------------------------------------------------------------------------*/
void IDMRenderer :: setDefaultSourceRenderer ( IDMSourceRenderer &source )
  {
  IDMRenderer::pSource = &source;
  }

/*------------------------------------------------------------------------------
| IDMRenderer::setDefaultTargetRenderer                                        |
|                                                                              |
| Set default target renderer pointer to point to argument renderer.           |
------------------------------------------------------------------------------*/
void IDMRenderer :: setDefaultTargetRenderer ( IDMTargetRenderer &target )
  {
  IDMRenderer::pTarget = &target;
  }


/*-------------------------------- SOURCE RENDERER ---------------------------*/

/*------------------------------------------------------------------------------
| IDMSourceRenderer::IDMSourceRenderer                                         |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMSourceRenderer :: IDMSourceRenderer(const char *pRMFs,
                                       const char *pTypes)
  : IDMRenderer(pRMFs, pTypes)
{
  IMODTRACE_DEVELOP("IDMSourceRenderer::ctor (rmfs, types)");
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::IDMSourceRenderer                                         |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMSourceRenderer :: IDMSourceRenderer ()
  : IDMRenderer()
{
  IMODTRACE_DEVELOP("IDMSourceRenderer::ctor ()");

  setSupportedRMFs(IString("(")                 +
                   IDM::rmLibrary               +
                   ")x("                        +
                   IDMItem::rfForThisProcess()  +
                   ","                          +
                   IDM::rfText                  +
                   ","                          +
                   IDM::rfSharedMem             +
                   "),("                        +
                   IDM::rmDiscard               +
                   ","                          +
                   IDM::rmPrint                 +
                   ")x("                        +
                   IDM::rfUnknown               +
                   "),<"                        +
                   IDM::rmFile                  +
                   ","                          +
                   IDM::rfText                  +
                   ">,<"                        +
                   IDM::rmObject                +
                   ","                          +
                   IDM::rfObject                +
                   ">");
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::~IDMSourceRenderer                                        |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IDMSourceRenderer :: ~IDMSourceRenderer ()
{
  IMODTRACE_DEVELOP("IDMSourceRenderer::dtor ()");
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::canRender                                                 |
|                                                                              |
| Returns true if the renderer can render items with the argument attribute.   |
| Used when the drag is initiated.                                             |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: canRender(const IString &itemTypes)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::canRender(types)" );

  /**********************************************************************/
  /* Check if any type is supported - IDM::any allows any type ...      */
  /**********************************************************************/
  if (!(supportedTypes().indexOf(IDM::any)))
  {
    unsigned startPos = 1;
    unsigned count = itemTypes.occurrencesOf(',') + 1;

    for (unsigned i=0; i < count; i++)
    {
      unsigned index = itemTypes.indexOf(',', startPos);
      IString type = (index ? itemTypes.subString(startPos, index-1)  :
                                        itemTypes.subString(startPos));
      /******************************************************************/
      /* Verify the type ...                                            */
      /******************************************************************/
      if (supportedTypes().indexOf(type))
        return(true);

      startPos = index + 1;
    }

    /********************************************************************/
    /* Type not supported if we reach this point ...                    */
    /********************************************************************/
    return(false);
  }

  return(true);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::sourceRender                                              |
|                                                                              |
| Called when a target requests that a source renderer render an item.         |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer ::
                   sourceRender(IDMSourceRenderEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::sourceRender" );

  /********************************************************************/
  /* Indicate source and target are not the same                      */
  /********************************************************************/
  event.dragItem()->bTargetTheSource = false;

  /********************************************************************/
  /* Check for our rendering mechanism first                          */
  /********************************************************************/
  if (event.selectedMechanism() == IDM::rmLibrary)
  {
    /******************************************************************/
    /* Check for shared memory format ...                             */
    /******************************************************************/
    if (event.selectedFormat().indexOf(IDM::rfSharedMem))
    {
      ITRACE_DEVELOP( "Using rmLibrary, rfSharedMem - Src" );

      /****************************************************************/
      /* Give source item opportunity to acquire data                 */
      /****************************************************************/
      event.dragItem()->sourceRender(event);

      /****************************************************************/
      /* Call default/derived renderer to make the data available     */
      /* for the target.                                              */
      /****************************************************************/
      doRender(event);

      /****************************************************************/
      /* Free the dragtransfer structure.                             */
      /****************************************************************/
      DrgFreeDragtransfer (event.dragTransfer());
    }
    else
    {
      /****************************************************************/
      /* Unknown rendering format ...                                 */
      /****************************************************************/
      ITRACE_DEVELOP( "Unknown format, rm=rmLibrary - Src" );
      ITHROWLIBRARYERROR(IC_UNKNOWN_RENDERING_FORMAT,
                         IErrorInfo::invalidRequest,
                         IException::recoverable);
    }
  }
  else
  {
    /******************************************************************/
    /* Check for file rendering mechanism (delayed file rendering)    */
    /* Do not check the format since it may be text or unknown ...    */
    /******************************************************************/
    if (event.selectedMechanism() == IDM::rmFile)
    {
      ITRACE_DEVELOP( "Using rmFile (delayed file rendering) - Src" );
      /****************************************************************/
      /* Give source item opportunity to acquire data                 */
      /****************************************************************/
      event.dragItem()->sourceRender( event );

      /****************************************************************/
      /* Let the target (probably the WPS) know we're finished ...    */
      /****************************************************************/
      informTargetOfCompletion( event );
    }
    else
    {
      /****************************************************************/
      /* Unknown rendering mechanism ...                              */
      /****************************************************************/
      ITHROWLIBRARYERROR(IC_UNKNOWN_RENDERING_MECHANISM,
                         IErrorInfo::invalidRequest,
                         IException::recoverable);
    }
  }

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::sourcePrepare                                             |
|                                                                              |
| Called when an item indicates that it requires preparation.                  |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer ::
                   sourcePrepare(IDMSourcePrepareEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::sourceRenderPrepare" );

  /********************************************************************/
  /* Give source item opportunity to perform preparation.             */
  /********************************************************************/
  event.dragItem()->sourcePrepare(event);

  /********************************************************************/
  /* Call default/derived renderer to perform any required            */
  /* preparation.                                                     */
  /********************************************************************/
  doRenderPrepare(event);

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::sourceEnd                                                 |
|                                                                              |
| Called when the target window informs the source window that it has finished |
| processing the data that was associated with the dropped item.               |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer :: sourceEnd(IDMSourceEndEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::sourceEnd" );

  /********************************************************************/
  /* Give item opportunity to perform clean-up                        */
  /********************************************************************/
  event.dragItem()->sourceEnd(event);

  /********************************************************************/
  /* Call default/derived renderer to do its source end event         */
  /* processing.                                                      */
  /********************************************************************/
  doRenderEnd(event);

  /********************************************************************/
  /* Obtain the drag operation from the item in order to access       */
  /* the operation's item collection per removal of item from         */
  /* the collection.  Also, reset the PM DRAGITEM's ulItemId field    */
  /* so we will not attempt to construct an IDMItem from it when we   */
  /* are processing DM_PRINTOBJECT or DM_DISCARDOBJECT.               */
  /********************************************************************/
  IDMSourceOperation *pSrcOp = event.dragItem()->sourceOperation();
  event.dragItem()->pPMDragItem->ulItemID = 0;
  pSrcOp->removeItem(event.dragItem());

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::sourceDiscard                                             |
|                                                                              |
| Called by the IDMSourceHandler::sourceDiscard() function when it was         |
| indicated that the source window has responsibility for discarding the item. |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer ::
                   sourceDiscard(IDMSourceDiscardEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::sourceDiscard" );

  /********************************************************************/
  /* Get the source operation                                         */
  /********************************************************************/
  IDMSourceOperation::Handle pSrcOpH = event.sourceOperation();

  /********************************************************************/
  /* Get the source item from the collection.                         */
  /********************************************************************/
  IDMItem::Handle pDIH = pSrcOpH->item(1);

  /********************************************************************/
  /* Give source item opportunity to perform discard processing.      */
  /********************************************************************/
  pDIH->sourceDiscard(event);

  /********************************************************************/
  /* Call default/derived renderer to do its source discard event     */
  /* processing.                                                      */
  /********************************************************************/
  doDiscard(event);

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::sourcePrint                                               |
|                                                                              |
| Called by the IDMSourceHandler::sourcePrint() function when it was           |
| indicated that the source window has responsibility for printing the item.   |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer ::
                   sourcePrint(IDMSourcePrintEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::sourcePrint" );

  /********************************************************************/
  /* Get the source operation                                         */
  /********************************************************************/
  IDMSourceOperation::Handle pSrcOpH = event.sourceOperation();

  /********************************************************************/
  /* Get the source item from the collection.                         */
  /********************************************************************/
  IDMItem::Handle pDIH = pSrcOpH->item(1);

  /********************************************************************/
  /* Give source item opportunity to perform print processing.        */
  /********************************************************************/
  pDIH->sourcePrint(event);

  /********************************************************************/
  /* Call default/derived renderer to do its source print event       */
  /* processing.                                                      */
  /********************************************************************/
  doPrint(event);

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::informTargetOfCompletion                                  |
|                                                                              |
| Informs the target window that source renderering has completed on the drag  |
| item.                                                                        |
------------------------------------------------------------------------------*/
IDMSourceRenderer& IDMSourceRenderer ::
                   informTargetOfCompletion(IDMSourceRenderEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::informTargetOfCompletion" );

  if (!(event.dragTransfer()))
  {
    ITHROWLIBRARYERROR(IC_DRAGTRANSFER_INVALID,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);
  }

  /********************************************************************/
  /* Tell the target we have finished renderering the item and the    */
  /* data is available.                                               */
  /*                                                                  */
  /* Note:  Only valid for source renderering                         */
  /********************************************************************/
  ITRACE_DEVELOP( "Posting DM_RENDERCOMPLETE" );
  if (!DrgPostTransferMsg (event.targetWindowHandle(),
                           DM_RENDERCOMPLETE,
                           event.dragTransfer(),
                           (unsigned short)event.completion(),
                           0,
                           event.canRetry()))
  {
    ITHROWGUIERROR("DrgPostTransferMsg");
  }

  /********************************************************************/
  /* Free the dragtransfer structure.                                 */
  /********************************************************************/
  DrgFreeDragtransfer (event.dragTransfer());

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::doRender                                                  |
|                                                                              |
| Called when a render event occurs and this renderer must render a drag item. |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: doRender(IDMSourceRenderEvent &event)
{
  void     *pBuffer,
           *pTempBuffer;

  IMODTRACE_DEVELOP( "IDMSourceRenderer::doRender" );

  /********************************************************************/
  /* Get the shared memory buffer address from the target information */
  /*                                                                  */
  /* NOTE:  The first 4 bytes of the buffer contain the buffer's      */
  /*        size, so we index around it.                              */
  /********************************************************************/
  pBuffer = (void *)event.targetInfo();
  pTempBuffer = (char *)pBuffer + 4;

  /********************************************************************/
  /* Get the data from the drag item and copy it to our transfer      */
  /* buffer.                                                          */
  /********************************************************************/
  IString tempStr = event.dragItem()->contents();
  memcpy(pTempBuffer, (char *)tempStr, (tempStr.length() + 1));

  /********************************************************************/
  /* Free the memory object and return true to indicate the render    */
  /* processed ...                                                    */
  /********************************************************************/
  DosFreeMem (pBuffer);

  return(true);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::doRenderPrepare                                           |
|                                                                              |
| Called when a render prepate event occurs and this renderer must render a    |
| drag item.                                                                   |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: doRenderPrepare(IDMSourcePrepareEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::doRenderPrepare" );

  /********************************************************************/
  /* Return the length of the source IString                          */
  /********************************************************************/
  event.setTargetInfo(event.dragItem()->contentsSize()+1);
  return(true);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::doRenderEnd                                               |
|                                                                              |
| Called when an end event occurs and this renderer must end the render        |
| operation.                                                                   |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: doRenderEnd(IDMSourceEndEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::doRenderEnd" );
  ITRACE_DEVELOP( "return false" );

  return(false);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::doDiscard                                                 |
|                                                                              |
| Called when a discard event occurs and this renderer must discard a drag     |
| item.                                                                        |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: doDiscard(IDMSourceDiscardEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::doDiscard" );
  ITRACE_DEVELOP( "return false" );

  return(false);
}

/*------------------------------------------------------------------------------
| IDMSourceRenderer::doPrint                                                   |
|                                                                              |
| Called when a print event occurs and this renderer must print a drag item.   |
------------------------------------------------------------------------------*/
Boolean IDMSourceRenderer :: doPrint(IDMSourcePrintEvent &event)
{
  IMODTRACE_DEVELOP( "IDMSourceRenderer::doPrint" );
  ITRACE_DEVELOP( "return false" );

  return(false);
}


/*-------------------------------- TARGET RENDERER ---------------------------*/

/*------------------------------------------------------------------------------
| IDMTargetRenderer::IDMTargetRenderer                                         |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMTargetRenderer :: IDMTargetRenderer(const char *pRMFs,
                                       const char *pTypes)
  : IDMRenderer(pRMFs, pTypes),
    pPMDragTransfer(0),
    fDMTargetRendererData( new IDMTargetRendererData() )
{
  IMODTRACE_DEVELOP("IDMTargetRenderer::ctor (rmfs, types)");
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::IDMTargetRenderer                                         |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMTargetRenderer :: IDMTargetRenderer()
  : IDMRenderer(),
    pPMDragTransfer(0),
    fDMTargetRendererData( new IDMTargetRendererData() )
{
  IMODTRACE_DEVELOP("IDMTargetRenderer::ctor ()");

  setSupportedRMFs(IString("(")                 +
                   IDM::rmLibrary               +
                   ")x("                        +
                   IDMItem::rfForThisProcess()  +
                   ","                          +
                   IDM::rfText                  +
                   ","                          +
                   IDM::rfSharedMem             +
                   "),<"                        +
                   IDM::rmFile                  +
                   ","                          +
                   IDM::rfText                  +
                   ">,<"                        +
                   IDM::rmObject                +
                   ","                          +
                   IDM::rfObject                +
                   ">");
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::~IDMTargetRenderer                                        |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IDMTargetRenderer :: ~IDMTargetRenderer()
{
  IMODTRACE_DEVELOP("IDMTargetRenderer::dtor ()");
  delete this->fDMTargetRendererData;
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::canRender                                                 |
|                                                                              |
| Returns a drop indicator with the renderer's ability to render the item with |
| the argument attributes.                                                     |
------------------------------------------------------------------------------*/
DropIndicator IDMTargetRenderer :: canRender(const IDMItem::Handle &pIDMItem)
{
  DropIndicator dropFlag = IDM::neverOk;

  IMODTRACE_DEVELOP( "IDMTargetRenderer::canRender(item)" );

  /**********************************************************************/
  /* Check if any type is supported - IDM::any allows any type ...      */
  /**********************************************************************/
  if (!(supportedTypes().indexOf(IDM::any)))
  {
    /********************************************************************/
    /* Pick thru the item's types to see if we support one of them.     */
    /********************************************************************/
    Boolean noTypesSupported = true;
    IString itemTypes = pIDMItem->types();
    unsigned count = itemTypes.occurrencesOf(',') + 1;
    unsigned startPos = 1;
    for (unsigned i=0; i < count; i++)
    {
      unsigned index = itemTypes.indexOf(',', startPos);
      IString type = (index ? itemTypes.subString(startPos, index-1)  :
                              itemTypes.subString(startPos));

      /******************************************************************/
      /* Verify the type                                                */
      /******************************************************************/
      if (supportedTypes().indexOf(type))
      {
        noTypesSupported = false;
        break;
      }
    }

    /********************************************************************/
    /* If no types supported, indicate that there is no need in sending */
    /* us more drag over events.                                        */
    /********************************************************************/
    if (noTypesSupported)
      return(IDM::neverOk);
  }

  /**********************************************************************/
  /* Verify the source RMFs, whereas we have to have a match on         */
  /* on at least 1 of the source RMFs                                   */
  /**********************************************************************/
  IString
    selected = IDMItem::matchingRMFs( pIDMItem->rmfs(),
                                      this->supportedRMFs(),
                                      true );

  if ( selected.length() )
    {
    // Yes, then we can render it...
    ITRACE_DEVELOP( "Match on " + selected );
    pIDMItem->setSelectedRMF( selected );
    dropFlag = IDM::ok;
    }

  return(dropFlag);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::supportsOperation                                         |
|                                                                              |
| Returns true if the default operation (for example, move copy, or link)      |
| is supported by the item.                                                    |
------------------------------------------------------------------------------*/
Boolean IDMTargetRenderer ::
        supportsOperation(const IDMItem::Handle &pIDMItem)
{
  unsigned long defaultOp = pIDMItem->targetOperation()->operation();
  unsigned long supportedOps = pIDMItem->supportedOperations();

  IMODTRACE_DEVELOP( "IDMTargetRenderer::supportsOperation" );

  /**********************************************************************/
  /* Exit stage right if unknown drag operation defined (must derive    */
  /* and define corresponding behavior).                                */
  /**********************************************************************/
  if (defaultOp >= IDMOperation::unknown)
    return(false);

  /**********************************************************************/
  /* If the drag operation set by the source is IDMOperation::drag,     */
  /* use the item's supported operations to see if we can use the       */
  /* CUA defined default which is "copy" ...                            */
  /**********************************************************************/
  if (defaultOp == IDMOperation::drag)
  {
    if (supportedOps & IDMItem::moveable)
      pIDMItem->targetOperation()->setOperation(IDMOperation::move);
    else
    {
      if (supportedOps & IDMItem::copyable)
        pIDMItem->targetOperation()->setOperation(IDMOperation::copy);
      else
      {
        if (supportedOps & IDMItem::linkable)
          pIDMItem->targetOperation()->setOperation(IDMOperation::link);
        else
        {
          ITRACE_DEVELOP( "can't choose default" );
          return(false);
        }
      }
    }
    ITRACE_DEVELOP( "chose default="
                    +
                    IString( pIDMItem->targetOperation()->operation() ) );
    return(true);
  }

  /**********************************************************************/
  /* See if we can find a match with the item's supported operations,   */
  /* to allow a drop (ok) condition ...                                 */
  /*                                                                    */
  /* Note:  Source defined drag operation will be used here!            */
  /**********************************************************************/
  Boolean bOpOK = false;
  if ((defaultOp == IDMOperation::copy) && (supportedOps & IDMItem::copyable))
    bOpOK=true;

  if ((defaultOp == IDMOperation::move) && (supportedOps & IDMItem::moveable))
    bOpOK=true;

  if ((defaultOp == IDMOperation::link) && (supportedOps & IDMItem::linkable))
    bOpOK=true;

  ITRACE_DEVELOP( "non-default OK?=" + IString( bOpOK ) );
  return(bOpOK);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::targetRender                                              |
|                                                                              |
| Called when a drop event occurs for a drag item and this renderer must       |
| render the item.                                                             |
------------------------------------------------------------------------------*/
IDMTargetRenderer& IDMTargetRenderer ::
                   targetRender(IDMTargetDropEvent &event,
                                const IDMItem::Handle &pTgtDIH)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::targetRender" );

  /**********************************************************************/
  /* Get the selected RMF from the drag item                            */
  /**********************************************************************/
  IString
    rmf = pTgtDIH->selectedRMF(),
    rm  = IDMItem::rmFrom( rmf ),
    rf  = IDMItem::rfFrom( rmf );
  ITRACE_DEVELOP( "Selected RMF=" + rmf );

  /**********************************************************************/
  /* Check for our rendering mechanism first                            */
  /**********************************************************************/
  if (rm == IDM::rmLibrary)
  {
    IMODTRACE_DEVELOP( "Using rmLibrary..." );
    /********************************************************************/
    /* Next key off the format to determine our next step,              */
    /* first checking for inter-process format ...                      */
    /********************************************************************/
    if (rf.includes( IDM::rfProcess ))
    {
      if (rf == IDMItem::rfForThisProcess())
      {
        /****************************************************************/
        /* Same process - we can acquire the data at this point         */
        /****************************************************************/
         ITRACE_DEVELOP( "Using rfProcess" );
         IDMItem::Handle pSrcDIH = IDMItem::sourceItemFor(pTgtDIH);
         pTgtDIH->setContents(pSrcDIH->contents());
         pTgtDIH->setObject(pSrcDIH->object());
      }
      else
      {
        //Throw exception
        ITRACE_DEVELOP( "Error - IDM::rfProcess used for different process" );
        ITHROWLIBRARYERROR(IC_RFPROCESS_FOR_SAME_PROCESS,
                           IErrorInfo::invalidRequest,
                           IException::recoverable);
        return(*this);
      }
    }
    else
    {
      /******************************************************************/
      /* Check for the text format and acquire the data if text         */
      /******************************************************************/
      if (rf == IDM::rfText)
      {
        ITRACE_DEVELOP( "Using rfText" );
        pTgtDIH->setContents(pTgtDIH->sourceName());
      }
      else
      {
        /****************************************************************/
        /* Check for shared memory format ...                           */
        /****************************************************************/
        if (rf == IDM::rfSharedMem)
        {
          ITRACE_DEVELOP( "Using rfSharedMem - Tgt" );
          /**************************************************************/
          /* This rmf requires source preparation...                    */
          /**************************************************************/
          targetRenderPrepare(event, pTgtDIH);

          /**************************************************************/
          /* Attempt source rendering.                                  */
          /**************************************************************/
          if (renderAtSource(event, pTgtDIH))
          {
            /************************************************************/
            /* Store the transferred data that has been made available  */
            /* by the source.                                           */
            /************************************************************/
            renderComplete(pTgtDIH, (void *)event.targetInfo());
          }
          else
          {
            //Throw exception
            ITRACE_DEVELOP( "Error in renderAtSource!" );
            ITHROWLIBRARYERROR(IC_RENDER_AT_SOURCE_FAILED,
                               IErrorInfo::invalidRequest,
                               IException::recoverable);
            return(*this);
          }
        }
        else
        {
          /**************************************************************/
          /* Unknown rendering format ...                               */
          /**************************************************************/
          ITHROWLIBRARYERROR(IC_UNKNOWN_RENDERING_FORMAT,
                             IErrorInfo::invalidRequest,
                             IException::recoverable);
          return(*this);
        }
      }
    }
  }
  else
  {
    /********************************************************************/
    /* Check for the file rendering mechanism                           */
    /********************************************************************/
    if (IDMItem::rmFrom(rmf) == IDM::rmFile)
    {
      ITRACE_DEVELOP( "Using rmFile" );
    }
    else
    {
      /******************************************************************/
      /* Check for the WPS object rendering mechanism                   */
      /******************************************************************/
      if (IDMItem::rmFrom(rmf) == IDM::rmObject)
      {
        ITRACE_DEVELOP( "Using rmObject" );
      }
      else
      {
        /****************************************************************/
        /* Unknown rendering mechanism ...                              */
        /****************************************************************/
        //Throw exception
        ITHROWLIBRARYERROR(IC_UNKNOWN_RENDERING_MECHANISM,
                           IErrorInfo::invalidRequest,
                           IException::recoverable);
        return(*this);
      }
    }
  }

  /**********************************************************************/
  /* Allow item to do any specific drop processing related to the       */
  /* derived IDMItem class.                                             */
  /**********************************************************************/
  Boolean bRc = pTgtDIH->targetDrop(event);

  /**********************************************************************/
  /* Signal source that we're done (unless processing                   */
  /* IDMTargetEndEvent)                                                 */
  /**********************************************************************/
  informSourceOfCompletion(pTgtDIH, ((bRc) ? IDM::targetSuccessful
                                           : IDM::targetFailed));

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::targetRenderPrepare                                       |
|                                                                              |
| Called when the source window requests that the target window generate a     |
| preparation event.                                                           |
------------------------------------------------------------------------------*/
IDMTargetRenderer& IDMTargetRenderer ::
                   targetRenderPrepare(IDMTargetDropEvent &event,
                                       const IDMItem::Handle &pTgtDIH)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::targetRenderPrepare" );

  /**********************************************************************/
  /* Request source rendering preparation ...                           */
  /**********************************************************************/
  prepareAtSource(event, pTgtDIH);

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::targetRenderComplete                                      |
|                                                                              |
| Called when a completion event occurs and this renderer must render a drag   |
| item.                                                                        |
------------------------------------------------------------------------------*/
IDMTargetRenderer& IDMTargetRenderer ::
                   targetRenderComplete(IDMTargetEndEvent &event)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::targetRenderComplete" );

  IDMItem::Handle pTgtDIH;
  pTgtDIH = event.dragItem();

  /**********************************************************************/
  /* Store pointer to the current drag transfer, so we can retrieve     */
  /* later for freeing when we use threaded source rendering.           */
  /**********************************************************************/
  fDMTargetRendererData->pCurrentDragTransfer = event.dragTransfer();

  /**********************************************************************/
  /* Store the transferred data that has been made available by the     */
  /* source.                                                            */
  /**********************************************************************/
  renderComplete( pTgtDIH,
            (void *)fDMTargetRendererData->pCurrentDragTransfer->ulTargetInfo );

  /**********************************************************************/
  /* Give item opportunity to diddle ...                                */
  /**********************************************************************/
  Boolean bRc = pTgtDIH->targetEnd(event);

  /**********************************************************************/
  /* Signal source that we're done ...                                  */
  /**********************************************************************/
  informSourceOfCompletion(pTgtDIH, ((bRc) ? IDM::targetSuccessful
                                           : IDM::targetFailed));

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::informSourceOfCompletion                                  |
|                                                                              |
| Informs the source window that the target window has finished processing the |
| data associated with the dropped item.  The source window is informed for    |
| both source and target rendering.  The render completion code is used to     |
| to indicate if the target was successful.                                    |
------------------------------------------------------------------------------*/
IDMTargetRenderer& IDMTargetRenderer ::
                   informSourceOfCompletion(const IDMItem::Handle &pTgtDIH,
                                            IDM::RenderCompletion code)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::informSourceOfCompletion" );

  if (!(pTgtDIH->PMDragItem()))
  {
    ITHROWLIBRARYERROR(IC_DRAGITEM_INVALID,
                       IErrorInfo::invalidRequest,
                       IException::recoverable);
  }

  /**********************************************************************/
  /* Tell the source that we are done communicating for a particular    */
  /* drag item.                                                         */
  /**********************************************************************/
  pTgtDIH->sourceWindowHandle().sendEvent (DM_ENDCONVERSATION,
                                   MPFROMLONG(pTgtDIH->PMDragItem()->ulItemID),
                                   MPFROMLONG(code));

  /**********************************************************************/
  /* Free the dragtransfer structure.                                   */
  /**********************************************************************/
  if (fDMTargetRendererData->pCurrentDragTransfer)
  {
    DrgFreeDragtransfer( fDMTargetRendererData->pCurrentDragTransfer );
    fDMTargetRendererData->pCurrentDragTransfer = 0;
  }

  /**********************************************************************/
  /* Remove the item from the target operation's collection ...         */
  /**********************************************************************/
  IDMTargetOperation::Handle pTgtOpH = IDMTargetOperation::targetOperation();
  pTgtOpH->removeItem(pTgtDIH);
  if (!(pTgtOpH->numberOfItems()))
  {
    /********************************************************************/
    /* When the count of items reaches 0, reset the static target       */
    /* operation pointer, which causes the destructor for               */
    /* IDMTargetOperation to be called ...                              */
    /********************************************************************/
    IDMTargetOperation::setTargetOperation (0);

    /********************************************************************/
    /* Reset the drag transfer count and set the pointer to the PM      */
    /* drag transfer structure to 0.                                    */
    /********************************************************************/
    fDMTargetRendererData->ulDragTransferCount = 0;
    this->pPMDragTransfer = 0;
  }

  return(*this);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::renderAtSource                                            |
|                                                                              |
| Sends a render event to the source window to request source rendering for    |
| the argument item.                                                           |
------------------------------------------------------------------------------*/
Boolean IDMTargetRenderer :: renderAtSource(IDMTargetDropEvent &event,
                                            const IDMItem::Handle &pTgtDIH)
{
  void         *pBuffer;
  unsigned long bufferSize;
  PID           spid;
  TID           tid;

  IMODTRACE_DEVELOP( "IDMTargetRenderer::renderAtSource" );

  IString
    rmf = pTgtDIH->selectedRMF(),
    rm  = IDMItem::rmFrom( rmf ),
    rf  = IDMItem::rfFrom( rmf );

  Boolean
    bSharedMem = (rf == IDM::rfSharedMem) ? true : false;

  if (bSharedMem)
  {
    ITRACE_DEVELOP( "Rendering via <" + rm + ",rfSharedMem>" );

    /********************************************************************/
    /* Get the length of the text string so we know how large to make   */
    /* the (shared) buffer.                                             */
    /********************************************************************/
    bufferSize = event.targetInfo();

    /********************************************************************/
    /* Allocate a shared memory buffer and give the target process      */
    /* access to it.  First obtain the source process id.               */
    /* If either of the memory calls fail...bail.                       */
    /********************************************************************/
    if (!WinQueryWindowProcess (pTgtDIH->sourceWindowHandle(), &spid, &tid))
    {
      ITHROWGUIERROR("WinQueryWindowProcess");
    }

    unsigned long rc;
    rc = DosAllocSharedMem (&pBuffer, 0, bufferSize + 4,
                            (OBJ_GIVEABLE | PAG_READ | PAG_WRITE | PAG_COMMIT));
    if (rc)
    {
      ITHROWSYSTEMERROR(rc,
                        "DosAllocSharedMem",
                        IErrorInfo::outOfMemory,
                        IException::recoverable);
    }

    rc = DosGiveSharedMem (pBuffer, spid, (PAG_READ | PAG_WRITE));
    if (rc)
    {
      ITHROWSYSTEMERROR(rc,
                        "DosGiveSharedMem",
                        IErrorInfo::accessError,
                        IException::recoverable);
    }

    /********************************************************************/
    /* Save the pointer to the buffer that will be used to transfer the */
    /* data between the source and the target.                          */
    /*                                                                  */
    /* NOTE:  The first 4 bytes of the buffer contain the buffer's      */
    /*        size, so we store the size here.                          */
    /********************************************************************/
    *(unsigned long *)pBuffer = bufferSize + 4;
    event.setTargetInfo((unsigned long)pBuffer);
  } //rfSharedMem

  /**********************************************************************/
  /* Get the dragtransfer structure and allocate it if not allocated    */
  /**********************************************************************/
  _DRAGTRANSFER *pDXfer = allocDragTransfer(event, pTgtDIH);

  /**********************************************************************/
  /* Store pointer to shared memory buffer in the _DRAGTRANSFER         */
  /**********************************************************************/
  pDXfer->ulTargetInfo = event.targetInfo();

  ITRACE_DEVELOP( "Sending DM_RENDER" );
  if ( !DrgSendTransferMsg (pTgtDIH->sourceWindowHandle(),
                            DM_RENDER,
                            MPFROMP(pDXfer), 0) )
  {
    ITHROWGUIERROR("DrgSendTransferMsg");
    return(false);
  }

  return(true);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::prepareAtSource                                           |
|                                                                              |
| Sends a prepare event to the source window to request source preparation for |
| the argument item.                                                           |
------------------------------------------------------------------------------*/
Boolean IDMTargetRenderer :: prepareAtSource(IDMTargetDropEvent &event,
                                             const IDMItem::Handle &pTgtDIH)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::prepareAtSource" );

  /********************************************************************/
  /* Get the dragtransfer structure and allocate it if not allocated. */
  /********************************************************************/
  _DRAGTRANSFER *pDXfer = allocDragTransfer(event, pTgtDIH);

  ITRACE_DEVELOP( "Sending DM_RENDERPREPARE" );
  DrgSendTransferMsg (pTgtDIH->sourceWindowHandle(),
                      DM_RENDERPREPARE,
                      MPFROMP(pDXfer), 0);

  /********************************************************************/
  /* Save the target information in the event.                        */
  /********************************************************************/
  event.setTargetInfo(pDXfer->ulTargetInfo);

  /*************************************************************************/
  /* Update the source window handle in the drag item, so the target will  */
  /* be aware of an alternate source window for source rendering.          */
  /*************************************************************************/
  pTgtDIH->setSourceWindowHandle( pDXfer->pditem->hwndItem );

  return(true);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::renderComplete                                            |
|                                                                              |
| Called to process a completion event for a drag item.                        |
------------------------------------------------------------------------------*/
Boolean IDMTargetRenderer :: renderComplete(const IDMItem::Handle &pTgtDIH,
                                            void* pBuffer)
{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::renderComplete" );

  IString
    rmf = pTgtDIH->selectedRMF(),
    rm  = IDMItem::rmFrom( rmf ),
    rf  = IDMItem::rfFrom( rmf );

  if (rf == IDM::rfSharedMem)
  {
    ITRACE_DEVELOP( "Completing rendering via <" + rm + ",rfSharedMem>" );

    unsigned long ulBufferSize = *(unsigned long *)pBuffer;
    void *pTempBuffer = (char *)pBuffer + 4;

    /**********************************************************************/
    /* Store the transferred data in the target drag item                 */
    /**********************************************************************/
    pTgtDIH->setContents( IString(pTempBuffer, ulBufferSize ) );

    /**********************************************************************/
    /* Free the memory used to hold the dragged data.                     */
    /**********************************************************************/
    DosFreeMem (pBuffer);
  }

  return (true);
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::renderToName                                              |
|                                                                              |
| Called to provide the name of the location where the data will be renderer   |
| to.  This name can be a filename, shared memory buffer, etc.                 |
------------------------------------------------------------------------------*/
IString IDMTargetRenderer :: renderToName (const IDMItem::Handle &pDI)
{
  return (pDI->targetName());
}

/*------------------------------------------------------------------------------
| IDMTargetRenderer::allocDragTransfer                                         |
|                                                                              |
| Create a dragtransfer structure.                                             |
------------------------------------------------------------------------------*/
_DRAGTRANSFER* IDMTargetRenderer ::
                allocDragTransfer(IDMTargetDropEvent &dropEvent,
                                  const IDMItem::Handle &pDI)

{
  IMODTRACE_DEVELOP( "IDMTargetRenderer::allocDragTransfer" );

  /********************************************************************/
  /* Allocate a drag transfer array whose size is based on the        */
  /* number of drag items, unless the array already exists.           */
  /********************************************************************/
  if (!this->pPMDragTransfer)
  {
    this->pPMDragTransfer =
      DrgAllocDragtransfer( pDI->targetOperation()->numberOfItems() );

    ITRACE_DEVELOP( "Number of DRAGTRANSFERs allocated = " +
                    IString( pDI->targetOperation()->numberOfItems() ) );

    if (!this->pPMDragTransfer)
      ITHROWGUIERROR("DrgAllocDragtransfer");
  }

  /********************************************************************/
  /* Test to see if the drag transfer pointer is set, and if so,      */
  /* check to see if we have already alloted the DRAGTRANSFER         */
  /* structure from the drag transfer array for this drag item object.*/
  /* If we have, we are done and we return the drag transfer.         */
  /********************************************************************/
  if ( fDMTargetRendererData->pCurrentDragTransfer )
  {
    if ( fDMTargetRendererData->pCurrentDragTransfer->pditem ==
         pDI->PMDragItem() )
    {
      return( fDMTargetRendererData->pCurrentDragTransfer );
    }
  }

  /********************************************************************/
  /* If we reach this point we need to allot a DRAGTRANSFER structure */
  /* for this drag item object from the array.  Use the drag transfer */
  /* count to determine the location of the DRAGTRANSFER structure to */
  /* use.                                                             */
  /********************************************************************/
  unsigned long
    ulCnt = fDMTargetRendererData->ulDragTransferCount * sizeof(_DRAGTRANSFER);

  _DRAGTRANSFER
   *pdxfer = (_DRAGTRANSFER *)(((char *)this->pPMDragTransfer) + ulCnt);

  /********************************************************************/
  /* Increment the count to track how many DRAGTRANSFER structures we */
  /* have allotted.                                                   */
  /********************************************************************/
  fDMTargetRendererData->ulDragTransferCount++;

  /********************************************************************/
  /* Fill in the fields of the DRAGTRANSFER structure.                */
  /********************************************************************/
  pdxfer->cb = sizeof(_DRAGTRANSFER);
  pdxfer->hwndClient = dropEvent.window()->handle();
  pdxfer->pditem = pDI->PMDragItem();
  pdxfer->hstrSelectedRMF = pDI->handleFromString(pDI->selectedRMF());
  pdxfer->hstrRenderToName = pDI->handleFromString(renderToName (pDI));
  pdxfer->ulTargetInfo = 0;
  pdxfer->usOperation = (IDMTargetOperation::targetOperation())->operation();
  pdxfer->fsReply = 0;                      // Reply, used later

  /********************************************************************/
  /* Store pointer to the current drag transfer, so we can retrieve   */
  /* later for freeing when we use non-threaded source rendering.     */
  /********************************************************************/
  fDMTargetRendererData->pCurrentDragTransfer = pdxfer;

  /********************************************************************/
  /* Return the drag transfer structure we alloted.                   */
  /********************************************************************/
  return(pdxfer);
}/* end IDMTargetRenderer :: dragTransfer(...) */

/*------------------------------------------------------------------------------
| IDMTargetRenderer::dragTransfer                                              |
|                                                                              |
| Returns the PM dragtransfer structure.                                       |
------------------------------------------------------------------------------*/
_DRAGTRANSFER* IDMTargetRenderer :: dragTransfer() const
{
  return (this->pPMDragTransfer);
}
