/*******************************************************************************
* FILE NAME: idrawcv.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in idrawcv.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C" {
  #include <iwindefs.h>
}

#include <idrawcv.hpp>
#include <ihandle.hpp>
#include <igrafctx.hpp>
#include <iglist.hpp>
#include <igraphic.hpp>
#include <ipainhdr.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDRAWCV_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public drawingarea styles                                                    |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IDrawingCanvas::Style
  IDrawingCanvas::useDefaultPaintHandler ( 0, ICNV_ATTACHDEFAULTHANDLER ),
  IDrawingCanvas::classDefaultStyle    ( (WS_VISIBLE | WS_CLIPSIBLINGS),
                                       ICNV_ATTACHDEFAULTHANDLER );
#pragma data_seg()

class IDefaultDrawHandler : public IPaintHandler
{
typedef IPaintHandler
  Inherited;
/*******************************************************************************
* The IDefaultDrawHandle is a paint handler that is used to handle paint       *
* events for the IDrawingCanvas class.  This handler is automatically created  *
* and added to an IDrawingCanvas object unless you create the IDrawingCanvas   *
* without the IDrawingCanvas style useDefaultPaintHandler.                       *
*                                                                              *
* The IDefaultDrawHandler overrides the paintWindow function.  When the paint  *
* window function is called, this handler will set the clip region to the      *
* windows update region and redraw the affected graphic in the graphic list    *
* of the draw canvas.                                                          *
*******************************************************************************/
public:
/*------------------------------- Constructor ----------------------------------
| The only way to construct an instance of this class is from the              |
| default constructor.                                                         |
------------------------------------------------------------------------------*/
  IDefaultDrawHandler();
virtual
  ~IDefaultDrawHandler();

protected:
/*----------------------------- Event Processing -------------------------------
| paintWindow - Processes paint events.  Iterate through graphic objects in    |
|               the drawing area and call the draw function for each.          |
------------------------------------------------------------------------------*/
virtual Boolean
  paintWindow( IPaintEvent& event );

};

/*------------------------------------------------------------------------------
  This class wrappers an object pointer and destroys the pointed
  to object when the wrapper is destructed.
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator IDefaultDrawHandler* ();
IDefaultDrawHandler
 *operator -> ()
    {
    return *this;
    }
  ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
IDefaultDrawHandler
 *ptr;
} defaultHandler;

/*------------------------------------------------------------------------------
| DefaultHandler::operator IDefaultDrawHandler*                                |
------------------------------------------------------------------------------*/
DefaultHandler :: operator IDefaultDrawHandler* ()
    {
    if ( !ptr )
      ptr = new IDefaultDrawHandler;
    return ptr;
    }


/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IDrawingCanvas::Style
  IDrawingCanvas::fgCurrentDefaultStyle ( (WS_VISIBLE | WS_CLIPSIBLINGS),
                                          ICNV_ATTACHDEFAULTHANDLER );


#pragma data_seg()

/*------------------------------------------------------------------------------
| IDrawingCanvas::IDrawingCanvas                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas::IDrawingCanvas( unsigned long windowId, IWindow* parent,
                                IWindow* owner, const IRectangle& initial,
                                const Style& style)
//: ICanvas( windowId, parent, owner, initial ),        $HT!d
  : IDrawingCanvas::Inherited( ),                    // $HT!a
    fList(0),
    fGc(0),
    fDrawingCanvasData(0)
{
  // Attach the default draw handler if the style is set.
  if ( style & useDefaultPaintHandler )
  {
    this -> addHandler( defaultHandler );
  }

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

//convertToGUIStyle( style );                           $HT!d
  // Create the presentation system window.             $HT!a
  this->initialize( windowId,                        // $HT!a
                    parent,                          // $HT!a
                    owner,                           // $HT!a
                    initial,                         // $HT!a
                    this->convertToGUIStyle( style ),   // $HT!a
                    0 );                             // $HT!a
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::~IDrawingCanvas                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas::~IDrawingCanvas( )
{
  // Remove and delete the default handler if it exists.
  if (extendedStyle() & useDefaultPaintHandler.asExtendedUnsignedLong())
  {
    this -> removeHandler( defaultHandler );
  }
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::defaultStyle                                                 |
|                                                                              |
| Return the current default style for the class.                              |
------------------------------------------------------------------------------*/
IDrawingCanvas::Style IDrawingCanvas::defaultStyle ( )
{
  return fgCurrentDefaultStyle;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::convertToGUIStyle                                            |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                                 |
------------------------------------------------------------------------------*/
unsigned long IDrawingCanvas :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                  Boolean bExtOnly) const
{
  // Obtain the style from the class (ICanvas) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // No styles are defined for the drawing area, so we do not mask.
    ulStyle |= guiStyle.asUnsignedLong();
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::graphicList                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList* IDrawingCanvas::graphicList( ) const
{
  return fList;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::setGraphicList                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas& IDrawingCanvas::setGraphicList( IGList* list )
{
  fList = list;
  setLayoutDistorted(IWindow::layoutChanged, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::setGraphicContext                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas& IDrawingCanvas::setGraphicContext( IGraphicContext* pGC )
{
  fGc = pGC;
  return *this;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::graphicContext                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext* IDrawingCanvas::graphicContext( ) const
{
  return fGc;
}

/*------------------------------------------------------------------------------
| ICanvas::calcMinimumSize                                                     |
|                                                                              |
| Calculate the minimum screen size needed by the canvas.  This is the size    |
| needed to display all child windows (as determined by the layout routine).   |
------------------------------------------------------------------------------*/
ISize  IDrawingCanvas::calcMinimumSize ( ) const
{
  IGraphicContext *gc(this->graphicContext());
  Boolean createdGC(false);

  // Get the minimum size of the child windows that
  // may be on the drawing area.
  ISize      controls(Inherited::calcMinimumSize());
  IRectangle listBoundingRect;
  ISize      minSize;

  if (!gc)
  {
    gc = new IGraphicContext( handle() );
    createdGC = true;
  }

  if (graphicList() && !graphicList()->isEmpty())
  {
    listBoundingRect = graphicList()->boundingRect( *gc );
    minSize.setWidth( listBoundingRect.maxX() );
    minSize.setHeight( listBoundingRect.maxY() );
  }

  if (createdGC)
  {
    delete gc;
    gc = 0;
  }

  minSize = minSize.maximum( controls );

  return minSize;
}

/*------------------------------------------------------------------------------
| IDefaultDrawHandler::IDefaultDrawHandler                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IDefaultDrawHandler::IDefaultDrawHandler( )
{ }

/*------------------------------------------------------------------------------
| IDefaultDrawHandler::~IDefaultDrawHandler                                    |
|                                                                              |
| Empty destructor for page tuning.                                            |
------------------------------------------------------------------------------*/
IDefaultDrawHandler::~IDefaultDrawHandler( )
{ }

/*------------------------------------------------------------------------------
| IDefaultDrawHandler::paintWindow                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IDefaultDrawHandler::paintWindow( IPaintEvent& paintEvent )
{
  Boolean          didDraw(false);
  Boolean          createdGC(false);
  IDrawingCanvas*  drawingCanvas((IDrawingCanvas*)paintEvent.window());
  IGraphicContext* gc(drawingCanvas->graphicContext());

  if ( gc )
  {
    paintEvent.setGraphicContext( *gc );
  }
  else
  {
    if (paintEvent.presSpaceHandle())
    {
      gc = new IGraphicContext(paintEvent.presSpaceHandle());
      createdGC = true;
    }
  }

  if (paintEvent.rect().area())
  {
    paintEvent.clearBackground(drawingCanvas->backgroundColor());
    if (drawingCanvas->graphicList())
      drawingCanvas->graphicList()->drawOn(*gc);

    didDraw = true;
  }

  if (createdGC)
  {
    delete gc;
    gc = 0;
  }

  return didDraw;
}
