/*******************************************************************************
* FILE NAME: ifont.cpp                                                         *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ifont.hpp.                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
   #define INCL_GPILCIDS
   #define INCL_GPIPRIMITIVES
   #define INCL_WINMLE
   #define INCL_WINSTDFONT
   #define INCL_WINSYS
   #define INCL_WINERRORS
   #include <iwindefs.h>
   #include <string.h>
   #include <stdlib.h>
   #include <ctype.h>
}

#include <ifont.hpp>
#include <ifntpriv.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iexcept.hpp>
#include <iwindow.hpp>
#include <ifontdlg.hpp>
#include <ithread.hpp>
#include <iwcname.hpp>
  #include <ipmbidi.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFONT_CPP_
  #include <ipagetun.h>
#endif

#define IC_TRACE_ALL
#include <itrace.hpp>

// Local non-PM flags used by this class

#define IC_BITMAP_FONTS_ONLY              0x01
#define IC_VECTOR_FONTS_ONLY              0x02
#define IC_NONPROPORTIONAL_FONTS_ONLY     0x04
#define IC_MUSTUSEVECTORFONT              0x08

const int iCharArrayWidth = 256;

/*------------------------------------------------------------------------------
| Note:                                                                        |
|                                                                              |
| Because of reference counting, any new or existing member function which     |
| causes a change to the font should call copyPrivate().  This ensures that    |
| existing font objects which point to the same font private data object       |
| will not be changed.                                                         |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IFontPrivateData::IFontPrivateData                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFontPrivateData::IFontPrivateData() : flClStyles(0),
                                       pfontmetCl(0),
                                       pfattrsCl(0),
                                       pntClAngle(1,0),
                                       pntClShear(0,1),
                                       plClCharWidths(0),
                                       fxClPointSize(0),
                                       fxClHeight(0),
                                       fxClWidth(0),
                                       lClFontDir(CHDIRN_DEFAULT),
                                       lClLCId(0),
                                       refs(1)
{}


/*------------------------------------------------------------------------------
| IFontPrivateData::IFontPrivateData                                           |
|                                                                              |
| Construct font private data based an input parameters.                       |
------------------------------------------------------------------------------*/
IFontPrivateData::IFontPrivateData( const IFontPrivateData& fpd )
  : flClStyles( fpd.flClStyles ),
    pfontmetCl( 0 ),
    pfattrsCl( 0 ),
    pntClAngle( fpd.pntClAngle ),
    pntClShear( fpd.pntClShear ),
    plClCharWidths( 0 ),
    fxClPointSize( fpd.fxClPointSize ),
    fxClHeight( fpd.fxClHeight ),
    fxClWidth( fpd.fxClWidth ),
    lClFontDir( fpd.lClFontDir ),
    lClLCId( 0 ),
    refs( 1 )
{
  /* Initialize the font metrics structure. */
  if (fpd.pfontmetCl)
  {
    pfontmetCl = new FONTMETRICS;
    *pfontmetCl = *(fpd.pfontmetCl);
  }

  /* Initialize font attributes structure */
  if (fpd.pfattrsCl)
  {
    pfattrsCl = new FATTRS;
    *pfattrsCl = *(fpd.pfattrsCl);
  }
}

/*------------------------------------------------------------------------------
| IFontPrivateData::removeRef                                                  |
|                                                                              |
| Decrement the ref count and delete the object if refcount is zero.           |
------------------------------------------------------------------------------*/
void IFontPrivateData::removeRef()
{
  if ( --refs == 0 )
  {
    delete pfontmetCl;
    delete [] plClCharWidths;
    delete pfattrsCl;
    delete this;
  }
}


/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
|                                                                              |
| Construct a font for the font being used in an IWindow                       |
------------------------------------------------------------------------------*/
IFont :: IFont (const IWindow* pwindow)
         : ppd(new IFontPrivateData())
{
   IMODTRACE_DEVELOP("IFont-Constructor1");
   IWindowHandle wh;
   if (pwindow == 0)
      wh = HWND_DESKTOP;
   else
      wh = pwindow->handle();

   /*******************************************************************/
   /* For most windows, we can get the font information needed from   */
   /* the presentation space handle                                   */
   /*******************************************************************/
   HPS hps = WinGetPS(wh);
   initialize(hps);
   IString fontName( 0, 120 );
   unsigned long ulPresParams;

   /*******************************************************************/
   /* If the window passed is non-null, then do some additional       */
   /* processing to set the font up correctly                         */
   /*******************************************************************/
   if (pwindow != 0)
   {
     /*****************************************************************/
     /* First check the presentation parameters for any additional    */
     /* attributes being used by the graphics engine to display font  */
     /*****************************************************************/
     if (WinQueryPresParam(pwindow->handle(),
                           PP_FONTNAMESIZE,
                           0,
                           &ulPresParams,
                           (ULONG)120,
                           (PVOID)(char*)fontName,
                           QPF_NOINHERIT) != 0)
     {
       if (fontName.includes(".Bold"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_BOLD;
       if (fontName.includes(".Italic"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_ITALIC;
       if (fontName.includes(".Outline"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_OUTLINE;
       if (fontName.includes(".Underscore"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_UNDERSCORE;
       if (fontName.includes(".Strikeout"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_STRIKEOUT;
     }

     /*****************************************************************/
     /* Then check if window is an MLE control                        */
     /*****************************************************************/
     IWindowHandle hMle = pwindow->handle();
     IWindowClassName className( hMle );
     if ( className == WC_MLE )
     {
       /***************************************************************/
       /* If it is MLE, need to do some special case processing to    */
       /* create an IFont with the correct information since you      */
       /* cannot get at the hps for an MLE (as info set above is bad) */
       /***************************************************************/
       hMle.sendEvent( MLM_QUERYFONT,
                       IEventParameter1( ppd->pfattrsCl ),
                       IEventParameter2(0) );
       HDC hdc(GpiQueryDevice(hps));
       long vRes;
       DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &vRes);

       // Add 0.5 before doing integer conversion to avoid truncation errors
       unsigned long currPointSize = FIXEDINT(
        (( MAKEFIXED(ppd->pfattrsCl->lMaxBaselineExt, 0 ) * 72) / vRes)+0x8000);
       ppd->fxClPointSize = MAKEFIXED( currPointSize, 0 );
     }
   }
   WinReleasePS(hps);
}


/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
|                                                                              |
| Construct font based an input parameters.                                    |
------------------------------------------------------------------------------*/
IFont :: IFont         (const char*             pszFaceName,
                        unsigned long           ulPointSize,
                        Boolean                 useFixedFont,
                        Boolean                 useVectorFont,
                        const IPresSpaceHandle& presSpaceHandle)
         : ppd(new IFontPrivateData())
{
   IMODTRACE_DEVELOP("IFont-cstor with specified font");
   /* initialize font attributes structure */
   ppd->pfattrsCl = new FATTRS;
   memset(ppd->pfattrsCl, 0, sizeof(FATTRS));
   (ppd->pfattrsCl)->usRecordLength = sizeof(FATTRS);

// If setName(), setPointSize(), useNonPropOnly(), useBitmapOnly(),
// and useVectorOnly() are called from here we'll end up calling
// checkForMatch excesively.
// So just do what these functions do, but don't call fontChanged()
// and checkForMatch() until the end.

     strcpy((ppd->pfattrsCl)->szFacename, pszFaceName);

     ppd->fxClPointSize = MAKEFIXED(ulPointSize, 0);
     if(useVectorFont) {
       setCharHeight(ulPointSize, presSpaceHandle);
       setCharWidth(ulPointSize, presSpaceHandle);
     }

     if (useFixedFont)
        ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;
     else
        ppd->flClStyles = ppd->flClStyles & ~IC_NONPROPORTIONAL_FONTS_ONLY;

   if (useVectorFont)
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
   else
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;

   if (!useVectorFont) {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      mustUseVector(FALSE);
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      mustUseVector(TRUE);
   } /* endif */

   checkForMatch(presSpaceHandle);

   // Insure that the fontmetrics information matches the font selected
   // by creating the font temporarily into the hps passed and then
   // querying for the accurate information (not the general info)
   if ( ppd->pfontmetCl )
   {
     delete ppd->pfontmetCl;
     ppd->pfontmetCl = 0;
   }
   makeFontMetrics( presSpaceHandle );
}

/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
| Construct font to one being used in Presentation Space                       |
------------------------------------------------------------------------------*/
IFont :: IFont(const IPresSpaceHandle& presSpace)
         : ppd(new IFontPrivateData())
{
   initialize(presSpace);
}

/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
| Copy constructor                                                             |
------------------------------------------------------------------------------*/
IFont :: IFont (const IFont& fntmCopy)
{
  fntmCopy.ppd->addRef();
  ppd = fntmCopy.ppd;
}


/*------------------------------------------------------------------------------
| IFont::~IFont                                                                |
------------------------------------------------------------------------------*/
IFont :: ~IFont ()
{
   ppd->removeRef();
   ppd = 0;
}

/*------------------------------------------------------------------------------
| IFont::operator=                                                             |
------------------------------------------------------------------------------*/
IFont& IFont::operator= (const IFont &aFont)
{
  if (this == &aFont )
    return *this;
  aFont.ppd->addRef();
  ppd->removeRef();
  ppd = aFont.ppd;
  return (*this);
}


/*------------------------------------------------------------------------------
| IFont::useNonPropOnly                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: useNonPropOnly (Boolean nonProportionalOnly,
                                const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (nonProportionalOnly) {
      ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;
      if (!isFixed()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_NONPROPORTIONAL_FONTS_ONLY;
      fontChanged();
   } /* endif */
   checkForMatch(presSpaceHandle);
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::useBitmapOnly                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: useBitmapOnly (Boolean bitmapOnly,
                              const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (bitmapOnly) {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      mustUseVector(FALSE);
      if (!isBitmap()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      fontChanged();
      mustUseVector(FALSE);
   } /* endif */
   checkForMatch(presSpaceHandle);
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::useVectorOnly                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: useVectorOnly (Boolean vectorOnly,
                               const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (vectorOnly) {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      mustUseVector(TRUE);
      if (isBitmap()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      fontChanged();
      mustUseVector(FALSE);
   } /* endif */
   checkForMatch(presSpaceHandle);
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setDirection                                                          |
|                                                                              |
| This method only affects text drawn using beginUsingFont.                    |
------------------------------------------------------------------------------*/
IFont& IFont :: setDirection( Direction lFontDir )
{
   copyPrivate();
   switch (lFontDir) {
      case leftToRight:
         ppd->lClFontDir = CHDIRN_LEFTRIGHT;
         break;
      case rightToLeft:
         ppd->lClFontDir = CHDIRN_RIGHTLEFT;
         break;
      case topToBottom:
         ppd->lClFontDir = CHDIRN_TOPBOTTOM;
         break;
      case bottomToTop:
         ppd->lClFontDir = CHDIRN_BOTTOMTOP;
         break;
      case defaultDir:
         ppd->lClFontDir = CHDIRN_DEFAULT;
         break;
   } /* endswitch */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setName                                                               |
| Set the faceName of the font.                                                |
------------------------------------------------------------------------------*/
IFont& IFont :: setName (const char* pszFaceName,
                         const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   strcpy((ppd->pfattrsCl)->szFacename, pszFaceName);
   fontChanged();
   checkForMatch(presSpaceHandle);
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setPointSize                                                          |
------------------------------------------------------------------------------*/
IFont& IFont ::  setPointSize (unsigned long ulSize,
                               const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   ppd->fxClPointSize = MAKEFIXED(ulSize, 0);
   if (isBitmapOnly()) {
      fontChanged();
      checkForMatch(presSpaceHandle);
   } else {
      // As a performance boost for a vector font just need
      // to change the size and not try querying all fonts
      // for a closest match
      setCharHeight(ulSize, presSpaceHandle);
      setCharWidth(ulSize, presSpaceHandle);
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setSizeFromHps                                                        |
------------------------------------------------------------------------------*/
IFont& IFont ::  setSizeFromHps (unsigned long ulSize,
                                 const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   ppd->fxClPointSize = MAKEFIXED(ulSize, 0);
   setWidthHeight(ulSize, presSpaceHandle);
   makeCharWidths(presSpaceHandle);
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setBold                                                               |
------------------------------------------------------------------------------*/
IFont& IFont :: setBold(Boolean bold)
{
    copyPrivate();
    if (bold)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_BOLD;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_BOLD;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setItalic                                                             |
------------------------------------------------------------------------------*/
IFont& IFont :: setItalic(Boolean italics)
{
    copyPrivate();
    if (italics)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_ITALIC;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_ITALIC;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setUnderscore                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: setUnderscore(Boolean underscore)
{
    copyPrivate();
    if (underscore)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_UNDERSCORE;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_UNDERSCORE;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setStrikeout                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setStrikeout(Boolean strikeout)
{
    copyPrivate();
    if (strikeout)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_STRIKEOUT;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_STRIKEOUT;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setOutline                                                            |
------------------------------------------------------------------------------*/
IFont& IFont :: setOutline(Boolean outline)
{
    copyPrivate();
    if (outline)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_OUTLINE;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_OUTLINE;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setAllEmphasis                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: setAllEmphasis(Boolean setOn)
{
    copyPrivate();
    if (setOn)
       ppd->pfattrsCl->fsSelection = (FATTR_SEL_BOLD |
                                      FATTR_SEL_ITALIC |
                                      FATTR_SEL_UNDERSCORE |
                                      FATTR_SEL_STRIKEOUT |
                                      FATTR_SEL_OUTLINE);
    else
       ppd->pfattrsCl->fsSelection = 0;

    return *this;
}

/*------------------------------------------------------------------------------
| IFont::isBitmap                                                              |
------------------------------------------------------------------------------*/
Boolean IFont :: isBitmap () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (((ppd->pfontmetCl)->fsDefn & FM_DEFN_OUTLINE) == 0);
}

/*------------------------------------------------------------------------------
| IFont::isFixed                                                               |
------------------------------------------------------------------------------*/
Boolean IFont :: isFixed () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (Boolean)((ppd->pfontmetCl)->fsType & FM_TYPE_FIXED);
}



/*------------------------------------------------------------------------------
| IFont::name                                                                  |
------------------------------------------------------------------------------*/
IString IFont :: name() const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   IString strTemp((char*)((ppd->pfontmetCl)->szFacename));
   return strTemp;
}


/*------------------------------------------------------------------------------
| IFont::isNonPropOnly                                                         |
------------------------------------------------------------------------------*/
Boolean IFont :: isNonPropOnly () const
{
   return (ppd->flClStyles & IC_NONPROPORTIONAL_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::isBitmapOnly                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isBitmapOnly () const
{
   return (ppd->flClStyles & IC_BITMAP_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::isVectorOnly                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isVectorOnly () const
{
   return (ppd->flClStyles & IC_VECTOR_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::pointSize                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: pointSize () const
{
   return FIXEDINT(ppd->fxClPointSize);
}


/*------------------------------------------------------------------------------
| IFont::isBold                                                                |
------------------------------------------------------------------------------*/
Boolean IFont :: isBold ( ) const
{
   return (ppd->pfattrsCl->fsSelection & FATTR_SEL_BOLD);
}

/*------------------------------------------------------------------------------
| IFont::isItalic                                                              |
------------------------------------------------------------------------------*/
Boolean IFont :: isItalic ( ) const
{
   return (ppd->pfattrsCl->fsSelection & FATTR_SEL_ITALIC);
}

/*------------------------------------------------------------------------------
| IFont::isUnderscore                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isUnderscore ( ) const
{
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_UNDERSCORE ;
}

/*------------------------------------------------------------------------------
| IFont::isStrikeout                                                           |
------------------------------------------------------------------------------*/
Boolean IFont :: isStrikeout ( ) const
{
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_STRIKEOUT ;
}

/*------------------------------------------------------------------------------
| IFont::isOutline                                                             |
------------------------------------------------------------------------------*/
Boolean IFont :: isOutline ( ) const
{
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_OUTLINE ;
}



/*------------------------------------------------------------------------------
| IFont::avgCharWidth                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgCharWidth() const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lAveCharWidth;
}


/*------------------------------------------------------------------------------
| IFont::maxSize                                                               |
------------------------------------------------------------------------------*/
ISize IFont :: maxSize () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lMaxCharInc, ppd->pfontmetCl->lMaxBaselineExt);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::maxCharHeight                                                         |
------------------------------------------------------------------------------*/
unsigned long  IFont :: maxCharHeight () const
{
   return maxSize().height();
}

/*------------------------------------------------------------------------------
| IFont::maxUppercaseSize                                                      |
------------------------------------------------------------------------------*/
ISize IFont :: maxUppercaseSize( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lEmInc, ppd->pfontmetCl->lMaxBaselineExt);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::superscriptSize                                                       |
------------------------------------------------------------------------------*/
ISize IFont :: superscriptSize( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSuperscriptXSize,
              ppd->pfontmetCl->lSuperscriptYSize);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::subscriptSize                                                         |
------------------------------------------------------------------------------*/
ISize IFont :: subscriptSize( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSubscriptXSize,
              ppd->pfontmetCl->lSubscriptYSize);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::superscriptOffset                                                     |
------------------------------------------------------------------------------*/
ISize IFont :: superscriptOffset( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSuperscriptXOffset,
              ppd->pfontmetCl->lSuperscriptYOffset);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::subscriptOffset                                                       |
------------------------------------------------------------------------------*/
ISize IFont :: subscriptOffset( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSubscriptXOffset,
              ppd->pfontmetCl->lSubscriptYOffset);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::avgLowercase                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgLowercase () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lXHeight;

}

/*------------------------------------------------------------------------------
| IFont::avgUppercase                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgUppercase () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lEmHeight;

}

/*------------------------------------------------------------------------------
| IFont::maxLowercaseAscender                                                  |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxLowercaseAscender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lLowerCaseAscent;

}

/*------------------------------------------------------------------------------
| IFont::maxLowercaseDescender                                                 |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxLowercaseDescender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lLowerCaseDescent;

}

/*------------------------------------------------------------------------------
| IFont::maxDescender                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxDescender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lMaxDescender;
}

/*------------------------------------------------------------------------------
| IFont::maxAscender                                                           |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxAscender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lMaxAscender;
}

/*------------------------------------------------------------------------------
| IFont::externalLeading                                                       |
------------------------------------------------------------------------------*/
unsigned long IFont :: externalLeading () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lExternalLeading;
}

/*------------------------------------------------------------------------------
| IFont::internalLeading                                                       |
------------------------------------------------------------------------------*/
unsigned long IFont :: internalLeading () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lInternalLeading;
}


/*------------------------------------------------------------------------------
| IFont::charWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: charWidth (char c) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeCharWidths();
   return ppd->plClCharWidths[c];
}


/*------------------------------------------------------------------------------
| IFont::textWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long  IFont :: textWidth (const char* pszText)  const
{
   unsigned long ulWidth = 0;
   if (pszText) {
      IFont* const localThis = (IFont* const)this;
      localThis->makeCharWidths();
      const char* Temp = pszText;
      while (*Temp) {
         ulWidth += ppd->plClCharWidths[*Temp];
         Temp++;
      } /* endwhile */
   } /* endif */

   return ulWidth;
}


/*------------------------------------------------------------------------------
| IFont::textWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long  IFont :: textWidth (const char*    pszText,
                                   IBidiSettings& settings)  const
{
   IString tmpText(pszText);
   LAYOUT_OBJECT layout = bidiCreateLayoutObject();
   unsigned long tilde = 0;
   bidiPrepareForDisplay(layout,tmpText,settings,tilde);
   bidiDeleteLayoutObject(layout);
   return textWidth(tmpText);
}



/*------------------------------------------------------------------------------
| IFont::textLines                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: textLines    (const char* pszLine,
                                     unsigned long ulLineWidth) const
{
   unsigned long ulLines = 0;
   if (ulLineWidth  &&
       this->minTextWidth( pszLine ) <= ulLineWidth)
   {
      ulLines = 1;
      unsigned long ulWidth = 0;
      char* LastSpace = 0;
      char* Ptr = (char*)pszLine;
      while (*Ptr) {
         if (ulWidth + charWidth( *Ptr ) > ulLineWidth)
         {                   // More text than will fit.
            if (isspace( *Ptr ))
            {                // Space between lines.
               LastSpace = Ptr;
            }
            else if (LastSpace)
            {                // Back up to start of word.
               Ptr = LastSpace;
            }
            Ptr++;
            ulLines++;       // Need another line.
            ulWidth = 0;
         }
         else if (*Ptr == '\n')
         {                   // New line character (fits on this line).
            Ptr++;
            if (*Ptr)
            {                // More characters for another line.
               ulLines++;    // Need another line.
               ulWidth = 0;
            }
         }
         else
         {                   // Adding to current line.
            ulWidth += charWidth(*Ptr);
            if (isspace(*Ptr)) {
               LastSpace = Ptr;
            }
            Ptr++;
         }
      } /* endwhile */
   } /* endif */
   return ulLines;
}

/*------------------------------------------------------------------------------
| IFont::minTextWidth                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: minTextWidth     (const char* pszLine) const
{
   if (pszLine) {
      char* Temp = new char[strlen(pszLine)+1];
      strcpy(Temp, pszLine);
      char* Start = Temp;
      unsigned long ulWidth = 0;
      while (*Start) {
         while (*Start && isspace(*Start)) {
            Start++;
         } /* endwhile */
         char* End = Start;
         while (*End && !isspace(*End)) {
            End++;
         } /* endwhile */
         char SaveCh = *End;
         *End = 0;
         unsigned long ulChkWidth = textWidth(Start);
         if (ulChkWidth > ulWidth) {
            ulWidth = ulChkWidth;
         } /* endif */
         *End = SaveCh;
         Start = End;
      } /* endwhile */
      delete [] Temp;
      return ulWidth;
   } else {
      return 0;
   } /* endif */
}



/*------------------------------------------------------------------------------
| IFont::beginUsingFont                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: beginUsingFont (const IPresSpaceHandle& hps)
{
   IMODTRACE_DEVELOP("IFont::beingUsingFont");
   if (!ppd->lClLCId) {
      ppd->lClLCId = 1;

      // Have to make sure our local Id is not already in use
      // First, get the total number of local Ids in use
      long NumLcId = GpiQueryNumberSetIds(hps);

      // Check if there are any in use
      if (NumLcId) {
         long Types[255];
         STR8 Names[255];
         long LcIds[255];

         // Get information about all local Ids
         Boolean RC = GpiQuerySetIds(hps, NumLcId, Types, Names, LcIds);
         if (!RC) {
            ITRACE_DEVELOP("Error getting ids.");
            ITHROWGUIERROR("GpiQuerySetIds");
         } /* endif */

         // Check all used Ids to see if ours is already used
         for (unsigned long LCNum = 0; LCNum < NumLcId ; LCNum++ ) {
            if (LcIds[LCNum] == ppd->lClLCId) {
               // Ours is already used, increment to next Id and restart search
               ppd->lClLCId++;
               LCNum = -1;

               // NewLCId must be in the range 1 to 254
               if (ppd->lClLCId > 254) {
                  // Error! No more free ids!
                  ITRACE_DEVELOP("Error no free ids.");
                  ITHROWGUIERROR("GpiQuerySetIds");
               } /* endif */
            } /* endif */
         } /* endfor */

      } /* endif */
   } /* endif */

   LONG match = GpiCreateLogFont(hps, 0, ppd->lClLCId, ppd->pfattrsCl);
   if (match==GPI_ERROR) {
      ITHROWGUIERROR("GpiCreateLogFont");
   } /* endif */
   if (!GpiSetCharSet(hps, ppd->lClLCId)) {
      ITHROWGUIERROR("GpiSetCharSet");
   }
          /* Check if the font found is an outline (vector) font      */
   if ((ppd->pfattrsCl)->fsFontUse & FATTR_FONTUSE_OUTLINE) {
            /* Set the Character mode so that the characters will     */
            /* be stretched properly.                                 */
      GpiSetCharMode(hps, CM_MODE2);

      SIZEF sizef;
      sizef.cx = ppd->fxClWidth;
      sizef.cy = ppd->fxClHeight;
            /* Set character box for the outline font to size above   */
      GpiSetCharBox( hps, &sizef );
      GRADIENTL gradlAngle;
      gradlAngle.x = ppd->pntClAngle.x();
      gradlAngle.y = ppd->pntClAngle.y();
      GpiSetCharAngle(hps, &gradlAngle);
      POINTL shearAngle = ppd->pntClShear.asPOINTL();
      GpiSetCharShear(hps, &shearAngle);
   } /* endif */
   else if ( ppd->pfontmetCl )
   {
      GpiSetCharMode(hps, CM_MODE2);
      SIZEF sizef = { MAKEFIXED(ppd->pfontmetCl->lEmInc,0),
                      MAKEFIXED(ppd->pfontmetCl->lEmHeight,0) };
      GpiSetCharBox( hps, &sizef );
   }
   GpiSetCharDirection(hps, ppd->lClFontDir);

   return *this;
}


/*------------------------------------------------------------------------------
| IFont::endUsingFont                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: endUsingFont (const IPresSpaceHandle& hps)
{
   IMODTRACE_DEVELOP("IFont::endUsingFont");
   GpiSetCharSet(hps, LCID_DEFAULT);
   if (ppd->lClLCId) {
      GpiDeleteSetId(hps, ppd->lClLCId);
      ppd->lClLCId = 0;
   } /* endif */
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::makeFontMetrics                                                       |
------------------------------------------------------------------------------*/
IFont& IFont :: makeFontMetrics  (const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::makeFontMetrics");
   if (!ppd->pfontmetCl) {
      HPS hps;
      if(presSpaceHandle)
        hps = presSpaceHandle;
      else
        hps = WinGetPS(HWND_DESKTOP);
      beginUsingFont(hps);
      ppd->pfontmetCl = new FONTMETRICS;
      GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), ppd->pfontmetCl);
      endUsingFont(hps);
      if (presSpaceHandle == 0)
        WinReleasePS(hps);
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::makeCharWidths                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: makeCharWidths (const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::makeCharWidths");
   if (!ppd->plClCharWidths) {
      HPS hps;
      if (presSpaceHandle)
        hps = presSpaceHandle;
      else
        hps = WinGetPS(HWND_DESKTOP);
      beginUsingFont(hps);
      ppd->plClCharWidths = new long[iCharArrayWidth];
      GpiQueryWidthTable(hps, 0, iCharArrayWidth, ppd->plClCharWidths);
      endUsingFont(hps);
      if (presSpaceHandle == 0)
        WinReleasePS(hps);
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::checkForMatch                                                         |
|                                                                              |
| Search for the closest match in all installed fonts.                         |
------------------------------------------------------------------------------*/
IFont& IFont :: checkForMatch (const IPresSpaceHandle& presSpaceHandle)
{

   unsigned long ulPtSize = FIXEDINT(ppd->fxClPointSize);
   long lNumFonts;
   long lReqFonts = 0;
   long lBitmapMatch = 0;
   long lNextBitmap = 0;
   long lVectorMatch = 0;
   long lHowClose = 99999;
   unsigned long
     Index = 0,
     BitmapIndex = 0,
     VectorIndex = 0,
     NextBitmapIndex = 0;

   HPS hps;
   if (presSpaceHandle) {
      hps = presSpaceHandle;
   } else {
      hps = WinGetPS(HWND_DESKTOP);
   } /* endif */

   lNumFonts = GpiQueryFonts(hps, QF_PUBLIC | QF_PRIVATE,
                  (PSZ)(ppd->pfattrsCl)->szFacename, &lReqFonts, 0L, 0);

   if (lNumFonts) {
      ITRACE_DEVELOP(IString(lNumFonts) + IString("Fonts found."));

      FONTMETRICS* pfm;

      pfm = new FONTMETRICS[lNumFonts];

      GpiQueryFonts(hps, QF_PUBLIC | QF_PRIVATE, (PSZ)(ppd->pfattrsCl)->szFacename,
         &lNumFonts, sizeof(FONTMETRICS), pfm);
      ITRACE_DEVELOP((IString(lNumFonts) + IString("Fonts found for ")) +
                      IString(ppd->pfattrsCl->szFacename) );
      long lHorzRes, lVertRes;
      HDC hdc = GpiQueryDevice(hps);
      DevQueryCaps(hdc, CAPS_HORIZONTAL_FONT_RES, 1L, &lHorzRes);
      DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &lVertRes);
      ITRACE_DEVELOP((IString("H&V resolution ") + IString(lHorzRes)) +
                      IString(lVertRes));

      for (Index = 0; Index < lNumFonts ; Index++)
      {
         // Skip font if it doesn't match the Proportional/Fixed specification
         // or the Vector/Bitmap specification
         if ( ((isNonPropOnly()) && (pfm[Index].fsType & FM_TYPE_FIXED))   ||
              ((!isNonPropOnly()) && (!(pfm[Index].fsType & FM_TYPE_FIXED))) ||
              ((isVectorOnly()) && (pfm[Index].fsDefn & FM_DEFN_OUTLINE))  ||
              ((isBitmapOnly()) && (!(pfm[Index].fsDefn & FM_DEFN_OUTLINE))) )
         {
            // Vector font doesn't need to check device resolution and point size
            if (pfm[Index].fsDefn & FM_DEFN_OUTLINE)
            {
               lVectorMatch = pfm[Index].lMatch;
               VectorIndex = Index;
            }
            else
              // Check bitmap for device resolution and pointSize
              if ((pfm[Index].sXDeviceRes == lHorzRes) &&
                  (pfm[Index].sYDeviceRes == lVertRes))
              {
                ITRACE_DEVELOP((IString(pfm[Index].sNominalPointSize) +
                                IString(" vs. ")) + IString(ulPtSize*10) );
                if (pfm[Index].sNominalPointSize == (ulPtSize*10))
                {
                   lBitmapMatch = pfm[Index].lMatch;
                   BitmapIndex = Index;
                }
                else
                {
                   long currentHowClose =
                       abs((ulPtSize*10) - pfm[Index].sNominalPointSize);
                   if (!lBitmapMatch &&
                       currentHowClose <= lHowClose)
                   {
                       lNextBitmap = pfm[Index].lMatch;
                       NextBitmapIndex = Index;
                       lHowClose = currentHowClose;
                   } /* endif */
                }
              }
         } /* endif for match on specification */

      } /* endfor */

      // We'll reset Index below to the position in pfm that's the best match.
      // That way we can save the FONTMETRICS.
      if (isBitmapOnly()) {
         if (!lBitmapMatch) {
            lBitmapMatch = lNextBitmap;
         } /* endif */
         (ppd->pfattrsCl)->lMatch = lBitmapMatch;
         ITRACE_DEVELOP("Exact bitmap match found.");
         (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
         Index = BitmapIndex;
      } else {
         if (isVectorOnly() || (ppd->flClStyles & IC_MUSTUSEVECTORFONT)) {
            (ppd->pfattrsCl)->lMatch = lVectorMatch;
            ITRACE_DEVELOP("Exact vector match found.");
            (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse | FATTR_FONTUSE_OUTLINE;
            Index = VectorIndex;
         } else {
            if (lBitmapMatch) {
               (ppd->pfattrsCl)->lMatch = lBitmapMatch;
               (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
               Index = BitmapIndex;
            } else {
               if (lVectorMatch) {
                  (ppd->pfattrsCl)->lMatch = lVectorMatch;
                  (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse | FATTR_FONTUSE_OUTLINE;
                  Index = VectorIndex;
               } else {
                  (ppd->pfattrsCl)->lMatch = lNextBitmap;
                  (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
                  Index = NextBitmapIndex;
               } /* endif */
            } /* endif */
         } /* endif */
      } /* endif */

      // Ok, now let's save the FONTMETRICS
      delete ppd->pfontmetCl;
      ppd->pfontmetCl = new FONTMETRICS;
      memcpy(ppd->pfontmetCl, &pfm[Index], sizeof(FONTMETRICS));

      // Set some FATTRS values from the FONTMETRICS
      (ppd->pfattrsCl)->idRegistry = (ppd->pfontmetCl)->idRegistry;
      (ppd->pfattrsCl)->usCodePage = (ppd->pfontmetCl)->usCodePage;
      if (!(ppd->pfontmetCl->fsDefn & FM_DEFN_OUTLINE))
      {
         (ppd->pfattrsCl)->lMaxBaselineExt = (ppd->pfontmetCl)->lMaxBaselineExt;
         (ppd->pfattrsCl)->lAveCharWidth = (ppd->pfontmetCl)->lAveCharWidth;
      }
      (ppd->pfattrsCl)->fsFontUse |= FATTR_FONTUSE_NOMIX;

      delete [] pfm;
   } /* endif */

   if ((isNonPropOnly() || isBitmapOnly() || isVectorOnly())
         && !lBitmapMatch && !lVectorMatch) {
      if (strcmp((ppd->pfattrsCl)->szFacename, (char*)"Courier")) {
         strcpy((ppd->pfattrsCl)->szFacename, (char*)"Courier");
         ITRACE_DEVELOP("Checking for match again.");
         checkForMatch(hps);
      } /* endif */
   } /* endif */

   if(presSpaceHandle == 0)
     WinReleasePS(hps);

   return *this;
}


/*------------------------------------------------------------------------------
| IFont::mustUseVector                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: mustUseVector (Boolean fDoIt)
{
   if (isBitmapOnly()) {
      ppd->flClStyles = ppd->flClStyles & ~IC_MUSTUSEVECTORFONT;
      return *this;
   } /* endif */
   if (ppd->pntClAngle.x() != 1 || ppd->pntClAngle.y() != 0 ||
       ppd->pntClShear.x() != 0 || ppd->pntClShear.y() != 1 || isVectorOnly()) {
      ppd->flClStyles = ppd->flClStyles | IC_MUSTUSEVECTORFONT;
      return *this;
   } /* endif */

   if (fDoIt) {
      ppd->flClStyles = ppd->flClStyles | IC_MUSTUSEVECTORFONT;
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_MUSTUSEVECTORFONT;
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::initialize                                                            |
------------------------------------------------------------------------------*/
IFont& IFont :: initialize(const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::initialize");
   ppd->pfattrsCl = new FATTRS;
   HPS hps;

   if (presSpaceHandle == 0)
      hps = WinGetPS(HWND_DESKTOP);
   else
      hps = presSpaceHandle;

   ppd->pfontmetCl = new FONTMETRICS;
   GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), ppd->pfontmetCl);

   // Set our internal flags for this font so that checkForMatch
   // will work properly.
   // Don't call the useNonPropOnly() type function since they'll
   // cause us to call checkForMatch() prematurely.

   if (ppd->pfontmetCl->fsType & FM_TYPE_FIXED)
     ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;

   if (ppd->pfontmetCl->fsDefn & FM_DEFN_OUTLINE)
     ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
   else
     ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;

   /* initialize font attributes structure */
   (ppd->pfattrsCl)->usRecordLength = sizeof(FATTRS);
   (ppd->pfattrsCl)->fsSelection = (ppd->pfontmetCl)->fsSelection;
   (ppd->pfattrsCl)->lMatch = (ppd->pfontmetCl)->lMatch;
   (ppd->pfattrsCl)->idRegistry = (ppd->pfontmetCl)->idRegistry;
   (ppd->pfattrsCl)->usCodePage = (ppd->pfontmetCl)->usCodePage;
   (ppd->pfattrsCl)->fsType = 0;
   (ppd->pfattrsCl)->fsFontUse = FATTR_FONTUSE_NOMIX;
   strcpy((ppd->pfattrsCl)->szFacename, (ppd->pfontmetCl)->szFacename);

   long currPointSize;
   if (ppd->flClStyles & IC_VECTOR_FONTS_ONLY) {
     (ppd->pfattrsCl)->lMaxBaselineExt = 0;
     (ppd->pfattrsCl)->lAveCharWidth = 0;

     long lyFontResolution;
     SIZEF sizef;
     HDC hDC = GpiQueryDevice(hps);
     DevQueryCaps( hDC, CAPS_VERTICAL_FONT_RES,
                   (long)1, &lyFontResolution);
     GpiQueryCharBox (hps, &sizef);
     // Add 0.5 before doing integer conversion to avoid truncation errors
     currPointSize = FIXEDINT((((sizef.cy) * 72) / lyFontResolution)+0x8000);
     }
   else {
     (ppd->pfattrsCl)->lMaxBaselineExt = (ppd->pfontmetCl)->lMaxBaselineExt;
     (ppd->pfattrsCl)->lAveCharWidth = (ppd->pfontmetCl)->lAveCharWidth;
     currPointSize = (ppd->pfontmetCl)->sNominalPointSize/10;
     }

   setSizeFromHps(currPointSize, hps);

   if(presSpaceHandle == 0)
      WinReleasePS(hps);

   return *this;
}


/*------------------------------------------------------------------------------
| IFont::fontChanged                                                           |
|                                                                              |
| Will cause a subsequent call to makeFontMetrics()                            |
------------------------------------------------------------------------------*/
IFont& IFont :: fontChanged ()
{
// IMODTRACE_DEVELOP("IFont::fontChanged");
   delete [] ppd->plClCharWidths;
   ppd->plClCharWidths = 0;
   delete ppd->pfontmetCl;
   ppd->pfontmetCl = 0;
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setWindowFont                                                         |
|                                                                              |
| Used by IControl::setFont()
------------------------------------------------------------------------------*/
Boolean IFont :: setWindowFont(IWindow* pwindow) const
{
   IMODTRACE_DEVELOP("IFont::setWindowFont");
      Boolean result;

      // Build string to pass on WinSetPresParam call
      IString presParm(pointSize());
      presParm += ".";
      presParm += name();

      if (isBold())
        presParm += ".Bold";

      if (isItalic())
        presParm += ".Italic";

      if (isOutline())
        presParm += ".Outline";
      if (isUnderscore())
        presParm += ".Underscore";
      if (isStrikeout())
        presParm += ".Strikeout";

      result = WinSetPresParam(pwindow->handle(),
                      PP_FONTNAMESIZE,
                      presParm.length()+1,
                      (char*)presParm);

      if (result == false)
        ITHROWGUIERROR("WinSetPresParam");

      return result;
}

/*------------------------------------------------------------------------------
| IFont::setCharWidth                                                          |
|                                                                              |
| Only affects chars draw after beginUsingFont                                 |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharWidth     (unsigned long ulWidth,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   // Only works for Vector Fonts
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   long  lxFontResolution;
   HDC   hDC;
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   HPS hps;
   if (presSpaceHandle)
     hps = presSpaceHandle;
   else
     hps = WinGetPS(HWND_DESKTOP);
   hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_HORIZONTAL_FONT_RES,
                 (long)1, &lxFontResolution);
         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClWidth = (FIXED)((MAKEFIXED(ulWidth, 0) / 72 ) *
                     lxFontResolution );
   delete [] ppd->plClCharWidths;
   ppd->plClCharWidths = 0;
   makeCharWidths(hps);

   if (presSpaceHandle == 0)
     WinReleasePS(hps);
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setCharHeight                                                         |
|                                                                              |
| Only affects chars draw after beginUsingFont                                 |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharHeight    (unsigned long ulHeight,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   // only valid for Vector fonts
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   long  lyFontResolution;
   HDC   hDC;
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   HPS hps;
   if(presSpaceHandle)
     hps = presSpaceHandle;
   else
     hps = WinGetPS(HWND_DESKTOP);
   hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_VERTICAL_FONT_RES,
                 (long)1, &lyFontResolution);
         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClHeight = (FIXED)(((MAKEFIXED(ulHeight, 0)) / 72 ) *
                      lyFontResolution );

   if (presSpaceHandle == 0)
     WinReleasePS(hps);
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setWidthHeight                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: setWidthHeight   (unsigned long ulPointSize,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
     return *this;

   long  devRes[2];
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   HPS hps;
   if(presSpaceHandle)
      hps = presSpaceHandle;
   else
      hps = WinGetPS(HWND_DESKTOP);
   HDC hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_HORIZONTAL_FONT_RES,
                 (long)2, devRes);
   if (presSpaceHandle == 0)
     WinReleasePS(hps);

         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClWidth = (FIXED)((MAKEFIXED(ulPointSize, 0) / 72 ) *
                     devRes[0]);
   ppd->fxClHeight = (FIXED)(((MAKEFIXED(ulPointSize, 0)) / 72 ) *
                     devRes[1]);
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setCharSize                                                           |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharSize (const ISize& size,
                             const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   setCharWidth(size.width(), presSpaceHandle);
   return setCharHeight(size.height(), presSpaceHandle);
}


/*------------------------------------------------------------------------------
| IFont::setFontAngle                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontAngle     (const IPoint& point,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   ppd->pntClAngle = point;
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setFontShear                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontShear     (const IPoint& point,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   ppd->pntClShear = point;
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setFontFromDialog                                                     |
|                                                                              |
| Change the IFont to correspond to the one specified by the fontDlg.          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontFromDialog (IFontDialog& fontDlg,
                                   const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::setFontFromDialog");

   copyPrivate();
   delete ppd->pfattrsCl;

   ppd->pfattrsCl = new FATTRS;
   memcpy(ppd->pfattrsCl, &(fontDlg.fData->fAttrs), sizeof(FATTRS));

   if ( ppd->pfattrsCl->fsFontUse & FATTR_FONTUSE_OUTLINE )
   {
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
   }
   else
   {
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
   }

   fontChanged();
   setSizeFromHps(fontDlg.pointSize(), presSpaceHandle);
   checkForMatch(presSpaceHandle);

   return *this;
}

/*------------------------------------------------------------------------------
| IFont::copyWidthsArray                                                       |
|                                                                              |
| Copies the widths array into the specified buffer.                           |
| Used as a performance boost for calculations done for ICanvas layout.        |
------------------------------------------------------------------------------*/
IFont& IFont :: copyWidthsArray(char* buf, unsigned long length)
{
   IASSERTPARM(buf && length <= iCharArrayWidth);
   memcpy(buf, ppd->plClCharWidths, (length * sizeof(long)));

   return *this;
}

/*------------------------------------------------------------------------------
| IFont::fattrs                                                                |
|                                                                              |
| Return a const pointer to the OS/2 FATTRS struct.                            |
------------------------------------------------------------------------------*/
const struct _FATTRS* IFont :: fattrs () const
{
   return ppd->pfattrsCl;
}

/*------------------------------------------------------------------------------
| IFont::fontmetrics                                                           |
|                                                                              |
| Return a const pointer to the OS/2 FONTMETRICS struct.                       |
------------------------------------------------------------------------------*/
const struct _FONTMETRICS* IFont :: fontmetrics () const
{
   return ppd->pfontmetCl;
}

/*------------------------------------------------------------------------------
| IFont::copyPrivate                                                           |
|                                                                              |
| Makes a new copy of the fonts private data, decrements the use count for     |
| the old private data and increments the count for the new private data.      |
------------------------------------------------------------------------------*/
IFont& IFont :: copyPrivate( )
{
  if ( ppd->refCount() > 1 )
  {
    IFontPrivateData* newppd = new IFontPrivateData( *ppd );
    newppd->addRef();
    ppd->removeRef();
    ppd = newppd;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::FaceNameCursor                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::FaceNameCursor::FaceNameCursor ( FontType fontType,
                                        const IPresSpaceHandle& presSpaceHandle)
                                      : index(0), count(0), facenames(0)
{
  IPresSpaceHandle hps(presSpaceHandle);

  if (!hps)
  {
    hps = WinGetPS( HWND_DESKTOP );
  }

  HDC hdc(GpiQueryDevice(hps));
  long hres,vres;
  DevQueryCaps(hdc, CAPS_HORIZONTAL_FONT_RES, 1L, &hres);
  DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &vres);

  long numRequired(0);
  long fontCount( GpiQueryFonts( hps,
                                 QF_PRIVATE | QF_PUBLIC,
                                 0,
                                 &numRequired,
                                 (long)sizeof(FONTMETRICS),
                                 0 ));
  if ( fontCount )
  {
    PFONTMETRICS fontmetrics = new FONTMETRICS[fontCount];

    count = GpiQueryFonts( hps,
                           QF_PRIVATE | QF_PUBLIC,
                           0,
                           &fontCount,
                           (long)sizeof(FONTMETRICS),
                           fontmetrics );
    count = fontCount;
    int deviceCount;
    int i;
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ( fontType == FaceNameCursor::both &&
         (((fontmetrics[i].sXDeviceRes == hres) &&
           (fontmetrics[i].sYDeviceRes == vres)) || (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE)))
        deviceCount++;
      else if ( fontType == FaceNameCursor::bitmap &&
              ((fontmetrics[i].sXDeviceRes == hres) &&
               (fontmetrics[i].sYDeviceRes == vres)) &&
              !(fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
      else if (fontType == FaceNameCursor::vector &&
               (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
    } /* endfor */
    facenames = new IString[deviceCount];
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ( fontType == FaceNameCursor::both &&
         (((fontmetrics[i].sXDeviceRes == hres) &&
           (fontmetrics[i].sYDeviceRes == vres)) || (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE)))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
      else if (fontType == FaceNameCursor::bitmap &&
              ((fontmetrics[i].sXDeviceRes == hres) &&
               (fontmetrics[i].sYDeviceRes == vres)) &&
              !(fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
      else if (fontType == FaceNameCursor::vector &&
               (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
    } /* endfor */
    delete [] fontmetrics;
    count = deviceCount;
  }
  if (!presSpaceHandle)
  {
    WinReleasePS( hps );
  }
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::~FacenameCursor                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::FaceNameCursor::~FaceNameCursor ()
{
  if (count)
  {
    delete [] facenames;
  }
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToFirst                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToFirst()
{
  if (count)
  {
    index = 0;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToNext                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToNext()
{
  Boolean foundDuplicate;
  int j(index);
  index++;

  while (index < count)
  {
    foundDuplicate = false;
    for (int i=j;i>=0 && !foundDuplicate;i--)
    {
      if ( facenames[i] == facenames[index] )
        foundDuplicate = true;
    } /* endfor */
    if (!foundDuplicate)
      break;
    index++;
  } /* endwhile */

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToPrevious                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToPrevious()
{

  Boolean foundDuplicate;
  int j(index);
  index--;

  while (index < count)
  {
    foundDuplicate = false;
    for (int i=j;i<count && !foundDuplicate;i++)
    {
      if ( facenames[i] == facenames[index] )
        foundDuplicate = true;
    } /* endfor */
    if (!foundDuplicate)
      break;
    index--;
  } /* endwhile */

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToLast                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToLast()
{

  if (count)
  {
    index = count - 1;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::isValid                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::isValid() const
{
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::invalidate                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IFont::FaceNameCursor::invalidate()
{
  index = count;
}

/*------------------------------------------------------------------------------
| IFont::faceNameAt                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IFont::faceNameAt( const FaceNameCursor& faceNameCursor )
{
  return faceNameCursor.facenames[faceNameCursor.index];
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::PointSizeCursor                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::PointSizeCursor::PointSizeCursor ( const char* faceName,
                                        const IPresSpaceHandle& presSpaceHandle)
                                        : index(0), count(0), pointsizes(0)
{
  IPresSpaceHandle hps(presSpaceHandle);

  if (!hps)
  {
    hps = WinGetPS( HWND_DESKTOP );
  }

  HDC hdc(GpiQueryDevice(hps));
  long hres,vres;
  DevQueryCaps(hdc, CAPS_HORIZONTAL_FONT_RES, 1L, &hres);
  DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &vres);

  long numRequired(0);
  long fontCount( GpiQueryFonts( hps,
                                 QF_PRIVATE | QF_PUBLIC,
                                 faceName,
                                 &numRequired,
                                 (long)sizeof(FONTMETRICS),
                                 0 ));
  if ( fontCount )
  {
    PFONTMETRICS fontmetrics = new FONTMETRICS[fontCount];

    count = GpiQueryFonts( hps,
                           QF_PRIVATE | QF_PUBLIC,
                           faceName,
                           &fontCount,
                           (long)sizeof(FONTMETRICS),
                           fontmetrics );
    count = fontCount;
    int deviceCount;
    int i;
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ((fontmetrics[i].sXDeviceRes == hres &&
           fontmetrics[i].sYDeviceRes == vres ) ||
          (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
    } /* endfor */
    pointsizes = new long[deviceCount];
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ((fontmetrics[i].sXDeviceRes == hres &&
           fontmetrics[i].sYDeviceRes == vres ) ||
          (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        pointsizes[deviceCount] = fontmetrics[i].sNominalPointSize/10;
        deviceCount++;
      }
    } /* endfor */
    delete [] fontmetrics;
    count = deviceCount;
  }
  if (!presSpaceHandle)
  {
    WinReleasePS( hps );
  }
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::~PointSizeCursor                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::PointSizeCursor::~PointSizeCursor ()
{
  if (count)
  {
    delete [] pointsizes;
  }
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToFirst                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToFirst()
{
  if (count)
  {
    index = 0;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToNext                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToNext()
{
  index++;

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToPrevious                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToPrevious()
{
  index--;
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToLast                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToLast()
{

  if (count)
  {
    index = count - 1;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::isValid                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::isValid() const
{
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::invalidate                                           |
|                                                                              |
| Invalidate the cursor.                                                       |
------------------------------------------------------------------------------*/
void IFont::PointSizeCursor::invalidate()
{
  index = count;
}

/*------------------------------------------------------------------------------
| IFont::pointsizeAt                                                           |
|                                                                              |
| Return the point size for the argument cursor.                               |
------------------------------------------------------------------------------*/
long IFont::pointSizeAt( const PointSizeCursor& pointSizeCursor )
{
  return pointSizeCursor.pointsizes[pointSizeCursor.index];
}
