/*******************************************************************************
* FILE NAME: iframhdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iframhdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINFRAMEMGR
  #define INCL_WININPUT                 // For WinQueryFocus.
  #define INCL_WINTRACKRECT
  #include <iwindefs.h>
  }

#include <iframhdr.hpp>

#include <iframprv.hpp>
#include <iframevt.hpp>
#include <iswp.hpp>
#include <itrace.hpp>
#include <icontrol.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <ihelp.hpp>
#include <icmdevt.hpp>
#include <iframe.hpp>
#include <inotifev.hpp>
#include <icoordsy.hpp>
#include <isysmenu.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFRAMHDR_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IFrameHandler::dispatchHandlerEvent                                          |
|                                                                              |
| Intercepts all window events and dispatches frame-specific events to the     |
| various handle event functions.                                              |
| Note: The following messages are of interest:                                |
|    For WM_CALCFRAMERECT:                                                     |
|      Dispatch the calcRect function.                                         |
|                                                                              |
|    For WM_QUERYFRAMECTLCOUNT:                                                |
|      1. Get the number of standard controls by invoking the default frame    |
|         window procedure.                                                    |
|      2. Get the number of frame extensions via a call to numberOfExtensions()|
|      3. Return the sum in the event result.                                  |
|                                                                              |
|    For WM_FORMATFRAME:                                                       |
|      Dispatch the format() function.                                         |
|                                                                              |
|    For WM_CLOSE:                                                             |
|      Dispatch the closed() function.                                         |
|                                                                              |
|    For WM_ACTIVATE:                                                          |
|      Dispatch either activated() or deactivated() based on mp1.              |
|                                                                              |
|    For WM_SAVEAPPLICATION:                                                   |
|      Dispatch saved().                                                       |
|                                                                              |
|    For WM_PAINT:                                                             |
|      Dispatch draw().                                                        |
|                                                                              |
|    For WM_COMMAND:                                                           |
|      Dispatch command();                                                     |
|                                                                              |
|    For WM_DESTROY:                                                           |
|      Call frame->removeFromWindowList();                                     |
|                                                                              |
|    For WM_DRAWITEM: or WM_MEASUREITEM:                                       |
|      Return true if already sent to client.                                  |
|                                                                              |
|    For WM_CONTROLPOINTER:                                                    |
|      Return the handle of the mouse pointer to use in the event result.      |
|                                                                              |
|    For WM_HELP:                                                              |
|      Display help for dynamically-assigned help IDs.                         |
|                                                                              |
|    This handler returns false for all other events.                          |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::dispatchHandlerEvent ( IEvent &event )
  {
  IMODTRACE_ALL("IFrameHandler::dispatchHandlerEvent");
  Boolean
    result = false;

  switch( event.eventId() )
    {
    case WM_CALCFRAMERECT:
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_CALCFRAMERECT" );

      IFrameEvent
        frameEvent( event );

      // Dispatch virtual function...
      result = this -> calcRect( frameEvent );

      // If handled, copy result to input event structure...
      if ( result )
        {
        event.setResult( frameEvent.result() );
        }
      }
      break;


    case WM_QUERYFRAMECTLCOUNT:
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_QUERYFRAMECTLCOUNT" );

      // Call default procedure to get number of standard controls...
      unsigned long
        stdControls = this -> defaultProcedure( event );

      // Compute additional controls count...
      unsigned long
        extensions = this -> numberOfExtensions(
                                         (IFrameWindow*)event.window() );

      // Result is the sum...
      event.setResult( stdControls + extensions );

      result = true;
      }
      break;


    case WM_FORMATFRAME:
      {
      IMODTRACE_DEVELOP( "IFrameHandler WM_FORMATFRAME" );

      IFrameFormatEvent
        fmtEvent( event );

      // Dispatch virtual function...
      result = this -> format( fmtEvent );

      // If handled, copy result to input event structure...
      if ( result )
        event.setResult( fmtEvent.result() );
      }
      break;

    case WM_SYSCOMMAND:
    case WM_CLOSE:
    {
      Boolean processClose = false;    // Don't treat most syscmds as close.

      if ( event.eventId() == WM_CLOSE )
      {       // Always close window on WM_CLOSE.
         processClose = true;
      }
      else if ( event.parameter1().number1() == SC_CLOSE )
      {       // Close system command (e.g. from system menu or accelerator).
         processClose = true;   // Almost always close the window.

         // Check if the "Close" system menu choice is disabled.
         // If so, ignore the close request.
         IFrameWindow
          *pFrame = (IFrameWindow*)( event.window() );
         IWindowHandle sysMenuHandle =
                         pFrame->handleFor( IFrameWindow::systemMenu );
         if ( sysMenuHandle )
         {         // Frame has a system menu.
            Boolean needToDelete = false;
            ISystemMenu* sysMenu =
              (ISystemMenu*)IWindow::windowWithHandle( sysMenuHandle );
            if ( !sysMenu )
            {
               sysMenu = new ISystemMenu( pFrame );
               needToDelete = true;
            }

            if ( !sysMenu->isItemEnabled( ISystemMenu::idClose ))
            {      // "Close" choice is disabled, ignore the request.
               processClose = false;
            }

            if ( needToDelete )
            {      // Memory clean-up.
               delete sysMenu;
            }
         }
      }

      if ( processClose )
        {
        IFrameEvent
          frameEvent( event );

        frameEvent.frame()->fFrameWindowData->flags |=
                                        IFrameWindowData::hasBeenClosed;

        result = this -> closed( frameEvent );

        if ( result )
          event.setResult( frameEvent.result() );
        }
      break;
    }

    case WM_ACTIVATE:
      {
      IFrameEvent
        frameEvent( event );


      if ( (unsigned long)frameEvent.parameter1() )
        {
        ITRACE_ALL(IString("ACTIVATE frame=")+
                   IString( (unsigned long)(void*)frameEvent.frame()).d2x() );
        frameEvent.frame() -> update();
        result = this -> activated( frameEvent );
        }
      else
        result = this -> deactivated( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;
    case WM_SAVEAPPLICATION:
      {
      IFrameEvent
        frameEvent( event );

      result = this -> saved( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;

    case WM_PAINT:
      {
      IFrameEvent
        frameEvent( event );

      result = this -> draw( frameEvent );

      if ( result )
        event.setResult( frameEvent.result() );
      }
      break;

    case WM_COMMAND:
      {
      ICommandEvent
        cmdEvent( event );

      result = this -> command( cmdEvent );

      if ( result )
        event.setResult( cmdEvent.result() );
      }
      break;

    case WM_DESTROY:
      {
      IFrameWindow
       *pFrame = (IFrameWindow*)( event.window() );
      // If frame wasn't "closed", do so now.
      if ( !( pFrame->fFrameWindowData->flags & IFrameWindowData::hasBeenClosed ) )
        {
        pFrame -> setDestroyOnClose( false ); // Prevent recursion.
        IFrameEvent
          frameEvent( event );
        this -> closed( frameEvent );
        }
      pFrame -> removeFromWindowList();

      //-----------------------------------------------------------------
      // Without special processing, the combination of                   
      // IFrameWindow::showModally and IWindow::setAutoDeleteObject       
      // will cause a trap, during manipulation of the deleted frame      
      // object at the end of showModally.  This code causes the          
      // auto-delete processing in IWindow to be bypassed, so the object 
      // deletion can be done by IFrameWindow instead.  By delaying the   
      // setting of the private data flag until now, there should be no   
      // out-of-synch conditions between the flag and the auto-delete     
      // setting of the frame object (for example, caused by handlers 
      // attached to the frame object).
      //-----------------------------------------------------------------
      if ( pFrame->fFrameWindowData->flags & IFrameWindowData::modal  &&  
           pFrame->isAutoDeleteObject() )                                 
        {                                                                 
         pFrame->fFrameWindowData->flags |= IFrameWindowData::modalAutoDelete; 
         pFrame->setAutoDeleteObject( false );                            
        }                                                                 
      }
      break;

    case WM_DRAWITEM:
    case WM_MEASUREITEM:
      result = ( (unsigned long)event.parameter1() == FID_CLIENT );
      break;

    case WM_CONTROLPOINTER:
      {
      IFrameWindow
       *pFrame = (IFrameWindow*)event.dispatchingWindow();

      if (pFrame->mousePointer())
      {
        event.setResult((unsigned long)pFrame->mousePointer());
      }
      else
      {
        event.setResult((unsigned long)event.parameter2());
      }
      result = true;
      }
      break;

    case WM_WINDOWPOSCHANGED:
      // If the frame is becoming active and this is a secondary window,
      // be sure that the owner window not minimized.  This is needed to
      // work around the PM behavior for primary and secondary windows
      // being restored from the task list.
      if ((event.window() != 0) &&
         (((unsigned long) event.parameter2()) & AWP_ACTIVATE))
      {
         IWindow* pOwner = event.window()->owner();
         if (pOwner && pOwner->isFrameWindow() &&
            ((IFrameWindow*)pOwner)->isMinimized())
         {
            ((IFrameWindow*)pOwner)->restore();
         }
      }
      break;

    case IC_UM_DESTROY_FRAME: 
       if ( !IDESTROYWINDOW( event.handle() ) )
       {
          ITHROWGUIERROR("IDESTROYWINDOW");
       }
       result = true;
       break;

    //---------------------------------------------------------------
    // Watch for private class library message to disable the
    // automatic update of frame extensions when the BIDI attributes
    // for the window change or for command to update extensions
    //---------------------------------------------------------------
    case IC_UM_BIDI_CHANGE:
       {
         IFrameWindow *pFrame = (IFrameWindow*)( event.window() );
         if ( event.parameter1().asUnsignedLong() == 0 )
         {
           // Set flag to disable frame extension update
           pFrame->fFrameWindowData->disableBidiChildUpdate = true;
         }
         else
         {
           // Frame extension should be updated
           IBidiSettings bidi(*pFrame);
           IWindow::ChildCursor cwCursor( *pFrame );
           for ( cwCursor.setToFirst(); cwCursor.isValid();
                 cwCursor.setToNext() )
           {
             IWindowHandle whChild = pFrame->childAt( cwCursor );
             IWindow* pwinChild = IWindow::windowWithHandle( whChild );
             if ( pwinChild && ( pwinChild->id() != FID_CLIENT ))
             {
               bidi.apply(*pwinChild);
             }
           }
         }
       }
       break;

    //---------------------------------------------------------------
    // Look for changes to the BIDI attributes by watching for the
    // WM_SETBIDIATTR message.  This message is not in the OS/2 2.1
    // toolkit, so it is hardcoded (0x0BD0) here.
    //---------------------------------------------------------------
    case 0x0BD0:
       {
         if (event.window()->isFrameWindow())
         {
           IFrameWindow *pFrame = (IFrameWindow*)( event.window() );
           if ( pFrame->fFrameWindowData->disableBidiChildUpdate )
           {
             // Frame extensions should not be updated
             pFrame->fFrameWindowData->disableBidiChildUpdate = false;
           }
           else
           {
             // Frame extension should be updated
             pFrame->postEvent(IC_UM_BIDI_CHANGE,true);
           }
         }
       }
       break;

    case WM_HELP:
    {
       ITRACE_DEVELOP( "WM_HELP" );
       // Search the parent window chain beginning from the focus window for
       // a window with a dynamically-assigned help panel ID.
       unsigned long
         helpPanelId = 0;
       IWindowHandle
         windowToCheck( IQUERYFOCUS( IWindow::desktopWindow()->handle() ) ),
         frameWindow( event.dispatchingWindow()->handle() );
       while ( ! helpPanelId  &&  windowToCheck )
       {
          IWindow
           *window = IWindow::windowWithHandle( windowToCheck );
          if ( window  &&  window->helpId() )
          {        // Found a help panel to display.
             helpPanelId = window->helpId();
          }
          else if ( windowToCheck == frameWindow )
          {        // Exhausted the parent window chain.
             break;
          }
          else
          {
             windowToCheck = IPARENTOF( windowToCheck );
          }
       }

       if ( helpPanelId )
       {  // Have contextual help to display.
     
          // Let other handlers see the WM_HELP, since they may have dependencies
          // on it (like calling IHelpWindow::setActiveWindow).
          result = event.dispatchingWindow()->dispatchRemainingHandlers( event, false );
          if ( ! result )
          {  // Help request not handled yet.
             IHelpWindow
              *helpWindow = IHelpWindow::helpWindow( event.dispatchingWindow() );
             helpWindow->show( IResourceId( helpPanelId ) );
          }
          result = true;
       }
       // Else allow help table processing occur.

       break ;
    }

    default:
      break;
    }
  return result;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::numberOfExtensions                                            |
|                                                                              |
| Return the number of frame extensions that the argument frame possesses.     |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::format                                                        |
|                                                                              |
| Fills an array of window position structures, one for each frame control.    |
| This function does this for the frame extensions that have been added to the |
| frame using IFrameWindow::addExtension.                                      |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::closed                                                        |
|                                                                              |
| Called when the frame window is closed.                                      |
| Notes:  If the frame is modal, dismiss it (presuming DID_CANCEL).            |
|         If there is a client, send it a WM_SYSCOMMAND/SC_CLOSE so that a     |
|           command handler attached to the client window can process the      |
|           close (but first turn on a special bit indicating that the event   |
|           is dispatched from the frame, so a canvas client will not route    |
|           the command event back to the frame).                              |
|         If the frame indicates it should destroy the window on close, then   |
|           WinDestroyWindow.                                                  |
|         Return true (so default window procedure never gets called)          |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::closed ( IFrameEvent &frameEvent )
  {
  IMODTRACE_DEVELOP( "IFrameHandler::closed" );

  if ( frameEvent.frame()->isModal() )
    frameEvent.frame() -> dismiss( DID_CANCEL );

  IWindowHandle
    client =  frameEvent.frame() -> clientHandle();

  /********************************************/
  /* Notify that the frame window is closing. */
  /********************************************/
  frameEvent.window()->notifyObservers(INotificationEvent(
                                       IFrameWindow::closeId,
                                       *(frameEvent.window()), false));

  if ( client )
    {                          // Client area to route events back to
    unsigned long ul2 = (frameEvent.eventId() == WM_SYSCOMMAND) ?
                           (unsigned long)frameEvent.parameter2() :
                           CMDSRC_OTHER;
    // Route event with special flag in mp2 on to client
    client.sendEvent( WM_SYSCOMMAND,
                      SC_CLOSE,
                      IEventParameter2( (unsigned short)(ul2 | 0x8000),
                                        false ) );
    }

  if ( frameEvent.frame()->willDestroyOnClose() )
    {
       frameEvent.frame()->postEvent(IC_UM_DESTROY_FRAME);
    }
  return true;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::activated                                                     |
|                                                                              |
| Called when the frame window is activated.                                   |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::activated ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::deactivated                                                   |
|                                                                              |
| Called when the frame is asked to give up the input focus.                   |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::deactivated ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::saved                                                         |
|                                                                              |
| Called when the application statis is to be saved.                           |
| Notes:  The default implementation simply returns false.                     |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::saved ( IFrameEvent &frameEvent )
  {
  return false;
  }

/*------------------------------------------------------------------------------
| IFrameHandler::draw                                                          |
|                                                                              |
| Called when the frame window needs to be redrawn.  This function draws the   |
| separator lines between frame extensions and returns false.                  |
| Notes: Iterate extensions, drawing each via IFrameExtension::drawSeparator.  |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::command                                                       |
|                                                                              |
| Called when the frame window receives a command event.  This function checks |
| to see if the frame has a client.  If so, it returns false (so the client    |
| can receive the event).  Otherwise, it returns true.                         |
| Notes:  If there's a client, send it the WM_COMMAND with a special bit       |
|           turned on indicating that the event is dispatched from the frame,  |
|           so that a canvas client area won't route it back to the frame.     |
|         Always return true so PM's frame window procedure won't try to route |
|           it again to the client  (without the bit turn on).                 |
------------------------------------------------------------------------------*/
Boolean IFrameHandler::command ( ICommandEvent &cmdEvent )
  {
  IWindowHandle
    client = IWINDOWFROMID( cmdEvent.handle(), FID_CLIENT );
  if ( client )
    {           // Tinker with WM_COMMAND before sending it to client
    unsigned long ul2 = (unsigned long)cmdEvent.parameter2();

    IEventResult evtres =
      client.sendEvent( WM_COMMAND,
                        cmdEvent.parameter1(),
                        IEventParameter2( (unsigned short)(ul2 | 0x8000),
                                          false ) );
                                  // Turn on special flag in mp2

    cmdEvent.setResult(evtres);
    }
  return true;   // Don't let WC_FRAME winproc also forward to client
  }

/*------------------------------------------------------------------------------
| IFrameHandler::calcRect                                                      |
|                                                                              |
| Iterate extensions, expand frame rectangle by height and width of            |
| each extension if given a client rectangle, shrink client rectangle          |
| by height and width of each extension if given a frame rectangle.            |
------------------------------------------------------------------------------*/
// Moved to iframe.cpp

/*------------------------------------------------------------------------------
| IFrameHandler::IFrameHandler                                                 |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IFrameHandler::IFrameHandler ( )
  {
  }

/*------------------------------------------------------------------------------
| IFrameHandler::~IFrameHandler                                                |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IFrameHandler::~IFrameHandler ( )
  {
  }
