/*******************************************************************************
* FILE NAME: igbitmap.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igbitmap.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
  {
  #define INCL_DEV                /* Device Function definitions  */
  #define INCL_GPICONTROL         /* GPI control Functions        */
  #define INCL_GPITRANSFORMS      /* GPI transform functions      */
  #define INCL_GPICORRELATION
  #define INCL_GPIBITMAPS
  #define INCL_GPIPRIMITIVES
  #define INCL_GPILOGCOLORTABLE
  #define INCL_WINPOINTERS
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  #include <math.h>
  #include <igbmodel.hpp>
  #undef word
  }

#include <igbitmap.hpp>
#include <icoordsy.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <iexcept.hpp>
#include <iresstat.hpp>
#include <ireslib.hpp>
#include <ithread.hpp>
#include <itrace.hpp>
#include <igrafctx.hpp>
#include <itrnsfrm.hpp>
#include <icconst.h>
#include <ibmpdata.hpp>
#include <igbundle.hpp>
#include <iwindow.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGBITMAP_CPP_
  #include <ipagetun.h>
#endif

/***************************************************************/
/* raster opetarors.                                           */
/***************************************************************/
#pragma data_seg(ICLStaticConst)
const long IGBitmap::black            (ROP_ZERO           );
const long IGBitmap::notSourceErase   (ROP_NOTSRCERASE    );
const long IGBitmap::halftone         (0x00FC             );
const long IGBitmap::invert           (ROP_NOTSRCCOPY     );
const long IGBitmap::sourceErase      (ROP_NOTSRCERASE    );
const long IGBitmap::destInvert       (ROP_DSTINVERT      );
const long IGBitmap::patternInvert    (ROP_PATINVERT      );
const long IGBitmap::sourceInvert     (ROP_SRCINVERT      );
const long IGBitmap::sourceAnd        (ROP_SRCAND         );
const long IGBitmap::mergePaint       (ROP_MERGEPAINT     );
const long IGBitmap::mergeCopy        (ROP_MERGECOPY      );
const long IGBitmap::normal           (ROP_SRCCOPY        );
const long IGBitmap::sourcePaint      (ROP_SRCPAINT       );
const long IGBitmap::patternCopy      (ROP_PATCOPY        );
const long IGBitmap::patternPaint     (ROP_PATPAINT       );
const long IGBitmap::white            (ROP_ONE            );
#pragma data_seg()

/*------------------------------------------------------------------------------
| This class wrappers an object pointer and destroys the pointed               |
| to object when the wrapper is destructed.                                    |
------------------------------------------------------------------------------*/
static struct IGBitmapStaticData {
IGBitmapStaticData ( )
{
  gc = 0;
}
~IGBitmapStaticData ( )
{
  if (gc)
  {
     if (gcBitmap)
     {
        GpiSetBitmap(gc->handle(), 0);
        GpiDeleteBitmap(gcBitmap);
     }
     delete gc;
     gc = 0;
  }
}
IGraphicContext
  *gc;
ISize
   gcSize;
IBitmapHandle
   gcBitmap;
} bitmapStaticData;

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
| Constructor from an bitmapHandle.                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IBitmapHandle& bitmapHandle )
  : fBitmap(bitmapHandle), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
| Constructor from a bitmap resource ID.                                       |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( unsigned long bitmapId )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
  IBitmapData *pBmpData(0);
  IResourceId resourceId(bitmapId);  //create a object wrapper for resouce ID.

  // Check if the bitmap has already been loaded in the resource lib.
  // returns a bitmap handle if it is already loaded.
  // bitmapSet() is a class member function of private class IResLibStaticPtr
  // which tracks all loaded bitmaps.
  pBmpData = IResLibStaticPtr::bitmapSet().find( bitmapId,
                                                 resourceId.resourceLibrary().fileName(),
                                                 ISize(0,0) );
  if ( pBmpData )  //bitmap has already been loaded
  {
    fBitmap = pBmpData->handle();   //assign bitmap handle as data member.
  }
  else //load bitmap.
  {
    unsigned long bmp;
    //load bitmap into resource library.
    bmp = loadBitmap( resourceId,
                      resourceId.resourceLibrary().handle(), ISize(0,0) );
    //create a new IBitmapData object.
    pBmpData = new IBitmapData (bitmapId, bmp, IString());
    // register bmp with IResLibStaticPtr
    IResLibStaticPtr::bitmapSet().add(pBmpData);
    fBitmap = bmp;   //assign bitmap handle as data member.
  }
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
| Constructor using an IResourceId object.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IResourceId& resourceId )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
  IBitmapData *pBmpData(0);

  // Check if the bitmap has already been loaded from file into the resource lib.
  // returns a bitmap handle if it is already loaded.
  // bitmapSet() is a class member function of private class IResLibStaticPtr
  // which tracks all loaded bitmaps.
  pBmpData = IResLibStaticPtr::bitmapSet().find( resourceId,
                                                 resourceId.resourceLibrary().fileName(),
                                                 ISize(0,0) );
  if ( pBmpData )   //bitmap has already been loaded
  {
    fBitmap = pBmpData->handle();  //assign bitmap handle as data member.
  }
  else
  {
    unsigned long bmp;
    //load bitmap into resource library.
    bmp = loadBitmap( resourceId,
                      resourceId.resourceLibrary().handle(), ISize(0,0) );
    //create a new IBitmapData object.
    pBmpData = new IBitmapData (resourceId, bmp, resourceId.resourceLibrary().fileName());
    // register bmp with IResLibStaticPtr
    IResLibStaticPtr::bitmapSet().add(pBmpData);
    fBitmap = bmp;  //assign bitmap handle as data member.
  }
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
| Constructor using an imageFileName.                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IString& ImageFilename )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
  loadFromFile( ImageFilename, UNKNOWN_IMAGETYPE );
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
| Constructor using an imageFileName with imageFormat.                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IString& ImageFilename, ImageFormat imageFormat )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(imageFormat), fMod(0)
{
  loadFromFile( ImageFilename, ft(imageFormat) );
}

/*------------------------------------------------------------------------------
| IGBitmap::loadFromFile                                                       |
| Load a bitmap from an imageFileName with imageFormat.                        |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::loadFromFile( const IString& filename, int ft )
{
  MOD_ERR me;
  long fgColor;
  long bgColor;
  unsigned long bmp;

  //create a new MOD.
  fMod = new MOD;
  me   = ModCreateFromFile(filename, &ft, (char*)"", fMod);

  // map from gbm filetype to IGBitmap enum
  switch (ft)
  {
    case 0:
      fIf = Bitmap;
      break;
    case 1:
      fIf = GIF;
      break;
    case 2:
      fIf = PCX;
      break;
    case 3:
      fIf = TIFF;
      break;
    case 4:
      fIf = Targa;
      break;
    case 5:
      fIf = Amiga;
      break;
    case 11:
      fIf = XBM;
      break;
    case 12:
      fIf = PSEG;
      break;
  } /* endswitch */

  if (me != MOD_ERR_OK)
  {
    ITHROWLIBRARYERROR(ModError(me),
                       IErrorInfo::accessError,
                       IException::recoverable);
  }

  if (!ModMapVisual( IThread::current().anchorBlock(), fMod, 0,
                     (HBITMAP*)&bmp, &bgColor, &fgColor ))
    ITHROWGUIERROR("Error mapping visual.");

  IBitmapData *pBmpData(new IBitmapData (999, bmp, filename ));
  IResLibStaticPtr::bitmapSet().add(pBmpData);
  fBitmap = bmp;

  IGraphicBundle currentBundle(this->graphicBundle());
  currentBundle.setPenColor( IColor((unsigned char)(fgColor >> 16),
                                    (unsigned char)(fgColor >> 8),
                                    (unsigned char)fgColor ));
  currentBundle.setBackgroundColor( IColor((unsigned char)(bgColor >> 16),
                                           (unsigned char)(bgColor >> 8),
                                           (unsigned char)bgColor ));
  this->setGraphicBundle( currentBundle );


  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::writeToFile                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::writeToFile( const IString& imageFilename,
                                 ImageFormat imageFormat )
{
  MOD_ERR me;
  Boolean fDelete(false);

  if (!fMod)
  {
    fDelete = true;
    fMod = new MOD;
    me = ModCreateFromHBitmap( fMod, IThread::current().anchorBlock(), fBitmap);
    if (me != MOD_ERR_OK)
    {
      ITHROWLIBRARYERROR(ModError(me),
                         IErrorInfo::accessError,
                         IException::recoverable);
    } /* endif */
  } /* endif */


  me = ModWriteToFile( fMod, ft(imageFormat), imageFilename, (char*)"");
  if (me != MOD_ERR_OK)
  {
    ITHROWLIBRARYERROR(ModError(me),
                       IErrorInfo::accessError,
                       IException::recoverable);
  } /* endif */

  if (fDelete)
  {
    ModDelete(fMod);
    delete fMod;
    fMod = 0;
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::imageFormat                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::ImageFormat IGBitmap::imageFormat( ) const
{
  return fIf;
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IGraphicContext& gc,
                    const IRectangle& windowRectangle )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
  IGraphicContext    memoryGC;
  BITMAPINFOHEADER2  bmih2;
  PBYTE              pBits;
  unsigned long      bmp;
  long               fmts[2];

  memset ( &bmih2, 0, sizeof(BITMAPINFOHEADER2) );
  GpiQueryDeviceBitmapFormats(memoryGC.handle(), 2, fmts);

  bmih2.cbFix = sizeof(BITMAPINFOHEADER2);
  bmih2.cx = windowRectangle.width();
  bmih2.cy = windowRectangle.height();
  bmih2.cPlanes   = (USHORT)fmts[0];
  bmih2.cBitCount = (USHORT)fmts[1];

  bmp = GpiCreateBitmap ( memoryGC.handle(),
                          &bmih2,
                          0,
                          0,
                          0 );
  if (!bmp)
    ITHROWGUIERROR("GpiCreateBitmap");

  GpiSetBitmap(memoryGC.handle(), bmp);

  POINTL aptl[3] = { 0,0, bmih2.cx, bmih2.cy,
                     windowRectangle.left(), windowRectangle.bottom()};

  if( GpiBitBlt( memoryGC.handle(),
                 gc.handle(),
                 3L, aptl, ROP_SRCCOPY, BBO_IGNORE) == GPI_ERROR)
    ITHROWGUIERROR("GpiBitBlt");

  GpiSetBitmap(memoryGC.handle(), 0);

  IBitmapData *pBmpData(new IBitmapData (999, bmp, IString()));
  IResLibStaticPtr::bitmapSet().add(pBmpData);
  fBitmap = bmp;
}

/*------------------------------------------------------------------------------
| IGBitmap::IGBitmap                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::IGBitmap( const IGBitmap& bitmap, const IRectangle& rectangle,
                    long rasterOperation )
  : fBitmap(0), fMask(0), fTransparent(0),
    fVo(IGBitmap::raw), fIf(IGBitmap::Bitmap), fMod(0)
{
  BITMAPINFOHEADER2 bmpInfo;
  IGraphicContext   memoryGC;
  unsigned long     bmp;

  ISize bitmapSize(bitmap.size());
  IASSERT((rectangle.top() <= bitmapSize.height() &&
           rectangle.right() <= bitmapSize.width() &&
           rectangle.bottom() >= 0 && rectangle.left() >= 0));

  bmpInfo.cbFix = sizeof(BITMAPINFOHEADER2);
  if (!GpiQueryBitmapInfoHeader(bitmap.handle(), &bmpInfo))
    ITHROWGUIERROR("GpiQueryBitmapInfoHeader");

  bmpInfo.cx = rectangle.width();
  bmpInfo.cy = rectangle.height();

  bmp = GpiCreateBitmap( memoryGC.handle(), &bmpInfo, 0, 0, 0);

  if ( !bmp )
    ITHROWGUIERROR("GpiCreateBitmap");

  GpiSetBitmap(memoryGC.handle(), bmp);

  POINTL aptl[4] = { 0,0, bmpInfo.cx-1, bmpInfo.cy-1,
                     rectangle.left(), rectangle.bottom(),
                     rectangle.right(), rectangle.top() };

  if (GpiWCBitBlt( memoryGC.handle(),
                   bitmap.handle(),
                   4L,
                   aptl,
                   rasterOperation,
                   BBO_IGNORE) == GPI_ERROR)
    ITHROWGUIERROR("GpiWCBitBlt");

  GpiSetBitmap(memoryGC.handle(), 0);

  IBitmapData *pBmpData(new IBitmapData (999, bmp, IString()));
  IResLibStaticPtr::bitmapSet().add(pBmpData);
  fBitmap = bmp;
}

/*------------------------------------------------------------------------------
| IGBitmap::~IGBitmap                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::~IGBitmap()
{
  if (fMask)
    GpiDeleteBitmap(fMask);

  if (fTransparent)
    delete fTransparent;

  if (fMod)
  {
    ModDelete(fMod);
    delete fMod;
  }
}

/*------------------------------------------------------------------------------
| IGBitmap::handle                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBitmapHandle IGBitmap::handle( ) const
{
  return IBitmapHandle( fBitmap );
}

/*------------------------------------------------------------------------------
| IGBitmap::handle                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointerHandle IGBitmap::asPointerHandle( ) const
{
  IGraphicContext    memoryGC;
  BITMAPINFOHEADER2  bmih2;
  PBYTE              pBits;
  ISize              bitmapSize(this->size());
  POINTERINFO        pointerInfo;
  HBITMAP            maskBitmap;
  HPOINTER           hptr;

  memset ( &bmih2, 0, sizeof(BITMAPINFOHEADER2) );
  bmih2.cbFix = sizeof(BITMAPINFOHEADER2);
  bmih2.cx = bitmapSize.width();
  bmih2.cy = bitmapSize.height() * 2;
  bmih2.cPlanes = 1;
  bmih2.cBitCount = 8;
  bmih2.cbImage = bmih2.cx * bmih2.cy;
  bmih2.ulColorEncoding = BCE_RGB;
  pBits = new BYTE[(bmih2.cx+1)*(bmih2.cy+1)];
  memset( pBits, 0, (bmih2.cx+1)*(bmih2.cy+1));

  maskBitmap = GpiCreateBitmap ( memoryGC.handle(),
                                 &bmih2,
                                 CBM_INIT,
                                 pBits,
                                 (PBITMAPINFO2) &bmih2 );
  if (!maskBitmap)
    ITHROWGUIERROR("GpiCreateBitmap");

  pointerInfo.fPointer   = true;
  pointerInfo.xHotspot   = 0;
  pointerInfo.yHotspot   = 0;
  pointerInfo.hbmColor   = this->fBitmap;
  pointerInfo.hbmPointer = maskBitmap;

  hptr = WinCreatePointerIndirect( IWindow::desktopWindow()->handle(),
                                   &pointerInfo );
  if (!hptr)
    ITHROWGUIERROR("WinCreatePointerIndirect");

  GpiDeleteBitmap( maskBitmap );

  delete [] pBits;

  return IPointerHandle( hptr );

}

/*------------------------------------------------------------------------------
| IGBitmap::size                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ISize IGBitmap::size( ) const
{
  BITMAPINFOHEADER pBmpInfo;

  pBmpInfo.cbFix=sizeof(BITMAPINFOHEADER);

  if (!GpiQueryBitmapParameters(fBitmap, &pBmpInfo))
  {
    ITHROWGUIERROR("GpiQueryBitmapParameters");
  }
  return ISize( pBmpInfo.cx, pBmpInfo.cy );
}

/*------------------------------------------------------------------------------
| IGBitmap::drawOn                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::drawOn( IGraphicContext& gc )
{
  drawOn( gc, IGBitmap::normal );
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::drawOn                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::drawOn( IGraphicContext& gc, long rasterOperation )
{
  POINTL aPtl[4];
  ISize  sizeBitmap(size());
  IPoint newPoint(fPos);

  if ( ICoordinateSystem::isConversionNeeded() &&
       (gc.recoordinationHeight() != 0) )
  {
    newPoint.setY( gc.recoordinationHeight() - newPoint.y() - sizeBitmap.height() );
  }

  aPtl[0] = newPoint.asPOINTL();
  aPtl[2].x = 0;
  aPtl[2].y = 0;
  aPtl[3].x = sizeBitmap.width();
  aPtl[3].y = sizeBitmap.height();
  aPtl[1].x = aPtl[0].x + aPtl[3].x - 1;
  aPtl[1].y = aPtl[0].y + aPtl[3].y - 1;

  return draw( gc, rasterOperation, (void*)aPtl, BBO_IGNORE );
}

/*------------------------------------------------------------------------------
| IGBitmap::drawOn                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::drawOn( IGraphicContext& gc,
                          const IPoint& targetBottomLeft,
                          const IPoint& targetTopRight,
                          const IPoint& sourceBottomLeft,
                          const IPoint& sourceTopRight,
                          long          rasterOperation,
                          CompressMode  compressMode )
{
  POINTL aPtl[4];
  ULONG  ulCompress;

  switch ( compressMode )
  {
    case and:
      ulCompress = BBO_AND;
      break;
    case or:
      ulCompress = BBO_OR;
      break;
    case ignore:
      ulCompress = BBO_IGNORE;
      break;
  } /* endswitch */

  aPtl[0] = targetBottomLeft.asPOINTL();
  aPtl[1] = targetTopRight.asPOINTL();
  aPtl[2] = sourceBottomLeft.asPOINTL();
  aPtl[3] = sourceTopRight.asPOINTL();

  if ( ICoordinateSystem::isConversionNeeded() &&
       (gc.recoordinationHeight() != 0) )
  {
    for (int i=0;i<3;i++)
    {
      aPtl[i].y = gc.recoordinationHeight() - aPtl[i].y;
    }
  }
  return draw( gc, rasterOperation, (void*)aPtl, ulCompress );
}

/*------------------------------------------------------------------------------
| IGBitmap::sizeTo                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::sizeTo( const ISize& newSize )
{
  if (newSize != this->size())
  {
    Boolean fDelete(false);
    unsigned long bmp;
    MOD modNew;
    MOD_ERR me;
    long bgColor,
         fgColor;

    if (!fMod)
    {
      fDelete = true;
      fMod = new MOD;
      me = ModCreateFromHBitmap( fMod, IThread::current().anchorBlock(), fBitmap);
      if (me != MOD_ERR_OK)
      {
        ITHROWLIBRARYERROR(ModError(me),
                           IErrorInfo::accessError,
                           IException::recoverable);
      }
    }

    if (fMask)
    {
      GpiDeleteBitmap(fMask);
      fMask = 0;
    }

    me = ModResize(fMod, newSize.width(), newSize.height(), &modNew);
    if (me != MOD_ERR_OK)
    {
      ITHROWLIBRARYERROR(ModError(me),
                         IErrorInfo::accessError,
                         IException::recoverable);
    } /* endif */

    if (!ModMapVisual( IThread::current().anchorBlock(), &modNew, (int)fVo,
                       (HBITMAP*)&bmp, &bgColor, &fgColor ))
    {
      ITHROWGUIERROR("Error mapping visual.");
    }

    // put fbitmap into bitmap set.
    IBitmapData *pBmpData(new IBitmapData (999, bmp, IString()));
    IResLibStaticPtr::bitmapSet().add(pBmpData);
    fBitmap = bmp;


    if (fDelete)
    {
      ModDelete(&modNew);
      ModDelete(fMod);
      delete fMod;
      fMod = 0;
    }
    else
    {
      ModDelete(fMod);
      ModMove(&modNew, fMod);
    }
  } /* endif */
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::reflectHorizontally                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::reflectHorizontally( )
{
  modReflect(ModReflectHorz);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::reflectVertically                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::reflectVertically( )
{
  modReflect(ModReflectVert);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::transposeXForY                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::transposeXForY( )
{
  modReflect(ModTranspose);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::rotateBy90                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::rotateBy90( )
{
  modReflect(ModRotate90);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::rotateBy180                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::rotateBy180( )
{
  modReflect(ModRotate180);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::rotateBy270                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::rotateBy270( )
{
  modReflect(ModRotate270);
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::setViewOption                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::setViewOption( ViewOption viewOption )
{
  if (fVo != viewOption)
  {
    MOD_ERR me;
    long fgColor;
    long bgColor;
    unsigned long bmp;

    if (!fMod)
    {
      fMod = new MOD;
      me = ModCreateFromHBitmap( fMod, IThread::current().anchorBlock(), fBitmap);
      if (me != MOD_ERR_OK)
      {
        ITHROWLIBRARYERROR(ModError(me),
                           IErrorInfo::accessError,
                           IException::recoverable);
      } /* endif */
    }

    if (fMask)
    {
      GpiDeleteBitmap(fMask);
      fMask = 0;
    }

    if (!ModMapVisual( IThread::current().anchorBlock(), fMod, (int)viewOption,
                       (HBITMAP*)&bmp, &bgColor, &fgColor ))
      ITHROWGUIERROR("Error mapping visual.");

    // put fbitmap into bitmap set.
    IBitmapData *pBmpData(new IBitmapData (999, bmp, IString()));
    IResLibStaticPtr::bitmapSet().add(pBmpData);
    fBitmap = bmp;

    fVo = viewOption;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::viewOption                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap::ViewOption IGBitmap::viewOption( ) const
{
  return fVo;
}

/*------------------------------------------------------------------------------
| IGBitmap::moveTo                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::moveTo ( const IPoint& point )
{
  fPos = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::position                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGBitmap::position ( ) const
{
  return fPos;
}

/*------------------------------------------------------------------------------
| IGBitmap::copy                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBitmapHandle IGBitmap::copy(const IBitmapHandle& bitmapHandle )
{
  IGraphicContext gc;
  return IBitmapHandle(copyBitmap(gc, bitmapHandle, ISize(0,0)));
}


/*------------------------------------------------------------------------------
| IGBitmap::setTransparentColor                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::setTransparentColor(const IColor& aColor)
{
  if (fTransparent)
  {
    delete fTransparent;
    fTransparent = 0;
  }
  fTransparent = new IColor(aColor);

  // If there is an old mask bitmap, delete it

  if (fMask)
  {
    GpiDeleteBitmap(fMask);
    fMask = 0;
  }

  // Go ahead and create the mask now (so we'll have it later)

  createMask();

  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::resetTransparentColor                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::resetTransparentColor()
{
  if (fTransparent)
  {
    delete fTransparent;
    fTransparent = 0;
    if (fMask)
    {
      GpiDeleteBitmap(fMask);
      fMask = 0;
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::transparentColor                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGBitmap::transparentColor() const
{
  if (fTransparent)
    return *fTransparent;
  else
  {
    // Need to talk about this and decide if we should throw an exception
    // or just return a color with index of 0
    return IColor(0);
  }
}

/*------------------------------------------------------------------------------
| IGBitmap::hasTransparentColor                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGBitmap::hasTransparentColor() const
{
  if (fTransparent)
    return true;
  else
    return false;
}

/*------------------------------------------------------------------------------
| IGBitmap::createMask                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
void IGBitmap::createMask ()
{

  BITMAPINFOHEADER2  bmih2;
  PBYTE              pBits;
  ISize              bmpsize;
  POINTL             ptl;
  unsigned long      index;
  unsigned long      hps;
  unsigned long      colorIndex;

  bmpsize = this->size();

  IGraphicContext gc;
  hps = gc.handle();

  // Create a 1x1 bitmap using the transparent color, set the
  // bitmap into the presentation space, and then query the
  // pixel value.  We can now use this value to determine which
  // pels should be masked.

  // Use point 0,0 for testing
  POINTL aptl[4];
  aptl[0] = aptl[1] = aptl[2] = IPoint(0,0).asPOINTL();
  aptl[3].x = 1;
  aptl[3].y = 1;

  // Create a test bitmap
  pBits = new BYTE[16];
  memset ( &bmih2, 0, sizeof(BITMAPINFOHEADER2) );
  bmih2.cbFix = sizeof(BITMAPINFOHEADER2);
  bmih2.cx = 1;
  bmih2.cy = 2;
  bmih2.cPlanes = 1;
  bmih2.cBitCount = 24;
  bmih2.cbImage = 4;
  bmih2.ulColorEncoding = BCE_RGB;
  pBits[0] = fTransparent->blueMix();
  pBits[1] = fTransparent->greenMix();
  pBits[2] = fTransparent->redMix();
  IBitmapHandle testBmp = GpiCreateBitmap ( hps,
                                            &bmih2,
                                            CBM_INIT,
                                            pBits,
                                            (PBITMAPINFO2) &bmih2 );

  // Set the test bitmap into the hps

  GpiSetBitmap ( hps, testBmp );

  // Query the color value to use when generating the mask

  colorIndex = GpiQueryPel( hps, aptl );

  // Clean-up the test bitmap data

  GpiSetBitmap ( hps, 0 );
  delete [] pBits;
  GpiDeleteBitmap ( testBmp );

  // Set up header information for mask bitmap

  memset ( &bmih2, 0, sizeof(BITMAPINFOHEADER2) );
  bmih2.cbFix = sizeof(BITMAPINFOHEADER2);
  bmih2.cx = bmpsize.width();
  bmih2.cy = bmpsize.height();
  bmih2.cPlanes = 1;
  bmih2.cBitCount = 24;
  bmih2.cbImage = bmih2.cx * bmih2.cy * 4;
  bmih2.ulColorEncoding = BCE_RGB;

  // Put source bitmap in presentation space so we can query it

  GpiSetBitmap ( hps, fBitmap );

  // Scan bitmap for transparent color to generate mask data
  //
  // Note: Each pel is represented by a 3 byte RGB structure.

  pBits = new BYTE[(bmih2.cx+1)*(bmih2.cy+1)*4];
  index = 0;
  for ( ptl.y = 0; ptl.y < bmih2.cy; ptl.y++ )
  {
    for ( ptl.x = 0; ptl.x < bmih2.cx; ptl.x++ )
    {
      LONG lColor = GpiQueryPel ( hps, &ptl );
      if ( lColor == colorIndex )
      {
        pBits[index++] = 0xFF;
        pBits[index++] = 0xFF;
        pBits[index++] = 0xFF;
      }
      else
      {
        pBits[index++] = 0x00;
        pBits[index++] = 0x00;
        pBits[index++] = 0x00;
      }
    }
    // PM requires each row of data to be aligned on a 4-byte boundary
    index = ((index + 3) >> 2) << 2;
  }
  // We are adding one to the height to work around a bug in a
  // device driver.
  bmih2.cy++;

  // We are finished querying data, so remove the bitmap from the ps

  GpiSetBitmap ( hps, 0 );

  // Create the mask bitmap from our data

  fMask = GpiCreateBitmap ( hps,
                            &bmih2,
                            CBM_INIT,
                            pBits,
                            (PBITMAPINFO2) &bmih2 );
  delete [] pBits;
}


/*------------------------------------------------------------------------------
| IGBitmap::draw                                                               |
|                                                                              |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::draw( IGraphicContext& gc, long lRasterOp,
                          void *pPoints, unsigned long compMode )
{
  IGraphicBundle::BackgroundMixMode oldMix;
  IColor            oldColor(0);
  long              rc;
  ITransformMatrix* currentTransform;
  IGraphicBundle*   currentBundle;

  // Check if device supports the source transparent mix mode and
  // create the mask bitmap if it does not
  if (hasGraphicBundle())
  {
    currentBundle = new IGraphicBundle(gc);
    gc.setGraphicBundle(graphicBundle());
  }

  if ( fTransparent && !fMask )
  {
    createMask();
  }

  if (hasTransformMatrix())
  {
    currentTransform = new ITransformMatrix(gc.worldTransformMatrix());
    gc.setWorldTransformMatrix(transformMatrix(), transformMethod());
  }

  if ((lRasterOp == normal) && fMask )
  {
    BITMAPINFOHEADER infoHeader;
    infoHeader.cbFix=sizeof(BITMAPINFOHEADER);
    GpiQueryBitmapParameters(fMask,&infoHeader);
    ISize bmpSize(size());
    if (ISize(infoHeader.cx,infoHeader.cy/2) == bmpSize)
    {
      IColor penColor(gc.penColor());
      IColor backColor(gc.backgroundColor());

      PPOINTL local((POINTL*)pPoints);
      local[2].y = bmpSize.height();
      local[3].y = local[2].y + bmpSize.height();

      gc.setPenColor( IColor::white ).setBackgroundColor( IColor::black );

      rc = GpiWCBitBlt( gc.handle(),
                        fMask,
                        4L,
                        local,
                        sourceAnd,BBO_IGNORE);
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");

      local[2].y = 0L;
      local[3].y = bmpSize.height();

      gc.setPenColor( penColor ).setBackgroundColor( backColor );

      rc = GpiWCBitBlt( gc.handle(),
                        fBitmap,
                        4L,
                        (PPOINTL)local,
                        sourcePaint, BBO_IGNORE );
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");
      else if (rc == GPI_HITS)
        this->setHitSelected(true);

      gc.setPenColor( IColor::white ).setBackgroundColor( IColor::black );

      rc = GpiWCBitBlt( gc.handle(),
                        fMask,
                        4L,
                        (PPOINTL)local,
                        sourceInvert, BBO_IGNORE );

      gc.setPenColor( penColor ).setBackgroundColor( backColor );
    }
    else
    {
      // Copy the points into a new structure before calling BitBlt
      // This is needed because some display will modify the values
      POINTL  ptl[4];
      memcpy ( ptl, pPoints, 4*sizeof(POINTL) );

      // Check if the current memory ps is big enough
      ISize bmpsize = this->size();
      if ( bitmapStaticData.gc )
      {
        if (( bitmapStaticData.gcSize.width() < bmpsize.width() ) ||
            ( bitmapStaticData.gcSize.height() < bmpsize.height() ))
        {
          if ( bitmapStaticData.gcBitmap )
          {
            GpiSetBitmap(bitmapStaticData.gc->handle(), 0);
            GpiDeleteBitmap(bitmapStaticData.gcBitmap);
            bitmapStaticData.gcBitmap = 0;
          }
          delete bitmapStaticData.gc;
          bitmapStaticData.gc = 0;
        }
      }

      // Create a memory ps (if we do not have one already)
      if ( !bitmapStaticData.gc )
      {
        if ( bitmapStaticData.gcSize.width() < bmpsize.width() )
          bitmapStaticData.gcSize.setWidth(bmpsize.width());
        if ( bitmapStaticData.gcSize.height() < bmpsize.height() )
          bitmapStaticData.gcSize.setHeight(bmpsize.height());

        bitmapStaticData.gc = new IGraphicContext(bitmapStaticData.gcSize);

        // Create a new bitmap for the graphic context so that we can
        // use the GpiWCBitBlt call later without creating a new bitmap
        // each time we draw
        BITMAPINFOHEADER2  bmih2;
        memset ( &bmih2, 0, sizeof(BITMAPINFOHEADER2) );
        bmih2.cbFix = sizeof(BITMAPINFOHEADER2);
        bmih2.cx = bitmapStaticData.gcSize.width();
        bmih2.cy = bitmapStaticData.gcSize.height();
        bmih2.cPlanes = 1;
        bmih2.cBitCount = 24;
        bitmapStaticData.gcBitmap = GpiCreateBitmap
           ( bitmapStaticData.gc->handle(), &bmih2, 0L, NULL, NULL );
        if (!bitmapStaticData.gcBitmap)
          ITHROWGUIERROR("GpiCreateBitmap");
        GpiSetBitmap(bitmapStaticData.gc->handle(),
                     bitmapStaticData.gcBitmap);
      }

      // Wipe out the screen contents that will not show through
      // by ANDing the mask with the screen
      rc = GpiWCBitBlt( gc.handle(),
                        fMask,
                        4L,
                        ptl,
                        sourceAnd, BBO_IGNORE );
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");

      // Invert the mask to the memory graphic context
      memcpy ( ptl, pPoints, 4*sizeof(POINTL) );
      ptl[0].x = ptl[0].y = 0;
      ptl[1].x = ptl[3].x - ptl[2].x - 1;
      ptl[1].y = ptl[3].y - ptl[2].y - 1;
      rc = GpiWCBitBlt( bitmapStaticData.gc->handle(),
                        fMask,
                        4L,
                        ptl,
                        ROP_NOTSRCCOPY, BBO_IGNORE );
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");

      // And the bitmap to the memory graphic context
      memcpy ( ptl, pPoints, 4*sizeof(POINTL) );
      ptl[0].x = ptl[0].y = 0;
      ptl[1].x = ptl[3].x - ptl[2].x - 1;
      ptl[1].y = ptl[3].y - ptl[2].y - 1;
      rc = GpiWCBitBlt( bitmapStaticData.gc->handle(),
                        fBitmap,
                        4L,
                        ptl,
                        sourceAnd, BBO_IGNORE );
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");

      // Or the memory graphic context with the screen
      // (merge the two images)
      memcpy ( ptl, pPoints, 4*sizeof(POINTL) );
      ptl[2].x = 0;
      ptl[2].y = 0;
      ptl[3].x = ptl[1].x - ptl[0].x + 1;
      ptl[3].y = ptl[1].y - ptl[0].y + 1;
      GpiSetBitmap(bitmapStaticData.gc->handle(), 0);
      rc = GpiWCBitBlt(gc.handle(),
                     bitmapStaticData.gcBitmap,
                     4L,
                     ptl,
                     sourcePaint, BBO_IGNORE );
      GpiSetBitmap(bitmapStaticData.gc->handle(),
                     bitmapStaticData.gcBitmap);
      if (rc == GPI_ERROR )
        ITHROWGUIERROR("GpiWCBitBlt");

    }
  }
  else
  {
    rc = GpiWCBitBlt( gc.handle(),
                      fBitmap,
                      4L,
                      (PPOINTL)pPoints,
                      lRasterOp, compMode );
  }

  if (rc == GPI_ERROR )
  {
    ITHROWGUIERROR("GpiWCBitBlt");
  }
  else if ( rc == GPI_HITS )
  {
    this->setHitSelected(true);
  }
  if (hasTransformMatrix())
  {
    gc.setWorldTransformMatrix(*currentTransform, IGraphic::replace );
    delete currentTransform;
  }

  if (hasGraphicBundle())
  {
    gc.setGraphicBundle(*currentBundle);
    delete currentBundle;
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IGBitmap::ft                                                                 |
|                                                                              |
| Return the file type based on the image format.                              |
------------------------------------------------------------------------------*/
int IGBitmap::ft( const ImageFormat& imageFormat )
{
  if (imageFormat == Bitmap)
    return 0;
  if (imageFormat == GIF)
    return 1;
  if (imageFormat == PCX)
    return 2;
  if (imageFormat == TIFF)
    return 3;
  if (imageFormat == Targa)
    return 4;
  if (imageFormat == Amiga)
    return 5;
  if (imageFormat == XBM)
    return 11;
  if (imageFormat == PSEG)
    return 12;
  return 0;
}

/*------------------------------------------------------------------------------
| IGBitmap::modReflect                                                         |
|                                                                              |
| General purpose reflector function.                                          |
------------------------------------------------------------------------------*/
IGBitmap& IGBitmap::modReflect(int (* _Optlink reflector)(_MOD*, _MOD*))
{
  Boolean fDelete(false);
  MOD modNew;
  long bgColor,
       fgColor;
  unsigned long bmp;
  MOD_ERR me;

  if (!fMod)
  {
    fDelete = true;
    fMod = new MOD;
    me = ModCreateFromHBitmap( fMod, IThread::current().anchorBlock(), fBitmap);
    if (me != MOD_ERR_OK)
    {
      ITHROWLIBRARYERROR(ModError(me),
                         IErrorInfo::accessError,
                         IException::recoverable);
    } /* endif */
  }

  if (fMask)
  {
    GpiDeleteBitmap(fMask);
    fMask = 0;
  }

  me = (*reflector)(fMod, &modNew);
  if (me != MOD_ERR_OK)
  {
    ITHROWLIBRARYERROR(ModError(me),
                       IErrorInfo::accessError,
                       IException::recoverable);
  } /* endif */

  if (!ModMapVisual( IThread::current().anchorBlock(), &modNew, (int)fVo,
                     (HBITMAP*)&bmp, &bgColor, &fgColor ))
  {
    ITHROWLIBRARYERROR1(ModError(me),
                        IErrorInfo::accessError,
                        IException::recoverable,
                        "Error mapping visual." );
  }

  // put fbitmap into bitmap set.
  IBitmapData *pBmpData(new IBitmapData (999, bmp, IString()));
  IResLibStaticPtr::bitmapSet().add(pBmpData);
  fBitmap = bmp;

  if (fDelete)
  {
    ModDelete(fMod);
    ModDelete(&modNew);
    delete fMod;
    fMod = 0;
  }
  else
  {
    ModDelete( fMod );
    ModMove(&modNew, fMod);
  }
  return *this;
}
