/*******************************************************************************
* FILE NAME: igline.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igline.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <igline.hpp>
#include <igrafctx.hpp>
#include <igbundle.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGLINE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGLine::IGLine                                                               |
| The copy constructor which constructs a IGLine object from an existing       |
| instance "line".                                                             |
------------------------------------------------------------------------------*/
IGLine::IGLine( const IGLine& line )
  : IGLine::Inherited(line), fStart(line.fStart), fEnd(line.fEnd)
{
}

/*------------------------------------------------------------------------------
| IGLine::IGLine                                                               |
| Construct a IGLine object which connects two points.                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGLine::IGLine( const IPoint& startPoint, const IPoint& endPoint )
  : IGLine::Inherited(), fStart(startPoint), fEnd(endPoint)
{
}

/*------------------------------------------------------------------------------
| IGLine::~IGLine                                                              |
| The virtual destructor.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGLine::~IGLine()
{
}

/*------------------------------------------------------------------------------
| IGLine::start                                                                |
| A IGLine object is internally defined by two points. The method returns the  |
| start point.                                                                 |
------------------------------------------------------------------------------*/
IPoint IGLine::startingPoint() const
{
  return fStart;
}

/*------------------------------------------------------------------------------
| IGLine::end                                                                  |
| A IGLine object is internally defined by two points. The method returns the  |
| end point.                                                                   |
------------------------------------------------------------------------------*/
IPoint IGLine::endingPoint() const
{
  return fEnd;
}

/*------------------------------------------------------------------------------
| IGLine::center                                                               |
| Return the center of the IGLine object.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGLine::centerPoint() const
{
  return IPoint((fStart.x()+fEnd.x())/2, (fStart.y()+fEnd.y())/2);
}

/*------------------------------------------------------------------------------
| IGLine::setStart                                                             |
| Set the start point of the IGLine object.                                    |
| Return itself as a reference.                                                |
------------------------------------------------------------------------------*/
IGLine& IGLine::setStartingPoint(const IPoint& point)
{
  fStart = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGLine::setEnd                                                               |
| Set the end point of the IGLine object.                                      |
| Return itself as a reference.                                                |
------------------------------------------------------------------------------*/
IGLine& IGLine::setEndingPoint(const IPoint& point)
{
  fEnd = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGLine::slope                                                                |
| Get the slope of the IGLine object.                                          |
| An exception will be thrown if fEnd.x() is equal to fStart.x()               |
------------------------------------------------------------------------------*/
double IGLine::slope() const
{
  return ((double)(fEnd.y()-fStart.y()))/((double)(fEnd.x()-fStart.x()));
}

/*------------------------------------------------------------------------------
| IGLine::drawOn                                                                |
| Draw the IGLine object on graphicContext, an ICLUI class which wraps the GPI  |
| device context. Return a reference of this object.                            |
-------------------------------------------------------------------------------*/
IGLine& IGLine::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGLine::operator=                                                            |
| Assignment operator which in turn calls the assignment of its base class     |
| Inherited, a typedef in the .hpp file. Return a reference of this object.    |
------------------------------------------------------------------------------*/
IGLine& IGLine::operator= ( const IGLine& line )
{
  // If assignment to itself, return a reference to itself.
  if (this == &line)
    return *this;

  // Call base class' overrided member function.
  this->Inherited::operator=( line );
  // A member-wise assignment.
  fStart = line.fStart;
  fEnd   = line.fEnd;
  return *this;
}

/*------------------------------------------------------------------------------
| IGLine::operator==                                                           |
| Equal operator which in turn calls the equal operator of its base class      |
| Inherited, a typedef in the .hpp file.                                       |
------------------------------------------------------------------------------*/
IBase::Boolean IGLine::operator== ( const IGLine& line ) const
{
  return (this->Inherited::operator==( line ) &&
          fStart == line.fStart &&
          fEnd == line.fEnd);
}

/*------------------------------------------------------------------------------
| IGLine::operator!=                                                           |
| Unequal operator which negates the result of calling operator==              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGLine::operator!= ( const IGLine& line ) const
{
  return !( *this == line );
}
