/*******************************************************************************
* FILE NAME: iglist.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iglist.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <iglist.hpp>
#include <igrafctx.hpp>
#include <ipoint.hpp>
#include <itrnsfrm.hpp>
#include <iseq.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGLIST_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGListSeq member functions                                                   |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IGListSeq class declaration                                                  |
| A sequence of IGraphic pointers.                                             |
|                                                                              |
------------------------------------------------------------------------------*/
class IGListSeq : public ISequence<IGraphic*>
{
public:
  IGListSeq( unsigned long items = 3000 );    //initial size = 3000.
  ~IGListSeq( );
};

/*------------------------------------------------------------------------------
| IGListSeq::IGListSeq                                                         |
| Default constructor.                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGListSeq::IGListSeq( unsigned long items )
  : ISequence<IGraphic*>(items)
{}

/*------------------------------------------------------------------------------
| IGListSeq::~IGListSeq                                                        |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGListSeq::~IGListSeq()
{}


/*------------------------------------------------------------------------------
| ObjectCursor member functions                                                |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor, a nested class.                                      |
| Constructor for iterating an IGList.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList& list )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::regular ), fId(0), fGc(0), fCtm(0), fBtm(0)
{
}

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor                                                       |
| Construct an IGList::Cursor which iterates through the graphic objects with  |
| the specified Identifier in the list.                                        |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList& list, unsigned long objectIdentifier )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::id ), fId(objectIdentifier),
    fGc(0), fCtm(0), fBtm(0)
{}

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor                                                       |
| Construct an IGList::Cursor which iterates through the graphic objects       |
| intersecting or containing hitPoint in the given graphicContext.             |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList&    list,
                        IGraphicContext& graphicContext,
                        const IPoint&    hitPoint )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::hit ), fId(0), fGc(&graphicContext), fCtm(0),
    fBtm(0)
{
  // Set hit point to the graphic context.
  fGc->setHitPoint( hitPoint );
  if (list.hasTransformMatrix())
  {
    // Save graphicContext's transformation matrix to fCtm.
    fCtm = new ITransformMatrix(fGc->worldTransformMatrix());
    // Set fGc's transformation matrix to list's.
    fGc->setWorldTransformMatrix(list.transformMatrix(), list.transformMethod());
  }
}

/*------------------------------------------------------------------------------
| IGList::Cursor::~Cursor                                                      |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::~Cursor()
{
  if (fCursorType == IGList::Cursor::hit)
  {
    if (fCursorList->hasTransformMatrix())
    {
      // Restore fGc's previous transformation matrix.
      fGc->setWorldTransformMatrix(*fCtm, IGraphic::replace );
      delete fCtm;
    }
    if (fBtm)
      delete fBtm;
  }
  delete (IGListSeq::Cursor*)fCursor;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToFirst                                                   |
| Set cursor to first element in the list.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToFirst()
{
  // Set the fCursor, a IGListSeq::Cursor to its first element.
  // Get the boolean result.
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToFirst());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular: //OK, cursor already at first.
    break;
    case IGList::Cursor::id:  //ID cursor. Find the first match with fId.
      if (rc)
      {
        IGraphic& graphic(fCursorList->graphicAt( *this ));
        if (graphic.id() == fId)
          rc = true;
        else
          rc = setToNext();
      }
    break;
    case IGList::Cursor::hit:  //Hit cursor. Find the first with a hit.
      if (rc)
      {
        // fGc starting hit testing.
        fGc->startHitTesting();

        IGraphic& graphic(fCursorList->graphicAt(*this));
        graphic.setHitSelected(false);

        if (graphic.isHitSelectable())
          graphic.drawOn(*fGc);        //Hit testing only.
        if (!graphic.isHitSelected())
        {
          rc = setToNext();      //Go to next (recursive).
          if (!rc)
            fGc->stopHitTesting();
        }
        else   //There is a hit.
        {
          // Replace the transformation matrix with the one having the hit.
          if (fBtm)
            delete fBtm;
          fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
          // fGc stopping hit testing.
          fGc->stopHitTesting();
        }
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToNext                                                    |
| Move the IGList::Cursor to the next element.                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToNext()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToNext());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    // ID cursor. Do an ID match recursively until
    // match is found or until the end of the list.
    case IGList::Cursor::id:
      if (rc)
      {
        rc = false;
        // ID cursor. Do an ID match recursively until
        // match is found or until the end of the list.
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToNext())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          if (graphic.id() == fId)
          {
            rc = true;
            break;
          }
        }
      }
    break;
    // Hit cursor. Recursively do a hit testing until
    // a hit is found or until the end of the list.
    case IGList::Cursor::hit:
      if (rc)
      {
        rc = false;
        // fGc starting hit testing.
        fGc->startHitTesting();

        // ID cursor. Do an ID match recursively until
        // match is found or until the end of the list.
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToNext())
        {
          // Get the IGraphic at the cursor.
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          // Reset hit selected flag.
          graphic.setHitSelected(false);

          if (graphic.isHitSelectable())
            graphic.drawOn(*fGc); //hit detecting without drawing.
          if (graphic.isHitSelected())  // found the next hit
          {
            rc = true;
            if (fBtm)  // Delete the old transformation matrix.
              delete fBtm;
            // Copy the current world transformation matrix.
            fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
            break;
          }
        } /* endfor */
        // fGc stopping hit testing.
        fGc->stopHitTesting();
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToPrevious                                                |
| Opposite to IGList::Cursor::setToNext.                                       |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToPrevious()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToPrevious());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    // ID cursor. Do an id match recursively until
    // match is found or until the beginning of the list.
    case IGList::Cursor::id:
      if (rc)
      {
        rc = false;
        // ID cursor. Do an ID match recursively until
        // match is found or until the beginning of the list.
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToPrevious())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          if (graphic.id() == fId)
          {
            rc = true;
            break;
          }
        }
      }
    break;
    // Hit cursor. Recursively do a hit testing until
    // a hit is found or until the beginning of the list.
    case IGList::Cursor::hit:
      if (rc)
      {
        rc = false;
        // fGc starting hit testing.
        fGc->startHitTesting();

        // Hit cursor. Recursively do a hit testing until
        // a hit is found or until the beginning of the list.
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToPrevious())
        {
          // Get the IGraphic at the cursor.
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          // Reset hit selected flag.
          graphic.setHitSelected(false);

          if (graphic.isHitSelectable())
            graphic.drawOn(*fGc);     //hit detecting without drawing.
          if (graphic.isHitSelected())  // found the previous hit.
          {
            rc = true;
            if (fBtm)
              delete fBtm;  // delete the old transformation matrix.
            // Copy the current world transformation matrix.
            fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
            break;
          }
        } /* endfor */
        // fGc stopping hit testing.
        fGc->stopHitTesting();
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToLast                                                    |
| Opposite to IGList::Cursor::setToFirst.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToLast()
{
  // Set the cursor to the last position.
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToLast());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    case IGList::Cursor::id:
      if (rc)
      {
        // Get the IGraphic at the cursor position.
        IGraphic& graphic(fCursorList->graphicAt( *this ));
        // Call setToPrevious() until you get an ID match.
        if (graphic.id() == fId)
          rc = true;
        else
          rc = setToPrevious();
      }
    break;
    case IGList::Cursor::hit:
      if (rc)
      {
        // fGc starting hit testing.
        fGc->startHitTesting();

        // Get the IGraphic at the cursor position.
        IGraphic& graphic(fCursorList->graphicAt(*this));
        // Reset hit selected flag.
        graphic.setHitSelected(false);

        if (graphic.isHitSelectable())
          graphic.drawOn(*fGc);    // Hit testing without drawing.
        if (!graphic.isHitSelected())
        {
          // Call setToPrevious() until you get a hit.
          rc = setToPrevious();
          if (!rc)
            fGc->stopHitTesting();
        }
        else
        {
          if (fBtm)
            delete fBtm;  // Delete the old transformation matrix.
          // Copy the current world transformation matrix.
          fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
          fGc->stopHitTesting();
        }
      }
    break;
  } /* endswitch */

  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::isValid                                                      |
| Is the cursor in a valid position.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::isValid() const
{
  return ((IGListSeq::Cursor*)fCursor)->isValid();
}

/*------------------------------------------------------------------------------
| IGList::Cursor::invalidate                                                   |
| Mark cursor as invalidate.                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
void IGList::Cursor::invalidate()
{
  ((IGListSeq::Cursor*)fCursor)->invalidate();
}


/*------------------------------------------------------------------------------
| IGList member functions                                                      |
------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
| Default constructor.                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList ( )
  : IGList::Inherited(), fList(new IGListSeq()) //IGList contains a IGListSeq
{
  fIsList = true; // IGList is derived from IGraphic.
                  // This object is not only an IGraphic but also an IGList.
}

/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
| Constrcut a IGList with "graphic" as its first element.                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList( IGraphic& graphic )
  : IGList::Inherited( ), fList(new IGListSeq())
{
  // IGList is derived from IGraphic.
  // This object is not only an IGraphic but also an IGList.
  fIsList = true;
  fList->add(&graphic);  // Add graphic as the first element.
}

/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
| Copy constrcutor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList ( const IGList& list )
  : IGList::Inherited( list ), fList(new IGListSeq())
{
  // IGList is derived from IGraphic.
  // This object is not only an IGraphic but also an IGList.
  fIsList = true;
  // Copy all elements from IGList: list.
  fList->addAllFrom(*list.fList);
}

/*------------------------------------------------------------------------------
| IGList::~IGList                                                              |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::~IGList ( )
{
  delete fList;
}

/*------------------------------------------------------------------------------
| IGList::drawOn                                                               |
| Draw the entire IGList on a given IGraphicContext.                           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::drawOn( IGraphicContext& fGc )
{
  fGc.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::numberOfGraphics                                                     |
| Return the number of graphics added into the IGList.                         |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGList::numberOfGraphics( ) const
{
  return fList->numberOfElements();
}

/*------------------------------------------------------------------------------
| IGList::isEmpty                                                              |
| Is the list empty?                                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::isEmpty( ) const
{
  return fList->isEmpty();
}

/*------------------------------------------------------------------------------
| IGList::isIGList                                                             |
| Return whether the element at the position pointed by the cursor is itself   |
| an IGList.                                                                   |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::isIGList( const Cursor& cursor ) const
{
  IGraphic& graphic(graphicAt( cursor ));
  return graphic.fIsList;
}

/*------------------------------------------------------------------------------
| IGList::firstGraphic                                                         |
| Get the first element in the IGList.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::firstGraphic( ) const
{
  IGraphic* graphic(fList->firstElement());
  return graphic;
}

/*------------------------------------------------------------------------------
| IGList::lastGraphic                                                          |
| Get the last element in the IGList.                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::lastGraphic( ) const
{
  IGraphic* graphic(fList->lastElement());
  return graphic;
}

/*------------------------------------------------------------------------------
| IGList::graphicAt                                                            |
| Get the graphic element pointed to by the cursor.                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic& IGList::graphicAt( const Cursor& cursor ) const
{
  IGraphic* graphic(fList->elementAt(*((IGListSeq::Cursor*)cursor.fCursor)));
  return *graphic;
}

/*------------------------------------------------------------------------------
| IGList::graphicAtPosition                                                    |
| Get the graphic element at position.                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic& IGList::graphicAtPosition( unsigned long position ) const
{
  IGraphic* graphic(fList->elementAtPosition(position));
  return *graphic;
}

/*------------------------------------------------------------------------------
| IGList::boundingRectAt                                                       |
| Get the bounding box of the graphic pointed by the cursor.                   |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGList::boundingRectAt( const Cursor& cursor ) const
{
  cursor.fGc->startBoundaryAccumulation();
  IGraphic& graphic(graphicAt(cursor));
  graphic.drawOn( *cursor.fGc );
  return cursor.fGc->stopBoundaryAccumulation();
}

/*------------------------------------------------------------------------------
| IGList::addAsFirst                                                           |
| Add graphic as the first element in the IGList.                              |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::addAsFirst( IGraphic& graphic )
{
  fList->addAsFirst(&graphic);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsLast                                                            |
| Add graphic as the last element in the IGList.                               |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::addAsLast( IGraphic& graphic )
{
  fList->addAsLast(&graphic);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsNext                                                            |
| Add graphic as the next element pointed by the cursor in the IGList.         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::addAsNext( const Cursor& cursor, IGraphic& graphic )
{
  fList->addAsNext( &graphic, *((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsPrevious                                                        |
| Add graphic as the previous element pointed by the cursor in the IGList.     |
| Return a reference of this object.                                          |
------------------------------------------------------------------------------*/
IGList& IGList::addAsPrevious( const Cursor& cursor, IGraphic& graphic )
{
  fList->addAsPrevious( &graphic, *((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAtPosition                                                        |
| Add graphic at position in the IGList.                                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::addAtPosition( unsigned long position, IGraphic& graphic )
{
  fList->addAtPosition( position, &graphic );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::replaceAt                                                            |
| Replace the element pointed to by the cursor in the IGList with graphic.     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::replaceAt( const Cursor& cursor, IGraphic& graphic )
{
  fList->replaceAt(*((IGListSeq::Cursor*)cursor.fCursor), &graphic );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeFirst                                                          |
| Remove the first element from the IGList.                                    |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeFirst()
{
  fList->removeFirst();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeLast                                                           |
| Remove the last element from the IGList.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeLast()
{
  fList->removeLast();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAtPosition                                                     |
| Remove the element at position in the IGList.                                |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeAtPosition( unsigned long position )
{
  fList->removeAtPosition(position);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAt                                                             |
| Remove the element pointed to by the cursor in the IGList.                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeAt( const Cursor& cursor )
{
  fList->removeAt(*((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAll                                                            |
| Remove all elements in the IGList.                                           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeAll( )
{
  fList->removeAll();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAllWithId                                                      |
| Remove all elements with ID identifier.                                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGList& IGList::removeAllWithId( unsigned long identifier )
{
  IGList::Cursor cursor( *this, identifier );
  for( cursor.setToFirst(); cursor.isValid(); )
  {
    removeAt( cursor );
    cursor.setToFirst();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::topGraphicUnderPoint                                                 |
| Return the top graphic which hits the point under the given graphicContext.  |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::topGraphicUnderPoint( const IPoint& point,
                                        IGraphicContext& graphicContext )
{
  IGraphic* hit(0);
  // Construct a hit detecting cursor.
  IGList::Cursor cursor( *this, graphicContext, point );

  // Move the hit detecting cursor to the last hitting graphic, which
  // should be at the top of all graphics satisfying the hit detection.
  if (cursor.setToLast())
    hit = &graphicAt( cursor );

  return hit;
}

/*-------------------------------------------------------------------------------
| IGList::bottomGraphicUnderPoint                                               |
| Return the bottom graphic which hits the point under the given graphicContext.|                                                                             |
|                                                                               |
-------------------------------------------------------------------------------*/
IGraphic* IGList::bottomGraphicUnderPoint( const IPoint& point,
                                           IGraphicContext& graphicContext )
{
  IGraphic* hit(0);
  // Construct a hit detecting cursor.
  IGList::Cursor cursor( *this, graphicContext, point );

  // Move the hit detecting cursor to the first hiting graphic, which
  // should be at the bottom of all graphics satisfying the hit detection.
  if (cursor.setToFirst())
    hit = &graphicAt( cursor );

  return hit;
}


/*------------------------------------------------------------------------------
| IGList::sort                                                                 |
| Sort the IGList in an order defined by the comparisonFunction, a pointer to  |               |
| a functon which returns a long with signature (IGraphic* const&,             |
| IGraphic* const&).                                                           |                 |
------------------------------------------------------------------------------*/
IGList& IGList::sort( long (*comparisonFunction)(IGraphic* const&,
                                                 IGraphic* const&))
{
  fList->sort( comparisonFunction );
  return *this;
}
