/*******************************************************************************
* FILE NAME: igpie.cpp                                                         *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igpie.hpp.                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <igpie.hpp>
#include <igrafctx.hpp>
#include <igbundle.hpp>

/*------------------------------------------------------------------------------
| Segment definitions                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
#ifdef IC_PAGETUNE
  #define _IGPIE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Data conversion macro between long and float numbers.                        |
|                                                                              |
------------------------------------------------------------------------------*/
#define      DBL2LONG(x)     ((long)  ((x)*65536) )
#define      LONG2DBL(x)     ((double) (((double)x)/(double)65536) )

/*------------------------------------------------------------------------------
| IGPie::IGPie                                                                 |
| Constructor.                                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IGPie::IGPie( const IRectangle& rectangle,
              double startAngle, double sweepAngle )
  : IGPie::Inherited(), fRect(rectangle), fStart(DBL2LONG(startAngle)),
    fSweep(DBL2LONG(sweepAngle))
{}

/*------------------------------------------------------------------------------
| IGPie::IGPie                                                                 |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGPie::IGPie( const IGPie& pie)
  : IGPie::Inherited(pie), fRect(pie.fRect), fStart(pie.fStart),
    fSweep(pie.fSweep)
{
}

/*------------------------------------------------------------------------------
| IGPie::~IGPie                                                                |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGPie::~IGPie( )
{ }

/*------------------------------------------------------------------------------
| IGPie::enclosingRect                                                         |
| Get the enclosing rectangle of the pie.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGPie::enclosingRect( ) const
{
  return fRect;
}

/*------------------------------------------------------------------------------
| IGPie::startAngle                                                            |
| Get the start angle of the pie.                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGPie::startAngle( ) const
{
  return LONG2DBL(fStart);
}

/*------------------------------------------------------------------------------
| IGPie::sweepAngle                                                            |
| Get the sweeping angle of the pie.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
double IGPie::sweepAngle( ) const
{
  return LONG2DBL(fSweep);
}

/*------------------------------------------------------------------------------
| IGPie::setEnclosingRect                                                      |
| Set the enclosing rectangle of the pie.                                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPie& IGPie::setEnclosingRect( const IRectangle& rectangle )
{
  fRect = rectangle;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPie::setStartAngle                                                         |
| Set the start angle of the pie.                                              |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPie& IGPie::setStartAngle( double startAngle )
{
  fStart = DBL2LONG(startAngle);
  return *this;
}

/*------------------------------------------------------------------------------
| IGPie::setSweepAngle                                                         |
| Set the sweeping angle of the pie.                                           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPie& IGPie::setSweepAngle( double sweepAngle )
{
  fSweep = DBL2LONG(sweepAngle);
  return *this;
}

/*------------------------------------------------------------------------------
| IGPie::drawOn                                                                |
| Draw the pie on a given IGraphicContext: graphicContext                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPie& IGPie::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPie:operator=                                                              |
| Assignment operator.                                                         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPie& IGPie::operator= (const IGPie& pie )
{
  // if assignment to itself, return a reference to itself.
  if ( this == &pie )
    return *this;

  // call base class' overrided member function.
  this->Inherited::operator=(pie);
  // a member-wise assignment.
  fRect    = pie.fRect;
  fStart   = pie.fStart;
  fSweep   = pie.fSweep;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPie:operator==                                                             |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGPie::operator== (const IGPie& pie ) const
{
  // a member-wise comparison.
  return ( fRect == pie.fRect && fStart == pie.fStart &&
           fSweep == pie.fSweep && this->Inherited::operator==(pie));
}

/*------------------------------------------------------------------------------
| IGPie:operator!=                                                             |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGPie::operator!= (const IGPie& pie ) const
{
  return !(*this == pie);
}

/*------------------------------------------------------------------------------
| IGChord::IGChord                                                             |
| Constructor.                                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IGChord::IGChord( const IRectangle& rect,
                  double startAngle, double sweepAngle )
  : IGChord::Inherited( rect, startAngle, sweepAngle )
{}

/*------------------------------------------------------------------------------
| IGChord::IGChord                                                             |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGChord::IGChord( const IGChord& chord )
  : IGChord::Inherited( chord )
{ }

/*------------------------------------------------------------------------------
| IGChord::~IGChord                                                            |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGChord::~IGChord( )
{ }

/*------------------------------------------------------------------------------
| IGChord::draw                                                                |
| Draw itself on a given IGraphicContext: graphicContext.                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGChord& IGChord::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}
