/*******************************************************************************
* FILE NAME: igpyline.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igpyline.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*******************************************************************************/

#include <igpyline.hpp>
#include <igrafctx.hpp>
#include <igbundle.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGPYLINE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGPolyline::IGPolyline                                                       |
| Copy constructor from an existing IGPloyline object. fPtArray is an          |
| IPointArray which stores all points along the polyline.                      |
------------------------------------------------------------------------------*/
IGPolyline::IGPolyline( const IGPolyline& polyline )
  : IGPolyline::Inherited(polyline), fPtArray( polyline.fPtArray )
{
}

/*------------------------------------------------------------------------------
| IGPolyline::IGPolyline                                                       |
| Constructs a IGPolyline from an array of points in IPointArray.              |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolyline::IGPolyline( const IPointArray& pointArray )
  : IGPolyline::Inherited(), fPtArray( pointArray )
{
}

/*------------------------------------------------------------------------------
| IGPolyline::~IGPolyline                                                      |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolyline::~IGPolyline()
{
}

/*------------------------------------------------------------------------------
| IGPolyline::drawOn                                                           |
| Draw the IGPolyline object on the IGraphicContext: graphicContext.           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::numberOfPoints                                                   |
| Return the number of points in the IGPolyline.                               |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGPolyline::numberOfPoints() const
{
  return fPtArray.size();
}

/*------------------------------------------------------------------------------
| IGPolyline::pointArray                                                       |
| Return the IPointArray which stores all the points along the IGPolyline.     |
|                                                                              |
------------------------------------------------------------------------------*/
const IPointArray& IGPolyline::pointArray() const
{
  return fPtArray;
}

/*------------------------------------------------------------------------------
| IGPolyline::pointArray                                                       |
| Indexing into the array of points                                            |
|                                                                              |
------------------------------------------------------------------------------*/
const IPoint& IGPolyline::point( unsigned long index ) const
{
  return fPtArray[index];
}

/*------------------------------------------------------------------------------
| IGPolyline::setPoints                                                        |
| Set the IGPolyline's IPointArray to pointArray                               |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::setPoints( const IPointArray& pointArray )
{
  fPtArray = pointArray;    //calling IPointArray's operator=
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::setPoints                                                        |
| Set point into IGPolyline's IPointArray by indexing.                         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::setPoint( unsigned long index, const IPoint& point )
{
  fPtArray[index] = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::insertPoint                                                      |
| Insert point into IGPolyline's IPointArray by indexing.                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::insertPoint(  unsigned long index, const IPoint& point )
{
  fPtArray.insert( index, point );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::addPoint                                                         |
| Add one more point to the end of the IPointArray.                            |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::addPoint( const IPoint& point )
{
  fPtArray.add( point );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::removePoint                                                      |
| Remove point from IPointArray at position "index".                           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::removePoint( unsigned long index )
{
  fPtArray.remove( index );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::reversePoints                                                    |
| Reverse all points in the IPointArray.                                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::reversePoints( )
{
  fPtArray.reverse();
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::operator=                                                        |
| Assignment operator which also calls its base class' operator=.              |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolyline& IGPolyline::operator= ( const IGPolyline& polyline )
{
  // If assignment to itself, return a reference to itself.
  if (this == &polyline)
    return *this;

  this->Inherited::operator=(polyline);  //Inherited is a typedef in .hpp
  // A member-wise assignment.
  fPtArray = polyline.fPtArray;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolyline::operator==                                                       |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGPolyline::operator== ( const IGPolyline& polyline ) const
{
  // A member-wise comparison.
  return ( fPtArray == polyline.fPtArray &&
           this->Inherited::operator==( polyline ));
}

/*------------------------------------------------------------------------------
| IGPolyline::operator!=                                                       |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGPolyline::operator!= ( const IGPolyline& polyline ) const
{
  return !( *this == polyline );
}

/*------------------------------------------------------------------------------
| IGPolygon::IGPolygon                                                         |
| Copy constructor. fFm is the IGPolygon's filling mode FillMode.              |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolygon::IGPolygon( const IGPolygon& polygon)
  : IGPolygon::Inherited( polygon ), fFm(polygon.fFm)
{
}

/*------------------------------------------------------------------------------
| IGPolygon::IGPolygon                                                         |
| Construct a IGPolygon from an IPointArray of points.                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolygon::IGPolygon( const IPointArray& pointArray )
  : IGPolygon::Inherited(pointArray), fFm(IGPolygon::alternate)
{
}

/*------------------------------------------------------------------------------
| IGPolygon::~IGPolygon                                                        |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolygon::~IGPolygon()
{
}

/*------------------------------------------------------------------------------
| IGPolygon::drawOn                                                            |
| Draw the IGPolygon on IGraphicContext: graphicContext.                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolygon& IGPolygon::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolygon::fillMode                                                          |
| Get the filling mode for the IGPolygon.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGPolygon::FillMode IGPolygon::fillMode() const
{
  return fFm;
}

/*------------------------------------------------------------------------------
| IGPolygon::setFillMode                                                       |
| Set the filling mode for the IGPolygon. FillMode is defined as a class-scoped|
| enum. Return a reference of this object.                                     |
------------------------------------------------------------------------------*/
IGPolygon& IGPolygon::setFillMode( IGPolygon::FillMode fillmode )
{
  fFm = fillmode;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolygon::operator=                                                         |
| Assignment operator.                                                         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGPolygon& IGPolygon::operator= ( const IGPolygon& polygon )
{
  // If assignment to itself, return a reference to itself.
  if ( this == &polygon )
    return *this;

  // Call base class' overrided member function.
  this->Inherited::operator=(polygon);
  // A member-wise assignment.
  fFm      = polygon.fFm;
  return *this;
}

/*------------------------------------------------------------------------------
| IGPolygon::operator==                                                        |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGPolygon::operator== ( const IGPolygon& polygon ) const
{
  // A member-wise comparison.
  return ( this->Inherited::operator==( polygon ) && fFm == polygon.fFm );
}

/*------------------------------------------------------------------------------
| IGPolygon::operator!=                                                        |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGPolygon::operator!= ( const IGPolygon& polygon ) const
{
  return !(*this == polygon);
}

