/*******************************************************************************
* FILE NAME: igraphic.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igraphic.cpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <igraphic.hpp>
#include <igrafctx.hpp>
#include <ipoint.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGRAPHIC_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGraphic::IGraphic                                                           |
| Default constructor.                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic::IGraphic( )
  : fId(0), fTm(IGraphic::rightMultiply), fBundle(0), fTransMat(0),
    fSelected(false), fSelectable(true), fIsList(false)
{
}

/*------------------------------------------------------------------------------
| IGraphic::IGraphic                                                           |
| Construct an IGraphic using an existing IGraphicBundle.                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic::IGraphic( const IGraphicBundle& graphicBundle )
  : fId(0), fTm(IGraphic::rightMultiply), fBundle(0), fTransMat(0),
    fSelected(false), fSelectable(true), fIsList(false)
{
  fBundle = new IGraphicBundle(graphicBundle);
}

/*------------------------------------------------------------------------------
| IGraphic::IGraphic                                                           |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic::IGraphic( const IGraphic& graphic )
  : fBundle(0), fTransMat(0)
{
  // a member-wise copy.
  if (graphic.hasGraphicBundle())
    this->setGraphicBundle(graphic.graphicBundle());
  if (graphic.hasTransformMatrix())
    this->setTransformMatrix(graphic.transformMatrix());

  this->fId         = graphic.fId; // id
  this->fTm         = graphic.fTm; // transformation matrix.
  this->fSelected   = graphic.fSelected;  //hit flag.
  this->fSelectable = graphic.fSelectable; // whether can be selected.
  this->fIsList     = graphic.fIsList;   // IGList flag.
}

/*------------------------------------------------------------------------------
| IGraphic::~IGraphic                                                          |
| Virtual destructor                                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic::~IGraphic( )
{
  if ( fBundle )
  {
    delete fBundle;
  }
  if ( fTransMat )
  {
    delete fTransMat;
  }
}

/*------------------------------------------------------------------------------
| IGraphic::boundingRect                                                       |
| Return the bounding rectangle in given graphicContext.                       |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGraphic::boundingRect( IGraphicContext& graphicContext )
{
  graphicContext.startBoundaryAccumulation();
  //draw invisibly in order to accumulate the bounding box.
  drawOn( graphicContext );
  return graphicContext.stopBoundaryAccumulation();
}

/*------------------------------------------------------------------------------
| IGraphic::setGraphicBundle                                                   |
| Set graphic bundle.                                                          |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setGraphicBundle( const IGraphicBundle& graphicBundle )
{
  if ( fBundle )              //fBundle !=0, bundle is set before.
    *fBundle = graphicBundle;
  else
    fBundle = new IGraphicBundle(graphicBundle);  // copy over
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::graphicBundle                                                      |
| Return the graphic bundle.                                                   |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle IGraphic::graphicBundle ( ) const
{
  if (fBundle)
    return IGraphicBundle(*fBundle);
  else
    return IGraphicBundle(); // return default bundle.
}

/*------------------------------------------------------------------------------
| IGraphic::removeGraphicBundle                                                |
| dettach the graphic bundle from this graphic.                                |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::removeGraphicBundle( )
{
  if ( fBundle )
  {
    delete fBundle;
    fBundle = 0;    // good habit! Set it 0.
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::rotateBy                                                           |
| Rotate a graphic by certain angle, pt in the center of rotation.             |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::rotateBy( double angle, const IPoint &pt )
{
  if (!fTransMat)
    fTransMat = new ITransformMatrix;

  fTransMat->rotateBy( angle, pt );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::scaleBy                                                            |
| Scale the graphic horizontally by x, vertically by y, and originate from     |
| IPoint pt. Return a reference of this object.                                |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::scaleBy( double xScale, double yScale, const IPoint& pt )
{
  if (!fTransMat)
    fTransMat = new ITransformMatrix;

  fTransMat->scaleBy( xScale, yScale, pt );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::translateBy                                                        |
| Move the graphic to pt.                                                      |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::translateBy( const IPoint& pt )
{
  if (!fTransMat)
    fTransMat = new ITransformMatrix;

  fTransMat->translateBy( pt );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::setTransformMatrix                                                 |
| Set transformation matrix.                                                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setTransformMatrix( const ITransformMatrix& transformMatrix)
{
  if (fTransMat)
    *fTransMat = transformMatrix;
  else
    fTransMat = new ITransformMatrix( transformMatrix );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::resetTransformMatrix                                               |
| Delete transformation matrix.                                                |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::resetTransformMatrix()
{
  if ( fTransMat )
  {
    delete fTransMat;
    fTransMat = 0;
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphic::transformMatrix                                                    |
| Return transformation matrix, if there is none, create one.                  |
|                                                                              |
------------------------------------------------------------------------------*/
const ITransformMatrix IGraphic::transformMatrix() const
{
  if ( fTransMat )
    return ITransformMatrix( *fTransMat );
  else
    return ITransformMatrix();
}

/*------------------------------------------------------------------------------
| IGraphic::contains                                                           |
| Whether the IGraphic contains a IPoint: pt.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::contains( IGraphicContext& graphicContext, const IPoint& pt )
{
  setHitSelected( false ); // reset hit falg.

  graphicContext.startHitTesting()  //set GPI hit detecting mode
                .setHitPoint(pt);   //set hit point
  drawOn( graphicContext );         //hit detecting
  graphicContext.stopHitTesting();  //stop hit detecting mode.

  return isHitSelected();
}

/*------------------------------------------------------------------------------
| IGraphic::setTransformMethod                                                 |
| Set transform method.                                                        |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setTransformMethod( TransformMethod transformMethod )
{
  fTm = transformMethod;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::setTransformMethod                                                 |
| Return transform method.                                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic::TransformMethod IGraphic::transformMethod( ) const
{
  return fTm;
}

/*------------------------------------------------------------------------------
| IGraphic::hasGraphicBundle                                                   |
| Is there any graphic bundle used by this graphic?                            |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::hasGraphicBundle() const
{
  return ((fBundle == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGraphic::hasTransformMatrix                                                 |
| Is there any transformation matrix used by this graphic?                     |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::hasTransformMatrix() const
{
  return ((fTransMat == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGraphic::setId                                                              |
| Set Graphic object ID.                                                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setId( unsigned long identifier )
{
  fId = identifier;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::id                                                                 |
| Get graphic obeject ID.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphic::id( ) const
{
  return fId;
}

/*------------------------------------------------------------------------------
| IGraphic::setHitSelected                                                     |
| There is a hit detected on this graphic.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setHitSelected( IBase::Boolean hitSelected )
{
  this->fSelected = hitSelected;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::isHitSelected                                                      |
| Is there a hit detected on this graphic?                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::isHitSelected( ) const
{
  return (fSelectable ? fSelected : false);
}

/*------------------------------------------------------------------------------
| IGraphic::setHitSelectable                                                   |
| Set hit selectable flag.                                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::setHitSelectable( IBase::Boolean hitSelectable )
{
  this->fSelectable = hitSelectable;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::isHitSelectable                                                    |
| Is hit selectable flag set?                                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::isHitSelectable( ) const
{
  return fSelectable;
}

/*------------------------------------------------------------------------------
| IGraphic::operator=                                                          |
| Assignment operator.                                                         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGraphic& IGraphic::operator= ( const IGraphic& graphic )
{
  // If assignment to itself, return a reference to itself.
  if ( this == &graphic )    //must have, canonial form.
    return *this;

  // Clear old graphic bundle and transformation matrix.
  this->removeGraphicBundle();
  this->resetTransformMatrix();

  // Copy over the new one from IGraphic: graphic
  if (graphic.hasGraphicBundle())
    this->setGraphicBundle(graphic.graphicBundle());
  if (graphic.hasTransformMatrix())
    this->setTransformMatrix(graphic.transformMatrix());
  // A member-wise assignment.
  this->fId         = graphic.fId;
  this->fTm         = graphic.fTm;
  this->fSelected   = graphic.fSelected;
  this->fSelectable = graphic.fSelectable;
  this->fIsList     = graphic.fIsList;

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphic::operator==                                                         |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::operator== ( const IGraphic& graphic ) const
{
  // A member-wise comparison.
  if ((this->hasGraphicBundle() && graphic.hasGraphicBundle()) &&
      (this->hasTransformMatrix() && graphic.hasTransformMatrix()) &&
      (this->transformMatrix() == graphic.transformMatrix()) &&
      (this->graphicBundle() == graphic.graphicBundle()) &&
      (this->fId == graphic.fId) &&
      (this->fTm == graphic.fTm) &&
      (this->fSelected == graphic.fSelected) &&
      (this->fSelectable == graphic.fSelectable))
    return true;
  else
    return false;
}

/*------------------------------------------------------------------------------
| IGraphic::operator!=                                                         |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphic::operator!= ( const IGraphic& graphic ) const
{
  return !(*this == graphic);
}
