/*******************************************************************************
* FILE NAME: igregion.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igregion.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
{
  #define INCL_GPI
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <iwindow.hpp>
#include <igregion.hpp>
#include <igrafctx.hpp>
#include <igrect.hpp>
#include <igpie.hpp>
#include <igpyline.hpp>
#include <igelipse.hpp>
#include <iexcept.hpp>
#include <itrnsfrm.hpp>
#include <igbundle.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGREGION_CPP_
  #include <ipagetun.h>
#endif

// Conversion macro from double to fixed.
#define      DBL2FIXED(x)     ((FIXED)  ((x)*65536) )

/*------------------------------------------------------------------------------
| IGRegion::IGRegion                                                           |
| Default constructor.                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::IGRegion( )
  : IGRegion::Inherited(), fRegion(0)
{
  // Copy over the desktopWindow's graphic context.
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  RECTL rcl;

  // Create a region.
  if ((fRegion = GpiCreateRegion(graphicContext.handle(), 0, &rcl)) == RGN_ERROR)
    ITHROWGUIERROR("GpiCreateRegion");
}

/*------------------------------------------------------------------------------
| IGRegion::IGRegion                                                           |
| Constructor using a IRegionHandle.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::IGRegion( const IRegionHandle& regionHandle )
  : IGRegion::Inherited(), fRegion(regionHandle)
{
}

/*------------------------------------------------------------------------------
| IGRegion::IGRegion                                                           |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::IGRegion( const IGRegion& region )
  : IGRegion::Inherited( region ), fRegion(0)
{
  // Create a graphic context using the top level window handle.
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  RECTL rcl;

  //Create a region.
  if ((fRegion = GpiCreateRegion(graphicContext.handle(), 0, &rcl)) == RGN_ERROR)
    ITHROWGUIERROR("GpiCreateRegion");

  // Combine two regions, here used to copy over the attributes.
  GpiCombineRegion( graphicContext.handle(), fRegion, region.fRegion,
                    region.fRegion, CRGN_COPY );
}

/*------------------------------------------------------------------------------
| IGRegion::IGRegion                                                           |
| Constructor using a given graphicContext and a rectangle.                    |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::IGRegion( const IGraphicContext& graphicContext,
                    const IRectangle& rectangle )
  : IGRegion::Inherited()
{
  RECTL rctl(rectangle.asRECTL());
  long  rectCount(1);

  if ( rectangle.size().width() == 0 || rectangle.size().height() == 0 )
  {
    rectCount = 0;     //no rectangle
  }
  // Create a region covered by a set of rectangles.
  fRegion = GpiCreateRegion(graphicContext.handle(), rectCount, &rctl );
  if ( fRegion == RGN_ERROR )
  {
    ITHROWGUIERROR2("GpiCreateRegion",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
}

/*------------------------------------------------------------------------------
| IGRegion::IGRegion                                                           |
| Constructor using a window handle.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::IGRegion( const IWindowHandle& windowHandle )
  : IGRegion::Inherited()
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  RECTL rcl;

  // Create a region.
  if ((fRegion = GpiCreateRegion(graphicContext.handle(), 0, &rcl)) == RGN_ERROR)
    ITHROWGUIERROR("GpiCreateRegion");

  // Query the attributes to fill in the new created fRegion.
  if (WinQueryUpdateRegion( windowHandle, fRegion ) == RGN_ERROR )
  {
    ITHROWGUIERROR2("WinQueryUpdateRegion",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
}


/*------------------------------------------------------------------------------
| IGRegion::~IGRegion                                                          |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::~IGRegion( )
{
  if (fRegion)
  {
    IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );

    GpiDestroyRegion( graphicContext.handle(), fRegion);
  }
}

/*------------------------------------------------------------------------------
| IGRegion::moveBy                                                             |
| Move the entire region to point.                                             |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::moveBy( const IPoint& point )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );

  POINTL ptl(point.asPOINTL());

  // Move a region.
  if (!GpiOffsetRegion( graphicContext.handle(), fRegion, &ptl ))
  {
    ITHROWGUIERROR2("GpiOffsetRegion",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::clear                                                              |
| Set the number of rectangles to zero in the region.                          |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::clear( )
{
  // Create a graphic context using the top level window handle.
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  RECTL rc;

  // Set the number of rectangles to zero in the region.
  if (!GpiSetRegion( graphicContext.handle(), fRegion, 0, &rc ))
  {
    ITHROWGUIERROR2("GpiSetRegion",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::drawOn                                                             |
| Draw itself on a given graphicContext.                                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::drawOn( IGraphicContext& graphicContext )
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::operator                                                           |
| Return the region handle of IGRegion.                                        |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion::operator const IRegionHandle( ) const
{
  return fRegion;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGRegion)                                               |
| Combine two regions and their attributes.                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGRegion& region )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, region.fRegion,
                    CRGN_OR );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGRegion)                                               |
| Take off region from IGRegion. Return the differences.                       |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGRegion& region )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, region.fRegion,
                    CRGN_DIFF );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGRegion)                                               |
| Get the intersection of two regions. Return a reference of this object.      |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGRegion& region )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, region.fRegion,
                    CRGN_AND );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGRegion)                                               |
| Get the exclusive OR of two regions. Return a reference of this object.      |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGRegion& region )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, region.fRegion,
                    CRGN_XOR );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor= (IGRegion)                                                |
| Assignment operator. Return a reference of this object.                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator= ( const IGRegion& region )
{
  if (this == &region)
    return *this;

  this->Inherited::operator=( region );
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, region.fRegion, region.fRegion,
                    CRGN_COPY );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGRectangle)                                            |
| Combine IGRegion with a region described by rect.                            |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGRectangle& rect )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doRect(graphicContext, rect));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_OR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGRectangle)                                            |
| Take off the region described by teh rect from IGRegion.                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGRectangle& rect )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doRect(graphicContext, rect));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_DIFF );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGRectangle)                                            |
| Get the intersection of IGRegion and the region described by rect.           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGRectangle& rect )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doRect(graphicContext, rect));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_AND );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGRectangle)                                            |
| Get the exclusive OR of IGRegion and the region described by rect.           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGRectangle& rect )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doRect(graphicContext, rect));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_XOR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGPie)                                                  |
| Add a pie into the region.                                                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGPie& pie )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPie(graphicContext, pie)); // create a pie region.
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_OR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGPie)                                                  |
| Take a pie off the region.                                                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGPie& pie )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPie(graphicContext, pie));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_DIFF );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGPie)                                                  |
| Take the intersection of this region and the pie.                            |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGPie& pie )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPie(graphicContext, pie));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_AND );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGPie)                                                  |
| Take the exclusive OR of this region and the pie.                            |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGPie& pie )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPie(graphicContext, pie));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_XOR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGChord)                                                |
| Add a chord into this region.                                                |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGChord& chord )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doChord(graphicContext, chord));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_OR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGChord)                                                |
| Take a chord region off the region.                                          |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGChord& chord )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doChord(graphicContext, chord));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_DIFF );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGChord)                                                |
| Get the intersection of this region and the chord region.                    |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGChord& chord )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doChord(graphicContext, chord));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_AND );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGChord)                                                |
| Return the exclusive OR of this region and the chord region.                 |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGChord& chord )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doChord(graphicContext, chord));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_XOR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGPolygon)                                              |
| Add the polygon region into this region.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGPolygon& polygon )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPolygon(graphicContext, polygon));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_OR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGPolygon)                                              |
| Take the polygon region off this region.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGPolygon& polygon )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPolygon(graphicContext, polygon));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_DIFF );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGPolygon)                                              |
| return the intersection of this region and the polygon region.               |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGPolygon& polygon )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPolygon(graphicContext, polygon));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_AND );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGPolygon)                                              |
| Return the exclusive OR of this region and the polygon region.               |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGPolygon& polygon )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doPolygon(graphicContext, polygon));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_XOR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor+= (IGEllipse)                                              |
| Add a ellipse region into this region.                                       |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator+= ( const IGEllipse& ellipse )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doEllipse(graphicContext, ellipse));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_OR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor-= (IGEllipse)                                              |
| Take the ellipse region off this region.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator-= ( const IGEllipse& ellipse )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doEllipse(graphicContext, ellipse));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_DIFF );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor&= (IGEllipse)                                              |
| Return the intersection of this region and the ellipse region.               |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator&= ( const IGEllipse& ellipse )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doEllipse(graphicContext, ellipse));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_AND );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::opertor^= (IGEllipse)                                              |
| return the exclusive OR of this region and ellipse region.                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::operator^= ( const IGEllipse& ellipse )
{
  IGraphicContext graphicContext( IWindow::desktopWindow()->handle() );
  IRegionHandle newRegion(doEllipse(graphicContext, ellipse));
  // GpiCombineRegion(PS handle, destination, source1, source2, method).
  GpiCombineRegion( graphicContext.handle(), fRegion, fRegion, newRegion, CRGN_XOR );
  GpiDestroyRegion( graphicContext.handle(), newRegion );
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::doRect                                                             |
| Create a rectangle region using a given graphicContect and an                |
| IRectangle: rect.  Return a region handle.                                   |
------------------------------------------------------------------------------*/
IRegionHandle IGRegion::doRect( IGraphicContext& graphicContext, const IGRectangle& rect )
{
  // Get the top right of the rectangle and convert it to a PM POINTL.
  POINTL ptl(rect.enclosingRect().topRight().asPOINTL());
  IRegionHandle newRegion;
  long rc;

  // Specify the start of a path.
  if (!GpiBeginPath(graphicContext.handle(), 1L ))
    ITHROWGUIERROR2("GpiBeginPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);

  // Move the the bottom left of the rect.
  graphicContext.setCurrentDrawingPosition( rect.enclosingRect().bottomLeft() );
  // Set the transformation matrix and graphic bundle of rect to the graphic context.
  if (rect.hasTransformMatrix())
  {
    graphicContext.setWorldTransformMatrix(rect.transformMatrix(), rect.transformMethod());
  }

  if (rect.hasGraphicBundle())
  {
    graphicContext.setGraphicBundle(rect.graphicBundle());
  }

  // Draw box.
  rc = GpiBox(graphicContext.handle(), DRO_OUTLINE, &ptl,
              rect.rounding().coord1(), rect.rounding().coord2() );
  if (rc == GPI_ERROR)
  {
    ITHROWGUIERROR2("GpiBox",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  // End the path.
  if (!GpiEndPath(graphicContext.handle()))
    ITHROWGUIERROR2("GpiEndPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Convert a path to a region. Fill the path using the winding rule.
  newRegion = GpiPathToRegion(graphicContext.handle(),1L,FPATH_WINDING);
  return newRegion;
}

/*------------------------------------------------------------------------------
| IGRegion::doPie                                                              |
| Converts a pie to a pie region. Return a region handle.                      |
|                                                                              |
------------------------------------------------------------------------------*/
IRegionHandle IGRegion::doPie( IGraphicContext& graphicContext, const IGPie& pie )
{
  IRegionHandle newRegion;
  unsigned long ps(graphicContext.handle());
  IRectangle rect(pie.enclosingRect());

  // The GPI arc parameters decide the shape and the orientation of the arc.
  // See GPI Programming Guide page 3-12 for details.
  ARCPARAMS arcp = { rect.rightCenter().x() - rect.center().x(),
                     rect.topCenter().y()   - rect.center().y(), 0,0 };
  POINTL    ptl( rect.center().asPOINTL() );

  // Specify the start of a path.
  if (!GpiBeginPath(graphicContext.handle(), 1L ))
    ITHROWGUIERROR2("GpiBeginPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Set the transformation matrix and graphic bundle of pie to the graphic context.
  if (pie.hasTransformMatrix())
  {
    graphicContext.setWorldTransformMatrix(pie.transformMatrix(), pie.transformMethod());
  }

  if (pie.hasGraphicBundle())
  {
    graphicContext.setGraphicBundle(pie.graphicBundle());
  }
  // Set the current arc parameters in PM.
  if (!GpiSetArcParams( ps, &arcp ))
  {
    ITHROWGUIERROR2("GpiSetArcParams",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  graphicContext.setCurrentDrawingPosition( rect.center() );

  if (GpiPartialArc( ps, &ptl, MAKEFIXED(1, 0),
                     DBL2FIXED(pie.startAngle()),
                     DBL2FIXED(pie.sweepAngle()) ) == GPI_ERROR)
  {
    ITHROWGUIERROR2("GpiPartialArc",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  // Draw the closing line.
  if ( GpiLine( ps, &ptl ) == GPI_ERROR )
  {
    ITHROWGUIERROR2("GpiLine",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  // Finish path.
  if (!GpiEndPath(graphicContext.handle()))
    ITHROWGUIERROR2("GpiEndPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Convert a pie path to a region. Fill the path using alternate rule.
  newRegion = GpiPathToRegion(ps,1L,FPATH_ALTERNATE);
  return newRegion;
}

/*------------------------------------------------------------------------------
| IGRegion::doChord                                                            |
| Converts a chord to a chord region. Return a region handle.                  |
|                                                                              |
------------------------------------------------------------------------------*/
IRegionHandle IGRegion::doChord( IGraphicContext& graphicContext, const IGChord& chord )
{
  IRegionHandle newRegion;
  unsigned long ps(graphicContext.handle());
  IRectangle rect(chord.enclosingRect());

  // The GPI arc parameters decide the shape and the orientation of the arc.
  // See GPI Programming Guide page 3-12 for details.
  ARCPARAMS arcp = { rect.rightCenter().x() - rect.center().x(),
                     rect.topCenter().y()   - rect.center().y(), 0,0 };
  POINTL    ptl( rect.center().asPOINTL() );

  // Set the transformation matrix and graphic bundle of chord to the graphic context.
  if (chord.hasTransformMatrix())
  {
    graphicContext.setWorldTransformMatrix(chord.transformMatrix(), chord.transformMethod());
  }

  if (chord.hasGraphicBundle())
  {
    graphicContext.setGraphicBundle(chord.graphicBundle());
  }
  if (!GpiSetArcParams( ps, &arcp ))
  {
    ITHROWGUIERROR2("GpiSetArcParams",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  graphicContext.setCurrentDrawingPosition( rect.center() );

  // Move the pen without actually drawing to the end of the chord.
  IGraphicBundle::PenType currentPenType(graphicContext.penType());
  graphicContext.setPenType( IGraphicBundle::invisible );
  if ( GpiPartialArc( ps, &ptl, MAKEFIXED(1, 0),
                      DBL2FIXED(chord.startAngle() + chord.sweepAngle()),
                      (FIXED)0 ) == GPI_ERROR )
  {
    ITHROWGUIERROR2("GpiPartialArc",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  graphicContext.setPenType( currentPenType );
  // Specify the start of a path.
  if (!GpiBeginPath(graphicContext.handle(), 1L ))
    ITHROWGUIERROR2("GpiBeginPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Draw the entire chord, including the arc, in one shot.
  if ( GpiPartialArc( ps, &ptl, MAKEFIXED(1, 0),
                      DBL2FIXED(chord.startAngle()),
                      DBL2FIXED(chord.sweepAngle()) ) == GPI_ERROR )
  {
    ITHROWGUIERROR2("GpiPartialArc",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  // Finish the path.
  if (!GpiEndPath(graphicContext.handle()))
    ITHROWGUIERROR2("GpiEndPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Converts the chord path to a region. Fill the path using alternate rule.
  newRegion = GpiPathToRegion(ps,1L,FPATH_ALTERNATE);
  return newRegion;
}

/*------------------------------------------------------------------------------
| IGRegion::doPolygon                                                          |
|  Converts a polygon to a chord region. Return a region handle.               |
|                                                                              |
------------------------------------------------------------------------------*/
IRegionHandle IGRegion::doPolygon( IGraphicContext& graphicContext, const IGPolygon& polygon )
{
  IRegionHandle newRegion;
  unsigned long ps(graphicContext.handle());
  // Retrieve the filling mode.
  LONG fillMode = (polygon.fillMode() == IGPolygon::alternate
                                          ? FPATH_ALTERNATE
                                          : FPATH_WINDING);
  // Specify the start of a path.
  if (!GpiBeginPath(graphicContext.handle(), 1L ))
    ITHROWGUIERROR2("GpiBeginPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  if (polygon.numberOfPoints() > 1)
  {
    unsigned long  arraySize(polygon.pointArray().size());
    PPOINTL        pPtl(new POINTL[arraySize-1]);  //array of POINTL.
    POINTL         startPt(polygon.point(0).asPOINTL());

    // Set the transformation matrix and graphic bundle of polygon to the graphic context.
    if (polygon.hasTransformMatrix())
    {
      graphicContext.setWorldTransformMatrix(polygon.transformMatrix(), polygon.transformMethod());
    }
    if (polygon.hasGraphicBundle())
    {
      graphicContext.setGraphicBundle(polygon.graphicBundle());
    }

    // Converts all IPoint to POINTL structure for PM.
    for (int i=1; i < arraySize; i++)
    {
      pPtl[i-1] = polygon.point(i).asPOINTL();
    } /* endfor */

    graphicContext.setCurrentDrawingPosition( polygon.pointArray()[0] );

    // GPI draw polygon.
    if (GpiPolyLine( ps, arraySize-1, pPtl ) == GPI_ERROR )
    {
      ITHROWGUIERROR2("GpiPolyLine",
                      IErrorInfo::invalidRequest,
                      IException::recoverable);
    }

    // Draw the closing line.
    if (GpiLine( ps, &startPt ) == GPI_ERROR )
    {
      ITHROWGUIERROR2("GpiLine",
                      IErrorInfo::invalidRequest,
                      IException::recoverable);
    }

    delete [] pPtl;
  }
  // Finish the path.
  if (!GpiEndPath(graphicContext.handle()))
    ITHROWGUIERROR2("GpiEndPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Converts the polygon path to a region. Fill the path using polygon's
  // filling rule.
  newRegion = GpiPathToRegion(ps,1L,fillMode);

  return newRegion;
}

/*------------------------------------------------------------------------------
| IGRegion::doEllipse                                                          |
|  Converts an ellipse to a chord region. Return a region handle.              |
|                                                                              |
------------------------------------------------------------------------------*/
IRegionHandle IGRegion::doEllipse( IGraphicContext& graphicContext, const IGEllipse& ellipse )
{
  IRegionHandle newRegion;
  unsigned long ps(graphicContext.handle());

  IRectangle bRect(ellipse.enclosingRect());
  POINTL     centerPt(bRect.center().asPOINTL());

  // Specify the start of a path.
  if (!GpiBeginPath(graphicContext.handle(), 1L ))
    ITHROWGUIERROR2("GpiBeginPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);

  // Set the transformation matrix and graphic bundle of ellipse to the graphic context.
  if (ellipse.hasTransformMatrix())
  {
    graphicContext.setWorldTransformMatrix(ellipse.transformMatrix(), ellipse.transformMethod());
  }

  if (ellipse.hasGraphicBundle())
  {
    graphicContext.setGraphicBundle(ellipse.graphicBundle());
  }

  GpiMove(ps, &centerPt);

  // Set the current arc parameters in PM.
  // The GPI arc parameters decide the shape and the orientation of the arc.
  // See GPI Programming Guide page 3-12 for details.
  ARCPARAMS arcp = { 0, 0, bRect.rightCenter().x() - bRect.center().x(),
                           bRect.topCenter().y() - bRect.center().y() };
  if (!GpiSetArcParams( ps, &arcp ))
  {
    ITHROWGUIERROR2("GpiSetArcParams",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  if (GpiFullArc( ps, DRO_OUTLINE, MAKEFIXED(1, 0)) == GPI_ERROR)
  {
    ITHROWGUIERROR2("GpiFullArc",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  }

  // Finish the path.
  if (!GpiEndPath(graphicContext.handle()))
    ITHROWGUIERROR2("GpiEndPath",
                    IErrorInfo::invalidRequest,
                    IException::recoverable);
  // Converts the chord path to a region. Fill the path using winding rule.
  newRegion = GpiPathToRegion(ps,1L,FPATH_WINDING);
  return newRegion;
}

// All functions below are private. The purpose of making these virtual
// functions private is to explicitly block the use of these functions in IGRegion.
/*------------------------------------------------------------------------------
| IGRegion::rotateBy                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::rotateBy( double angle, const IPoint &pt )
{
  angle = angle;
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::scaleBy                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::scaleBy( double xScale, double yScale, const IPoint& pt )
{
  xScale = yScale;
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::translateBy                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::translateBy( const IPoint& pt )
{
  return *this;
}

/*------------------------------------------------------------------------------
| IGRegion::setTransformMatrix                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGRegion& IGRegion::setTransformMatrix( const ITransformMatrix& transformMatrix )
{
  return *this;
}
