/*******************************************************************************
* FILE NAME: igroupbx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igroupbx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINERRORS
  #define INCL_WINSTATICS
  #define INCL_GPIPRIMITIVES
  #define INCL_WINSYS
  #include <iwindefs.h>
}

#ifndef _IGROUPBX_
  #include <igroupbx.hpp>
#endif
#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _ICOLOR_
  #include <icolor.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFONT_
  #include <ifont.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGROUPBX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public group box styles                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IGroupBox::Style IGroupBox::classDefaultStyle = WS_VISIBLE;
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new group box objects (initial value).                     |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IGroupBox::Style IGroupBox::currentDefaultStyle = WS_VISIBLE;
#pragma data_seg()


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to create a group box control on a standard window.              |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(unsigned long ulId,
                       IWindow* pwndParent,
                       IWindow* pwndOwner,
                       const IRectangle& rectInit,
                       const Style& gbsStyle)
                     : fGroupBoxData( 0 )
{
  IASSERTPARM(pwndParent!=0);          //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | gbsStyle.asExtendedUnsignedLong() );

  //Create group box control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle whGroupBox =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( gbsStyle ),
                      WC_STATIC,
                      pwndParent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0 );


  startHandlingEventsFor(whGroupBox);
}


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to create instance for a template window.                        |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(unsigned long ulId,
                       IWindow* pdlgwndParent)
                     : fGroupBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to instantiate from an existing group box control.               |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(const IWindowHandle& wh)
                     : fGroupBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(wh);
}


/*------------------------------------------------------------------------------
| IGroupBox::~IGroupBox                                                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IGroupBox :: ~IGroupBox()
{
  // delete fGroupBoxData;
}


/*------------------------------------------------------------------------------
| IGroupBox::defaultStyle                                                      |
|                                                                              |
| Return the default style for new group box objects.                          |
------------------------------------------------------------------------------*/
IGroupBox::Style  IGroupBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IGroupBox::setDefaultStyle                                                   |
|                                                                              |
| Replace the default style for new group box objects.                         |
------------------------------------------------------------------------------*/
void  IGroupBox :: setDefaultStyle(const IGroupBox::Style& gbsStyle)
{
  currentDefaultStyle = gbsStyle;
}


/*------------------------------------------------------------------------------
| IGroupBox::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IGroupBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean styles flow thru and add the groupbox style
    ulStyle |= guiStyle.asUnsignedLong();
    ulStyle |= SS_GROUPBOX;
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IGroupBox::calcMinimumSize                                                   |
------------------------------------------------------------------------------*/
ISize  IGroupBox :: calcMinimumSize() const
{
  ISize siz = displaySize(text());

  /*----------------------------------------------------------------------------
  | We need to pad the width of 2 characters to allow for the border           |
  ----------------------------------------------------------------------------*/

  int pad = IFont(this).avgCharWidth() << 1;
  siz.setWidth(siz.width() + pad);


  return siz;
}

/*------------------------------------------------------------------------------
| IGroupBox::foregroundColor                                                   |
|                                                                              |
| Returns the foreground color of the IGroupBox.                               |
------------------------------------------------------------------------------*/
IColor IGroupBox::foregroundColor () const
{
  return (IWindow::color(PP_FOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::windowStaticText)));
}

/*------------------------------------------------------------------------------
| IGroupBox::setText                                                           |
|                                                                              |
| Changes the text used by the group box.                                      |
------------------------------------------------------------------------------*/
IGroupBox &  IGroupBox :: setText ( const char* pszText )
{
  Inherited::setText(pszText);
  setLayoutDistorted(IWindow::minimumSizeChanged, 0);
  return *this;
}


/*------------------------------------------------------------------------------
| IGroupBox::setText                                                           |
|                                                                              |
| Changes the group box text from a string table.                              |
------------------------------------------------------------------------------*/
IGroupBox &  IGroupBox :: setText ( const IResourceId & residText )
{
  Inherited::setText(residText);
  return *this;
}

/*------------------------------------------------------------------------------
| IGroupBox::visibleRectangle                                                  |
|                                                                              |
| Returns the area painted by the group box.                                   |
------------------------------------------------------------------------------*/
IRectangle IGroupBox :: visibleRectangle () const
{
  return IRectangle(position(),ISize(0,0));
}

