/*******************************************************************************
* FILE NAME: igstring.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igstring.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <igstring.hpp>
#include <igrafctx.hpp>
#include <igbundle.hpp>
#include <ihandle.hpp>
#include <ireslib.hpp>
#include <ifont.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGSTRING_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IGString& graphicString )
  : IGString::Inherited( graphicString ), fString( graphicString.fString ),
    fLocation( graphicString.fLocation ), fFont(0), fClip(0)
{
  if (graphicString.fFont)   // If font is not set...
    fFont = new IFont(*(graphicString.fFont)); // Copy IFont over.
}

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
| Constructor using an IString and a position point.                           |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IString& string, const IPoint& point )
  : IGString::Inherited(), fString( string ), fLocation( point ), fFont(0), fClip(0)
{
}

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
| Constructor using an IString, a position point, and a font.                  |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IString& string, const IPoint& point,
                    const IFont& font )
  : IGString::Inherited(), fString( string ), fLocation( point ),
    fFont( new IFont( font )), fClip(0)
{
}

/*------------------------------------------------------------------------------
| IGString::~IGString                                                          |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::~IGString()
{
  if (fFont)
  {
    delete fFont;
  }
}


/*------------------------------------------------------------------------------
| IGString::drawOn                                                             |
| Draw itself on a given graphicContext. Return a reference of this object.    |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::drawOn(IGraphicContext& graphicContext)
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::font                                                               |
| Return an IFont object used by the IGString.                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IFont IGString::font() const
{
  if ( fFont )
    return IFont( *fFont );
  else
    return IFont();      // Default font
}

/*------------------------------------------------------------------------------
| IGString::setFont                                                            |
| Set IString's font. Return a reference of this object.                       |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setFont( const IFont& font )
{
  if ( fFont )
  {
    delete fFont;
  }
  fFont = new IFont(font);
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::hasFont                                                            |
| Return whether there is a predefined IFont associated with the IGString.     |
|-----------------------------------------------------------------------------*/
IBase::Boolean IGString::hasFont() const
{
  return ((fFont == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGString::removeFont                                                         |
| Release the associated IFont. Return a reference of this object.             |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::removeFont()
{
  if ( fFont )
  {
    delete fFont;
    fFont = 0;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::moveTo                                                             |
| Move the IGString position to point. Return a reference of this object.      |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::moveTo ( const IPoint& point )
{
  fLocation = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::position                                                           |
| Return the position of the IGString. Return a reference of this object.      |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGString::position( ) const
{
  return fLocation;
}

/*------------------------------------------------------------------------------
| IGString::setClippingRect                                                    |
| Set the clipping rectangle of IGString. Return a reference of this object.   |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setClippingRect   ( const IRectangle& rect )
{
  if (fClip)
    delete fClip;  // Delete the old clip rect.
  fClip = new IRectangle( rect );
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::clippingRect                                                       |
| Get the clipping rectangle of IGString.                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGString::clippingRect       ( ) const
{
  if (fClip)
    return *fClip;

  return IRectangle();  // Default clipping rect.

}

/*------------------------------------------------------------------------------
| IGString::resetClippingRect                                                  |
| Reset or remove the clipping rectangle of IString.                           |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
IGString& IGString::resetClippingRect ( )
{
  if (fClip)
  {
    delete fClip;
    fClip = 0;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::isClippingRectSet                                                  |
| Test whether the clipping rectangle is set.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::isClippingRectSet  ( ) const
{
  return ((fClip == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGString::operator=                                                          |
| Assignment operator. Return a reference of this object.                      |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::operator=( const IGString& graphicString )
{
  // If assignment to itself, return a reference to itself.
  if ( this == &graphicString )
    return *this;

  // Call base class' overrided member function.
  this->Inherited::operator=( graphicString );
  // Member-wise assignment.
  fString = graphicString.fString;
  fLocation = graphicString.fLocation;
  if ( fFont )   // Clear old font.
  {
    delete fFont;
    fFont = 0;
  }
  if ( graphicString.fFont )  // Assign new font.
  {
    fFont = new IFont(*graphicString.fFont);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::operator==                                                         |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::operator==( const IGString& graphicString ) const
{
  if ( fFont && graphicString.fFont )
  {
    return ( fString   == graphicString.fString    &&
             fLocation == graphicString.fLocation  &&
             this->Inherited::operator==( graphicString ));
  }
  else if (!fFont && !graphicString.fFont)
  {
    return ( fString    == graphicString.fString    &&
             fLocation  == graphicString.fLocation  &&
             this->Inherited::operator==( graphicString ));
  }
  else
  {
    return false;
  }
}

/*------------------------------------------------------------------------------
| IGString::operator!=                                                         |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::operator!=( const IGString& graphicString ) const
{
  return !( *this == graphicString );
}

/*------------------------------------------------------------------------------
| IGString::setText                                                            |
| Set text to IGString. Return a reference of this object.                     |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setText( const char* text )
{
  fString = text;
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::setText                                                            |
| Set a text resource id to the IGString. Return a reference of this object.   |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setText( const IResourceId& residText )
{
  // Load a text string from resource library.
  IString str = residText.resourceLibrary().loadString(residText);
  setText((char*)str);
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::text                                                               |
| Return the text string of IGString.                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IString IGString::text( ) const
{
  return fString;
}
