/*******************************************************************************
* FILE NAME: ihandle.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ihandle.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINPOINTERS
  #define INCL_GPIBITMAPS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  #include <malloc.h>
  }

#include <ihandle.hpp>

#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

#ifndef _ISTREAM_
  #include <istream.hpp>
#endif

#ifndef _ITHREAD_
  #include <ithread.hpp>
#endif
#ifndef _IWINDOW_
  #include <iwindow.hpp>
#endif

#ifndef _IWCNAME_
  #include <iwcname.hpp>
#endif

#ifndef _ICCONST_
  #include <icconst.h>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHANDLE_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
 isValid - Is this a valid window handle for the presentation system?
------------------------------------------------------------------------------*/
Boolean IWindowHandle::isValid ( ) const
{
  Boolean valid = IQUERYANCHOR( this->handle );
  if ( ! valid )
  {
     // Handle the slim chance that the application is relying on this
     // function to call WinInitialize via ICurrentThread::anchorBlock.
     valid = IISWINDOW( IThread::current().anchorBlock(), this->handle );
  }
  return valid;
}


/*------------------------------------------------------------------------------
 sendEvent - send a window system message/event to the window
 represented by this handle, using the specified parameters.
------------------------------------------------------------------------------*/
IEventResult IWindowHandle::sendEvent  ( unsigned long eventId,
            const IEventParameter1 &parm1, const IEventParameter2 &parm2 ) const
{
   MRESULT rs = ISENDMESSAGE(this->handle, eventId, parm1, parm2);
   return IEventResult((int)rs); //JJB NEED TO ADD A LONG CONSTRUCTOR
}

/*------------------------------------------------------------------------------
 sendEvents -  send a window system message/event to the children
 or decendents of the window represented by this handle,
 using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::sendEvents ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2,
         BroadcastTo value ) const
{
   unsigned long ulCmd = BMSG_SEND;
   switch(value)
   {
     case descendants:
       ulCmd |= BMSG_DESCENDANTS;
       break;
     case frames:
       ulCmd |= BMSG_FRAMEONLY;
       break;
     default:
       ulCmd |= (BMSG_FRAMEONLY | BMSG_DESCENDANTS);
       break;
   } /* endSwitch */
   if (!WinBroadcastMsg(this->handle, eventId, parm1, parm2, ulCmd))
      ITHROWGUIERROR2("WinBroadcastMsg", IErrorInfo::invalidParameter, IException::recoverable);
}


/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to the window
 represented by this handle, using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::postEvent  ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2 ) const
{
   if (!IPOSTMESSAGE(this->handle, eventId, parm1, parm2))
      ITHROWGUIERROR2("WinPostMsg", IErrorInfo::invalidParameter, IException::recoverable);
}

/*------------------------------------------------------------------------------
 postEvents -  post (asynchronously)  a window system message/event to
 the  children/decendents of the window represented by this handle,
 using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::postEvents ( unsigned long eventId,
      const IEventParameter1 &parm1, const IEventParameter2 &parm2,
      BroadcastTo value) const
{
   unsigned long ulCmd = BMSG_POST;
   switch(value)
   {
     case descendants:
       ulCmd |= BMSG_DESCENDANTS;
       break;
     case frames:
       ulCmd |= BMSG_FRAMEONLY;
       break;
     default:
       ulCmd |= (BMSG_FRAMEONLY | BMSG_DESCENDANTS);
       break;
   } /* endSwitch */

   if (!WinBroadcastMsg(this->handle, eventId, parm1, parm2, ulCmd))
      ITHROWGUIERROR2("WinBroadcastMsg", IErrorInfo::invalidParameter, IException::recoverable);
}

/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to this
 message queue , using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvent  ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2) const
{
   if (!IPOSTQUEUEMESSAGE(this->handle, eventId, parm1, parm2))
      ITHROWGUIERROR2("WinPostQueueMsg", IErrorInfo::invalidParameter, IException::recoverable);
}
/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to the
 specified message queue handle, using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvent  ( const IMessageQueueHandle& handle,
         unsigned long eventId, const IEventParameter1 &parm1,
         const IEventParameter2 &parm2)
{
   if (!WinPostQueueMsg(handle, eventId, parm1, parm2))
      ITHROWGUIERROR2("WinPostQueueMsg", IErrorInfo::invalidParameter, IException::recoverable);
}

/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to
 all threads that have a message queue, using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvents ( unsigned long eventId,
      const IEventParameter1 &parm1, const IEventParameter2 &parm2)
{
   if (!WinBroadcastMsg(0, eventId, parm1, parm2, BMSG_POSTQUEUE))
      ITHROWGUIERROR2("WinBroadcastMsg", IErrorInfo::invalidParameter, IException::recoverable);
}
