/*******************************************************************************
* FILE NAME: imenubar.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imenubar.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINFRAMEMGR
  #define INCL_WINWINDOWMGR
  #define INCL_WINMENUS
  #include <iwindefs.h>
}

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif
#ifndef _IFRAME_
  #include <iframe.hpp>
#endif
#ifndef _IMENU_
  #include <imenubar.hpp>
#endif
#ifndef _IEXCEPT_
 #include <iexcept.hpp>
#endif
#ifndef _ITRACE_
 #include <itrace.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMENUBAR_CPP_
  #include <ipagetun.h>
#endif

#pragma data_seg(ICLStaticConst)
const IMenuBar::Style
  IMenuBar::empty             ( 0, IMBS_EMPTY ),
  IMenuBar::wrapper           ( 0, IMBS_WRAPPER ),
  IMenuBar::classDefaultStyle ( 0, IMBS_EMPTY );
#pragma data_seg()

#pragma data_seg(ICLNonConst)
  IMenuBar::Style
              IMenuBar::currentDefaultStyle ( 0, IMBS_EMPTY );
#pragma data_seg()

/*------------------------------------------------------------------------------
| IMenuBar::IMenuBar                                                           |
|  Create empty or wrapper                                                     |
------------------------------------------------------------------------------*/
IMenuBar::IMenuBar( IFrameWindow*      owner,
                    const Style&       style )
                  : fMenuBarData( 0 )
{
   IASSERTPARM(owner != 0);

   // Set extended style
   setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

   /* Check for both empty and wrapper being set,
    * or neither set.
    */
   if ( ( (style & wrapper) && (style & empty) ) ||       // both set
        ( !(style & (wrapper | empty)) ) )     // neither set
   {
     ITHROWLIBRARYERROR(IC_INVALIDSTYLE,
                        IErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   HWND frameMenuBar = handleWithParent(FID_MENU, owner->handle());

   if (style & empty)
   {
     IWindowHandle whMenu =
          this -> create( FID_MENU,
                          0,
                          style.asUnsignedLong() | MS_ACTIONBAR |
                          WS_VISIBLE | WS_SYNCPAINT,
                          WC_MENU,
                          owner->handle(),
                          owner->handle(),
                          IRectangle(),
                          0,
                          0,
                          IWindow::onTopOfSiblings );

     if (frameMenuBar)
        WinDestroyWindow(frameMenuBar);

     owner->postEvent(WM_UPDATEFRAME,            // Cause frame to update
                      IEventData(FCF_MENU, 0),   // to include menu.
                      IEventData());

     startHandlingEventsFor(whMenu);
   }
   else    // it's a wrapper
   {
     if (frameMenuBar == 0) {
       ITHROWLIBRARYERROR(IC_NO_FRAME_MENUBAR,
                          IErrorInfo::invalidRequest,
                          IException::recoverable);
     } /* endif */

     setAutoDestroyWindow(false);
     startHandlingEventsFor(frameMenuBar);

   }
}

/*------------------------------------------------------------------------------
| IMenuBar::IMenuBar                                                           |
|  Create the menu from resource file                                          |
------------------------------------------------------------------------------*/
IMenuBar::IMenuBar( const IResourceId& menuResId,
                    IFrameWindow*      owner )
                  : fMenuBarData( 0 )
{
   IMODTRACE_DEVELOP("IMenuBar::ctor(2)");
   IASSERTPARM(owner != 0);

   HWND frameMenuBar = handleWithParent(FID_MENU, owner->handle());
   IWindowHandle whMenu =
     (menuResId.resourceLibrary()).loadMenu(menuResId, owner);

   if(frameMenuBar)                     //Destroy the old FID_MENU
      WinDestroyWindow(frameMenuBar);

   owner->postEvent(WM_UPDATEFRAME,            // Cause frame to update
                    IEventData(FCF_MENU, 0),   // to include menu.
                    IEventData());

   startHandlingEventsFor(whMenu);
}


/*------------------------------------------------------------------------------
| IMenuBar::~IMenuBar                                                          |
------------------------------------------------------------------------------*/
IMenuBar::~IMenuBar()
{}

/*------------------------------------------------------------------------------
| IMenuBar::setMenu                                                            |
|  Set a new menu resource ID. The previous menu-bar is destroyed.             |
------------------------------------------------------------------------------*/
IMenuBar& IMenuBar::setMenu(const IResourceId& menuResId)
{
   IWindowHandle hwndMenu =
     (menuResId.resourceLibrary()).loadMenu(menuResId, owner());

   if (id() != FID_MENU) {
      WinSetWindowUShort(hwndMenu,  //Menu window ID is create with
                         QWS_ID,    //FID_MENU. Change it.
                         (unsigned short)id());
   } else {
     IWindow* powner = owner();
     if (isValid())
         WinDestroyWindow(handle());
      powner->postEvent(WM_UPDATEFRAME,              // Cause frame to update
                       IEventData(FCF_MENU, 0),     // to include menu.
                       IEventData());
   } /* endif */
   startHandlingEventsFor(hwndMenu);
   return *this;
}

/*------------------------------------------------------------------------------
| IMenuBar::Style IMenuBar::defaultStyle                                       |
------------------------------------------------------------------------------*/
IMenuBar::Style IMenuBar::defaultStyle()
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| IMenuBar::convertToGUIStyle                                                  |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IMenuBar::convertToGUIStyle( const IBitFlag& guiStyle,
                                           Boolean         bExtOnly ) const
{
  // Obtain the style from the class (IMenu) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to return extended styles in the user defined range and
    // extended styles for IMenuBar, IMBS_, which have a one-to-one
    // correspondence to our style bits
    ulStyle |= extendedStyle() & IMBS_EXTMASK;
  }
  else
  {
    // No base styles for IMenuBar, but use mask for future expandability
    ulStyle |= guiStyle.asUnsignedLong() & IMBS_MASK;
  }

  return( ulStyle );
}

