/*******************************************************************************
* FILE NAME: inotebk.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                              *
*     IPageHandle                                                              *
*     INotebook                                                                *
*                                                                              *
*   See inotebk0.cpp for the implementation of the platform independent        *
*   nested classes:                                                            *
*     INotebook::Cursor                                                        *
*     INotebook::PageSettings                                                  *
*                                                                              *
* NOTES:                                                                       *
*  1. This class is used in conjunction with IPage.                            *
*  2. Methods that have a 0 default for an IPage pointer will use the top      *
*     page as the default.                                                     *
*  3. When a pages is removed from a notebook, the IPage object still          *
*     exists...  However, the following data (normally maintained by the       *
*     notebook) is lost:                                                       *
*      - status text (or length)                                               *
*      - tab text (or length)                                                  *
*      - tab bitmap                                                            *
*     Therefore, removing a page from a notebook, and adding it back again     *
*     currently requires that the user reset the bitmap, status and tab text.  *
*     bitmap, status and tab text.                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_DOSMODULEMGR
  #define INCL_WINSYS
  #define INCL_WINENTRYFIELDS
  #define INCL_WINCLIPBOARD
  #define INCL_WININPUT             // WM_CHAR, etc.
  #define INCL_WINMESSAGEMGR        // for WNDPARAM struct
  #define INCL_NLS                  // ES_ANY, ES_SBCS, ES_DBCS, ES_MIXED
  #define INCL_WINSTDDRAG
  #define INCL_WINWINDOWMGR         // WinQueryWindowULong
  #define INCL_WINSTDBOOK
  #define INCL_WINDIALOGS           // Needed for PDLGTEMPLATE in BOOKPAGEINFO
  #define INCL_GPI
  #include <iwindefs.h>
}

#ifndef _INOTEBK_
  #include <inotebk.hpp>
#endif
#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _ICOLOR_
  #include <icolor.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFRAME_
  #include <iframe.hpp>
#endif
#ifndef _IMPHDR_
  #include <imphdr.hpp>
#endif
#ifndef _INOTEHDR_
  #include <inotehdr.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _INOTEBK_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public notebook styles.                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const INotebook::Style
  INotebook::backPagesBottomRight = BKS_BACKPAGESBR,            // 0x00000001
  INotebook::backPagesBottomLeft  = BKS_BACKPAGESBL,            // 0x00000002
  INotebook::backPagesTopRight    = BKS_BACKPAGESTR,            // 0x00000004
  INotebook::backPagesTopLeft     = BKS_BACKPAGESTL,            // 0x00000008
  INotebook::majorTabsRight       = BKS_MAJORTABRIGHT,          // 0x00000010
  INotebook::majorTabsLeft        = BKS_MAJORTABLEFT,           // 0x00000020
  INotebook::majorTabsTop         = BKS_MAJORTABTOP,            // 0x00000040
  INotebook::majorTabsBottom      = BKS_MAJORTABBOTTOM,         // 0x00000080
  INotebook::squareTabs           ( 0, IBKS_SQUARETABS ),       // 0x00000001 *
  INotebook::roundedTabs          = BKS_ROUNDEDTABS,            // 0x00000100
  INotebook::polygonTabs          = BKS_POLYGONTABS,            // 0x00000200
  INotebook::solidBinding         ( 0, IBKS_SOLIDBIND ),        // 0x00000002 *
  INotebook::spiralBinding        = BKS_SPIRALBIND,             // 0x00000400
  INotebook::statusTextLeft       ( 0, IBKS_STATUSTEXTLEFT ),   // 0x00000004 *
  INotebook::statusTextRight      = BKS_STATUSTEXTRIGHT,        // 0x00001000
  INotebook::statusTextCenter     = BKS_STATUSTEXTCENTER,       // 0x00002000
  INotebook::tabTextLeft          ( 0, IBKS_TABTEXTLEFT ),      // 0x00000008 *
  INotebook::tabTextRight         = BKS_TABTEXTRIGHT,           // 0x00004000
  INotebook::tabTextCenter        = BKS_TABTEXTCENTER,          // 0x00008000
  INotebook::pmCompatible         ( 0, IBKS_PMCOMPATIBLE),      // 0x00000010 *
  INotebook::classDefaultStyle  ( BKS_BACKPAGESBR |
                                  BKS_MAJORTABRIGHT |
                                  BKS_TABTEXTCENTER |
                                  WS_VISIBLE ,
                                  IBKS_SQUARETABS |
                                  IBKS_SOLIDBIND |
                                  IBKS_STATUSTEXTLEFT );
                               // 0x00000001 backPagesBottomRight
                               // 0x00000010 majorTabsRight
                               // 0x00008000 tabTextCenter
                               // 0x80000000 visible
                                             // 0x00000001 squareTabs
                                             // 0x00000002 solidBinding
                                             // 0x00000004 statusTextLeft

const INotebook::PageSettings::Attribute
  INotebook::PageSettings::noAttribute  = 0,                    // 0x00000000
  INotebook::PageSettings::statusTextOn = BKA_STATUSTEXTON,     // 0x00000001
  INotebook::PageSettings::majorTab     = BKA_MAJOR,            // 0x00000040
  INotebook::PageSettings::minorTab     = BKA_MINOR,            // 0x00000080
  INotebook::PageSettings::autoPageSize = BKA_AUTOPAGESIZE;     // 0x00000100

const INotebook::clrFlags
  INotebook::bgnPageColor  = IBKA_BACKGROUNDPAGECOLOR,          // 0x00000001 *
  INotebook::bgnMajorColor = IBKA_BACKGROUNDMAJORCOLOR,         // 0x00000002 *
  INotebook::bgnMinorColor = IBKA_BACKGROUNDMINORCOLOR,         // 0x00000004 *
  INotebook::fgnMajorColor = IBKA_FOREGROUNDMAJORCOLOR,         // 0x00000008 *
  INotebook::fgnMinorColor = IBKA_FOREGROUNDMINORCOLOR;         // 0x00000010 *
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new notebook objects (initial value).                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  INotebook::Style
       INotebook::currentDefaultStyle ( BKS_BACKPAGESBR |
                                        BKS_MAJORTABRIGHT |
                                        BKS_TABTEXTCENTER |
                                        WS_VISIBLE ,
                                        IBKS_SQUARETABS |
                                        IBKS_SOLIDBIND |
                                        IBKS_STATUSTEXTLEFT );
#pragma data_seg()

  #define UPAGE   (unsigned long)page

/*------------------------------------------------------------------------------
| INotebookHandler                                                             |
|                                                                              |
| This class wrappers an object pointer and destroys the pointed to object     |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator INotebookHandler* ( );
INotebookHandler
 *operator -> ( )
    {
    return *this;
    }
 ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
INotebookHandler
 *ptr;
} defaultHandler;

DefaultHandler :: operator INotebookHandler* ( )
    {
    if ( !ptr )
      ptr = new INotebookHandler;
    return ptr;
    }

/*------------------------------------------------------------------------------
| ITabBitmapMgr                                                                |
|                                                                              |
| This class is used to manage the tab bitmaps for a notebook's pages.         |
------------------------------------------------------------------------------*/
class ITabBitmapMgr {
public:
  ITabBitmapMgr ( const IBitmapHandle& tabBitmap,
                  const IPageHandle&   pgHandle );
virtual
 ~ITabBitmapMgr ( );
void
  setNext ( ITabBitmapMgr* );
ITabBitmapMgr
 *next    ( );
private:
IBitmapHandle
  bitmapHandle;
IPageHandle
  pageHandle;
ITabBitmapMgr
 *nextInList;
};

/*------------------------------------------------------------------------------
| ITabBitmapMgr::ITabBitmapMgr                                                 |
------------------------------------------------------------------------------*/
ITabBitmapMgr :: ITabBitmapMgr ( const IBitmapHandle& tabBitmap,
                                 const IPageHandle&   pgHandle )
{
  bitmapHandle = tabBitmap;
  pageHandle = pgHandle;
  nextInList = 0;
}

/*------------------------------------------------------------------------------
| ITabBitmapMgr::~ITabBitmapMgr                                                |
------------------------------------------------------------------------------*/
ITabBitmapMgr :: ~ITabBitmapMgr ( )
{ }

/*------------------------------------------------------------------------------
| ITabBitmapMgr::setNext                                                       |
------------------------------------------------------------------------------*/
void ITabBitmapMgr :: setNext ( ITabBitmapMgr* nextOne )
{
  nextInList = nextOne;
}

/*------------------------------------------------------------------------------
| ITabBitmapMgr::next                                                          |
------------------------------------------------------------------------------*/
ITabBitmapMgr* ITabBitmapMgr :: next ( )
{
  return nextInList;
}

/*------------------------------------------------------------------------------
| IPageHandle::IPageHandle                                                     |
|                                                                              |
| Class constructor                                                            |
------------------------------------------------------------------------------*/
IPageHandle :: IPageHandle ( Value pageId )
  : IHandle( pageId )
{ }


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Construct a notebook from a window id, parent, and owner window handles,     |
| and initial size rectangle, and a style.                                     |
------------------------------------------------------------------------------*/
INotebook :: INotebook ( unsigned long windowId,
                         IWindow* parent,
                         IWindow* pOwner,
                         const IRectangle& initial,
                         const Style& style )
  : fNotebookData(0),
    bmClTabBitmapMgr(0),
    pnotebookColors(0),
    colorFlags(0),
    ulClValidate(0)
{
  // assertions on input parms
  IASSERTPARM(parent!=0);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  IWindowHandle owner = (pOwner == 0) ? IWindowHandle(0) : pOwner->handle();


  IWindowHandle whNB =
      this -> create( windowId,
                      0,
                      convertToGUIStyle( style ),
                      WC_NOTEBOOK,
                      parent->handle(),
                      owner,
                      initial,
                      0,
                      0 );

   startHandlingEventsFor(whNB);
   this -> addHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

   // Default Tab Dimensions
   SHORT MajorTabWidth  = 50;
   SHORT MajorTabHeight = 30;
   SHORT MinorTabWidth  = 50;
   SHORT MinorTabHeight = 30;

   // Default DogEar Dimensions
   SHORT DogEarWidth   = 30;
   SHORT DogEarHeight  = 20;

   // Set the various notebook dimensions
   setPageButtonSize( ISize(DogEarWidth,DogEarHeight) );
   setMajorTabSize( ISize(MajorTabWidth,MajorTabHeight) );
   setMinorTabSize( ISize(MinorTabWidth,MinorTabHeight) );
}


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Construct a notebook from a window id and the window handle which serves     |
| as both parent and owner.                                                    |
------------------------------------------------------------------------------*/
INotebook :: INotebook( unsigned long windowId,
                        IWindow* parentAndOwner )
  : fNotebookData(0),
    bmClTabBitmapMgr(0),
    pnotebookColors(0),
    colorFlags(0),
    ulClValidate(0)
{
   IASSERTPARM( parentAndOwner != 0 );

   setAutoDestroyWindow(false);
   startHandlingEventsFor(windowId, parentAndOwner);
   this -> addHandler( defaultHandler );
  IMousePointerHandler::defaultHandler()->handleEventsFor(this);

   // Default Tab Dimensions
   SHORT MajorTabWidth  = 50;
   SHORT MajorTabHeight = 30;
   SHORT MinorTabWidth  = 50;
   SHORT MinorTabHeight = 30;

   // Default DogEar Dimensions
   SHORT DogEarWidth   = 30;
   SHORT DogEarHeight  = 20;

   // Set the various notebook dimensions
   setPageButtonSize( ISize(DogEarWidth,DogEarHeight) );
   setMajorTabSize( ISize(MajorTabWidth,MajorTabHeight) );
   setMinorTabSize( ISize(MinorTabWidth,MinorTabHeight) );
}


/*------------------------------------------------------------------------------
| INotebook::INotebook                                                         |
|                                                                              |
| Constructor to instantiate from an existing push button control.             |
------------------------------------------------------------------------------*/
INotebook :: INotebook ( const IWindowHandle& handle )
  : fNotebookData(0),
    bmClTabBitmapMgr(0),
    pnotebookColors(0),
    colorFlags(0),
    ulClValidate(0)
{
   setAutoDestroyWindow(false);
   startHandlingEventsFor(handle);
   this -> addHandler( defaultHandler );
   IMousePointerHandler::defaultHandler()->handleEventsFor(this);
}


/*------------------------------------------------------------------------------
| INotebook::~INotebook                                                        |
|                                                                              |
| Constructor to instantiate from an existing notebook control.                |
------------------------------------------------------------------------------*/
INotebook :: ~INotebook ( )
{
   this -> removeHandler( defaultHandler );
   IMousePointerHandler::defaultHandler()->stopHandlingEventsFor(this);
   delete (pnotebookColors);

   while (bmClTabBitmapMgr)            // Remove the associated bitmap managers
   {
     ITabBitmapMgr *tmpTabBitmapMgr = bmClTabBitmapMgr;
     bmClTabBitmapMgr = bmClTabBitmapMgr->next();
     delete tmpTabBitmapMgr;
   }
}


/*------------------------------------------------------------------------------
| INotebook::defaultStyle                                                      |
|                                                                              |
| Return the default style for new notebook objects.                           |
------------------------------------------------------------------------------*/
INotebook::Style INotebook :: defaultStyle ( )
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| INotebook::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook::convertToGUIStyle ( const IBitFlag& guiStyle,
                                             Boolean bExtOnly ) const
{
  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // BKS_ styles use all but one of their available bits in the lower
    // word of the base GUI style.  Therefore, obtain that portion asis,
    // masking out the upper word. Four of the styles were denoted as
    // 0 (OS/2 PM), due to the lack of available bits.  We are representing
    // those as extended bits so we or the developer can accurately test
    // for them later.  Since their native values are 0, we do not have to
    // do any additional processing here.
    ulStyle |= guiStyle.asUnsignedLong() & IBKS_MASK;

  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| INotebook::setBinding                                                        |
|                                                                              |
| Set the binding style of the notebook.                                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setBinding ( Binding binding )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulExtStyle = extendedStyle();

  if (binding == spiral)
  {
    ulStyle |= spiralBinding.asUnsignedLong();
    ulExtStyle &= ~solidBinding.asExtendedUnsignedLong();
  }
  else if (binding == solid)
  {
    ulStyle &= ~spiralBinding.asUnsignedLong();
    ulExtStyle |= solidBinding.asExtendedUnsignedLong();
  }
  else
  {
    ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,IErrorInfo::invalidParameter,
                       IException::recoverable);
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();

    // Update extended style ...
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setOrientation                                                    |
|                                                                              |
| Set the orientation style of the notebook.                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setOrientation ( Orientation orientation )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulMask = ~(backPagesBottomRight.asUnsignedLong()  |
                           backPagesBottomLeft.asUnsignedLong()   |
                           backPagesTopRight.asUnsignedLong()     |
                           backPagesTopLeft.asUnsignedLong()      |
                           majorTabsRight.asUnsignedLong()        |
                           majorTabsLeft.asUnsignedLong()         |
                           majorTabsTop.asUnsignedLong()          |
                           majorTabsBottom.asUnsignedLong());
  ulStyle &= ulMask;
  switch (orientation)
  {
    case backpagesBottomTabsRight:
      ulStyle |= (backPagesBottomRight.asUnsignedLong() |
                  majorTabsRight.asUnsignedLong());
      break;
    case backpagesTopTabsRight:
      ulStyle |= (backPagesTopRight.asUnsignedLong() |
                  majorTabsRight.asUnsignedLong());
      break;
    case backpagesBottomTabsLeft:
      ulStyle |= (backPagesBottomLeft.asUnsignedLong() |
                  majorTabsLeft.asUnsignedLong());
      break;
    case backpagesTopTabsLeft:
      ulStyle |= (backPagesTopLeft.asUnsignedLong() |
                  majorTabsLeft.asUnsignedLong());
      break;
    case backpagesRightTabsTop:
      ulStyle |= (backPagesTopRight.asUnsignedLong() |
                  majorTabsTop.asUnsignedLong());
      break;
    case backpagesLeftTabsTop:
      ulStyle |= (backPagesTopLeft.asUnsignedLong() |
                  majorTabsTop.asUnsignedLong());
      break;
    case backpagesRightTabsBottom:
      ulStyle |= (backPagesBottomRight.asUnsignedLong() |
                  majorTabsBottom.asUnsignedLong());
      break;
    case backpagesLeftTabsBottom:
      ulStyle |= (backPagesBottomLeft.asUnsignedLong() |
                  majorTabsBottom.asUnsignedLong());
      break;
    default:
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,IErrorInfo::invalidParameter,
                         IException::recoverable);
      break;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();
    notifyObservers(INotificationEvent(INotebook::orientationId,
                                       *this, true, (void*)orientation));
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setTabShape                                                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabShape ( TabShape tabShape )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulExtStyle = extendedStyle();
  unsigned long ulMask = ~(roundedTabs.asUnsignedLong()  |
                           polygonTabs.asUnsignedLong());
  ulStyle &= ulMask;
  ulExtStyle &= ~squareTabs.asExtendedUnsignedLong();

  switch (tabShape)
  {
    case square:
      ulExtStyle |= squareTabs.asExtendedUnsignedLong();
      break;
    case rounded:
      ulStyle |= roundedTabs.asUnsignedLong();
      break;
    case polygon:
      ulStyle |= polygonTabs.asUnsignedLong();
      break;
    default:
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,IErrorInfo::invalidParameter,
                         IException::recoverable);
      break;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();

    // Update extended style ...
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setStatusTextAlignment                                            |
|                                                                              |
| Set the alignment style of the text in the notebook's status line.           |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusTextAlignment ( TextAlignment align )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulExtStyle = extendedStyle();
  unsigned long ulMask = ~(statusTextRight.asUnsignedLong()  |
                           statusTextCenter.asUnsignedLong());
  ulStyle &= ulMask;
  ulExtStyle &= ~statusTextLeft.asExtendedUnsignedLong();

  switch (align)
  {
    case left:
      ulExtStyle |= statusTextLeft.asExtendedUnsignedLong();
      break;
    case right:
      ulStyle |= statusTextRight.asUnsignedLong();
      break;
    case center:
      ulStyle |= statusTextCenter.asUnsignedLong();
      break;
    default:
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,IErrorInfo::invalidParameter,
                         IException::recoverable);
      break;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();

    // Update extended style ...
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| INotebook::setTabTextAlignment                                               |
|                                                                              |
| Set the alignment of the text in the notebook's tabs.                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabTextAlignment ( TextAlignment align )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulExtStyle = extendedStyle();
  unsigned long ulMask = ~(tabTextRight.asUnsignedLong()  |
                           tabTextCenter.asUnsignedLong());
  ulStyle &= ulMask;
  ulExtStyle &= ~statusTextLeft.asExtendedUnsignedLong();

  switch (align)
  {
    case left:
      ulExtStyle |= tabTextLeft.asExtendedUnsignedLong();
      break;
    case right:
      ulStyle |= tabTextRight.asUnsignedLong();
      break;
    case center:
      ulStyle |= tabTextCenter.asUnsignedLong();
      break;
    default:
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,IErrorInfo::invalidParameter,
                         IException::recoverable);
      break;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();

    // Update extended style ...
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::binding                                                           |
------------------------------------------------------------------------------*/
INotebook::Binding INotebook :: binding ( ) const
{
  return( (style() & spiralBinding.asUnsignedLong()) ?  INotebook::spiral
                                                     :  INotebook::solid );
}


/*------------------------------------------------------------------------------
| INotebook::orientation                                                       |
------------------------------------------------------------------------------*/
INotebook::Orientation INotebook :: orientation ( ) const
{
  unsigned long ulStyle = style();
  INotebook::Orientation currentOrientation(backpagesRightTabsBottom);

  if (ulStyle & majorTabsRight.asUnsignedLong())
  {
    if (ulStyle & backPagesBottomRight.asUnsignedLong())
      currentOrientation = INotebook::backpagesBottomTabsRight;
    else
      currentOrientation = INotebook::backpagesTopTabsRight;
  }
  else if (ulStyle & majorTabsLeft.asUnsignedLong())
  {
    if (ulStyle & backPagesBottomLeft.asUnsignedLong())
      currentOrientation = INotebook::backpagesBottomTabsLeft;
    else
      currentOrientation = INotebook::backpagesTopTabsLeft;
  }
  else if (ulStyle & majorTabsTop.asUnsignedLong())
  {
    if (ulStyle & backPagesTopLeft.asUnsignedLong())
      currentOrientation = INotebook::backpagesLeftTabsTop;
    else
      currentOrientation = INotebook::backpagesRightTabsTop;
  }
  else if (ulStyle & majorTabsBottom.asUnsignedLong())
  {
    if (ulStyle & backPagesBottomLeft.asUnsignedLong())
      currentOrientation = INotebook::backpagesLeftTabsBottom;
    else
      currentOrientation = INotebook::backpagesRightTabsBottom;
  }
  return currentOrientation;
}


/*------------------------------------------------------------------------------
| INotebook::tabShape                                                          |
------------------------------------------------------------------------------*/
INotebook::TabShape INotebook :: tabShape ( ) const
{
  unsigned long ulStyle = style();
  if (ulStyle & roundedTabs.asUnsignedLong())
    return INotebook::rounded;
  else if (ulStyle & polygonTabs.asUnsignedLong())
    return INotebook::polygon;
  else
    return INotebook::square;
}


/*------------------------------------------------------------------------------
| INotebook::statusTextAlignment                                               |
|                                                                              |
| Return the alignment of the status text of the notebook.                     |
------------------------------------------------------------------------------*/
INotebook::TextAlignment INotebook :: statusTextAlignment ( ) const
{
  unsigned long ulStyle = style();
  if (ulStyle & statusTextRight.asUnsignedLong())
    return INotebook::right;
  else if (ulStyle & statusTextCenter.asUnsignedLong())
    return INotebook::center;
  else
    return INotebook::left;
}


/*------------------------------------------------------------------------------
| INotebook::tabTextAlignment                                                  |
|                                                                              |
| Return the alignment of the tab text of the notebook.                        |
------------------------------------------------------------------------------*/
INotebook::TextAlignment INotebook :: tabTextAlignment ( ) const
{
  unsigned long ulStyle = style();
  if (ulStyle & tabTextRight.asUnsignedLong())
    return INotebook::right;
  else if (ulStyle & tabTextCenter.asUnsignedLong())
    return INotebook::center;
  else
    return INotebook::left;
}


/*------------------------------------------------------------------------------
| INotebook::setMajorTabSize                                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setMajorTabSize ( const ISize& sizeMajorTab )
{
   Boolean retVal = sendEvent(BKM_SETDIMENSIONS,
                              MPFROM2SHORT(sizeMajorTab.width(),
                                           sizeMajorTab.height()),
                              BKA_MAJORTAB);
   if (!retVal)
   {
     ITHROWGUIERROR("BKM_SETDIMENSIONS");
   }
   return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setMinorTabSize                                                   |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setMinorTabSize ( const ISize& sizeMinorTab )
{
   Boolean retVal = sendEvent(BKM_SETDIMENSIONS,
                              MPFROM2SHORT(sizeMinorTab.width(),
                                           sizeMinorTab.height()),
                              BKA_MINORTAB);
   if (!retVal)
   {
     ITHROWGUIERROR("BKM_SETDIMENSIONS");
   }
   return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setPageButtonSize                                                 |
|                                                                              |
| Set the size of the arrow buttons used to turn the notebook's pages, in      |
| pixels.                                                                      |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setPageButtonSize ( const ISize& sizePageButton )
{
   Boolean retVal = sendEvent(BKM_SETDIMENSIONS,
                              MPFROM2SHORT(sizePageButton.width(),
                                           sizePageButton.height()),
                              BKA_PAGEBUTTON);
   if (!retVal)
   {
     ITHROWGUIERROR("BKM_SETDIMENSIONS");
   }
   return *this;
}


/*------------------------------------------------------------------------------
| INotebook::refreshTabs                                                       |
|                                                                              |
| Causes all tabs in the notebook to be invalidated (and thus repainted).      |
------------------------------------------------------------------------------*/
INotebook& INotebook :: refreshTabs ( )
{
   Boolean retVal = sendEvent(BKM_INVALIDATETABS, 0, 0);
   if (!retVal)
   {
     ITHROWGUIERROR("BKM_INVALIDATETABS");
   }
   return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::backgroundColor                                                   |
|                                                                              |
| Returns the background color of the notebook.                                |
------------------------------------------------------------------------------*/
IColor INotebook::backgroundColor ( ) const
{
  return (IWindow::color(PP_BACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::defaultControl)));
}

/*------------------------------------------------------------------------------
| INoteBook::hiliteBackgroundColor                                             |
|                                                                              |
| Returns the hilite background color of the notebook.                         |
------------------------------------------------------------------------------*/
IColor INotebook::hiliteBackgroundColor ( ) const
{
  return (IWindow::color(PP_HILITEBACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::defaultButton)));
}

/*------------------------------------------------------------------------------
| INoteBook::pageBackgroundColor                                               |
|                                                                              |
| Returns the background page color of the notebook.                           |
------------------------------------------------------------------------------*/
IColor INotebook::pageBackgroundColor ( ) const
{
  /********************************************************************/
  /* If this color was set previously, retrieve it from where we saved*/
  /* it and return it.  Otherwise, if a color has not been previously */
  /* set, return the default.                                         */
  /********************************************************************/
  if (colorFlags & INotebook::bgnPageColor)
  {
    if (pnotebookColors)
    {
      return (UnsignedLongAsRGB(pnotebookColors->pageBackgroundColor));
    }
  }
  return IGUIColor(IGUIColor::notebookPageBgnd);
}

/*------------------------------------------------------------------------------
| INoteBook::majorTabBackgroundColor                                           |
|                                                                              |
| Returns the major tab background color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::majorTabBackgroundColor ( ) const
{
  /********************************************************************/
  /* If this color was set previously, retrieve it from where we saved*/
  /* it and return it.  Otherwise, if a color has not been previously */
  /* set, return the default.                                         */
  /********************************************************************/
  if (colorFlags & INotebook::bgnMajorColor)
  {
    if (pnotebookColors)
    {
      return (UnsignedLongAsRGB(pnotebookColors->majorTabBackgroundColor));
    }
  }
  return IGUIColor(IGUIColor::notebookPageBgnd);
}

/*------------------------------------------------------------------------------
| INoteBook::minorTabBackgroundColor                                           |
|                                                                              |
| Returns the minor tab background color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::minorTabBackgroundColor ( ) const
{
  /********************************************************************/
  /* If this color was set previously, retrieve it from where we saved*/
  /* it and return it.  Otherwise, if a color has not been previously */
  /* set, return the default.                                         */
  /********************************************************************/
  if (colorFlags & INotebook::bgnMinorColor)
  {
    if (pnotebookColors)
    {
      return (UnsignedLongAsRGB(pnotebookColors->minorTabBackgroundColor));
    }
  }
  return IGUIColor(IGUIColor::notebookPageBgnd);
}

/*------------------------------------------------------------------------------
| INoteBook::majorTabForegroundColor                                           |
|                                                                              |
| Returns the major tab foreground color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::majorTabForegroundColor ( ) const
{
  /********************************************************************/
  /* If this color was set previously, retrieve it from where we saved*/
  /* it and return it.  Otherwise, if a color has not been previously */
  /* set, return the default.                                         */
  /********************************************************************/
  if (colorFlags & INotebook::fgnMajorColor)
  {
    if (pnotebookColors)
    {
      return (UnsignedLongAsRGB(pnotebookColors->majorTabForegroundColor));
    }
  }
  return IGUIColor(IGUIColor::windowText);
}

/*------------------------------------------------------------------------------
| INoteBook::minorTabForegroundColor                                           |
|                                                                              |
| Returns the minor tab foreground color.                                      |
------------------------------------------------------------------------------*/
IColor INotebook::minorTabForegroundColor ( ) const
{
  /********************************************************************/
  /* If this color was set previously, retrieve it from where we saved*/
  /* it and return it.  Otherwise, if a color has not been previously */
  /* set, return the default.                                         */
  /********************************************************************/
  if (colorFlags & INotebook::fgnMinorColor)
  {
    if (pnotebookColors)
    {
      return (UnsignedLongAsRGB(pnotebookColors->minorTabForegroundColor));
    }
  }
  return IGUIColor(IGUIColor::windowText);
}

/*------------------------------------------------------------------------------
| INoteBook::setPageBackgroundColor                                            |
|                                                                              |
| Sets the page background color in a notebook.                                |
------------------------------------------------------------------------------*/
INotebook& INotebook::setPageBackgroundColor ( const IColor &color )
{
  setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDPAGECOLOR);
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMajorTabBackgroundColor                                        |
|                                                                              |
| Sets the major tab background color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMajorTabBackgroundColor ( const IColor &color )
{
  setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDMAJORCOLOR);
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMinorTabBackgroundColor                                        |
|                                                                              |
| Sets the minor tab background color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMinorTabBackgroundColor ( const IColor &color )
{
  setNotebookColors (color.asRGBLong(), BKA_BACKGROUNDMINORCOLOR);
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMajorTabForegroundColor                                        |
|                                                                              |
| Sets the major tab foreground color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMajorTabForegroundColor ( const IColor &color )
{
  setNotebookColors (color.asRGBLong(), BKA_FOREGROUNDMAJORCOLOR);
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::setMinorTabForegroundColor                                        |
|                                                                              |
| Sets the minor tab foreground color.                                         |
------------------------------------------------------------------------------*/
INotebook& INotebook::setMinorTabForegroundColor ( const IColor &color )
{
  setNotebookColors (color.asRGBLong(), BKA_FOREGROUNDMINORCOLOR);
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetPageBackgroundColor                                          |
|                                                                              |
| Resets the page background color by undoing a previous set.                  |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetPageBackgroundColor ( )
{
  colorFlags &= ~bgnPageColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMajorTabBackgroundColor                                      |
|                                                                              |
| Resets the major tab background color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMajorTabBackgroundColor ( )
{
  colorFlags &= ~bgnMajorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMinorTabBackgroundColor                                      |
|                                                                              |
| Resets the minor tab background color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMinorTabBackgroundColor ( )
{
  colorFlags &= ~bgnMinorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMajorTabForegroundColor                                      |
|                                                                              |
| Resets the major tab foreground color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMajorTabForegroundColor ( )
{
  colorFlags &= ~fgnMajorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::resetMinorTabForegroundColor                                      |
|                                                                              |
| Resets the minor tab foreground color by undoing a previous set.             |
------------------------------------------------------------------------------*/
INotebook& INotebook::resetMinorTabForegroundColor ( )
{
  colorFlags &= ~fgnMinorColor;
  return *this;
}

/*------------------------------------------------------------------------------
| INoteBook::isPMCompatible                                                    |
|                                                                              |
| Returns true if the notebook is displayed as the CUA '91 notebook.           |
------------------------------------------------------------------------------*/
Boolean INotebook::isPMCompatible ( ) const
{
  return( true );
}

//
// PageSettings and Cursor classes now implemented in inotebk0.cpp
//

/*------------------------------------------------------------------------------
| INotebook::insertPageInfo                                                    |
|                                                                              |
| Set the values in the PageSettings into the notebook.                        |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: insertPageInfo ( const PageSettings& pageInfo,
                                          const IPageHandle&  referencePage,
                                          IWindow*            window,
                                          unsigned long       fPosition )
{
  IPageHandle pageHandle =
        insertPage( MPFROMLONG( referencePage ),
                    MPFROM2SHORT( (USHORT)pageInfo.pageStyle.asUnsignedLong(),
                                  fPosition) );

  if (window && pageHandle)
  {
    IRectangle pageRect;

    /**************************************************************************/
    /* If the page window is a frame, we need to set the owner to NULL.  The  */
    /* notebook will set the parent to the page window (id 8006) and then we  */
    /* can set the owner to the notebook.  (We cannot set the owner to the    */
    /* notebook initially because this might cause the frame to have the same */
    /* parent and owner and cause problems).  If the owner of the frame is    */
    /* not the notebook, a focus loop problem will occur.                     */
    /**************************************************************************/
    if ( window->isFrameWindow() )
      window->setOwner( NULL );

    /**************************************************************************/
    /* save the page window rect                                              */
    /**************************************************************************/
    if ( pageInfo.isAutoSize() )
      pageRect = window->rect();

    Boolean returnBool = sendEvent(BKM_SETPAGEWINDOWHWND,
                                   (unsigned long)pageHandle,
                                   (unsigned long)window->handle());

    if (!returnBool)
    {
      ITHROWGUIERROR("BKM_SETPAGEWINDOWHWND");
    }
    else
    {
      /************************************************************************/
      /* restore the page window size                                         */
      /************************************************************************/
      if ( pageInfo.isAutoSize() )
        window->sizeTo( ISize( pageRect.width(), pageRect.height() ));

      /************************************************************************/
      /* If there is no owner for the page, set it to the notebook.           */
      /************************************************************************/
      if ( window->owner() == NULL )
        window->setOwner( this );
    }
  }
  if (pageHandle)
  {
    if (pageInfo.savedStatusText.size() != 0)
      setStatusText(pageHandle, pageInfo.savedStatusText);

    if (pageInfo.savedTabText.size() != 0)
      setTabText(pageHandle, pageInfo.savedTabText);

    if (pageInfo.savedTabBitmap != 0)
      setTabBitmap(pageHandle, pageInfo.savedTabBitmap);

    setUserData(pageHandle, pageInfo.savedUserData);
  }
  return pageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::addFirstPage                                                      |
|                                                                              |
| Add a given window to the notebook as the first page, using the given        |
| page settings.                                                               |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addFirstPage ( const PageSettings& pageInfo,
                                        IWindow* window )
{
  return insertPageInfo(pageInfo, IPageHandle(0), window, BKA_FIRST);
}


/*------------------------------------------------------------------------------
| INotebook::addLastPage                                                       |
|                                                                              |
| Add a given window to the notebook as the last page, using the given         |
| page settings.                                                               |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addLastPage ( const PageSettings& pageInfo,
                                       IWindow* window )
{
  return insertPageInfo(pageInfo, IPageHandle(0), window, BKA_LAST);
}


/*------------------------------------------------------------------------------
| INotebook::addPageBefore                                                     |
|                                                                              |
| Add a given window to the notebook before the referenced page, using the     |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageBefore ( const PageSettings& newPageInfo,
                                         const IPageHandle& referencePage,
                                         IWindow* window )
{
  return insertPageInfo(newPageInfo, referencePage, window, BKA_PREV);
}


/*------------------------------------------------------------------------------
| INotebook::addPageBefore                                                     |
|                                                                              |
| Add a given window to the notebook before the referenced page, using the     |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageBefore ( const PageSettings& newPageInfo,
                                         const Cursor& cursor,
                                         IWindow* window )
{
  return insertPageInfo(newPageInfo, cursor.current(), window, BKA_PREV);
}


/*------------------------------------------------------------------------------
| INotebook::addPageAfter                                                      |
|                                                                              |
| Add a given window to the notebook after the referenced page, using the      |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageAfter ( const PageSettings& newPageInfo,
                                        const IPageHandle& referencePage,
                                        IWindow* window )
{
  return insertPageInfo(newPageInfo, referencePage, window, BKA_NEXT);
}


/*------------------------------------------------------------------------------
| INotebook::addPageAfter                                                      |
|                                                                              |
| Add a given window to the notebook after the referenced page, using the      |
| given page settings.                                                         |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: addPageAfter ( const PageSettings& newPageInfo,
                                        const Cursor& cursor,
                                        IWindow* window )
{
  return insertPageInfo(newPageInfo, cursor.current(), window, BKA_NEXT);
}


/*------------------------------------------------------------------------------
| INotebook::insertPage                                                        |
------------------------------------------------------------------------------*/
unsigned long INotebook :: insertPage ( void* mp1, void* mp2 )
{
  unsigned long retVal = sendEvent(BKM_INSERTPAGE, mp1, mp2);
  if (!retVal)
  {
    ITHROWGUIERROR("BKM_INSERTPAGE");
  }
  ulClValidate++;
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::removePage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removePage ( const IPageHandle& page )
{
  deletePage (MPFROMLONG (page), MPFROMSHORT (BKA_SINGLE));
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::removePage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removePage ( const Cursor& cursor )
{
  return removePage(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::removeAllPages                                                    |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeAllPages ( )
{
  deletePage (0, MPFROMSHORT (BKA_ALL));
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::removeTabSection                                                  |
|                                                                              |
| If the page specified is a major tab page, the specified page and all        |
| subsequent pages up to the next major tab page are removed.                  |
| If the page specified is a minor tab page, the specified page and all        |
| subsequent pages up to the next page with any tab are removed.               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeTabSection ( const IPageHandle& page )
{
  unsigned short pageStyle = sendEvent( BKM_QUERYPAGESTYLE,
                                        MPFROMLONG( UPAGE ),
                                        0 );

  IASSERTSTATE( ( pageStyle & (BKA_MAJOR | BKA_MINOR) ) != 0 );

  deletePage (MPFROMLONG( UPAGE ), MPFROMSHORT( BKA_TAB ));
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::removeTabSection                                                  |
|                                                                              |
| If the page cursored is a major tab page, remove the cursored page and       |
| all subsequent pages up to the next major tab page.                          |
| If the page cursored is a minor tab page, remove the cursored page and       |
| all subsequent pages up to the next page with any tab.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: removeTabSection ( const Cursor& cursor )
{
  return removeTabSection(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::deletePage                                                        |
------------------------------------------------------------------------------*/
Boolean INotebook :: deletePage ( void* mp1, void* mp2 )
{
  Boolean retVal = sendEvent(BKM_DELETEPAGE, mp1, mp2);
  if (!retVal)
  {
    ITHROWGUIERROR("BKM_DELETEPAGE");
  }
  ulClValidate--;
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::turnToPage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: turnToPage ( const IPageHandle& page )
{
  Boolean retVal = sendEvent(BKM_TURNTOPAGE, UPAGE, 0);
  if (!retVal)
  {
    ITHROWGUIERROR("BKM_TURNTOPAGE");
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::turnToPage                                                        |
------------------------------------------------------------------------------*/
INotebook& INotebook :: turnToPage ( const Cursor& cursor )
{
  return turnToPage(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::topPage                                                           |
|                                                                              |
| Returns an IPageHandle object that references the current top page of the    |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: topPage ( ) const
{
  unsigned long retVal;
  IPageHandle topPageHandle (0);

  retVal = queryPage (0, MPFROM2SHORT (BKA_TOP, 0));
  if (retVal)
    topPageHandle = retVal;

  return topPageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::firstPage                                                         |
|                                                                              |
| Returns an IPageHandle object that references the first page of the notebook.|
------------------------------------------------------------------------------*/
IPageHandle INotebook :: firstPage ( ) const
{
  unsigned long retVal;
  IPageHandle firstPageHandle (0);

  retVal = queryPage (0, MPFROM2SHORT (BKA_FIRST, 0));
  if (retVal)
    firstPageHandle = retVal;

  return firstPageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::lastPage                                                          |
|                                                                              |
| Returns an IPageHandle object that references the last page of the notebook. |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: lastPage ( ) const
{
  unsigned long retVal;
  IPageHandle lastPageHandle (0);

  retVal = queryPage (0, MPFROM2SHORT (BKA_LAST, 0));
  if (retVal)
    lastPageHandle = retVal;

  return lastPageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::nextPage                                                          |
|                                                                              |
| Returns an IPageHandle object that references the next page of the notebook. |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: nextPage ( const IPageHandle& page ) const
{
  IPageHandle nextPageHandle (0);

  if (page)
  {
    unsigned long retVal = queryPage( MPFROMLONG( UPAGE ),
                                      MPFROM2SHORT( BKA_NEXT, 0 ));
    if (retVal)
      nextPageHandle = retVal;
  }
  return nextPageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::previousPage                                                      |
|                                                                              |
| Returns an IPageHandle object that references the page preceding the         |
| referenced page.                                                             |
------------------------------------------------------------------------------*/
IPageHandle INotebook :: previousPage ( const IPageHandle& page ) const
{
  IPageHandle prevPageHandle (0);

  if (page)
  {
    unsigned long retVal = queryPage( MPFROMLONG( UPAGE ),
                                      MPFROM2SHORT( BKA_PREV, 0 ));
    if (retVal)
      prevPageHandle = retVal;
  }
  return prevPageHandle;
}


/*------------------------------------------------------------------------------
| INotebook::queryPage                                                         |
------------------------------------------------------------------------------*/
unsigned long INotebook :: queryPage ( void* mp1, void* mp2 ) const
{
  long retVal = (long)sendEvent(BKM_QUERYPAGEID, mp1, mp2);

  if (retVal == BOOKERR_INVALID_PARAMETERS)
  {
    ITHROWGUIERROR("BKM_QUERYPAGEID:INVALID_PARAMETERS");
  }
  return (unsigned long)retVal;
}


/*------------------------------------------------------------------------------
| INotebook::window                                                            |
|                                                                              |
| Returns the window used by the notebook page.                                |
------------------------------------------------------------------------------*/
IWindow* INotebook :: window ( const INotebook::Cursor& cursor ) const
{
  return window(cursor.current());     // Return window for current page
}


/*------------------------------------------------------------------------------
| INotebook::window                                                            |
|                                                                              |
| Returns the window used by the notebook page.                                |
------------------------------------------------------------------------------*/
IWindow* INotebook :: window ( const IPageHandle& page ) const
{
  IWindow* pwinReturn = 0;             // Assume no window

  IWindowHandle wh =
    (unsigned long)sendEvent(BKM_QUERYPAGEWINDOWHWND, UPAGE, 0);
  if (((long)(unsigned long)wh) == BOOKERR_INVALID_PARAMETERS)
  {
     ITHROWGUIERROR("BKM_QUERYPAGEWINDOWHWND");
  }
  else if (wh != 0)
  {
     pwinReturn = windowWithHandle(wh);
  }

  return pwinReturn;
}


/*------------------------------------------------------------------------------
| INotebook::pageSettings                                                      |
|                                                                              |
| Create and return a PageSettings object representing the page.               |
------------------------------------------------------------------------------*/
INotebook::PageSettings INotebook :: pageSettings
                          ( const INotebook::Cursor& cursor ) const
{
   return pageSettings(cursor.current());
}

/*------------------------------------------------------------------------------
| INotebook::pageSettings                                                      |
|                                                                              |
| Create and return a PageSettings object representing the page.               |
------------------------------------------------------------------------------*/
INotebook::PageSettings INotebook :: pageSettings
                          ( const IPageHandle& page ) const
{
  INotebook::PageSettings newPageSettings (0);
  BOOKTEXT bookText;
  char bookBuffer [100];

  unsigned long ulStyle =
         (unsigned long)sendEvent(BKM_QUERYPAGESTYLE, UPAGE, 0);

  if (ulStyle & PageSettings::autoPageSize.asUnsignedLong())
    newPageSettings.pageStyle |= (PageSettings::autoPageSize);
  if (ulStyle & PageSettings::statusTextOn.asUnsignedLong())
    newPageSettings.pageStyle |= (PageSettings::statusTextOn);
  if (ulStyle & PageSettings::majorTab.asUnsignedLong())
    newPageSettings.pageStyle |= (PageSettings::majorTab);
  if (ulStyle & PageSettings::minorTab.asUnsignedLong())
    newPageSettings.pageStyle |= (PageSettings::minorTab);

  newPageSettings.savedUserData =
                sendEvent(BKM_QUERYPAGEDATA, UPAGE, 0);

  bookText.textLen = sizeof (bookBuffer);
  bookText.pString = bookBuffer;
  short usRetVal   = sendEvent( BKM_QUERYSTATUSLINETEXT,
                                UPAGE,
                                &bookText );

  if (usRetVal && (usRetVal != (short)BOOKERR_INVALID_PARAMETERS))
    newPageSettings.savedStatusText = IString (bookBuffer);

  usRetVal = sendEvent(BKM_QUERYTABTEXT, UPAGE, &bookText);
  if (usRetVal && (usRetVal != (short)BOOKERR_INVALID_PARAMETERS))
    newPageSettings.savedTabText = IString (bookBuffer);

  HBITMAP hTabBitmap =
               sendEvent(BKM_QUERYTABBITMAP, UPAGE, 0);
  if (hTabBitmap && ((long)hTabBitmap != BOOKERR_INVALID_PARAMETERS))
    newPageSettings.savedTabBitmap = IBitmapHandle (hTabBitmap);

  return newPageSettings;
}


/*------------------------------------------------------------------------------
| INotebook::setStatusText                                                     |
|                                                                              |
| Change the status text of the specified page in the notebook.                |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusText ( const IPageHandle& page,
                                        const char* text )
{
  PageSettings settings = this->pageSettings( page );

  if ( ! settings.isStatusTextOn() )
  {           // Ensure the page has a status text area.
     ITHROWLIBRARYERROR( IC_NOSTATUSTEXTLINE,
                         IErrorInfo::invalidRequest,
                         IException::recoverable );
  }
  else if ( settings.statusText() != text )
  {
    Boolean retVal =
              this->sendEvent( BKM_SETSTATUSLINETEXT, UPAGE, (void*)text );
    if ( ! retVal )
    {
       ITHROWGUIERROR( "BKM_SETSTATUSLINETEXT" );
    }
  }

  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setStatusText                                                     |
|                                                                              |
| Change the status text of the specified page in the notebook.                |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setStatusText ( const IPageHandle& page,
                                        const IResourceId& resId )
{
  return setStatusText(page,
                       resId.resourceLibrary().loadString(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setTabText                                                        |
|                                                                              |
| Change the tab text of the specified notebook page.                          |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabText ( const IPageHandle& page,
                                     const char* text )
{
  Boolean retVal = sendEvent( BKM_SETTABTEXT, UPAGE, (void*)text);
  if (!retVal)
  {
     ITHROWGUIERROR("BKM_SETTABTEXT");
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setTabText                                                        |
|                                                                              |
| Change the tab text of the specified notebook page.                          |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabText ( const IPageHandle& page,
                                     const IResourceId& resId )
{
  return setTabText(page,
                    resId.resourceLibrary().loadString(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setTabBitmap                                                      |
|                                                                              |
| Change the tab bitmap for the specified notebook page.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabBitmap ( const IPageHandle& page,
                                       const IBitmapHandle& bitmap )
{
  Boolean retVal = sendEvent( BKM_SETTABBITMAP, UPAGE,
                              (unsigned long)bitmap);
  if (!retVal)
  {
     ITHROWGUIERROR("BKM_SETTABBITMAP");
  }
  ITabBitmapMgr *tmpBitmapMgr = new ITabBitmapMgr(bitmap, page);
  tmpBitmapMgr->setNext(bmClTabBitmapMgr);
  bmClTabBitmapMgr = tmpBitmapMgr;
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setTabBitmap                                                      |
|                                                                              |
| Change the tab bitmap for the specified notebook page.                       |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setTabBitmap ( const IPageHandle& page,
                                       const IResourceId& resId )
{
  return setTabBitmap(page,
                      resId.resourceLibrary().loadBitmap(resId.id()));
}


/*------------------------------------------------------------------------------
| INotebook::setUserData                                                       |
|                                                                              |
| Change the user data in the specified notebook page.                         |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setUserData ( const IPageHandle& page,
                                      unsigned long ulData )
{
  Boolean retVal = sendEvent( BKM_SETPAGEDATA, UPAGE, ulData );
  if (!retVal)
  {
     ITHROWGUIERROR("BKM_SETPAGEDATA");
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setWindow                                                         |
|                                                                              |
| Associates the specified page with the window.                               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setWindow ( const IPageHandle& page,
                                    IWindow*           window )
{
  INotebook::PageSettings pageInfo = this->pageSettings( page );
  IRectangle pageRect;

  /****************************************************************************/
  /* If the page window is a frame, we need to set the owner to NULL.  The    */
  /* notebook will set the parent to the page window (id 8006) and then we    */
  /* can set the owner to the notebook.  (We cannot set the owner to the      */
  /* notebook initially because this might cause the frame to have the same   */
  /* parent and owner and cause problems).  If the owner of the frame is not  */
  /* the notebook, a focus loop problem will occur.                           */
  /****************************************************************************/
  if ( window->isFrameWindow() )
    window->setOwner( NULL );

  /*****************************************************************************
  * save the page window rect
  *****************************************************************************/
  if ( pageInfo.isAutoSize() )
    pageRect = window->rect();

  unsigned long temphwnd = (window ? (unsigned long)window->handle() : 0);
  Boolean returnBool = sendEvent(BKM_SETPAGEWINDOWHWND, UPAGE, temphwnd );

  if (!returnBool)
  {
     ITHROWGUIERROR("BKM_SETPAGEWINDOWHWND");
  }
  else
  {
     /*************************************************************************/
     /* restore the page window size                                          */
     /*************************************************************************/
     if ( pageInfo.isAutoSize() )
        window->sizeTo( ISize( pageRect.width(), pageRect.height() ));

     /*************************************************************************/
     /* If there is no owner for the page, set it to the notebook.            */
     /*************************************************************************/
     if ( window->owner() == NULL )
        window->setOwner( this );
  }
  return *this;
}


/*------------------------------------------------------------------------------
| INotebook::setWindow                                                         |
|                                                                              |
| Associates the specified page with the window.                               |
------------------------------------------------------------------------------*/
INotebook& INotebook :: setWindow ( const Cursor&  cursor,
                                    IWindow*       window )
{
  return setWindow(cursor.current(), window);
}


/*------------------------------------------------------------------------------
| INotebook::notebookSize                                                      |
|                                                                              |
| Return the size of the notebook needed to display a page.                    |
------------------------------------------------------------------------------*/
ISize INotebook :: notebookSize ( const IPageHandle& page ) const
{
  IWindow *pageWindow = this->window( page );
  RECTL pageSize = { 0, 0, 100, 100 };

  if ( pageWindow )
  {
     IRectangle pageRect = pageWindow->rect();
     if ( pageRect.size() == ISize() )
     {        // Use minimum size instead.
        pageRect.sizeTo( pageWindow->minimumSize() );
     }
     pageSize = pageRect.asRECTL();
  }

  Boolean res = handle().sendEvent( BKM_CALCPAGERECT,
                                    &pageSize, false );
  if ( res == false )
  {
     ITHROWGUIERROR( "BKM_CALCPAGERECT" );
  }
  return ISize( pageSize );
}

/*------------------------------------------------------------------------------
| INotebook::pageSize                                                          |
|                                                                              |
| Return the size of the "viewport" in which the notebook draws the pages      |
| it contains.                                                                 |
------------------------------------------------------------------------------*/
ISize INotebook :: pageSize ( ) const
{
  RECTL notebookSize = rect().asRECTL();
  Boolean res = handle().sendEvent(BKM_CALCPAGERECT,
                                   &notebookSize, true);
  if (res == false)
  {
     ITHROWGUIERROR("BKM_CALCPAGERECT");
  }
  return ISize(notebookSize);
}


/*------------------------------------------------------------------------------
| INotebook::totalPages                                                        |
|                                                                              |
| Return the total number of pages in the notebook.                            |
------------------------------------------------------------------------------*/
unsigned long INotebook :: totalPages ( ) const
{
  unsigned long retVal =
           (unsigned long) queryPageCount (0, MPFROMSHORT (BKA_END));
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMajorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a major tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMajorTab ( const IPageHandle& page ) const
{
  unsigned long retVal =
                 (unsigned long)queryPageCount( MPFROMLONG( UPAGE ),
                 MPFROMSHORT( PageSettings::majorTab.asUnsignedLong()));
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMajorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a major tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMajorTab ( const Cursor& cursor) const
{
  return pagesToMajorTab(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMinorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a minor tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMinorTab ( const IPageHandle& page ) const
{
  unsigned long retVal =
                 (unsigned long)queryPageCount( MPFROMLONG( UPAGE ),
                 MPFROMSHORT( PageSettings::minorTab.asUnsignedLong()));
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::pagesToMinorTab                                                   |
|                                                                              |
| Return the number of pages between the specified page and the next page with |
| a minor tab.                                                                 |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToMinorTab ( const Cursor& cursor ) const
{
  return pagesToMinorTab(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::pagesToEnd                                                        |
|                                                                              |
| Return the number of pages between the specified page and the end of the     |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToEnd ( const IPageHandle& page ) const
{
  unsigned long retVal =
           (unsigned long) queryPageCount (MPFROMLONG (UPAGE),
                                           MPFROMSHORT (BKA_END));
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::pagesToEnd                                                        |
|                                                                              |
| Return the number of pages between the specified page and the end of the     |
| notebook.                                                                    |
------------------------------------------------------------------------------*/
unsigned long INotebook :: pagesToEnd ( const Cursor& cursor ) const
{
  return pagesToEnd(cursor.current());
}


/*------------------------------------------------------------------------------
| INotebook::isEmpty                                                           |
|                                                                              |
| Queries whether the notebook has any pages.                                  |
------------------------------------------------------------------------------*/
Boolean INotebook :: isEmpty ( ) const
{
  if (totalPages())
    return false;
  else
    return true;
}


/*------------------------------------------------------------------------------
| INotebook::queryPageCount                                                    |
|                                                                              |
| Queries whether the notebook has any pages.                                  |
------------------------------------------------------------------------------*/
short INotebook :: queryPageCount ( void* mp1, void* mp2 ) const
{
  short retVal = sendEvent(BKM_QUERYPAGECOUNT, mp1, mp2);

  if (retVal == BOOKERR_INVALID_PARAMETERS)
  {
     ITHROWGUIERROR("BKM_QUERYPAGECOUNT");
  }
  return retVal;
}


/*------------------------------------------------------------------------------
| INotebook::calcMinimumSize                                                   |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The height is      |
| based on the minimum size of the superclass control, with the notebook       |
| drawing around it.                                                           |
------------------------------------------------------------------------------*/
ISize INotebook :: calcMinimumSize ( ) const
{
  ISize szLargestBook( 0, 0 );
  RECTL pageRect = this->rect().asRECTL();  // Get notebook rectangle.

  for ( IPageHandle pageHandle = this->firstPage();
        pageHandle;
        pageHandle = this->nextPage( pageHandle ))
  {
     ISize szNoteBook = this->notebookSize( pageHandle );
     szLargestBook = szLargestBook.maximum( szNoteBook );
  }

  if ( (szLargestBook.height() == 0) ||
       (szLargestBook.width()  == 0) )
  {
     ISize pageSize = this->Inherited::calcMinimumSize(); // Default min size.
     Boolean retVal = this->sendEvent( BKM_CALCPAGERECT, &pageRect, true );
     if ( ! retVal )
     {
        ITHROWGUIERROR( "BKM_CALCPAGERECT" );
     }

     RECTL newPageRect =
              IRectangle( IPoint( pageRect.xLeft, pageRect.yBottom ),
                          pageSize ).asRECTL();

     retVal = this->sendEvent( BKM_CALCPAGERECT, &newPageRect, false );
     if ( ! retVal )
     {
        ITHROWGUIERROR( "BKM_CALCPAGERECT" );
     }

     szLargestBook = ISize( newPageRect );
  }

  return szLargestBook;
}

/*------------------------------------------------------------------------------
| INotebook::setNotebookColors                              *** private ***    |
|                                                                              |
| This function handles the setting of the notebook colors that are not        |
| identified by system defined PP_* values.  The notebook has some special     |
| color areas defined that can only be set using a specific notebook message.  |
| In addition, the presentation manager notebook does not provide a way to     |
| query these colors.                                                          |
------------------------------------------------------------------------------*/
void INotebook::setNotebookColors ( unsigned long color,
                                    unsigned long colorArea )
{
  Boolean colorChanged = false;

  /********************************************************************/
  /* Check to see if the current color is different from the color we */
  /* are trying to set.                                               */
  /********************************************************************/
  switch (colorArea)
  {
    case BKA_BACKGROUNDPAGECOLOR:
      if (pageBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_BACKGROUNDMAJORCOLOR:
      if (majorTabBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_BACKGROUNDMINORCOLOR:
      if (minorTabBackgroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_FOREGROUNDMAJORCOLOR:
      if (majorTabForegroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;

    case BKA_FOREGROUNDMINORCOLOR:
      if (minorTabForegroundColor() != UnsignedLongAsRGB(color))
      {
        colorChanged = true;
      }
    break;
  }

  if (colorChanged)
  {
    /* Set the color on the notebook. */
    Boolean retVal = sendEvent(BKM_SETNOTEBOOKCOLORS, color, colorArea);
    if (!retVal)
    {
      ITHROWGUIERROR("BKM_SETNOTEBOOKCOLORS");
    }
  }
}

/*------------------------------------------------------------------------------
| INotebook::storeNotebookColors                              *** private ***  |
|                                                                              |
| This function handles the storing of the notebook colors that are not        |
| identified by system defined PP_* values.  This enables the colors to be     |
| querying later since the system notebook does not provide a query mechanism  |
| of its own.                                                                  |
------------------------------------------------------------------------------*/
void INotebook::storeNotebookColors ( unsigned long color,
                                      unsigned long colorArea )
{
  if (!pnotebookColors)
  {
    pnotebookColors = new NotebookColors;
  }

  /*******************************************************************/
  /* Set a flag to indicate which color we are setting and store the */
  /* color we set so it can be queried later.                        */
  /*******************************************************************/
  if (colorArea == BKA_BACKGROUNDPAGECOLOR)
  {
    pnotebookColors->pageBackgroundColor = color;
    colorFlags |= INotebook::bgnPageColor;
  }
  else if (colorArea == BKA_BACKGROUNDMAJORCOLOR)
  {
    pnotebookColors->majorTabBackgroundColor = color;
    colorFlags |= INotebook::bgnMajorColor;
  }
  else if (colorArea == BKA_BACKGROUNDMINORCOLOR)
  {
    pnotebookColors->minorTabBackgroundColor = color;
    colorFlags |= INotebook::bgnMinorColor;
  }
  else if (colorArea == BKA_FOREGROUNDMAJORCOLOR)
  {
    pnotebookColors->majorTabForegroundColor = color;
    colorFlags |= INotebook::fgnMajorColor;
  }
  else if (colorArea == BKA_FOREGROUNDMINORCOLOR)
  {
    pnotebookColors->minorTabForegroundColor = color;
    colorFlags |= INotebook::fgnMinorColor;
  }
}

/*------------------------------------------------------------------------------
| INotebook::UnsignedLongAsRGB                            *** private ***      |
|                                                                              |
| This function converts an unsigned long to an RGB value and returns it as    |
| an IColor object.                                                            |
------------------------------------------------------------------------------*/
IColor INotebook::UnsignedLongAsRGB ( unsigned long color ) const
{
  /* Convert the long color to an RGB value then to an IColor object. */
  unsigned char blue  = (unsigned char)(color);
  unsigned char green = (unsigned char)(color >> 8);
  unsigned char red   = (unsigned char)(color >> 16);
  return IColor(red, green, blue);
}
