/*******************************************************************************
* FILE NAME: ipainevt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ipainevt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_GPILOGCOLORTABLE
  #define INCL_GPIPRIMITIVES
  #define INCL_GPIREGIONS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  #include <string.h>
}

#include <ipainevt.hpp>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <iwindow.hpp>
#include <ireslib.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IPAINEVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IPaintEvent::IPaintEvent                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent::IPaintEvent ( const IEvent& evt )
  : IEvent(evt), hpsCl(), rectCl(), setGC(false)
{
   RECTL rcl;
   WinQueryUpdateRect( this->handle(), &rcl );
   if ((rcl.xLeft==rcl.xRight) || (rcl.yBottom==rcl.yTop))    // null rectangle
   {
      rectCl = IRectangle(0,0,0,0);   // Make it really null (both dimensions!)
   }
   else
   {
      rectCl = IRectangle( rcl.xLeft, rcl.yBottom,
                           rcl.xRight, rcl.yTop );
   } /* endif */

   //--------------------------------------------------------------------------
   // On OS/2 BIDI systems, we need to copy the BIDI attributes from the
   // window to the HPS (since OS/2 does not do this for us)
   //--------------------------------------------------------------------------
   IWindow* win = this->window();
   if ( IBidiSettings::isBidiSupported() )
   {
     IBidiSettings settings(*win);
     settings.apply(this->presSpaceHandle());
   }
}

/*------------------------------------------------------------------------------
| IPaintEvent::~IPaintEvent                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent::~IPaintEvent ( )
{
  HRGN oldrgn;
  if ( hpsCl )
  {
     if ( GpiSetClipRegion( hpsCl, NULLHANDLE, &oldrgn ) == RGN_ERROR )
        ITHROWGUIERROR( "GpiSetClipRegion" );
     if ( oldrgn )
     {
       if ( !( GpiDestroyRegion( hpsCl, oldrgn )))
          ITHROWGUIERROR( "GpiDestroyRegion" );
     }
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */
}

/*------------------------------------------------------------------------------
| IPaintEvent::presSpaceHandle                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPresSpaceHandle IPaintEvent::presSpaceHandle ( )
{
  if ( rectCl.area() )
  {
     if (!hpsCl)
     {
        hpsCl = window()->presSpace();
        this->initializeGC();
     }
     return hpsCl;
  }
  else
  {
     return 0;
  }
}

/*------------------------------------------------------------------------------
| IPaintEvent::setGraphicContext                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent&
  IPaintEvent::setGraphicContext ( const IPresSpaceHandle& presHandle,
                                   Boolean setClipRegion )
{
  HRGN oldrgn;
  if ( hpsCl )
  {
     if ( GpiSetClipRegion( hpsCl, NULLHANDLE, &oldrgn ) == RGN_ERROR )
        ITHROWGUIERROR( "GpiSetClipRegion" );
     if ( oldrgn )
     {
       if ( !( GpiDestroyRegion( hpsCl, oldrgn )))
          ITHROWGUIERROR( "GpiDestroyRegion" );
     }
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */

  hpsCl = presHandle;
  if ( setClipRegion )
  {
     this->initializeGC();
  }
  setGC = true;
  return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::setGraphicContext                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::setGraphicContext ( const IGraphicContext& gc,
                                              Boolean setClipRegion )
{
  return this->setGraphicContext( gc.handle(), setClipRegion );
}

/*------------------------------------------------------------------------------
| IPaintEvent::clearBackground                                                 |
|                                                                              |
| Clear and fill a given rectangle area                                        |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::clearBackground ( const IRectangle& rect,
                                            const IColor& clr )
{
   RECTL rcl = rect.asRECTL();
   if ( GpiQueryLogColorTable( hpsCl, LCOLOPT_INDEX, 0, 0, 0 ) == QLCT_RGB )
     WinFillRect( hpsCl, &rcl, clr.asRGBLong() );
   else
     WinFillRect( hpsCl, &rcl, clr.index() );
   return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::drawText                                                        |
|                                                                              |
| Draw a line of text starting at a given position.                            |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::drawText ( const char* pszText,
                                     const IPoint& atPoint,
                                     const IColor& clr )
{
   POINTL ptl = atPoint.asPOINTL();
   GpiSetMix( hpsCl, FM_OVERPAINT );
   if ( GpiQueryLogColorTable( hpsCl, LCOLOPT_INDEX, 0, 0, 0 ) == QLCT_RGB )
     GpiSetColor( hpsCl, clr.asRGBLong() );
   else
     GpiSetColor( hpsCl, clr.index() );
   GpiCharStringAt( hpsCl, &ptl, strlen(pszText), (PSZ)pszText );
   return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::initializeGC                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::initializeGC ( )
{
  if ( rectCl.area() )
  {
    HRGN  oldrgn, hrgn;
    RECTL rcl = rectCl.asRECTL();

    if ( (hrgn = GpiCreateRegion( hpsCl, 0, &rcl )) == RGN_ERROR )
       ITHROWGUIERROR( "GpiCreateRegion" );
    if ( WinQueryUpdateRegion( this->handle(), hrgn ) == RGN_ERROR )
       ITHROWGUIERROR( "WinQueryUpdateRegion" );
    if ( GpiSetClipRegion( hpsCl, hrgn, &oldrgn ) == RGN_ERROR )
       ITHROWGUIERROR( "GpiSetClipRegion" );
    if ( oldrgn == HRGN_ERROR )
       ITHROWGUIERROR( "GpiSetClipRegion" );
    if ( oldrgn != NULLHANDLE  &&  !( GpiDestroyRegion( hpsCl, oldrgn )))
       ITHROWGUIERROR( "GpiDestroyRegion" );
  }
  return *this;
}
