/*******************************************************************************
* FILE NAME: ipushbut.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ipushbut.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINBUTTONS
  #define INCL_WINERRORS
  #define INCL_WINFRAMEMGR        // WM_QUERYFOCUSCHAIN
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR       // QWL_xxx
  #define INCL_GPIPRIMITIVES
  #include <iwindefs.h>
}

#ifndef _IPUSHBUT_
  #include <ipushbut.hpp>
#endif
#ifndef _ICTLEVT_
  #include <ictlevt.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IFONT_
    #include <ifont.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IHANDLER_
  #include <ihandler.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif


#ifdef IC_PAGETUNE
  #define _IPUSHBUT_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public push button styles.                                                   |
------------------------------------------------------------------------------*/

#pragma data_seg(ICLStaticConst)
const IPushButton::Style
  IPushButton::help              = BS_HELP,
  IPushButton::systemCommand     = BS_SYSCOMMAND,
  IPushButton::defaultButton     = BS_DEFAULT,
  IPushButton::noBorder          = BS_NOBORDER,
  IPushButton::classDefaultStyle = WS_VISIBLE;
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  IPushButton::Style IPushButton::currentDefaultStyle = WS_VISIBLE;
#pragma data_seg()



/*------------------------------------------------------------------------------
| IPushButton::IPushButton                                                     |
|                                                                              |
| Constructor to create a push button control on a standard window.            |
------------------------------------------------------------------------------*/
IPushButton :: IPushButton ( unsigned long ulId,
                             IWindow* parent,
                             IWindow* pwndOwner,
                             const IRectangle& rectInit,
                             const Style& pbsStyle )
  : fPushButtonData( 0 )
{
  IASSERTPARM(parent!=0);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | pbsStyle.asExtendedUnsignedLong() );

  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle whPushButton =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( pbsStyle ),
                      WC_BUTTON,
                      parent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0 );

  startHandlingEventsFor(whPushButton);
}


/*------------------------------------------------------------------------------
| IPushButton::IPushButton                                                     |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
IPushButton :: IPushButton ( unsigned long ulId,
                             IWindow* pdlgwndParent )
  : fPushButtonData( 0 )
{
  setAutoDestroyWindow  ( false );
  startHandlingEventsFor( ulId,
                          pdlgwndParent );
}


/*------------------------------------------------------------------------------
| IPushButton::IPushButton                                                     |
|                                                                              |
| Constructor to instantiate from an existing push button control.             |
------------------------------------------------------------------------------*/
IPushButton :: IPushButton ( const IWindowHandle& wh )
  : fPushButtonData( 0 )
{
  setAutoDestroyWindow  ( false );
  startHandlingEventsFor( wh );
}

/*------------------------------------------------------------------------------
| IPushButton::~IPushButton                                                    |
------------------------------------------------------------------------------*/
IPushButton :: ~IPushButton ( )
{
  // delete fPushButtonData;
}


/*------------------------------------------------------------------------------
| IPushButton::defaultStyle                                                    |
------------------------------------------------------------------------------*/
IPushButton::Style  IPushButton :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IPushButton::setDefaultStyle                                                 |
------------------------------------------------------------------------------*/
void  IPushButton :: setDefaultStyle( const Style& pbsStyle )
{
  currentDefaultStyle = pbsStyle;
}


/*------------------------------------------------------------------------------
| IPushButton::convertToGUIStyle                                               |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IPushButton::convertToGUIStyle( const IBitFlag& guiStyle,
                                              Boolean bExtOnly ) const
{
  // Obtain the style from the class (IButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // Set the GUI pushbutton style


      ulStyle |= BS_PUSHBUTTON
      ;
  }


  return ulStyle;
}


/*------------------------------------------------------------------------------
| IPushButton::enableDefault                                                   |
|                                                                              |
| Sets the push button to be the default button.                               |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: enableDefault ( Boolean enable )
{
  // Give the button the default style, as appropriate.

  IWindowHandle hwndButton = this->handle();
  unsigned long ulStyle = style();
  if (enable)
  {
    if (!(ulStyle & defaultButton.asUnsignedLong()))
    {
      hwndButton.sendEvent( BM_SETDEFAULT, IEventParameter1( TRUE ),
                                           IEventParameter2( 0 ));
    }
  }
  else if (ulStyle & defaultButton.asUnsignedLong())
  {
    hwndButton.sendEvent( BM_SETDEFAULT, IEventParameter1( FALSE ),
                                         IEventParameter2( 0 ));
  }

  // Update the default push button maintained by the frame.
  IEventResult mr =          // Find the frame window.
            hwndButton.sendEvent( WM_QUERYFOCUSCHAIN,
                                  IEventParameter1( QFC_FRAME ),
                                  IEventParameter2( 0 ));
  IWindowHandle hwndFrame( mr.asUnsignedLong() );
  if (hwndFrame)
  {                          // Found the frame window.
     // Update frame's default, as appropriate.
     if (enable)
     {                       // Store this default button.
        ISETWINDOWULONG( hwndFrame, QWL_DEFBUTTON, hwndButton );
     }
     else if (IQUERYWINDOWULONG( hwndFrame, QWL_DEFBUTTON )
                                                 == hwndButton)
     {                       // Remove this default from frame.
        ISETWINDOWULONG( hwndFrame, QWL_DEFBUTTON, 0 );
     }
     // (Can't use IWindow::windowULong or IWindow::setWindowData,
     // because they are protected.)
  }

  return *this;
}


/*------------------------------------------------------------------------------
| IPushButton::disableDefault                                                  |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: disableDefault()
{
  enableDefault(false);
  return *this;
}


/*------------------------------------------------------------------------------
| IPushButton::isDefault                                                       |
------------------------------------------------------------------------------*/
Boolean  IPushButton :: isDefault() const
{
  return( (style() & defaultButton.asUnsignedLong()) ? true : false );
}

/*------------------------------------------------------------------------------
| IPushButton::enableHelp                                                      |
|                                                                              |
| Sets the help style which causes a help event to generated when the push     |
| button is pressed.                                                           |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: enableHelp(Boolean bTurnOn)
{
  unsigned long ulStyle = style();
  unsigned long helpStyle = help.asUnsignedLong();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle |= helpStyle;
  }
  else
  {
    ulStyle &= ~helpStyle;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);

  }
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::disableHelp                                                     |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: disableHelp()
{
  enableHelp(false);
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::isHelp                                                          |
------------------------------------------------------------------------------*/
Boolean  IPushButton :: isHelp() const
{
  return( (style() & help.asUnsignedLong()) ? true : false );
}

/*------------------------------------------------------------------------------
| IPushButton::enableSystemCommand                                             |
|                                                                              |
| Sets the system command style which causes a system command to be generated  |
| when the push button is pressed.                                             |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: enableSystemCommand(Boolean bTurnOn)
{
  unsigned long ulStyle = style();
  unsigned long newStyle = systemCommand.asUnsignedLong();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle |= newStyle;
  }
  else
  {
    ulStyle &= ~newStyle;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::disableSystemCommand                                            |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: disableSystemCommand()
{
  enableSystemCommand(false);
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::isSystemCommand                                                 |
------------------------------------------------------------------------------*/
Boolean  IPushButton :: isSystemCommand() const
{
  return( (style() & systemCommand.asUnsignedLong()) ? true : false );
}

/*------------------------------------------------------------------------------
| IPushButton::addBorder                                                       |
|                                                                              |
| Sets the noBorder style indicating the presence of a border around the       |
| push button.                                                                 |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: addBorder( Boolean add )
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;

  if ( add )
  {
    ulStyle &= ~noBorder.asUnsignedLong();
  }
  else
  {
    ulStyle |= noBorder.asUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();                           // Force refresh of control window
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::removeBorder                                                    |
------------------------------------------------------------------------------*/
IPushButton&  IPushButton :: removeBorder( )
{
  addBorder(false);
  return *this;
}

/*------------------------------------------------------------------------------
| IPushButton::hasBorder                                                       |
------------------------------------------------------------------------------*/
Boolean  IPushButton :: hasBorder ( ) const
{
  return( (style() & noBorder.asUnsignedLong()) ? false : true );
}


/*------------------------------------------------------------------------------
| IPushButton::calcMinimumSize                                                 |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The height is      |
| based on the font height and width is based on the average font width        |
| multiplied by the width of the push button text.                             |
------------------------------------------------------------------------------*/
ISize  IPushButton :: calcMinimumSize() const
{
  ISize size = displaySize(text());
  unsigned long ulWidth = size.width();
  unsigned long i=textLength();

  if (i==0)
     ulWidth += (2 + (9 * IQUERYSYSVALUE( SV_CXBORDER )));
  else
     ulWidth += (ulWidth/i  + 2 + (9 * IQUERYSYSVALUE( SV_CXBORDER )));
  unsigned long ulHeight = size.height() + (9 * IQUERYSYSVALUE( SV_CYBORDER ));
  return ISize(ulWidth, ulHeight);
}

