/*******************************************************************************
* FILE NAME: iradiobt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iradiobt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINBUTTONS
  #define INCL_WINERRORS
  #define INCL_GPIBITMAPS
  #define INCL_WINPOINTERS
  #define INCL_WINSYS
  #define INCL_GPIPRIMITIVES
  #include <iwindefs.h>
}

#ifndef _IRADIOBT_
  #include <iradiobt.hpp>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _IFONT_
    #include <ifont.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IRADIOBT_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Public radio button styles.                                                  |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IRadioButton::Style
  IRadioButton::noCursorSelect    = BS_NOCURSORSELECT,
  IRadioButton::autoSelect        ( 0, IBS_AUTORADIOBUTTON ),
  IRadioButton::classDefaultStyle ( WS_VISIBLE, IBS_AUTORADIOBUTTON );

const unsigned long
  IRadioButton::noSelections      = (unsigned long)ULONG_MAX;
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  IRadioButton::Style
        IRadioButton::currentDefaultStyle ( WS_VISIBLE, IBS_AUTORADIOBUTTON );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Class:  IRadioButtonData                                                     |
|                                                                              |
| Store private data used by IRadioButton.                                     |
------------------------------------------------------------------------------*/
class IRadioButtonData : public IBase {        
public:                                        
static ISize                                   
  gBitmapSize;                                 
}; // IRadioButtonData                         

#pragma data_seg(ICLNonConst)                  
  ISize IRadioButtonData::gBitmapSize ( 0, 0 );
#pragma data_seg()                             


/*------------------------------------------------------------------------------
| IRadioButton::IRadioButton                                                   |
|                                                                              |
| Constructor to create a radio button on a standard window.                   |
------------------------------------------------------------------------------*/
IRadioButton :: IRadioButton ( unsigned long ulId,
                               IWindow* parent,
                               IWindow* pwndOwner,
                               const IRectangle& rectInit,
                               const Style& rbsStyle )
  : fRadioButtonData( 0 )
{
  IASSERTPARM(parent!=0);             //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | rbsStyle.asExtendedUnsignedLong() );

  //Create radio button control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle whRadioButton =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( rbsStyle ),
                      WC_BUTTON,
                      parent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0 );

  startHandlingEventsFor(whRadioButton);
}


/*------------------------------------------------------------------------------
| IRadioButton::IRadioButton                                                   |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
IRadioButton :: IRadioButton ( unsigned long id,
                               IWindow* parentDialog)
                               : fRadioButtonData( 0 )
{
  setAutoDestroyWindow( false );
  startHandlingEventsFor(id, parentDialog);
}


/*------------------------------------------------------------------------------
| IRadioButton::IRadioButton                                                   |
|                                                                              |
| Constructor to instantiate from an existing radio button control.            |
------------------------------------------------------------------------------*/
IRadioButton :: IRadioButton ( const IWindowHandle& wh )
                               : fRadioButtonData( 0 )
{
  setAutoDestroyWindow( false );
  startHandlingEventsFor( wh );
}


/*------------------------------------------------------------------------------
| IRadioButton:~IRadioButton                                                   |
------------------------------------------------------------------------------*/
IRadioButton :: ~IRadioButton ( )
{
  // delete fRadioButtonData;
}


/*------------------------------------------------------------------------------
| IRadioButton:defaultStyle                                                    |
|                                                                              |
| Return the default style for new radio button objects.                       |
------------------------------------------------------------------------------*/
IRadioButton::Style  IRadioButton :: defaultStyle ( )
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IRadioButton:setDefaultStyle                                                 |
|                                                                              |
| Replace the default style for new radio button objects.                      |
------------------------------------------------------------------------------*/
void IRadioButton :: setDefaultStyle ( const Style& rbsStyle )
{
  currentDefaultStyle = rbsStyle;
}


/*------------------------------------------------------------------------------
| IRadioButton::convertToGUIStyle                                              |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IRadioButton::convertToGUIStyle(const IBitFlag& guiStyle,
                                              Boolean bExtOnly) const
{
  // Obtain the style from the class (ISettingButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // BS_ styles in the lower 4 bits of the style overlap.  Therefore, map
    // our extended style bits, back to these values.
    unsigned long ulTmp = extendedStyle() & IBS_EXTRADIOBTN;

    // Handle overlapping styles
    if (ulTmp & IBS_AUTORADIOBUTTON)
      ulStyle |= BS_AUTORADIOBUTTON;
    else
      ulStyle |= BS_RADIOBUTTON;


  }


  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IRadioButton:enableAutoSelect                                                |
|                                                                              |
| Sets or removes the autoSelect style for the radio button control.           |
------------------------------------------------------------------------------*/
IRadioButton&  IRadioButton :: enableAutoSelect ( Boolean enable )
{
  unsigned long ulOldStyle = this->style();
  unsigned long ulStyle = ( ulOldStyle & (unsigned long)(~BS_PRIMARYSTYLES) );
                                       // Use masking style from pmwin.h.

  if ( enable )
  {
     ulStyle |= BS_AUTORADIOBUTTON;
  }
  else
  {
     ulStyle |= BS_RADIOBUTTON;
  }

  if ( ulStyle != ulOldStyle )
  {
     this->setStyle( ulStyle );
     this->refresh();                  // Force repaint.
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IRadioButton:disableAutoSelect                                               |
------------------------------------------------------------------------------*/
IRadioButton&  IRadioButton :: disableAutoSelect()
{
  enableAutoSelect(false);
  return *this;
}


/*------------------------------------------------------------------------------
| IRadioButton:isAutoSelect                                                    |
------------------------------------------------------------------------------*/
Boolean  IRadioButton :: isAutoSelect() const
{
  return( (style() & BS_AUTORADIOBUTTON & (unsigned long)(~BS_RADIOBUTTON)) ? true : false );
}


/*------------------------------------------------------------------------------
| IRadioButton:enableCursorSelect                                              |
|                                                                              |
| Removes or sets the noCursorSelect style.  The noCursorSelect style causes   |
| the radio button not to select itself when given the focus as the result     |
| of an arrow key or tab key.  This style only applies to autoSelect radio     |
| buttons.                                                                     |
------------------------------------------------------------------------------*/
IRadioButton&  IRadioButton :: enableCursorSelect(Boolean bTurnOn)
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle &= ~noCursorSelect.asUnsignedLong();
  }
  else
  {
    ulStyle |= noCursorSelect.asUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IRadioButton:disableCursorSelect                                             |
------------------------------------------------------------------------------*/
IRadioButton&  IRadioButton :: disableCursorSelect()
{
  enableCursorSelect(false);
  return *this;
}


/*------------------------------------------------------------------------------
| IRadioButton:isCursorSelect                                                  |
------------------------------------------------------------------------------*/
Boolean  IRadioButton :: isCursorSelect() const
{
  unsigned long ulStyle = style();

  if ((ulStyle & BS_AUTORADIOBUTTON & (unsigned long)(~BS_RADIOBUTTON)) &&
      (ulStyle & noCursorSelect.asUnsignedLong()))
    return false;
  else
    return true;
}

/*------------------------------------------------------------------------------
| IRadioButton:selectedIndex                                                   |
|                                                                              |
| Returns the zero-based index of a selected radio button in a group.          |
| NOTE:  noSelections (-1) is returned if nothing is selected.                 |
------------------------------------------------------------------------------*/
unsigned long  IRadioButton :: selectedIndex() const
{
  unsigned long index = noSelections;
  IEventResult evt = handle().sendEvent(BM_QUERYCHECKINDEX,
                             IEventParameter1(0),
                             IEventParameter2(0));
  if (evt.asLong() != -1)
  {                // A button is selected, so return its index.
     index = evt.asUnsignedLong();
  }

  return index;
}


/*------------------------------------------------------------------------------
| IRadioButton:calcMinimumSize                                                 |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The minimum size   |
| is based on the maximum height and width of the font of the radio button     |
| text (plus the size of the radio button itself).                             |
------------------------------------------------------------------------------*/
ISize  IRadioButton :: calcMinimumSize() const
{
   long bmpWidth = IRadioButtonData::gBitmapSize.width();
   if ( bmpWidth == 0 )
   {            // Find size of radio button bitmap.
      BITMAPINFOHEADER pbmpData;
      pbmpData.cbFix=sizeof(BITMAPINFOHEADER);
      HBITMAP hbit = (HBITMAP) IGETSYSBITMAP(SBMP_CHECKBOXES);
      IQUERYBITMAPPARAMETERS(hbit, &pbmpData);
      IDELETEBITMAP(hbit);

      bmpWidth = pbmpData.cx;                                      
      IRadioButtonData::gBitmapSize.setWidth( bmpWidth );          
      IRadioButtonData::gBitmapSize.setHeight( pbmpData.cy );      
   }                                                            

   ISize size = displaySize(text());
   unsigned long ulWidth = size.width();
   unsigned long i=textLength();

   if (i==0)
     ulWidth += ((bmpWidth/4) + 2 + (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   else
     ulWidth += (ulWidth/i + (bmpWidth/4) + 2 +
                                    (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   unsigned long ulHeight = size.height() +
                                     (2 * IQUERYSYSVALUE(SV_CYBORDER));
   return ISize(ulWidth, ulHeight);
}
