/*******************************************************************************
* FILE NAME: irecohdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in irecohdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINWINDOWMGR      // WinSetMultWindowPos, WinQueryWindow
#define INCL_WINFRAMEMGR       // WM_WINDOWPOSCHANGED
#define INCL_WINMESSAGEMGR          // for WM_CALCVALIDRECTS
extern "C" {
  #include <iwindefs.h>
}

#include <irecohdr.hpp>
#include <icoordsy.hpp>
#include <ievent.hpp>
#include <iexcept.hpp>
#include <iframe.hpp>
#include <ihandle.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <iwindow.hpp>
#include <itrace.hpp>
#include <ithread.hpp>
#include <iswp.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IRECOHDR_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Default IRecoordHandler object                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointed               |
| to object when the wrapper is destructed.                                    |
------------------------------------------------------------------------------*/
static struct DefaultHandler
{
  operator IRecoordHandler* ( );

IRecoordHandler
 *operator -> ( )
  {
    return( *this );
  }

  ~DefaultHandler ( )
   {
     if (ptr)
     {
       delete ptr;
       ptr = 0;
     }
   }

IRecoordHandler
 *ptr;

} gdefaultHandler;

/*------------------------------------------------------------------------------
| DefaultHandler::operator IRecoordHandler*                                    |
------------------------------------------------------------------------------*/
DefaultHandler :: operator IRecoordHandler* ( )
{
  if (!ptr)
    ptr = new IRecoordHandler;

  return( ptr );
}

/*------------------------------------------------------------------------------
| IRecoordHandler::defaultHandler                                              |
------------------------------------------------------------------------------*/
IRecoordHandler*  IRecoordHandler::defaultHandler( )
{
   return gdefaultHandler;
}

/*------------------------------------------------------------------------------
| IRecoordHandler::IRecoordHandler                                             |
------------------------------------------------------------------------------*/
IRecoordHandler :: IRecoordHandler ( )
  : IRecoordHandler::Inherited ( )
{}

/*------------------------------------------------------------------------------
| IRecoordHandler::~IRecoordHandler                                            |
------------------------------------------------------------------------------*/
IRecoordHandler :: ~IRecoordHandler ( )
{}


/************************************************************************/
/* dispatchhandlerEvent - Dispatch command functions for this handler.  */
/************************************************************************/
IBase::Boolean  IRecoordHandler::dispatchHandlerEvent ( IEvent& evt )
{
   Boolean bRc = false;
   // IRecoordHandler needs to  see resize events to work properly.
   // In PM, it is possible that a user provided IResizeHandler could
   // stop the WM_WINDOWPOSCHANGED message that we look for.  Therefore,
   // IResizeHandler generates a substitute in this case.
   // Note that this logic relies on the fact that there is
   // no ICLUI handler of WM_WINDOWPOSCHANGED which returns
   // true.  This means that the IC_UM... message ONLY occurs
   // if a user supplied handler returns true before the
   // IRecoordHandler has run.
   // This is not a problem in Windows since we use different
   // messages in IRecoordHandler and IResizeHandler.
   if ((evt.eventId() == WM_WINDOWPOSCHANGED ) ||
       (evt.eventId() == IC_UM_WINDOWPOSCHANGED ) )
      {
      // Got a CALCSIZE message
      SWP* pswp =  (SWP*)
         (void*)evt.parameter1();
      if ( (pswp->fl & SWP_SIZE) )
         {
         // resizing occurred
         SWP*    pswp = (SWP*)(void*)evt.parameter1();
         IRectangle newRect( IPoint(pswp[0].x, pswp[0].y),
                             ISize(pswp[0].cx, pswp[0].cy)  );
         IRectangle oldRect( IPoint(pswp[1].x, pswp[1].y),
                             ISize(pswp[1].cx, pswp[1].cy)  );
         bRc = windowSizeChanged( evt.handle(), newRect, oldRect );
         }  // resizing occurred
      if ( bRc )
         {                        // event has been processed
         /*******************************************************/
         /* The handler has said not to pass on the event, so   */
         /* don't pass it to another handler.  However, we do   */
         /* go ahead and pass it to the window procedure of the */
         /* current window, since it is only a notification     */
         /* message and we could break something by not doing   */
         /* this.                                               */
         /*******************************************************/
         defaultProcedure(evt);
         }  // event processed
      } //

   return bRc;
}

/*------------------------------------------------------------------------------
| IRecoordHandler::windowSizeChanged                                           |
------------------------------------------------------------------------------*/
IBase::Boolean IRecoordHandler::windowSizeChanged (
                      IWindowHandle     handle,
                      const IRectangle& newWindowRect,
                      const IRectangle& oldWindowRect  )
{
   IMODTRACE_ALL("IRecoordHandler::windowSizeChanged");
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      // Only need to do this if operating in non-native coordinates
      // delta is value to be added to current y position of child windows
      long delta = newWindowRect.height() - oldWindowRect.height();

      // Determine if the window is a frame.  Need this later for
      // handling extensions.
      IWindow* pwin = IWindow::windowWithHandle( handle );
      Boolean isFrame = false;
      if (pwin)
         isFrame = pwin->isFrameWindow();
      HENUM henum = WinBeginEnumWindows( handle );
      IWindowHandle childHwnd = WinGetNextWindow( henum ) ;
      unsigned long maxChildren = 10;
      ISWP* swpArray = new ISWP[maxChildren];
      unsigned long numChildren = 0;

      while (childHwnd != 0)
         {
         Boolean processChild = true;
         // We need to exclude frame extensions and the client window
         // from adjustment by this handler.   They are handled by
         // the frame extensions code and/or the system.
         if (isFrame)
            {
            unsigned long id = IIDOF( childHwnd );
            switch (id)
               {
               case FID_SYSMENU     :
               case FID_TITLEBAR    :
               case FID_MINMAX      :
               case FID_VERTSCROLL  :
               case FID_HORZSCROLL  :
               case FID_MENU        :
               case FID_CLIENT      :
                  // skip system extensions and client
                  processChild = false;
                  break;
               default:
                  {
                  // check for IFrameWindow extensions
                  IWindow* childWin = IWindow::windowWithHandle( childHwnd );
                  if (childWin)
                     {
                     if ( ((IFrameWindow*)pwin)->isAnExtension( childWin ) )
                        processChild = false;
                     }
                  }
                  break;
               }
            }

         if (processChild)
            {
         // get current position of child window
            if ( maxChildren <= numChildren+1 )
               {
               // reallocate the swpArray
               maxChildren += 10;
               ISWP*  temp = swpArray;
               swpArray = new ISWP[maxChildren];
               unsigned long i;
               for ( i=0; i < numChildren; i++)
                  swpArray[i] = temp[i];
               delete [] temp;
               }
            if (! WinQueryWindowPos( childHwnd,
                                     (SWP*)&(swpArray[numChildren]) ) )
               {
               ITHROWGUIERROR2("WinQueryWindowPos",
                               IErrorInfo::invalidParameter,
                               IException::unrecoverable);
               }
            // MOVE & SIZE on in return value of WinQueryWindowPos.  Not
            // resizing or messing with Z order
            swpArray[numChildren].setSizeFlag(false);

            // adjust origin to keep relative to logical origin
            IPoint   pos = swpArray[numChildren].position();
            swpArray[numChildren].setPosition( IPoint(pos.x(), pos.y() + delta) );

            ITRACE_ALL( IString("WindowPos hwnd=") +
                        IString( childHwnd.asUnsigned() ).d2x() +
                        IString(" delta=") + IString(delta) +
                        IString(" origin=") +
                        swpArray[numChildren].position().asString()  );

            numChildren++;
            }  // if process child

         childHwnd =
            WinGetNextWindow( henum ) ;
         }   // while

      if (numChildren)
         {
         if (! WinSetMultWindowPos(IThread::current().anchorBlock(),
                                   (SWP*)swpArray,
                                   numChildren) )
            {                         // Error from WinSetMultWindowPos
            ITHROWGUIERROR2("WinSetMultWindowPos",
                            IErrorInfo::invalidParameter,
                            IException::unrecoverable);
            }
         }

      delete [] swpArray;
      WinEndEnumWindows( henum );
      }
   return false;      //  pass on to other handlers
}
