/*******************************************************************************
* FILE NAME: iscrlevt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iscrlevt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WINMESSAGEMGR
#define INCL_WINSCROLLBARS
extern "C" {
  #include <iwindefs.h>
}

#ifndef _ISCRLEVT_
  #include <iscrlevt.hpp>
#endif

#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _ISCROLL_
  #include <iscroll.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISCRLEVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IScrollEvent::IScrollEvent                                                   |
------------------------------------------------------------------------------*/
IScrollEvent :: IScrollEvent ( const IEvent& event )
  : IScrollEvent::Inherited ( event ),
    ulClPrevScrollBoxPos ( 0 )
{
  IScrollBar * scrollWindow;

  /*****************************************************************/
  /* Try to keep scrollAmount() independent of when/if the user    */
  /* repositions the scroll box.  The case where this is still a   */
  /* problem is if the scroll box is repositioned by one handler,  */
  /* and the second handler needs to call scrollAmount(), since    */
  /* the old scroll box position has now been lost so the delta    */
  /* can't be computed.                                            */
  /*****************************************************************/
  scrollWindow = scrollBarWindow();
  if (scrollWindow)
    ulClPrevScrollBoxPos = scrollWindow->scrollBoxPosition();
}


/*------------------------------------------------------------------------------
| IScrollEvent::~IScrollEvent                                                  |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IScrollEvent :: ~IScrollEvent ( )
{ ; }


/*------------------------------------------------------------------------------
| IScrollEvent::scrollBarWindow                                                |
|                                                                              |
| Return the scroll bar object.                                                |
------------------------------------------------------------------------------*/
IScrollBar  *IScrollEvent :: scrollBarWindow ( ) const
{
  HWND hwndCtl = IWINDOWFROMID(handle(), parameter1().number1());
  IWindow* pwin = 0;

  if (hwndCtl)
  {                                    //Valid handle
     pwin = IWindow::windowWithHandle(hwndCtl);
  }
  else
  {                                   //Parent was not the owner
     IWindow* owner = IWindow::windowWithHandle(handle());
     if (owner)
     {
        pwin = IWindow::windowWithOwner(parameter1().number1(), owner);
     }
  }
  return (IScrollBar*)pwin;
}


/*------------------------------------------------------------------------------
| IScrollEvent::scrollAmount                                                   |
|                                                                              |
| Return the number of units the scroll box has moved.                         |
------------------------------------------------------------------------------*/
long  IScrollEvent :: scrollAmount ( ) const
{
   long lScrolledUnits = 0;
   IScrollBar* pscrbar = scrollBarWindow();

   if (pscrbar != 0 )
   {
      switch (parameter2().number2())
      {
         case SB_LINEUP:               //(same as SB_LINELEFT)
            lScrolledUnits = -pscrbar->minScrollIncrement();
            break;
         case SB_LINEDOWN:             //(same as SB_LINERIGHT)
            lScrolledUnits = pscrbar->minScrollIncrement();
            break;
         case SB_PAGEUP:               //(same as SB_PAGELEFT)
            lScrolledUnits = -pscrbar->pageScrollIncrement();
            break;
         case SB_PAGEDOWN:             //(same as SB_PAGERIGHT)
            lScrolledUnits = pscrbar->pageScrollIncrement();
            break;
         case SB_SLIDERTRACK:
         case SB_SLIDERPOSITION:
            lScrolledUnits = newScrollBoxPosition() -
                                 ulClPrevScrollBoxPos;
            break;
         case SB_ENDSCROLL:
            lScrolledUnits = 0;        //All scrolling is done
            break;
         default:
         {                             //Unknown scroll type
            ITHROWLIBRARYERROR(IC_INVALIDEVENT,
                               IErrorInfo::invalidRequest,
                               IException::recoverable);

            break;
         }
      }

      /*************************************************************/
      /* Check for an out-of-scroll-unit bound condition and       */
      /* adjust the position of the scroll box accordingly.        */
      /*************************************************************/
      IRange rangeScrollBox = pscrbar->scrollBoxRange();
      long lMaxScroll =
              rangeScrollBox.lowerBound() - ulClPrevScrollBoxPos;
      if (lScrolledUnits < lMaxScroll)
      {                                //Hitting minimum limit
         lScrolledUnits = lMaxScroll;
      }
      else
      {                               //Scrolling past top/right?
         lMaxScroll =
            rangeScrollBox.upperBound() - ulClPrevScrollBoxPos;
         if (lScrolledUnits > lMaxScroll)
         {                            //Hitting maximum limit
            lScrolledUnits = lMaxScroll;
         }
      }
   }

   return lScrolledUnits;
}


/*------------------------------------------------------------------------------
| IScrollEvent::newScrollBoxPosition                                           |
|                                                                              |
| Return the new scroll box position.  Only valid for SB_SLIDERTRACK and       |
| SB_SLIDERPOSITION.                                                           |
------------------------------------------------------------------------------*/
unsigned long  IScrollEvent :: newScrollBoxPosition ( ) const
{
  switch (parameter2().number2())
  {                                    //Switch on scroll bar event
     case SB_SLIDERPOSITION:
     case SB_SLIDERTRACK:              //Above are valid cases
        break;
     case SB_LINEUP:                   //(Same as SB_LINELEFT)
     case SB_LINEDOWN:                 //(Same as SB_LINERIGHT)
     case SB_PAGEUP:                   //(Same as SB_PAGELEFT)
     case SB_PAGEDOWN:                 //(Same as SB_PAGERIGHT)
     case SB_ENDSCROLL:
     default:                          //Above are invalid cases
     {
        ITHROWLIBRARYERROR(IC_INVALIDEVENT,
                           IErrorInfo::invalidRequest,
                           IException::recoverable);
        break;
     }
  }

  return parameter2().number1();
}
