/*******************************************************************************
* FILE NAME: ispinnum.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ispinnum.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINSTDSPIN
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <ispinnum.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <istring.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISPINNUM_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public spin button styles.                                                   |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const INumericSpinButton::Style
  INumericSpinButton::padWithZeros      = SPBS_PADWITHZEROS,
  INumericSpinButton::classDefaultStyle ( WS_VISIBLE    |
                                          SPBS_MASTER   |
                                          SPBS_JUSTLEFT |
                                          SPBS_NUMERICONLY,
                                          0 );
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
INumericSpinButton::Style
  INumericSpinButton::currentDefaultStyle ( WS_VISIBLE    |
                                            SPBS_MASTER   |
                                            SPBS_JUSTLEFT |
                                            SPBS_NUMERICONLY,
                                            0 );
#pragma data_seg()

/*------------------------------------------------------------------------------
| INumericSpinButton::INumericSpinButton                                       |
|                                                                              |
| Constructor to create a spin button control on a standard window.            |
------------------------------------------------------------------------------*/
INumericSpinButton :: INumericSpinButton ( unsigned long id,
                                           IWindow* parent,
                                           IWindow* pwndOwner,
                                           const IRectangle& initial,
                                           const Style& style )
                          : IBaseSpinButton()
{
  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  /****************************************************************/
  /* Borland has bug with embedded ? operator so below is needed. */
  /****************************************************************/
  IWindowHandle owner = (pwndOwner == 0) ? IWindowHandle(0)
                                         : pwndOwner->handle();
  initialize ( id,
               parent->handle(),
               owner,
               convertToGUIStyle( style ),
               initial );
}

/*------------------------------------------------------------------------------
| INumericSpinButton::INumericSpinButton                                       |
|                                                                              |
| Constructor to create a spin button control on a standard window.            |
------------------------------------------------------------------------------*/
INumericSpinButton :: INumericSpinButton ( unsigned long id, IWindow* parent )
      :  IBaseSpinButton()
{
   IWindowHandle
     hwnd = IWindow::handleWithParent( id, parent->handle() );

   if ( !(ISTYLEOF(hwnd) & SPBS_NUMERICONLY) )
      ITHROWLIBRARYERROR(IC_NUMSPIN_WRAP_INVALID,
                         IErrorInfo::invalidRequest, IException::recoverable);
   setAutoDestroyWindow(false);
   initialize();
   startHandlingEventsFor( hwnd );
}

/*------------------------------------------------------------------------------
| INumericSpinButton::INumericSpinButton                                       |
|                                                                              |
| Constructor to instantiate spin button object from window handle.            |
------------------------------------------------------------------------------*/
INumericSpinButton :: INumericSpinButton ( const IWindowHandle& handle )
     :  IBaseSpinButton()
{
   if ( !(ISTYLEOF(handle) & SPBS_NUMERICONLY) )
      ITHROWLIBRARYERROR(IC_NUMSPIN_WRAP_INVALID,
                         IErrorInfo::invalidRequest, IException::recoverable);
   setAutoDestroyWindow(false);
   initialize();
   startHandlingEventsFor(handle);
}

/*------------------------------------------------------------------------------
| INumericSpinButton::~INumericSpinButton                                      |
------------------------------------------------------------------------------*/
INumericSpinButton::~INumericSpinButton ( )
{
}

/*------------------------------------------------------------------------------
| INumericTextSpinButton::defaultStyle                                         |
|                                                                              |
| Return the default style for new spin button objects.                        |
------------------------------------------------------------------------------*/
INumericSpinButton::Style  INumericSpinButton :: defaultStyle()
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| INumericSpinButton::convertToGUIStyle                                        |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long INumericSpinButton :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                      Boolean bExtOnly) const
{
  // Obtain the style from the class that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean SPBS_ styles, flow thru
    // Add SPBS_NUMERICONLY since we only want to do numeric spin buttons
    ulStyle |= guiStyle.asUnsignedLong() & ISPBS_MASK;
    ulStyle |= SPBS_NUMERICONLY;
    unsigned long ulTmp = extendedStyle();

    // Strip the master style bit if the servant extended style bit is defined
    if (ulTmp & servant.asExtendedUnsignedLong())
    {
      ulStyle &= ~master.asUnsignedLong();
    }

    // Add SPBS_JUSTCENTER if the extended style bit is defined
    if (ulTmp & centerAlign.asExtendedUnsignedLong())
    {
      ulStyle |= SPBS_JUSTCENTER;
    }
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| INumericSpinButton::setRange                                                 |
|                                                                              |
| Sets the lower and upper range of the spin button numbers.  If the value in  |
| the spin field does not match the range, PM sets the spin field to the       |
| closest limit.  If the range is empty, PM puts 0 in the spin field.  If      |
| override is set and the current value in the spin field is outside the       |
| new range, the current value in the spin field will not change.              |
------------------------------------------------------------------------------*/
INumericSpinButton& INumericSpinButton :: setRange ( const IRange& aNewRange,
                                                     Boolean override )
{
   IASSERTPARM(aNewRange.lowerBound() <= aNewRange.upperBound());
                         //If lower is greater than upper, operation will fail.

   if (override)
      handle().sendEvent(SPBM_OVERRIDESETLIMITS,
                         IEventParameter1((int)aNewRange.upperBound()),
                         IEventParameter2((int)aNewRange.lowerBound()));
   else
      handle().sendEvent(SPBM_SETLIMITS,
                         IEventParameter1((int)aNewRange.upperBound()),
                         IEventParameter2((int)aNewRange.lowerBound()));

   return *this;
}

/*------------------------------------------------------------------------------
| INumericSpinButton::range                                                    |
|                                                                              |
| Query lower and upper bounds of the numeric range.                           |
------------------------------------------------------------------------------*/
IRange INumericSpinButton :: range ( ) const
{
   long lLow = 0;
   long lHigh = 0;
   unsigned long ulOk = handle().sendEvent(SPBM_QUERYLIMITS,
                                           IEventParameter1((void*)&lHigh),
                                           IEventParameter2((void*)&lLow));
   if (!ulOk)
      ITHROWGUIERROR2("SPBM_QUERYLIMITS",
                      IErrorInfo::accessError,
                      IException::recoverable);
   return IRange(lLow, lHigh);
}

/*------------------------------------------------------------------------------
| INumericSpinButton::value                                                    |
|                                                                              |
| Return the number in the spin field.                                         |
------------------------------------------------------------------------------*/
long INumericSpinButton :: value ( ) const
{
   IMODTRACE_DEVELOP("INumericSpinButton::value");

   IWindowHandle ef = IWindow::handleWithParent( this->id(), this->handle() );

   // Work-around for window id breakage in Warp GA for the child
   // entry field of a spin button control (it should be the window
   // id of the parent spin button, but is 1).
   if (!ef)
   {            // Can't find a child entry field of a spin button.
      ef = IWindow::handleWithParent( 1, this->handle() );
   }            // Try again with the one Warp GA uses.
   unsigned long ulClBufLen =
         (unsigned long)IQUERYWINDOWTEXTLENGTH(ef) + 1;
   char* pszClBuf = new char[ulClBufLen];
   IQUERYWINDOWTEXT(ef, ulClBufLen, pszClBuf);
   IString str(pszClBuf);
   delete [] pszClBuf;
   return str.asInt();
}

/*------------------------------------------------------------------------------
| INumericSpinButton::setValue                                                 |
|                                                                              |
| Set the spin field with the specified value.                                 |
------------------------------------------------------------------------------*/
INumericSpinButton& INumericSpinButton :: setValue ( long aValue )
{
  IMODTRACE_DEVELOP("INumericSpinButton::setValue");

  if (aValue != value())
  {
    IWindowHandle ef = IWindow::handleWithParent( id(), this->handle() );

    // Work-around for window id breakage in Warp GA for the child
    // entry field of a spin button control (it should be the window
    // id of the parent spin button, but is 1).
    if (!ef)
    {            // Can't find a child entry field of a spin button.
       ef = IWindow::handleWithParent( 1, this->handle() );
    }            // Try again with the one Warp GA uses.

    ITRACE_DEVELOP(IString("Value to be set is ") + IString(aValue));
    ISETWINDOWTEXT(ef, (const char*)IString(aValue));
  }
  return *this;
}

/*------------------------------------------------------------------------------
| INumericSpinButton::spinTo                                                   |
|                                                                              |
| Set the spin field to the specified numeric value.                           |
------------------------------------------------------------------------------*/
INumericSpinButton& INumericSpinButton :: spinTo ( long    aValue,
                                                   Boolean spinToClosest )
{
   unsigned long ulOk =
        handle().sendEvent(SPBM_SETCURRENTVALUE,
                           IEventParameter1((int)aValue),
                           IEventParameter2(0));
   if (!ulOk) {
      if (spinToClosest) {
         long lValue;
         ulOk = handle().sendEvent(SPBM_QUERYVALUE,
                                   IEventParameter1((void*)&lValue),
                                   IEventParameter2(0,
                                                    SPBQ_ALWAYSUPDATE));
      } else
         ITHROWGUIERROR2("SPBM_SETCURRENTVALUE",
                         IErrorInfo::invalidParameter,
                         IException::recoverable);
   }
   return *this;
}

/*------------------------------------------------------------------------------
| INumericSpinButton::isSpinFieldValid                                         |
|                                                                              |
| Validate that the contents of the spin field are in the spin field number    |
| range.                                                                       |
------------------------------------------------------------------------------*/
IBase::Boolean INumericSpinButton :: isSpinFieldValid (
                                                   Boolean caseSensitive) const
{
   long lValue;
   unsigned long ulOk;
   ulOk = handle().sendEvent(SPBM_QUERYVALUE,
                             IEventParameter1((void*)&lValue),
                             IEventParameter2(0,
                                              SPBQ_DONOTUPDATE));

   return (ulOk) ? true : false;
}
