/*******************************************************************************
* FILE NAME: ispintxt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ispintxt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINSTDSPIN
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <ispintxt.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <inotifev.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <itrace.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ISPINTXT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public spin button styles.                                                   |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ITextSpinButton::Style
  ITextSpinButton::classDefaultStyle ( WS_VISIBLE    |
                                       SPBS_MASTER   |
                                       SPBS_JUSTLEFT |
                                       SPBS_ALLCHARACTERS,
                                       0 );
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
ITextSpinButton::Style
  ITextSpinButton::currentDefaultStyle ( WS_VISIBLE    |
                                         SPBS_MASTER   |
                                         SPBS_JUSTLEFT |
                                         SPBS_ALLCHARACTERS,
                                         0 );
#pragma data_seg()

/*------------------------------------------------------------------------------
| The spin button private data class.                                          |
------------------------------------------------------------------------------*/
class ITextSpinButtonData {
public:
  ITextSpinButtonData ( );
 ~ITextSpinButtonData ( );

unsigned long
  textCount,
  fChangeCount;
IString
  spinText;
char
 **pList;
}; // ITextSpinButtonData

/*------------------------------------------------------------------------------
| ITextSpinButtonData::ITextSpinButtonData                                     |
------------------------------------------------------------------------------*/
ITextSpinButtonData::ITextSpinButtonData( )
  : textCount(0), fChangeCount(0), spinText(), pList(0)
{ }

/*------------------------------------------------------------------------------
| ITextSpinButtonData::~ITextSpinButtonData                                    |
------------------------------------------------------------------------------*/
ITextSpinButtonData::~ITextSpinButtonData( )
{
  if (pList)
    delete [] pList;
}

/*------------------------------------------------------------------------------
| The spin button cursor private data class.                                   |
------------------------------------------------------------------------------*/
class ITextSpinButtonCursorData {
public:
  ITextSpinButtonCursorData (const ITextSpinButton& spinbt)
    : lCurrent(0), fValidCount(0), rSpBut(spinbt)
  { }
long
  lCurrent;
unsigned long
  fValidCount;
const ITextSpinButton
 &rSpBut;
private:
  ITextSpinButtonCursorData ( const ITextSpinButtonCursorData& cursor );
ITextSpinButtonCursorData
 &operator=  ( const ITextSpinButtonCursorData& cursor );
}; // ITextSpinButtonCursorData

/*------------------------------------------------------------------------------
| ITextSpinButton::ITextSpinButton                                             |
|                                                                              |
| Constructor to create a spin button control on a standard window.            |
------------------------------------------------------------------------------*/
ITextSpinButton :: ITextSpinButton ( unsigned long id,
                                     IWindow* parent,
                                     IWindow* pwndOwner,
                                     const IRectangle& initial,
                                     const Style& style )
               :   IBaseSpinButton()
{
  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  if (style.asUnsignedLong() & SPBS_NUMERICONLY)
     ITHROWLIBRARYERROR(IC_TXTSPIN_WRAP_INVALID,
                        IErrorInfo::invalidRequest, IException::recoverable);

  /****************************************************************/
  /* Borland has bug with embedded ? operator so below is needed. */
  /****************************************************************/
  IWindowHandle owner = (pwndOwner == 0) ? IWindowHandle(0)
                                         : pwndOwner->handle();
   initialize ( id,
                parent->handle(),
                owner,
                convertToGUIStyle( style ),
                initial );
   fTextSpinButtonData = new ITextSpinButtonData();
}

/*------------------------------------------------------------------------------
| ITextSpinButton::ITextSpinButton                                             |
|                                                                              |
| Constructor to create a spin button control on a standard window.            |
------------------------------------------------------------------------------*/
ITextSpinButton :: ITextSpinButton ( unsigned long id, IWindow* parentDialog )
      :  IBaseSpinButton()
{
   IWindowHandle
     hwnd = IWindow::handleWithParent( id, parentDialog->handle() );

   if ( ISTYLEOF(hwnd) & SPBS_NUMERICONLY )
      ITHROWLIBRARYERROR(IC_TXTSPIN_WRAP_INVALID,
                         IErrorInfo::invalidRequest, IException::recoverable);
   setAutoDestroyWindow(false);
   initialize();
   fTextSpinButtonData = new ITextSpinButtonData();
   startHandlingEventsFor( hwnd );
}

/*------------------------------------------------------------------------------
| ITextSpinButton::ITextSpinButton                                             |
|                                                                              |
| Constructor to instantiate spin button object from window handle.            |
------------------------------------------------------------------------------*/
ITextSpinButton :: ITextSpinButton ( const IWindowHandle& handle )
     :  IBaseSpinButton()
{
   if ( ISTYLEOF(handle) & SPBS_NUMERICONLY )
      ITHROWLIBRARYERROR(IC_TXTSPIN_WRAP_INVALID,
                         IErrorInfo::invalidRequest, IException::recoverable);
   setAutoDestroyWindow(false);
   initialize();
   fTextSpinButtonData = new ITextSpinButtonData();
   startHandlingEventsFor(handle);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::~ITextSpinButton                                            |
------------------------------------------------------------------------------*/
ITextSpinButton::~ITextSpinButton ( )
{
   delete fTextSpinButtonData;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::defaultStyle                                                |
|                                                                              |
| Return the default style for new spin button objects.                        |
------------------------------------------------------------------------------*/
ITextSpinButton::Style  ITextSpinButton :: defaultStyle()
{
  return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::convertToGUIStyle                                           |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ITextSpinButton :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                   Boolean bExtOnly) const
{
  // Obtain the style from the class that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= extendedStyle() & IS_EXTMASK;
  }
  else
  {
    // Let the clean SPBS_ styles, flow thru
    // No need to OR SPBS_ALLCHARACTERS since it is defined as 0
    ulStyle |= guiStyle.asUnsignedLong() & ISPBS_MASK;
    unsigned long ulTmp = extendedStyle();

    // Strip the master style bit if the servant extended style bit is defined
    if (ulTmp & servant.asExtendedUnsignedLong())
    {
      ulStyle &= ~master.asUnsignedLong();
    }

    // Add SPBS_JUSTCENTER if the extended style bit is defined
    if (ulTmp & centerAlign.asExtendedUnsignedLong())
    {
      ulStyle |= SPBS_JUSTCENTER;
    }
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| ITextSpinButton::text                                                        |
|                                                                              |
| Query the spin field text                                                    |
------------------------------------------------------------------------------*/
IString ITextSpinButton :: text  ( ) const
{
   IWindowHandle ef = IWindow::handleWithParent( this->id(), this->handle() );

   // Work-around for window id breakage in Warp GA for the child
   // entry field of a spin button control (it should be the window
   // id of the parent spin button, but is 1).
   if (!ef)
   {            // Can't find a child entry field of a spin button.
      ef = IWindow::handleWithParent( 1, this->handle() );
   }            // Try again with the one Warp GA uses.

   unsigned long ulClBufLen =
         (unsigned long)IQUERYWINDOWTEXTLENGTH(ef) + 1;

   char* pszClBuf = new char[ulClBufLen];
   IQUERYWINDOWTEXT(ef, ulClBufLen, pszClBuf);
   IString str(pszClBuf);
   delete [] pszClBuf;
   return str;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::setText                                                     |
|                                                                              |
| Set the spin field text.                                                     |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: setText ( const char* string )
{
   IString strOrigText = this->text();
   if (strOrigText != string)
   {
      IWindowHandle ef =
             IWindow::handleWithParent( this->id(), this->handle() );

      // Work-around for window id breakage in Warp GA for the child
      // entry field of a spin button control (it should be the window
      // id of the parent spin button, but is 1).
      if (!ef)
      {            // Can't find a child entry field of a spin button.
         ef = IWindow::handleWithParent( 1, this->handle() );
      }            // Try again with the one Warp GA uses.

      ISETWINDOWTEXT(ef, string);
   }
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::setText                                                     |
|                                                                              |
| Set the spin field text.                                                     |
------------------------------------------------------------------------------*/
ITextSpinButton&  ITextSpinButton :: setText(const IResourceId& residText)
{
  IString str = residText.resourceLibrary().loadString(residText);
  return setText((char*)str);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::isSpinFieldValid                                            |
|                                                                              |
| Validate that the contents of the spin field match one of the spin text      |
| array values.                                                                |
------------------------------------------------------------------------------*/
IBase::Boolean ITextSpinButton :: isSpinFieldValid ( Boolean caseSensitive ) const
{
   long lValue;
   unsigned long valid = 0;
   if (caseSensitive)
   {
      if (fTextSpinButtonData->textCount)
         for (int i = 0; i < fTextSpinButtonData->textCount && !valid; i++)
            if (elementAt(i) == text())
               valid = true;
   }
   else
     valid = handle().sendEvent(SPBM_QUERYVALUE,
                               IEventParameter1((void*)&lValue),
                               IEventParameter2(0,SPBQ_DONOTUPDATE));

   return (valid) ? true : false;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::elementAt                                                   |
|                                                                              |
| Return the text element at cursor position.                                  |
------------------------------------------------------------------------------*/
IString ITextSpinButton::elementAt ( const Cursor& cursor ) const
{
   IMODTRACE_DEVELOP("ITextSpinButton::elementAt");
   IASSERTPARM(cursor.isValid());
   unsigned long lIndex = findStringIndex(cursor);
   return IString((char*)fTextSpinButtonData->spinText.subString(lIndex));
}

/*------------------------------------------------------------------------------
| ITextSpinButton::elementAt                                                   |
|                                                                              |
| Return the text element at the index position.                               |
------------------------------------------------------------------------------*/
IString ITextSpinButton::elementAt ( unsigned long index ) const
{
   IMODTRACE_DEVELOP("ITextSpinButton::elementAt");
   IASSERTPARM(index < fTextSpinButtonData->textCount);
   unsigned long lIndex = findStringIndex(index);
   return IString((char*)fTextSpinButtonData->spinText.subString(lIndex));
}

/*------------------------------------------------------------------------------
| ITextSpinButton::replaceAt                                                   |
|                                                                              |
| Replace the text element at the cursor position.                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::replaceAt  ( const IResourceId& resid,
                                               Cursor &cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return replaceAt(newText,cursor);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::replaceAt                                                   |
|                                                                              |
| Replace the text element at the cursor position.                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::replaceAt  ( const char* newString,
                                               Cursor &cursor )
{
   IMODTRACE_DEVELOP("ITextSpinButton::replaceAt");
   ITRACE_DEVELOP( "Cursor at item: " + IString( cursor.fCursorData->lCurrent ));

   IASSERTPARM(cursor.isValid());
   //   remove the item at the cursor
   unsigned long lIndex = findStringIndex(cursor);
   IString subStr = (char*)fTextSpinButtonData->spinText.subString(lIndex);
   fTextSpinButtonData->spinText.remove(lIndex, subStr.length()+1);
   --fTextSpinButtonData->textCount;
   if ( cursor.fCursorData->lCurrent > fTextSpinButtonData->textCount )
   {          // Replacing the last item.
      addAsLast(newString);
      cursor.setToLast();
   }
   else if (fTextSpinButtonData->textCount == 0 || cursor.fCursorData->lCurrent == 1)
   {          // Replacing the first item.
      addAsFirst(newString);
      cursor.setToFirst();
   }
   else
   {          // Replacing an item in the middle somewhere.
      --cursor.fCursorData->lCurrent;
      addAsNext(newString, cursor);
   }

   // Undo the add's increment of the change count, since we don't
   // want other cursors to be invalidated by a replace.
   --fTextSpinButtonData->fChangeCount;
   cursor.fCursorData->fValidCount = fTextSpinButtonData->fChangeCount;  // Keep cursor valid.

   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::removeAt                                                    |
|                                                                              |
| Remove the text element at cursor position.                                  |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::removeAt( Cursor& cursor )
{
   IMODTRACE_DEVELOP("ITextSpinButton::removeAt");
   IASSERTPARM(cursor.isValid());
   unsigned long lIndex = findStringIndex(cursor);
   IString subStr = (char*)fTextSpinButtonData->spinText.subString(lIndex);
   fTextSpinButtonData->spinText.remove(lIndex, subStr.length()+1);
   --fTextSpinButtonData->textCount;            // One less element now.
   ++fTextSpinButtonData->fChangeCount;

   if ( cursor.fCursorData->lCurrent <= fTextSpinButtonData->textCount )
   {          // Did not remove last element.
      cursor.fCursorData->fValidCount = fTextSpinButtonData->fChangeCount;  // Keep cursor valid.
      ITRACE_DEVELOP( "Cursor at  = " + IString( cursor.fCursorData->lCurrent ));
   }
   else
   {     // Let cursor remain invalidated, since we removed the last element.
      ITRACE_DEVELOP( "Removed item " + IString( cursor.fCursorData->lCurrent ));
      ITRACE_DEVELOP( "Cursor now invalid." );
   }

   ITRACE_DEVELOP( "New number of items = " + IString( fTextSpinButtonData->textCount ));
   refreshText();
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::removeAll                                                   |
|                                                                              |
| Remove all spin text elements.                                               |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: removeAll ( )
{
   char* nullList[1];
   nullList[0] = "";
   unsigned long ulOk =
        handle().sendEvent(SPBM_SETARRAY,
                           IEventParameter1(nullList),
                           IEventParameter2(1));
   if (!ulOk)
      ITHROWGUIERROR2("SPBM_SETARRAY",
                      IErrorInfo::accessError,
                      IException::recoverable);

   fTextSpinButtonData->spinText = "";
   fTextSpinButtonData->textCount = 0;
   if (fTextSpinButtonData->pList) {
      delete [] fTextSpinButtonData->pList;
      fTextSpinButtonData->pList = 0;
   }
   ++fTextSpinButtonData->fChangeCount;

   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::add                                                         |
|                                                                              |
| Add a new item at the cursor location by pushing all the following items     |
| after the cursor.                                                            |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: add ( const char* string, Cursor& cursor )
{
   IMODTRACE_DEVELOP("ITextSpinButton::add");
   IASSERTPARM(cursor.isValid());
   ITRACE_DEVELOP( "Previous number of items = " + IString( fTextSpinButtonData->textCount ));
   ITRACE_DEVELOP( "Cursor previously at item: " + IString( cursor.fCursorData->lCurrent ));

   if (cursor.fCursorData->lCurrent == 1) {
      addAsFirst(string);
      cursor.setToFirst();
   }
   else {
      unsigned long ulLen1 = findStringIndex(cursor) - 1;
      unsigned long ulLen2 = strlen(string) + 1;
      unsigned long ulLen3 = fTextSpinButtonData->spinText.length() - ulLen1;
      fTextSpinButtonData->spinText = IString((char*)fTextSpinButtonData->spinText,
                                ulLen1,
                                string,
                                ulLen2,
                                (char*)fTextSpinButtonData->spinText.subString(ulLen1 + 1),
                                ulLen3);
      ++fTextSpinButtonData->textCount;                     // Increment counts.
      ++fTextSpinButtonData->fChangeCount;
      cursor.fCursorData->fValidCount = fTextSpinButtonData->fChangeCount;  // Keep cursor valid.
      refreshText();
   }
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::add                                                         |
|                                                                              |
| Add a new item at the cursor location by pushing all the following items     |
| after the cursor.                                                            |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: add ( const IResourceId& resid,
                                          Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return add(newText, cursor);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::add                                                         |
|                                                                              |
| Add a new item at the index location by pushing all the following items      |
| after the index.                                                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: add ( const char* string,
                                          unsigned long index )
{
   IMODTRACE_DEVELOP("ITextSpinButton::add");
   if ( index >= fTextSpinButtonData->textCount ) {
      addAsLast(string);
   }
   else {
      unsigned long ulLen1 = findStringIndex(index) - 1;
      unsigned long ulLen2 = strlen(string) + 1;
      unsigned long ulLen3 = fTextSpinButtonData->spinText.length() - ulLen1;
      fTextSpinButtonData->spinText = IString((char*)fTextSpinButtonData->spinText,
                                ulLen1,
                                string,
                                ulLen2,
                                (char*)fTextSpinButtonData->spinText.subString(ulLen1 + 1),
                                ulLen3);
      ++fTextSpinButtonData->textCount;
      ++fTextSpinButtonData->fChangeCount;
      refreshText();
   }
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::add                                                         |
|                                                                              |
| Add a new item at the index location by pushing all the following items      |
| after the index.                                                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: add ( const IResourceId& resid,
                                          unsigned long index )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return add(newText, index);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::add                                                         |
|                                                                              |
| Add a list of new items at the index location by pushing all the following   |
| items after the index.                                                       |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: add ( const char* const * stringArray,
                                          unsigned long index,
                                          unsigned long count )
{
   IMODTRACE_DEVELOP("ITextSpinButton::add");

   if (count == 0)
      return *this;

   IString newString;
   for (int i=0; i < count; i++ ) {
      newString+=stringArray[i];
      newString+=IString('\0');
   }
   if ( index >= fTextSpinButtonData->textCount )
      fTextSpinButtonData->spinText += newString;
   else if ( index == 0)
      fTextSpinButtonData->spinText = newString + fTextSpinButtonData->spinText;
   else {
      unsigned long ulLen1 = findStringIndex(index) - 1;
      unsigned long ulLen2 = newString.length();
      unsigned long ulLen3 = fTextSpinButtonData->spinText.length() - ulLen1;
      fTextSpinButtonData->spinText = IString((char*)fTextSpinButtonData->spinText,
                                ulLen1,
                                (char*)newString,
                                ulLen2,
                                (char*)fTextSpinButtonData->spinText.subString(ulLen1 + 1),
                                ulLen3);
   }
   fTextSpinButtonData->textCount += count;
   ++fTextSpinButtonData->fChangeCount;
   refreshText();
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::addAsNext                                                   |
|   Add a text next to the current cursor position.                            |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: addAsNext  ( const char*   string,
                                                 Cursor& cursor )
{
   if (cursor.fCursorData->lCurrent == fTextSpinButtonData->textCount) {
      addAsLast(string);
      cursor.setToLast();
   }
   else {
      ++cursor.fCursorData->lCurrent;   // bump the cursor.
      add(string, cursor);
   }
   return *this;
}

ITextSpinButton& ITextSpinButton :: addAsNext  ( const IResourceId& resid,
                                                 Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return add(newText,cursor);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::addAsFirst                                                  |
|                                                                              |
| Adds the text as the first item.                                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::addAsFirst ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsFirst(newText);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::addAsFirst                                                  |
|                                                                              |
| Adds the text as the first item.                                             |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::addAsFirst ( const char* string )
{
   IMODTRACE_DEVELOP("ITextSpinButton::addAsFirst");
   unsigned long ulLen1 = strlen(string) + 1;
   unsigned long ulLen2 = fTextSpinButtonData->spinText.length();
   if (ulLen2)
      fTextSpinButtonData->spinText = IString(string, ulLen1,
                          (char*)fTextSpinButtonData->spinText, ulLen2);
   else
      fTextSpinButtonData->spinText = IString(string, ulLen1);
   ++fTextSpinButtonData->textCount;
   ++fTextSpinButtonData->fChangeCount;
   refreshText();
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::addAsLast                                                   |
|                                                                              |
| Adds the text as the last item.                                              |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::addAsLast  ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsLast(newText);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::addAsLast                                                   |
|                                                                              |
| Adds the text as the last item.                                              |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::addAsLast  ( const char* string )
{
   IMODTRACE_DEVELOP("ITextSpinButton::addAsLast");
   unsigned long ulLen1 = fTextSpinButtonData->spinText.length();
   unsigned long ulLen2 = strlen(string) + 1;
   if (ulLen1)
      fTextSpinButtonData->spinText = IString(
          (char*)fTextSpinButtonData->spinText, ulLen1, string, ulLen2);
   else
      fTextSpinButtonData->spinText = IString(string, ulLen2);
   ++fTextSpinButtonData->textCount;
   ++fTextSpinButtonData->fChangeCount;
   refreshText();
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::spinTo                                                      |
|                                                                              |
|   Set the current text to the text at the cursor position.                   |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: spinTo ( const Cursor& cursor )
{
   IASSERTPARM(cursor.isValid());
   return (spinTo((unsigned long)(cursor.fCursorData->lCurrent - 1)));
}

/*------------------------------------------------------------------------------
| ITextSpinButton::spinTo                                                      |
|                                                                              |
|   Set the current text to the text at the index position.                    |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: spinTo ( unsigned long index )
{
   IMODTRACE_DEVELOP("ITextSpinButton::spinTo");
   handle().sendEvent(SPBM_SETCURRENTVALUE,
                      IEventParameter1(index),
                      IEventParameter2(0));
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::spinTo                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton :: spinTo ( const char* string, Boolean caseSensitive )
{
   IMODTRACE_DEVELOP("ITextSpinButton::spinTo");
   IString arrayString;
   IString inputString(string);
   if (!caseSensitive)
       inputString.upperCase();
   int j;
   Boolean bFound = false;
   for (int i=0; i < fTextSpinButtonData->textCount && bFound==false; i++ )
   {
       arrayString = elementAt(i);
       if (!caseSensitive)
           arrayString.upperCase();
       if (arrayString == inputString) {
          j=i;
          bFound = true;
       }
   }
   if (bFound)
      handle().sendEvent(SPBM_SETCURRENTVALUE,
                         IEventParameter1(j),
                         IEventParameter2(0));
   else
      ITHROWLIBRARYERROR(IC_SPINTO_TEXT_INVALID,
                         IErrorInfo::invalidParameter,
                         IException::recoverable);
   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::refreshText                                                 |
|                                                                              |
| Adds or replaces elements in the spin button.                                |
------------------------------------------------------------------------------*/
ITextSpinButton& ITextSpinButton::refreshText ( )
{
   IMODTRACE_DEVELOP("ITextSpinButton::refreshText");

   if ( fTextSpinButtonData->textCount ) {

      if (fTextSpinButtonData->pList) {
         delete [] fTextSpinButtonData->pList;
      }

      fTextSpinButtonData->pList = new char*[fTextSpinButtonData->textCount];
      unsigned int lIndex = 0;

      for (int ix = 0; ix < fTextSpinButtonData->textCount ; ix++ ) {
                                       //Use the string buffer address to setup
                                       //pList char* address.
         fTextSpinButtonData->pList[ix] = (char*)fTextSpinButtonData->spinText + lIndex;
         lIndex = fTextSpinButtonData->spinText.indexOf('\0', lIndex + 1);
      }

      unsigned long ulOk =
           handle().sendEvent(SPBM_SETARRAY,
                              IEventParameter1((char*)fTextSpinButtonData->pList),
                              IEventParameter2(fTextSpinButtonData->textCount));
      if (!ulOk)
          ITHROWLIBRARYERROR(IC_SET_SPIN_ARRAY_FAIL,
                             IErrorInfo::invalidRequest,
                             IException::recoverable);
   } else                              //If we have removed all of the
                                       //elements simply call removeAll
      removeAll();

   return *this;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::findStringIndex                                             |
|                                                                              |
| Locate the index of the string corresponding to the cursor position.         |
------------------------------------------------------------------------------*/
unsigned long ITextSpinButton::findStringIndex(const Cursor& cursor) const
{
   // Assume cursor is validated by the caller.
   return this->findStringIndex( cursor.fCursorData->lCurrent - 1 );
}

/*------------------------------------------------------------------------------
| ITextSpinButton::findStringIndex                                             |
|                                                                              |
| Locate the index of the string corresponding to the index position.          |
------------------------------------------------------------------------------*/
unsigned long ITextSpinButton::findStringIndex( unsigned long index ) const
{
   IMODTRACE_DEVELOP("ITextSpinButton::findStringIndex");
   /***************************************************/
   /* Loop to locate index of string to parse string. */
   /***************************************************/
   long lIndex = 1;
   for (int i=1; i < index + 1; i++ ) {
     lIndex = fTextSpinButtonData->spinText.indexOf('\0', lIndex);
     ++lIndex;
     ITRACE_DEVELOP( "Index = " + IString( lIndex ));
   }
   return lIndex;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::Cursor                                              |
|                                                                              |
| Cursor constructor.                                                          |
------------------------------------------------------------------------------*/
ITextSpinButton::Cursor::Cursor(const ITextSpinButton& spBut )
{
   IMODTRACE_DEVELOP("ITextSpinButton::Cursor::ctor");
   fCursorData = new ITextSpinButtonCursorData(spBut);
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::~Cursor                                             |
------------------------------------------------------------------------------*/
ITextSpinButton::Cursor::~Cursor ( )
{
   delete fCursorData;
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::setToFirst                                          |
|                                                                              |
| Set the cursor to the first element.                                         |
------------------------------------------------------------------------------*/
IBase::Boolean ITextSpinButton::Cursor::setToFirst ()
{
  if ( fCursorData->rSpBut.fTextSpinButtonData->textCount )
  {      // Spin button has a first element to point to.
     fCursorData->fValidCount =
                  fCursorData->rSpBut.fTextSpinButtonData->fChangeCount;
     fCursorData->lCurrent = 1;
     return true;
  } else {
     fCursorData->lCurrent = 0;
     return false;
  }
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::setToNext                                           |
|                                                                              |
| Set the cursor to the next element.                                          |
------------------------------------------------------------------------------*/
IBase::Boolean ITextSpinButton::Cursor :: setToNext ()
{
  if ( ! this->isValid() )
  {          // Reset invalid cursor to first item.
     return this->setToFirst();
  }
  else if ( fCursorData->lCurrent <
                    fCursorData->rSpBut.fTextSpinButtonData->textCount )
  {          // Point to next item.
     ++fCursorData->lCurrent;
     return true;
  }
  else
  {          // No more items, so invalidate the cursor.
     fCursorData->lCurrent = 0;
     return false;
  }
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::setToPrevious                                       |
|                                                                              |
| Set the cursor to the previous element.                                      |
------------------------------------------------------------------------------*/
IBase::Boolean ITextSpinButton::Cursor :: setToPrevious ()
{
  if ( ! this->isValid() )
  {          // Reset invalid cursor to last item.
     return this->setToLast();
  }
  else if ( fCursorData->lCurrent > 1 )
  {          // Point to previous item.
     --fCursorData->lCurrent;
     return true;
  }
  else
  {          // No more items, so invalidate the cursor.
     fCursorData->lCurrent = 0;
     return false;
  }
}

/*------------------------------------------------------------------------------
| ITextSpinButton::Cursor::setToLast                                           |
|                                                                              |
| Set the cursor to the last element.                                          |
------------------------------------------------------------------------------*/
IBase::Boolean ITextSpinButton::Cursor :: setToLast ()
{
  if ( fCursorData->rSpBut.fTextSpinButtonData->textCount )
  {      // Spin button has a last element to point to.
     fCursorData->fValidCount =
                  fCursorData->rSpBut.fTextSpinButtonData->fChangeCount;
     fCursorData->lCurrent =
                  fCursorData->rSpBut.fTextSpinButtonData->textCount;
     return true;
  } else {
     fCursorData->lCurrent = 0;
     return false;
  }
}

void ITextSpinButton::Cursor :: invalidate ()
/*------------------------------------------------------------------------------
  Invalidate the cursor
------------------------------------------------------------------------------*/
{
   fCursorData->lCurrent = 0;
}

IBase::Boolean ITextSpinButton::Cursor :: isValid () const
/*------------------------------------------------------------------------------
  Is the cursor valid?
------------------------------------------------------------------------------*/
{
   return ( fCursorData->lCurrent > 0 && fCursorData->fValidCount ==
       fCursorData->rSpBut.fTextSpinButtonData->fChangeCount) ? true : false;
}
