/*******************************************************************************
* FILE NAME: itextctl.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in itextctl.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C"
{
  #define INCL_GPIPRIMITIVES
  #define INCL_WINCLIPBOARD
  #define INCL_WINSYS
  #include <iwindefs.h>
  #include <string.h>
}

#include <ibase.hpp>
#include <itextctl.hpp>
#include <itxtcprv.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <ipoint.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <ithread.hpp>
#include <inotifev.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ITEXTCTL_CPP_
  #include <ipagetun.h>
#endif

#define MIN(a, b) ( ( (a) <= (b) ) ? (a) : (b) )
#define MAX(a, b) ( ( (a) >= (b) ) ? (a) : (b) )


/*------------------------------------------------------------------------------
| ITextControlData::ITextControlData                                           |
|                                                                              |
| This class contains special flags used by the text control                   |
------------------------------------------------------------------------------*/
ITextControlData::ITextControlData()
  : setTextFunction(false),
    containsMnemonic(false)
{}


/*------------------------------------------------------------------------------
| ITextControlData::~ITextControlData                                          |
|                                                                              |
| Destructor here for page tuning.                                             |
------------------------------------------------------------------------------*/
ITextControlData::~ITextControlData()
{}


/*------------------------------------------------------------------------------
| ITextControl::ITextControl                                                   |
|                                                                              |
| Default constructor here for page tuning.                                    |
------------------------------------------------------------------------------*/
ITextControl :: ITextControl ( )
  : fTextControlData(new ITextControlData())
{}


/*------------------------------------------------------------------------------
| ITextControl::~ITextControl                                                  |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
ITextControl :: ~ITextControl ( )
{
  delete fTextControlData;
}


/*------------------------------------------------------------------------------
| ITextControl::setText                                                        |
|                                                                              |
| Treat a font change like a minimum size change.                              |
------------------------------------------------------------------------------*/
ITextControl&  ITextControl :: setText(const char* pszText)
{
  IString strOrigText = text();

  Boolean newTextMnemonic = false;
  IString newText = pszText;
  if ( newText.includes('~') )
    newTextMnemonic = true;

  // If a mnemonic is involved in either then new or old string or
  // the strings are not the same, then update the text value
  if ( newTextMnemonic || fTextControlData->containsMnemonic ||
      (strOrigText != pszText))
  {
     fTextControlData->setTextFunction = true;
     ISize siz          = size ( );
     Boolean wasvisible = isVisible ( );

     if (wasvisible && ((siz.width()==0) || (siz.height()==0)))
        hide();

     ISETWINDOWTEXT(handle(), (PSZ)pszText);

     if (wasvisible && ((siz.width()==0) || (siz.height()==0)))
        show();

     this->notifyObservers(INotificationEvent(
                           ITextControl::textId, *this));

     // Modify the mnemonic flag based on previous check of new text string
     fTextControlData->containsMnemonic = newTextMnemonic;
     fTextControlData->setTextFunction = false;
  }
  return *this;
}


/*------------------------------------------------------------------------------
| ITextControl::setText                                                        |
|                                                                              |
| Sets the control window text using string table.                             |
------------------------------------------------------------------------------*/
ITextControl&  ITextControl :: setText(const IResourceId& residText)
{
  IString str = residText.resourceLibrary().loadString(residText);
  setText((char*)str);
  return *this;
}


/*------------------------------------------------------------------------------
| ITextControl::text                                                           |
|                                                                              |
| Returns the control window's text.                                           |
------------------------------------------------------------------------------*/
IString  ITextControl :: text() const
{
  unsigned long ulClBufLen =
           (unsigned long)IQUERYWINDOWTEXTLENGTH(handle()) + 1;

  char* pszClBuf = new char[ulClBufLen];
  IQUERYWINDOWTEXT(handle(), ulClBufLen, (PSZ)pszClBuf);

  IString str(pszClBuf);

  delete [] pszClBuf;
  return str;
}


/*------------------------------------------------------------------------------
| ITextControl::textLength                                                     |
|                                                                              |
| Returns the current length of the control window text, in bytes.             |
------------------------------------------------------------------------------*/
unsigned long  ITextControl :: textLength() const
{
  return (unsigned long)IQUERYWINDOWTEXTLENGTH(handle());
}


/*------------------------------------------------------------------------------
| ITextControl::displaySize                                                    |
|                                                                              |
| Returns the width and height of the rectangle that encloses the string       |
| when it is displayed on the window's presentation space.  This function      |
| takes into account the font attribute of the presentation space, including   |
| sheared or rotated string.                                                   |
------------------------------------------------------------------------------*/
ISize ITextControl:: displaySize(const char* pszText) const
{
   IString theText(pszText);

   if (pszText == 0) {
      theText = text();
   }

   POINTL aptl[TXTBOX_COUNT];
   POINTL ptl = {0,0};
   IPresSpaceHandle hps = presSpace();

   if ( IQUERYTEXTBOX( hps, strlen((char*)theText),
                       (char*)theText, aptl) == FALSE )
   {
      releasePresSpace(hps);
      ITHROWGUIERROR("GpiQueryTextBox");
   }

   releasePresSpace(hps);

   /***************************************************************/
   /* Determine the min and max of 'x' for each corner. This      */
   /* will enable support for sheared or rotated string.          */
   /***************************************************************/

   long lLeft = MIN(aptl[TXTBOX_TOPLEFT].x, aptl[TXTBOX_BOTTOMLEFT].x);
   lLeft = MIN(lLeft, aptl[TXTBOX_TOPRIGHT].x);
   lLeft = MIN(lLeft, aptl[TXTBOX_BOTTOMRIGHT].x);

   long lRight = MAX(aptl[TXTBOX_TOPLEFT].x, aptl[TXTBOX_BOTTOMLEFT].x);
   lRight = MAX(lRight, aptl[TXTBOX_TOPRIGHT].x);
   lRight = MAX(lRight, aptl[TXTBOX_BOTTOMRIGHT].x);

   long lBottom = MIN(aptl[TXTBOX_TOPLEFT].y, aptl[TXTBOX_BOTTOMLEFT].y);
   lBottom = MIN(lBottom, aptl[TXTBOX_TOPRIGHT].y);
   lBottom = MIN(lBottom, aptl[TXTBOX_BOTTOMRIGHT].y);

   long lTop = MAX(aptl[TXTBOX_TOPLEFT].y, aptl[TXTBOX_BOTTOMLEFT].y);
   lTop = MAX(lTop, aptl[TXTBOX_TOPRIGHT].y);
   lTop = MAX(lTop, aptl[TXTBOX_BOTTOMRIGHT].y);

   //The plus one is since where are returning a size and the coordinates
   //are zero based.
   //The plus 4 is because of an OS/2 bug, in that it doesn't return
   //the right size for vector fonts.  Opened an APAR on this.
   return ISize(lRight-lLeft + 1 + 4, lTop-lBottom + 1 + 4);
}


/*------------------------------------------------------------------------------
| ITextControl::clipboardHasTextFormat                                         |
|                                                                              |
| Queries whether the clipboard is in text format.                             |
------------------------------------------------------------------------------*/
Boolean  ITextControl :: clipboardHasTextFormat ( )
{
  unsigned long userInfo;

  Boolean bOpen =
              IQUERYCLIPBRDFMTINFO(IThread::current().anchorBlock(),
                                   CF_TEXT, &userInfo) ?
                                               true : false;
  return bOpen;
}


/*------------------------------------------------------------------------------
| ITextControl::setLayoutDistorted                                             |
|                                                                              |
| Queries whether the clipboard is in text format.                             |
------------------------------------------------------------------------------*/
ITextControl &  ITextControl :: setLayoutDistorted(
                                              unsigned long layoutAttributeOn,
                                              unsigned long layoutAttributeOff )
{
  unsigned long ulFlagsOn = layoutAttributeOn;
  if (layoutAttributeOn & IWindow::fontChanged)
  {                         // set flag for IControl::characterSize()
     ulFlagsOn |= IWindow::minimumSizeChanged;  // flag for ICanvas
  }
  Inherited::setLayoutDistorted(ulFlagsOn, layoutAttributeOff);

  return *this;
}

/*------------------------------------------------------------------------------
| ITextControl::fromSetText                                                    |
|                                                                              |
| Queries whether communication was from setText() function.                   |
------------------------------------------------------------------------------*/
Boolean ITextControl :: fromSetText ( ) const
{
  return fTextControlData->setTextFunction;
}
