/*******************************************************************************
* FILE NAME: ititle.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ititle.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
   #define INCL_WINSWITCHLIST
   #define INCL_WINFRAMEMGR
   #define INCL_WINSYS
   #include <iwindefs.h>
   #include <string.h>
   #include <stdlib.h>
}
#include <ititle.hpp>
#include <icconst.h>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <inotifev.hpp>
#include <istring.hpp>
#include <iwindow.hpp>


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ITITLE_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| titleBarHwnd                                                                 |
|  Returns FID_TITLEBAR HWND for owner.                                        |
|  Throws an exception if owner doesn't have FID_TITLEBAR.                     |
------------------------------------------------------------------------------*/
static IWindowHandle titleBarHwnd ( IWindow* owner )
{
  IWindowHandle hwnd =
     IWindow::handleWithParent( FID_TITLEBAR, owner->handle() );
  if ( hwnd == 0 )
  {
     ITHROWLIBRARYERROR( IC_NO_TITLEBAR,
                         IErrorInfo::invalidRequest,
                         IException::recoverable );
  }
  return hwnd;
}


/*------------------------------------------------------------------------------
| ITitle::ITitle                                                               |
------------------------------------------------------------------------------*/
ITitle :: ITitle ( IWindow*      owner,
                   const char*   objectName,
                   const char*   viewName,
                   unsigned long viewNumber )
  : pwinClOwner( owner ),
    fTitleData( 0 )
{
  // Title bar not a window in Windows.
  // IC_NOTYET ... need to override some of the ITextControl functions
  // in this class because of this or find a way to fake the window handle.
  this->setAutoDestroyWindow( false );
  this->startHandlingEventsFor( titleBarHwnd( this->pwinClOwner ));

  if ( objectName )
  {
     this->setTitleText( objectName, viewName, viewNumber );
  }
}

/*------------------------------------------------------------------------------
| ITitle::ITitle                                                               |
------------------------------------------------------------------------------*/
ITitle :: ITitle ( IWindow*           owner,
                   const IResourceId& objectName,
                   const IResourceId& viewName,
                   unsigned long      viewNumber )
  : pwinClOwner( owner ),
    fTitleData( 0 )
{
  // Title bar not a window in Windows.
  this->setAutoDestroyWindow( false);
  this->startHandlingEventsFor( titleBarHwnd( this->pwinClOwner ));

  IString strObjName = objectName.resourceLibrary().loadString( objectName );

  IString strViewName;
  if ( viewName != 0 )
     strViewName = viewName.resourceLibrary().loadString( viewName );

  this->setTitleText( strObjName, strViewName, viewNumber );
}

/*------------------------------------------------------------------------------
| ITitle::~ITitle                                                              |
------------------------------------------------------------------------------*/
ITitle :: ~ITitle ( )
{ }

/*------------------------------------------------------------------------------
| ITitle::objectText                                                           |
------------------------------------------------------------------------------*/
IString ITitle :: objectText ( ) const
{
  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );
  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  IString strObj;
  if ( uLen > 1 )
  {
     CHAR* pszClBuf;                         // To store entire title string.
     pszClBuf = new CHAR[ uLen ];
     *pszClBuf = '\0';

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     strObj = IString( pszClBuf );
     unsigned pos = strObj.indexOf(" - ");   // Locate " - ".
     if ( pos )
        strObj.remove( pos );

     delete [] pszClBuf;
  } /* endif */

  return strObj;
}

/*------------------------------------------------------------------------------
| ITitle::viewText                                                             |
------------------------------------------------------------------------------*/
IString ITitle :: viewText ( ) const
{
  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );

  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  IString strView;

  if ( uLen > 1 )
  {
     char* pszClBuf = new char[ uLen ];

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     IString strObj = IString( pszClBuf );
     unsigned pos = strObj.indexOf(" - ");   // Locate " - ".
     if ( pos )
     {
        strView = strObj.subString( pos + 3 ); // View name.
        pos = strView.lastIndexOf( ':' );      // Get ptr at last ":".
        if ( pos && ( strView.subString(pos+1).isDigits() ) )
        {
           // Remove : and numeric part.
           strView.remove( pos );
        }
     } /* endif */

     delete [] pszClBuf;
  } /* endif */

  return strView;
}

/*------------------------------------------------------------------------------
| ITitle::viewNumber                                                           |
------------------------------------------------------------------------------*/
unsigned long ITitle :: viewNumber ( ) const
{
  unsigned long ulViewNum = 0;

  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );
  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  if ( uLen > 1 )
  {
     char* pszClBuf = new char[ uLen ];

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     IString strObj = IString( pszClBuf );
     unsigned delimPos = strObj.indexOf(" - ");
     unsigned viewPos  = strObj.lastIndexOf( ':' );
     if ( (delimPos) &&                     // Found view delimiter.
          (viewPos > delimPos) &&           // Found view number.
          (viewPos < strObj.length()) )     // : not last char.
     {
        ulViewNum = strObj.subString( viewPos + 1 ).asUnsigned();
     }

     delete [] pszClBuf;
  } /* endif */

  return ulViewNum;
}

/*------------------------------------------------------------------------------
| ITitle::setObjectText                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setObjectText ( const char* objectName )
{
  IString strObj = this->objectText();
  if ( objectName )
  {
     if (strcmp( (char*)strObj, objectName ))
     {
        this->setTitleText( objectName, this->viewText(), this->viewNumber() );
        this->notifyObservers( INotificationEvent( ITitle::objectTextId,
                                                   *this,
                                                   true,
                                                   (char*)objectName ));
     }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setObjectText                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setObjectText ( const IResourceId& objectNameResId )
{
  IString strObjName =
     objectNameResId.resourceLibrary().loadString( objectNameResId );
  if ( strObjName.length() )
     this->setObjectText( strObjName );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewText                                                          |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewText ( const char* viewName )
{
  IString strView = this->viewText();

  if ( strcmp( (char*)strView, viewName ))
  {
     this->setTitleText( this->objectText(), viewName, this->viewNumber() );
     this->notifyObservers( INotificationEvent( ITitle::viewTextId,
                                                *this,
                                                true,
                                                (char*)viewName ));
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewText                                                          |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewText ( const IResourceId& viewNameResId )
{
  IString strViewName =
     (viewNameResId.resourceLibrary()).loadString( viewNameResId );
  if ( strViewName.length() )
     this->setViewText( strViewName );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewNumber                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewNumber ( unsigned long viewNum )
{
  /****************************************/
  /* Only change and notify if different. */
  /****************************************/
  if ( viewNum != this->viewNumber() )
  {
     this->setTitleText( this->objectText(), this->viewText(), viewNum );
     this->notifyObservers( INotificationEvent( ITitle::viewNumberId,
                                                *this,
                                                true,
                                                (void*)viewNum ));
  }
  return *this;
}


/*------------------------------------------------------------------------------
| ITitle::setTitleText                                                         |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setTitleText ( const char*   objectName,
                                 const char*   viewName,
                                 unsigned long viewNum )
{
  IString strTitleText;

  // Check to insure a non-NULL string pointer was passed
  if ( objectName )
  {
     strTitleText = objectName;

     // If object name exists, then append view name and number
     if ( *objectName != '\0' )
     {
        if ( viewName && *viewName != '\0' )
           strTitleText += IString( " - " ) + IString( viewName );

        if ( viewNum )
           strTitleText += IString( ":" ) + IString( viewNum );
     }
  }
  this->setText( strTitleText );

  // Update Task List or Window List.
  HSWITCH hsw = WinQuerySwitchHandle( this->pwinClOwner->handle(), (PID)0 );
  if ( hsw )
  {
     SWCNTRL swctl;
     WinQuerySwitchEntry( hsw, &swctl );

     // Length of title cannot be greater than MAXNAMEL (60) + 1,
     // because SWCNTRL.szSwtitle is only that big.
     if ( strTitleText.length() + 1 > MAXNAMEL )
     {
       strTitleText = strTitleText.subString( 1, MAXNAMEL + 1 );
     }
     else if ( !strTitleText.length() )
       strTitleText = " ";

     strcpy( swctl.szSwtitle, strTitleText );
     WinChangeSwitchEntry( hsw, &swctl );
  } /* endif */

  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setTitleText                                                         |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setTitleText ( const IResourceId& objectNameResId,
                                 const IResourceId& viewNameResId,
                                 unsigned long      viewNum )
{
  IString strObjName =
     objectNameResId.resourceLibrary().loadString( objectNameResId );

  IString strViewName;
  if ( viewNameResId )
     strViewName = viewNameResId.resourceLibrary().loadString( viewNameResId );
  else
     strViewName = "";

  // Enforcement of requiring an object name string will be done below
  this->setTitleText( strObjName, strViewName, viewNum );

  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::borderColor                                                         |
|                                                                             |
| Returns the border color for the ITitle.                                    |
------------------------------------------------------------------------------*/
IColor ITitle::borderColor ( ) const
{
  return IWindow::color( PP_BORDERCOLOR,
                         IGUIColor( IGUIColor::titleLine ));
}

/*------------------------------------------------------------------------------
| ITitle::activeColor                                                         |
|                                                                             |
| Returns the active color for the ITitle.                                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeColor ( ) const
{
  return IWindow::color( PP_ACTIVECOLOR,
                         IGUIColor( IGUIColor::activeTitleBgnd ));
}

/*------------------------------------------------------------------------------
| ITitle::inactiveColor                                                       |
|                                                                             |
| Returns the inactive color for the ITitle.                                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveColor ( ) const
{
  return IWindow::color( PP_INACTIVECOLOR,
                         IGUIColor( IGUIColor::inactiveTitleBgnd ));
}

/*------------------------------------------------------------------------------
| ITitle::activeTextForegroundColor                                           |
|                                                                             |
| Returns the active text foreground color for the ITitle.                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeTextForegroundColor ( ) const
{
  return IWindow::color( PP_ACTIVETEXTFGNDCOLOR,
                         IGUIColor( IGUIColor::activeTitleText ));
}

/*------------------------------------------------------------------------------
| ITitle::activeTextBackgroundColor                                           |
|                                                                             |
| Returns the active text background color for the ITitle.                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeTextBackgroundColor ( ) const
{
  return IWindow::color( PP_ACTIVETEXTBGNDCOLOR,
                         IGUIColor( IGUIColor::activeTitleTextBgnd ));
}

/*------------------------------------------------------------------------------
| ITitle::inactiveTextForegroundColor                                         |
|                                                                             |
| Returns the inactive text foreground color for the ITitle.                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveTextForegroundColor ( ) const
{
  return IWindow::color( PP_INACTIVETEXTFGNDCOLOR,
                         IGUIColor( IGUIColor::inactiveTitleText ));
}

/*------------------------------------------------------------------------------
| ITitle::inactiveTextBackgroundColor                                         |
|                                                                             |
| Returns the inactive text background color for the ITitle.                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveTextBackgroundColor ( ) const
{
  return IWindow::color( PP_INACTIVETEXTBGNDCOLOR,
                         IGUIColor( IGUIColor::inactiveTitleTextBgnd ));
}

/*------------------------------------------------------------------------------
| ITitle::setActiveTextForegroundColor                                        |
|                                                                             |
| Sets the active text foreground color for the ITitle.                       |
------------------------------------------------------------------------------*/
ITitle& ITitle::setActiveTextForegroundColor ( const IColor &color )
{
  IWindow::setColor( PP_ACTIVETEXTFGNDCOLOR, color );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setActiveTextBackgroundColor                                        |
|                                                                             |
| Sets the active text background color for the ITitle.                       |
------------------------------------------------------------------------------*/
ITitle& ITitle::setActiveTextBackgroundColor ( const IColor &color )
{
  IWindow::setColor( PP_ACTIVETEXTBGNDCOLOR, color );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setInactiveTextForegroundColor                                      |
|                                                                             |
| Sets the inactive text foreground color for the ITitle.                     |
------------------------------------------------------------------------------*/
ITitle& ITitle::setInactiveTextForegroundColor ( const IColor &color )
{
  IWindow::setColor( PP_INACTIVETEXTFGNDCOLOR, color );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setInactiveTextBackgroundColor                                      |
|                                                                             |
| Sets the inactive text background color for the ITitle.                     |
------------------------------------------------------------------------------*/
ITitle& ITitle::setInactiveTextBackgroundColor ( const IColor &color )
{
  IWindow::setColor( PP_INACTIVETEXTBGNDCOLOR, color );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetActiveTextForegroundColor                                      |
|                                                                             |
| Resets the active text foreground color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetActiveTextForegroundColor ( )
{
  this->resetColor( PP_ACTIVETEXTFGNDCOLOR );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetActiveTextBackgroundColor                                      |
|                                                                             |
| Resets the active text background color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetActiveTextBackgroundColor ( )
{
  this->resetColor( PP_ACTIVETEXTBGNDCOLOR );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetInactiveTextForegroundColor                                    |
|                                                                             |
| Resets the active text foreground color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetInactiveTextForegroundColor ( )
{
  this->resetColor( PP_INACTIVETEXTFGNDCOLOR );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetInactiveTextBackgroundColor                                    |
|                                                                             |
| Resets the active text background color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetInactiveTextBackgroundColor ( )
{
  this->resetColor( PP_INACTIVETEXTBGNDCOLOR );
  return *this;
}
