/*******************************************************************************
* FILE NAME: itrnsfrm.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in itrnsfrm.cpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"        //Declare C constructs.
{
  #define INCL_WINWINDOWMGR
  #define INCL_GPITRANSFORMS
  #include <iwindefs.h>
}

#include <itrnsfrm.hpp>
#include <itrnpriv.hpp>
#include <ihandle.hpp>
#include <iexcept.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ITRNSFRM_CPP_
  #include <ipagetun.h>
#endif

// Data conversion macro between long and float numbers.
#define      DBL2FIXED(x)     ((FIXED)  ((x)*65536) )
#define      FIXED2DBL(x)     ((double) (((double)x)/(double)65536) )

/*------------------------------------------------------------------------------
| ITransformData::ITransformData                                               |
| Default constructor for transformation matrix, a identity.                   |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformData::ITransformData()
  : fElement11(UNITY), fElement12(0), fElement21(0), fElement22(UNITY),
    fElement31(0), fElement32(0)
{}

/*------------------------------------------------------------------------------
| ITransformData::ITransformData                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformData::ITransformData( FIXED f11, FIXED f12, FIXED f21, FIXED f22,
                                long l31, long l32 )
  : fElement11(f11), fElement12(f12), fElement21(f21),
    fElement22(f22), fElement31(l31), fElement32(l32)
{}

/*------------------------------------------------------------------------------
| ITransformData::ITransformData                                               |
| Constructor for transformation matrix.                                       |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformData::ITransformData( const ITransformData& tData )
  : fElement11(tData.fElement11), fElement12(tData.fElement12),
    fElement21(tData.fElement21), fElement22(tData.fElement22),
    fElement31(tData.fElement31), fElement32(tData.fElement32)
{}


// OS/2 PM amd Window common code.
/*------------------------------------------------------------------------------
| ITransformData::operator=                                                    |
| Assignment operator. Return a reference of this object.                      |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformData& ITransformData::operator=( const ITransformData& transformData )
{
  // If assignment to itself, return a reference to itself.
  if ( this == &transformData )
    return *this;

  // Member-wise assignment.
  fElement11 = transformData.fElement11;
  fElement12 = transformData.fElement12;
  fElement21 = transformData.fElement21;
  fElement22 = transformData.fElement22;
  fElement31 = transformData.fElement31;
  fElement32 = transformData.fElement32;
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformData::operator==                                                   |
| Boolean equal operator.                                                      |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean ITransformData::operator==( const ITransformData& transformData ) const
{
  // member-wise comparison.
  return ( fElement11 == transformData.fElement11 &&
           fElement12 == transformData.fElement12 &&
           fElement21 == transformData.fElement21 &&
           fElement22 == transformData.fElement22 &&
           fElement31 == transformData.fElement31 &&
           fElement32 == transformData.fElement32 );
}

/*------------------------------------------------------------------------------
| ITransformData::operator!=                                                   |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean ITransformData::operator!=( const ITransformData& transformData ) const
{
  return !( *this == transformData );
}

/*------------------------------------------------------------------------------
| ITransformMatrix::ITransformMatrix                                           |
| Default constructor.                                                         |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix::ITransformMatrix()
  : fTransformData(new ITransformData)
{ }

/*------------------------------------------------------------------------------
| ITransformMatrix::ITransformMatrix                                           |
| Constructor.                                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix::ITransformMatrix( double element11, double element12,
                                    double element21, double element22,
                                    long   element31, long   element32 )
  : fTransformData( new ITransformData( DBL2FIXED(element11), DBL2FIXED(element12),
                                        DBL2FIXED(element21), DBL2FIXED(element22),
                                        element31, element32 ))
{ }

/*------------------------------------------------------------------------------
| ITransformMatrix::ITransformMatrix                                           |
| Constructor from struct _MATRIXLF.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix::ITransformMatrix( const struct _MATRIXLF &matrixlf )
  : fTransformData( new ITransformData( matrixlf.fxM11, matrixlf.fxM12,
                                        matrixlf.fxM21, matrixlf.fxM22,
                                        matrixlf.lM31,  matrixlf.lM32 ))
{}



/*------------------------------------------------------------------------------
| ITransformMatrix::ITransformMatrix                                           |
| Copy constructor.                                                            |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix::ITransformMatrix( const ITransformMatrix& transformMatrix )
  : fTransformData( new ITransformData(*transformMatrix.fTransformData))
{ }

/*------------------------------------------------------------------------------
| ITransformMatrix::~ITransformMatrix                                          |
| Virtual destructor.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix::~ITransformMatrix()
{
  delete fTransformData;  //delete private data member.
}

/*------------------------------------------------------------------------------
| ITransformMatrix::asMATRIXLF                                                 |
| Return the matrix as an OS/2 Developers Toolkit MATRIXLF structure.          |
|                                                                              |
------------------------------------------------------------------------------*/
struct _MATRIXLF ITransformMatrix::asMATRIXLF() const
{
  MATRIXLF mat;
  mat.fxM11 = fTransformData->fElement11;
  mat.fxM12 = fTransformData->fElement12;
  mat.lM13  = 0;
  mat.fxM21 = fTransformData->fElement21;
  mat.fxM22 = fTransformData->fElement22;
  mat.lM23  = 0;
  mat.lM31  = fTransformData->fElement31;
  mat.lM32  = fTransformData->fElement32;
  mat.lM33  = 1;
  return mat;
}


/*------------------------------------------------------------------------------
| ITransformMatrix::setElement11                                               |
| Set element11 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement11( double value )
{
  fTransformData->fElement11 = DBL2FIXED( value );
  return *this;
}
/*------------------------------------------------------------------------------
| ITransformMatrix::setElement12                                               |
| Set element12 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement12( double value )
{
  fTransformData->fElement12 = DBL2FIXED( value );
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::setElement21                                               |
| Set element21 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement21( double value )
{
  fTransformData->fElement21 = DBL2FIXED( value );
  return *this;
}
/*------------------------------------------------------------------------------
| ITransformMatrix::setElement22                                               |
| Set element22 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement22( double value )
{
  fTransformData->fElement22 = DBL2FIXED( value );
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::setElement31                                               |
| Set element31 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement31( long value )
{
  fTransformData->fElement31 = value;
  return *this;
}
/*------------------------------------------------------------------------------
| ITransformMatrix::setElement32                                               |
| Set element32 of the transformation matrix.                                  |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setElement32( long value )
{
  fTransformData->fElement32 = value;
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::setToIdentity                                              |
| Set transformation matrix to Identity.                                       |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::setToIdentity()
{
  fTransformData->fElement11 = UNITY;
  fTransformData->fElement12 = 0;
  fTransformData->fElement21 = 0;
  fTransformData->fElement22 = UNITY;
  fTransformData->fElement31 = 0;
  fTransformData->fElement32 = 0;
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element11                                                  |
| Get element11 of the transformation matrix.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
double ITransformMatrix::element11() const
{
  return FIXED2DBL(fTransformData->fElement11);
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element12                                                  |
| Get element12 of the transformation matrix.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
double ITransformMatrix::element12() const
{
  return FIXED2DBL(fTransformData->fElement12);
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element21                                                  |
| Get element21 of the transformation matrix.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
double ITransformMatrix::element21() const
{
  return FIXED2DBL(fTransformData->fElement21);
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element22                                                  |
| Get element22 of the transformation matrix.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
double ITransformMatrix::element22() const
{
  return FIXED2DBL(fTransformData->fElement22);
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element31                                                  |
| Get element131 of the transformation matrix.                                 |
|                                                                              |
------------------------------------------------------------------------------*/
long ITransformMatrix::element31() const
{
  return fTransformData->fElement31;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::element32                                                  |
| Get element32 of the transformation matrix.                                  |
|                                                                              |
------------------------------------------------------------------------------*/
long ITransformMatrix::element32() const
{
  return fTransformData->fElement32;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::scaleBy                                                    |
| Scale operation on transformation matrix.                                    |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::scaleBy( double xScale, double yScale,
                                             const IPoint& pt )
{
  //convert double to fix. ICLUI object to PM structure.
  FIXED afixed[2] = { DBL2FIXED( xScale ), DBL2FIXED( yScale ) };
  POINTL ptl(pt.asPOINTL());
  MATRIXLF mat(asMATRIXLF());

  // Get a cache presentation space handle.
  IPresSpaceHandle ps = WinGetPS( HWND_DESKTOP );
  // Apply scale on mat through ps.
  if (!GpiScale(ps, &mat, TRANSFORM_ADD, afixed, &ptl ))
  {
    ITHROWGUIERROR("GpiScale");
  }
  else  //Assign elements back to transformation matrix.
  {
    fTransformData->fElement11 = mat.fxM11;
    fTransformData->fElement12 = mat.fxM12;
    fTransformData->fElement21 = mat.fxM21;
    fTransformData->fElement22 = mat.fxM22;
    fTransformData->fElement31 = mat.lM31;
    fTransformData->fElement32 = mat.lM32;
  }
  WinReleasePS( ps );  //Release cached ps.
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::rotateBy                                                   |
| Rotate operation on transformation matrix.                                   |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::rotateBy( double angle, const IPoint& pt )
{
  // Convert ICLUI object to PM structure.
  POINTL ptl(pt.asPOINTL());
  MATRIXLF mat(asMATRIXLF());

  // Get a cache presentation space handle.
  IPresSpaceHandle ps = WinGetPS( HWND_DESKTOP );
  // Apply rotate on mat through ps.
  if (!GpiRotate(ps, &mat, TRANSFORM_ADD, DBL2FIXED(angle), &ptl ))
  {
    ITHROWGUIERROR("GpiRotate");
  }
  else  // Assign elements back to transformation matrix.
  {
    fTransformData->fElement11 = mat.fxM11;
    fTransformData->fElement12 = mat.fxM12;
    fTransformData->fElement21 = mat.fxM21;
    fTransformData->fElement22 = mat.fxM22;
    fTransformData->fElement31 = mat.lM31;
    fTransformData->fElement32 = mat.lM32;
  }
  WinReleasePS( ps );  //Release cached PS.
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::translateBy                                                |
| Move operation on transformation matrix.                                     |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::translateBy( const IPoint& pt )
{
  // Convert ICLUI object to PM structure.
  POINTL ptl(pt.asPOINTL());
  MATRIXLF mat(asMATRIXLF());

  // Get a cache presentation space handle.
  IPresSpaceHandle ps = WinGetPS( HWND_DESKTOP );
  // Apply translate on mat through ps.
  if (!GpiTranslate(ps, &mat, TRANSFORM_ADD, &ptl ))
  {
    ITHROWGUIERROR("GpiTranslate");
  }
  else    //Assign elements back to transformation matrix.
  {
    fTransformData->fElement11 = mat.fxM11;
    fTransformData->fElement12 = mat.fxM12;
    fTransformData->fElement21 = mat.fxM21;
    fTransformData->fElement22 = mat.fxM22;
    fTransformData->fElement31 = mat.lM31;
    fTransformData->fElement32 = mat.lM32;
  }
  WinReleasePS( ps );  //Release cached ps.
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::operator=                                                  |
| Assignment operator.                                                         |
| Return a reference of this object.                                           |
------------------------------------------------------------------------------*/
ITransformMatrix& ITransformMatrix::operator= (
                                       const ITransformMatrix& transformMatrix )
{
  // If assignment to itself, return a reference to itself.
  if ( this == &transformMatrix )
    return *this;

  // Copy transform data.
  *fTransformData = *transformMatrix.fTransformData;
  return *this;
}

/*------------------------------------------------------------------------------
| ITransformMatrix::operator==                                                 |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean ITransformMatrix::operator== ( const ITransformMatrix& transformMatrix ) const
{
  // A member-wise comparison.
  return ( *fTransformData == *transformMatrix.fTransformData );
}

/*------------------------------------------------------------------------------
| ITransformMatrix::operator!=                                                 |
| Boolean unequal operator.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean ITransformMatrix::operator!= ( const ITransformMatrix& transformMatrix ) const
{
  return !( *this == transformMatrix );
}
