/*******************************************************************************
* FILE NAME: iwindow7.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions used for the    *
*   IWindow notifcation handler in OS/2 and Windows.                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define INCL_WINFRAMEMGR
#define INCL_WINSYS

extern "C" {
   #include <iwindefs.h>
}



#include <iwinnhdr.hpp>
#include <iwindow.hpp>
#include <iexcept.hpp>
#include <ievent.hpp>
#include <inotifev.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IWINDOW7_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IWindowNotifyHandler::IWindowNotifyHandler                                   |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IWindowNotifyHandler::IWindowNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IWindowNotifyHandler::~IWindowNotifyHandler                                  |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IWindowNotifyHandler::~IWindowNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IWindowNotifyHandler::dispatchHandlerEvent                                   |
|                                                                              |
| This notifies IWindows of the following events:                              |
|                                                                              |
| size(Id)                                                                     |
| position(Id)                                                                 |
| visible(Id)                                                                  |
| enable(Id)                                                                   |
| destroy(Id)                                                                  |
| command(Id)                                                                  |
| sysCommand(Id)                                                               |
------------------------------------------------------------------------------*/

Boolean IWindowNotifyHandler::dispatchHandlerEvent( IEvent& anEvent)
{
   IWindow* thisWindow = anEvent.window();

   if (thisWindow)
   {
     switch (anEvent.eventId())
     {
       case WM_WINDOWPOSCHANGED:
       {
         PSWP pswpNew = (PSWP)(char*)anEvent.parameter1();
         if (pswpNew->fl & SWP_SIZE)
         {
            // Send Size changed notification
            ISize siz(pswpNew->cx, pswpNew->cy);
            thisWindow->notifyObservers(
                        INotificationEvent(IWindow::sizeId,
                                           *thisWindow, true, &siz));
         }
         if (pswpNew->fl & SWP_MOVE)
         {
            // Send Move notification
            IPoint pt(pswpNew->x, pswpNew->y);
            thisWindow->notifyObservers(
                   INotificationEvent(IWindow::positionId,
                                      *thisWindow, true, &pt));
         }
         if (pswpNew->fl & SWP_SHOW)
         {
            // Send Show notification
            thisWindow->notifyObservers(
                   INotificationEvent(IWindow::visibleId,
                                      *thisWindow, true, true));
         }
         else if (pswpNew->fl & SWP_HIDE)
         {
            // Send Hide notification
            thisWindow->notifyObservers(
                   INotificationEvent(IWindow::visibleId,
                                      *thisWindow, true, false));
         }
         break;
       } /* endcase */
       case WM_ENABLE:
       {
          Boolean fEnable = anEvent.parameter1().number1();
          thisWindow->notifyObservers(
                 INotificationEvent(IWindow::enableId,
                                    *thisWindow, true, (void*)fEnable));
          break;
       } /* endcase */

       case WM_SETFOCUS:
       {
          Boolean focus = anEvent.parameter2().number1();
          thisWindow->notifyObservers(
                   INotificationEvent(IWindow::focusId,
                                      *thisWindow, false, (void*)focus));
          break;
       } /* endcase */

       case WM_PRESPARAMCHANGED:
       {
         if (anEvent.parameter1() == PP_FOREGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::foregroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_BACKGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::backgroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_BORDERCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::borderColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_SHADOW)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::shadowColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_DISABLEDFOREGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::disabledForegroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_DISABLEDBACKGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::disabledBackgroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_ACTIVECOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::activeColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_INACTIVECOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::inactiveColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_HILITEFOREGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::hiliteForegroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_HILITEBACKGROUNDCOLOR)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                   IWindow::hiliteBackgroundColorId,
                                   *thisWindow));
         }
         else if (anEvent.parameter1() == PP_FONTNAMESIZE)
         {
           thisWindow->notifyObservers(INotificationEvent(
                                       IWindow::fontId, *thisWindow));
         }
         break;
       }

       case WM_COMMAND:
       {
          thisWindow->notifyObservers(
                   INotificationEvent(IWindow::commandId,
                                      *thisWindow, false, (void*)&anEvent));
          break;
       } /* endcase */

       case WM_SYSCOMMAND:
       {
          thisWindow->notifyObservers(
                   INotificationEvent(IWindow::systemCommandId,
                                      *thisWindow, false, (void*)&anEvent));
          break;
       } /* endcase */

       default:
       break;
     }  /* end switch */
   }

   return false;  // We never "process" an event
}




