/*NOSHIP*/
/*******************************************************************************
* FILE NAME: ilistbx4.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in both ilistbox.hpp and icombobx.hpp.                                     *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#ifdef IC_USE_CB
  #define IC_CLASSNAME IComboBox
  #define IC_CURSOR_CLASSNAME IBaseComboBox
#else
  #define IC_CLASSNAME IListBox
  #define IC_CURSOR_CLASSNAME IBaseListBox
#endif

/*------------------------------------------------------------------------------
| IC_CLASSNAME::add                                                            |
|                                                                              |
| Add an item to the list using a string resource id.                          |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::add ( unsigned long lIndex,
                                  const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return add(lIndex,newText);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::add                                                            |
|                                                                              |
| Add an item to the list using a const char *.  This method is also used by   |
| the other add item methods to actually insert the item.                      |
------------------------------------------------------------------------------*/
unsigned long   IC_CLASSNAME::add(unsigned long lIndex, const char* pszString)
{
//   needValidation++;   // add may result in an invalid cursor.
   incrementChangeCount( );   // add may result in an invalid cursor.
   unsigned long anIndex = count();
   if ((long)anIndex > (long)lIndex)
      anIndex = lIndex;
   IEventResult evt = handle().sendEvent(LM_INSERTITEM,
                               IEventParameter1(anIndex),
                               IEventParameter2((unsigned long)pszString));
   switch (evt.asLong())
   {
      case LIT_MEMERROR:
      case LIT_ERROR:
         ITHROWGUIERROR("LM_INSERTITEM");
         break;
      default:
         break;
   } /* endswitch */
   notifyObservers(INotificationEvent(IC_CLASSNAME::addId,
                                      *this, true, (void *)1));
   return evt.asUnsignedLong();  // Return index of inserted item.
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::add                                                            |
|                                                                              |
| Add a list of items to the listbox.  An array of character strings are       |
| passed in.  Return index of first inserted item.                             |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::add ( unsigned long index,
                                  const char* const * itemList,
                                  unsigned long count )
{
  int  i;
  unsigned long rc, firstIndex;

  // assertions on input parms
  IASSERTPARM(count != 0 && itemList != 0);
  for (i=0; i < count; i++)
  {
    rc = add( index+i, (char*)itemList[i] );
    if (!i)
      firstIndex = rc;
  }
  return firstIndex;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsFirst                                                     |
|                                                                              |
| Add an item at the beginning of the list using a string resource id.         |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAsFirst ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsFirst(newText);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsFirst                                                     |
|                                                                              |
| Add an item at the beginning of the list using a const char *.               |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAsFirst ( const char* pszString )
{
   return add(0, pszString);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsLast                                                      |
|                                                                              |
| Add an item at the end of the list using a string resource id.               |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAsLast ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsLast(newText);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsLast                                                      |
|                                                                              |
| Add an item at the end of the list using a const char *.                     |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAsLast ( const char* pszString )
{
   return add(LIT_END, pszString);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAscending                                                   |
|                                                                              |
| Add an item to the list in ascending order using a string resource id.       |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAscending ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAscending(newText);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAscending                                                   |
|                                                                              |
| Add an item to the list in ascending order using a const char *.             |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addAscending ( const char* pszString )
{
   return add(LIT_SORTASCENDING, pszString);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addDescending                                                  |
|                                                                              |
| Add an item to the list in descending order using a string resource id.      |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addDescending ( const IResourceId& resid )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addDescending(newText);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addDescending                                                  |
|                                                                              |
| Add an item to the list in descending order using a const char *.            |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::addDescending ( const char* pszString )
{
   return add(LIT_SORTDESCENDING, pszString);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::removeAll                                                      |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::removeAll ( )
{
   IEventResult evt = handle().sendEvent(LM_DELETEALL,
                                         IEventParameter1(0),
                                         IEventParameter2(0));
   if (!(evt.asUnsignedLong()))
      ITHROWGUIERROR("LM_DELETEALL");
// needValidation++;  // removeAll may result in an invalid cursor.
   incrementChangeCount( );   // add may result in an invalid cursor.
   notifyObservers(INotificationEvent(IC_CLASSNAME::removeId,
                                      *this, true, (void *)0));
   return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::remove                                                         |
|                                                                              |
| Remove one item from the list and return the number of remaining items.      |
------------------------------------------------------------------------------*/
unsigned long IC_CLASSNAME::remove ( unsigned long lIndex )
{
// needValidation++;  // remove may result in an invalid cursor.
   incrementChangeCount( );   // add may result in an invalid cursor.
   IEventResult evt = handle().sendEvent(LM_DELETEITEM,
                                         IEventParameter1(lIndex),
                                         IEventParameter2(0));
   notifyObservers(INotificationEvent(IC_CLASSNAME::removeId,
                                      *this, true, (void *)1));
   return evt.asUnsignedLong();
}


/*------------------------------------------------------------------------------
| IC_CLASSNAME::removeAt                                                       |
|                                                                              |
| Remove the item at the cursor's position.                                    |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::removeAt ( IC_CURSOR_CLASSNAME::Cursor& cursor )
{
  IASSERTSTATE(cursor.isValid());
  unsigned long i = cursor.asIndex();
  cursor.setToNext();
  if (cursor.isValid())
  {
     remove(i);
     cursor.setToIndex(cursor.asIndex() - 1);
  }
  else
     remove(i);

  return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::replaceAt                                                      |
|                                                                              |
| Replace the item at the cursor's position using a string resource id.        |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::replaceAt ( const IResourceId& resid,
                                        const IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return replaceAt(newText, cursor);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::replaceAt                                                      |
|                                                                              |
| Replace the item at the cursor's position using a const char *.              |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::replaceAt ( const char* strNew,
                                        const IC_CURSOR_CLASSNAME::Cursor& cursor )
{
  IASSERTSTATE(cursor.isValid());
  setItemText(cursor.asIndex(), strNew);
  return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::add                                                            |
|                                                                              |
| Add an item at the cursor's position using a string resource id.  The cursor |
| is set to the new item.                                                      |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::add ( const IResourceId& resid,
                                  IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return add(newText, cursor);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::add                                                            |
|                                                                              |
| Add an item at the cursor's position using a const char *.  The cursor is    |
| set to the new item.                                                         |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::add ( const char* strItem,
                                  IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IASSERTSTATE(cursor.isValid());
   unsigned long index = add(cursor.asIndex(), strItem);
   cursor.setToIndex(index);
   return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsFirst                                                     |
|                                                                              |
| Add an item at the beginning of the list using a string resource id.  The    |
| cursor is set to the new item.                                               |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsFirst ( const IResourceId& resid,
                                         IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsFirst(newText, cursor);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsFirst                                                     |
|                                                                              |
| Add an item at the beginning of the list using a const char *.  The cursor   |
| is set to the new item.                                                      |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsFirst ( const char* strItem,
                                         IC_CURSOR_CLASSNAME::Cursor& cursor )
{
  unsigned long index = addAsFirst(strItem);
  cursor.setToIndex(index);
  return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsLast                                                      |
|                                                                              |
| Add an item at the end of the list using a string resource id.  The cursor   |
| is set to the new item.                                                      |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsLast ( const IResourceId& resid,
                                        IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsLast(newText, cursor);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsLast                                                      |
|                                                                              |
| Add an item at the end of the list using a const char *.  The cursor is set  |
| to the new item.                                                             |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsLast ( const char* strItem,
                                        IC_CURSOR_CLASSNAME::Cursor& cursor )
{
  unsigned long index = addAsLast(strItem);
  cursor.setToIndex(index);
  return *this;
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsNext                                                      |
|                                                                              |
| Add an item after the item at the current cursor position using a string     |
| resource id.  The cursor is set to the new item.                             |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsNext ( const IResourceId& resid,
                                        IC_CURSOR_CLASSNAME::Cursor& cursor )
{
   IString newText = resid.resourceLibrary().loadString(resid);
   return addAsNext(newText, cursor);
}

/*------------------------------------------------------------------------------
| IC_CLASSNAME::addAsNext                                                      |
|                                                                              |
| Add an item after the item at the current cursor position using a const      |
| char *.  The cursor is set to the new item.                                  |
------------------------------------------------------------------------------*/
IC_CLASSNAME& IC_CLASSNAME::addAsNext ( const char* strItem,
                                        IC_CURSOR_CLASSNAME::Cursor& cursor )
{
  IASSERTSTATE(cursor.isValid());
  unsigned long index = cursor.asIndex();
  long lElems = count();

  if (index >= 0  &&  index < lElems)
  {
    if (index == lElems - 1)
       index = LIT_END;
    else
       index++;
    index = add(index, strItem);
    cursor.setToIndex(index);
  }
  return *this;
}

