/* NOSHIP */
#ifndef _IWPOSBUF_
#define _IWPOSBUF_
/*******************************************************************************
* FILE NAME: iwposbuf.hpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Declaration of the classes:                                                *
*     IWindowPosBuffer                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Forward declarations for other classes:
class IPoint;
class ISize;
class IWindow;
class IWinPosSet;

#include <ibase.hpp>
#include <ihandle.hpp>
#include <irect.hpp>
#include <iswp.hpp>

#pragma pack(4)

class IWindowPosBuffer : public IBase {
typedef IBase Inherited;
/*******************************************************************************
* This class is used to perform a buffered change of window sizes, positions,  *
* visibility, and z-ordering for children of a specified parent window.  The   *
* resulting collection is used to call WinSetMultWindowPos.                    *
*******************************************************************************/
public:
/*------------------------ Constructor/Destructor ------------------------------
| There are two ways to construct instances of this class:                     |
| 1. From an instance of IWindow                                               |
|    The IWindow instance must be the parent of all windows that will be       |
|    sized, moved, or otherwise changed through use of this class.             |
| 2. From an existing IWindowPosBuffer object                                  |
|    The copy constructor creates another IWindowPosBuffer object.             |
------------------------------------------------------------------------------*/
  IWindowPosBuffer  ( IWindow* parent );
  IWindowPosBuffer  ( const IWindowPosBuffer& original );
  ~IWindowPosBuffer ( );

/*----------------------- Position and Size Accessors --------------------------
| These functions provide means of getting and setting the position and size   |
| values of child windows:                                                     |
|   position   - Returns the proposed position of the specified child window.  |
|                If the child is not in the collection, the actual window      |
|                position is returned.                                         |
|   size       - Returns the proposed size of the specified child window.      |
|                If the child is not in the collection, the actual window      |
|                size is returned.                                             |
|   rect       - Returns the proposed position and size of the specified child |
|                window. If the child is not in the collection, the actual     |
|                window position and size is returned.                         |
|   moveTo     - Specifies a new position for the child window.                |
|   sizeTo     - Specifies a new size for the child window.                    |
|   moveSizeTo - Specifies a new position and size for the child window.       |
------------------------------------------------------------------------------*/
IPoint
  position     ( const IWindow* child ) const;
ISize
  size         ( const IWindow* child ) const;
IRectangle
  rect         ( const IWindow* child ) const;
IWindowPosBuffer
 &moveTo       ( IWindow* child, const IPoint& aPoint ),
 &moveTo       ( const IWindowHandle& child, const IPoint& aPoint ),
 &sizeTo       ( IWindow* child, const ISize& aSize ),
 &sizeTo       ( const IWindowHandle& child, const ISize& aSize ),
 &moveSizeTo   ( IWindow* child, const IRectangle& aRect ),
 &moveSizeTo   ( const IWindowHandle& child, const IRectangle& aRect );

/*---------------------------- Visibility Accessors ----------------------------
| These functions provide means of setting the visible attributes of child     |
| windows:                                                                     |
|   show   - Specifies that a child window should be made visible.             |
|   hide   - Specifies that a child window should be made invisible.           |
------------------------------------------------------------------------------*/
IWindowPosBuffer
 &show         ( IWindow* child ),
 &show         ( const IWindowHandle& child ),
 &hide         ( IWindow* child ),
 &hide         ( const IWindowHandle& child );

/*---------------------------- Z-Order Accessors -------------------------------
| These functions provide means of changing the z-order of the child windows:  |
|   orderAfter   - Specifies the child window to be moved after another child  |
|                  in the z-order.                                             |
|   orderAsFirst - Specifies the child window to be made first in the z-order. |
|   orderAsLast  - Specifies the child window to be made last in the z-order.  |
------------------------------------------------------------------------------*/
IWindowPosBuffer
 &orderAfter   ( IWindow* child, IWindow* sibling ),
 &orderAfter   ( const IWindowHandle& child, const IWindowHandle& sibling ),
 &orderAsFirst ( IWindow* child ),
 &orderAsFirst ( const IWindowHandle& child ),
 &orderAsLast  ( IWindow* child ),
 &orderAsLast  ( const IWindowHandle& child );

/*------------------------- Collection Manipulation ----------------------------
| These functions operate on all child windows added to the collection:        |
|   apply  - Causes all proposed window changes to be made effective.          |
|   empty  - Removes all window change entries from the collection.            |
------------------------------------------------------------------------------*/
IWindowPosBuffer
 &apply        ( Boolean checkBidi = true ),
 &empty        ( );

private:
/*--------------------------------- Private ----------------------------------*/
IWindowPosBuffer
 &setOrder     ( const IWindowHandle& child, const IWindowHandle& sibling );
Boolean
  isParentOk   ( const IWindowHandle& child ) const;
IWindow*
  pwndClParent;
IWinPosSet*
  pClBufSet;

class Item : public ISWP {
friend IWindowPosBuffer;
public:
  ~Item ( );
protected:
  Item  ( const IWindowHandle& hwnd );
  Item  ( const Item& original );

// handle stored as unsigned long.  This is for the key function to
// use, since it returns a reference to the key.
unsigned long
  fkeyHwnd;

friend unsigned long const
 &key          ( Item* const& element );
}; // IWindowPosBuff::Item

Item
 *createOrLocate           ( const IWindowHandle& hwnd );

friend Boolean
  IWindowPosBufferDeleter  ( Item* const& element,
                             void* anything );
}; // IWindowPosBuffer

/*----------------------------------------------------------------------------*/
/* Resume compiler default packing.                                           */
/*----------------------------------------------------------------------------*/
#pragma pack()

/*----------------------------- Inline Functions -----------------------------*/
#ifndef I_NO_INLINES
  #include <iwposbuf.inl>
#endif

#endif  /* _IWPOSBUF_  */
