/*******************************************************************************
* FILE NAME: idmimage.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMImage                                                                 *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1995       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #include <string.h>
  // define what subsections of PM headers to include
  #define INCL_WINSTDDRAG       // Standard drag stuff
  #include <os2.h>
}

#ifndef _IDMCOMM_
  #include <idmcomm.hpp>
#endif

#ifndef _IDMIMAGE_
  #include <idmimage.hpp>
#endif

#ifndef _ICCONST_
  #include <icconst.h>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDMIMAGE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
  Static operation definitions.

  These values exploit the specific values used for the PM toolkit
  DRG_* values.
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IDMImage::Style
  IDMImage::noStyle           = 0,
  IDMImage::ptr               = DRG_ICON,
  IDMImage::bmp               = DRG_BITMAP,
  IDMImage::polygon           = DRG_POLYGON,
  IDMImage::stretch           = DRG_STRETCH,
  IDMImage::transparent       = DRG_TRANSPARENT,
  IDMImage::closed            = DRG_CLOSED,
  IDMImage::classDefaultStyle = 0;
#pragma data_seg()

#pragma data_seg(ICLNonConst)
  IDMImage::Style IDMImage::currentDefaultStyle = 0;
#pragma data_seg()

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IDMImage :: IDMImage ()
  : pointerImage(0),
    bitmapImage(0),
    fadeImage(ISystemPointerHandle::dragFade),
    ptrOffset(ISize(0,0)),
    imageStretchSize(ISize(0,0)),
    imgStyle(0),
    pPointArray(0),
    points(0)
{
}

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Constructor - Construct this class from an icon or bitmap resource id.       |
------------------------------------------------------------------------------*/
IDMImage :: IDMImage (const IResourceId& resourceId,
                      Boolean iconResource,
                      const ISize& ptrOffset,
                      const ISize& stretchSize,
                      const Style& style)
  : ptrOffset(ptrOffset),
    fadeImage(ISystemPointerHandle::dragFade),
    imageStretchSize(stretchSize),
    imgStyle(style),
    pPointArray(0),
    points(0)
{
  if (iconResource)
  {
    this->pointerImage = IPointerHandle(
                         resourceId.resourceLibrary().loadIcon(resourceId));
    this->bitmapImage = 0;
    this->imgStyle |= IDMImage::ptr;
  }
  else
  {
    this->bitmapImage = IBitmapHandle(
                        resourceId.resourceLibrary().loadBitmap(resourceId));
    this->pointerImage = 0;
    this->imgStyle |= IDMImage::bmp;
  }

  if (this->imageStretchSize != 0)
  {
    this->imgStyle |= IDMImage::stretch;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Constructor - Construct this class from an icon handle.                      |
------------------------------------------------------------------------------*/
IDMImage :: IDMImage   (const IPointerHandle& icon,
                        const ISize& ptrOffset,
                        const ISize& stretchSize,
                        const Style& style)
  : bitmapImage(0),
    fadeImage(ISystemPointerHandle::dragFade),
    ptrOffset(ptrOffset),
    imageStretchSize(stretchSize),
    imgStyle(style),
    pPointArray(0),
    points(0)
{
  this->pointerImage = icon;
  this->imgStyle |= IDMImage::ptr;
  if (this->imageStretchSize != 0)
  {
    this->imgStyle |= IDMImage::stretch;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Constructor - Construct this class from a bitmap handle.                     |
------------------------------------------------------------------------------*/
IDMImage :: IDMImage (const IBitmapHandle& bitmap,
                      const ISize& ptrOffset,
                      const ISize& stretchSize,
                      const Style& style)
  : pointerImage(0),
    fadeImage(ISystemPointerHandle::dragFade),
    ptrOffset(ptrOffset),
    imageStretchSize(stretchSize),
    imgStyle(style),
    pPointArray(0),
    points(0)
{
  this->bitmapImage = bitmap;
  this->imgStyle |= IDMImage::bmp;
  if (this->imageStretchSize != 0)
  {
    this->imgStyle |= IDMImage::stretch;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Constructor - Construct this class from an array of points.                  |
------------------------------------------------------------------------------*/
IDMImage :: IDMImage (unsigned long pointArr[],
                      const unsigned long numberOfPoints,
                      const ISize& ptrOffset,
                      const Style& style)
  : pointerImage(0),
    bitmapImage(0),
    fadeImage(ISystemPointerHandle::dragFade),
    ptrOffset(ptrOffset),
    imageStretchSize(ISize(0,0)),
    imgStyle(style),
    pPointArray(0),
    points(numberOfPoints)
{
  this->imgStyle |= IDMImage::polygon;
  if (points)
  {
     unsigned long ulNumberOfArrayElements(points*2) ;
     pPointArray = new unsigned long[ulNumberOfArrayElements];
     memcpy((void *) pPointArray,
            (void *) pointArr,
            (sizeof(unsigned long) * ulNumberOfArrayElements )) ;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::IDMImage                                                           |
|                                                                              |
| Copy Constructor - Construct this class from an existing IDMImage.           |
------------------------------------------------------------------------------*/
IDMImage::IDMImage (const IDMImage& image)
: pointerImage(image.pointerImage),
  bitmapImage(image.bitmapImage),
  fadeImage(image.fadeImage),
  ptrOffset(image.ptrOffset),
  imageStretchSize(image.imageStretchSize),
  imgStyle(image.imgStyle),
  pPointArray(0),
  points(image.points)
{
  if (points)
  {
     unsigned long ulNumberOfArrayElements(points*2) ;
     pPointArray = new unsigned long[ulNumberOfArrayElements];
     memcpy((void *) pPointArray,
            (void *) image.pPointArray,
            (sizeof(unsigned long) * ulNumberOfArrayElements )) ;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::operator=                                                          |
|                                                                              |
| assignment operator.                                                         |
------------------------------------------------------------------------------*/
IDMImage& IDMImage::operator=(const IDMImage& image)
{
  if (this != &image)
  {
    pointerImage     = image.pointerImage;
    bitmapImage      = image.bitmapImage;
    fadeImage        = image.fadeImage;
    ptrOffset        = image.ptrOffset;
    imageStretchSize = image.imageStretchSize;
    imgStyle         = image.imgStyle;
    points           = image.points;
    if (pPointArray)
    {
       delete[] pPointArray ;
       pPointArray = 0 ;
    }

    if (points)
    {
       unsigned long ulNumberOfArrayElements(points*2) ;
       pPointArray = new unsigned long[ulNumberOfArrayElements];
       memcpy((void *) pPointArray,
              (void *) image.pPointArray,
              (sizeof(unsigned long) * ulNumberOfArrayElements )) ;
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IDMImage::~IDMImage                                                          |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IDMImage :: ~IDMImage ()
{
  if (pPointArray)
  {
    delete [] pPointArray;
    pPointArray = 0;
  }
}

/*------------------------------------------------------------------------------
| IDMImage::Style IDMImage::defaultStyle                                       |
------------------------------------------------------------------------------*/
IDMImage::Style IDMImage::defaultStyle()
{
   return currentDefaultStyle;
}

/*------------------------------------------------------------------------------
| IDMImage::setDefaultStyle                                                    |
------------------------------------------------------------------------------*/
void IDMImage::setDefaultStyle(const Style& style)
{
  currentDefaultStyle = style;
}

/*------------------------------------------------------------------------------
| IDMImage::setPointer                                                         |
|                                                                              |
| Sets the pointer image handle.                                               |
------------------------------------------------------------------------------*/
IDMImage& IDMImage::setPointer (const IPointerHandle& ptrImage)
{
  this->pointerImage = ptrImage;
  IDMImage::Style imageStyle = style();
  imageStyle |= IDMImage::ptr;
  setStyle(imageStyle);
  return (*this);
}

/*------------------------------------------------------------------------------
| IDMImage::setBitmap                                                          |
|                                                                              |
| Sets the bitmap image handle.                                                |
------------------------------------------------------------------------------*/
IDMImage& IDMImage::setBitmap (const IBitmapHandle& bmpImage)
{
  this->bitmapImage = bmpImage;
  IDMImage::Style imageStyle = style();
  imageStyle |= IDMImage::bmp;
  setStyle(imageStyle);
  return (*this);
}

/*------------------------------------------------------------------------------
| IDMImage::setStretchSize                                                     |
|                                                                              |
| Sets the size of the image when stretched.                                   |
------------------------------------------------------------------------------*/
IDMImage& IDMImage::setStretchSize (const ISize& stretchSize)
{
  this->imageStretchSize = stretchSize;
  IDMImage::Style imageStyle = style();
  imageStyle |= IDMImage::stretch;
  setStyle(imageStyle);
  return (*this);
}

/*------------------------------------------------------------------------------
| IDMImage::setPointArray                                                      |
|                                                                              |
| Sets the point array to be used to draw the image.                           |
------------------------------------------------------------------------------*/
IDMImage& IDMImage::setPointArray (unsigned long pointArr[],
                                   unsigned long numberOfPoints)
{
  points = numberOfPoints;
  if (pPointArray)
     delete [] pPointArray;
  if (points)
  {
     unsigned long ulNumberOfArrayElements(points*2) ;
     pPointArray = new unsigned long[ulNumberOfArrayElements];
     memcpy((void *) pPointArray,
            (void *) pointArr,
            (sizeof(unsigned long) * ulNumberOfArrayElements )) ;
    IDMImage::Style imageStyle = style();
    imageStyle |= IDMImage::polygon;
    setStyle(imageStyle);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IDMImage::asPMDragImage                                                      |
|                                                                              |
| Fills in a PM DRAGIMAGE structure.                                           |
------------------------------------------------------------------------------*/
void IDMImage :: asPMDragImage(_DRAGIMAGE *pPMDragImage,
                               Boolean useSystemIcons, Boolean fadeIcon)
{
  pPMDragImage->cb = sizeof (_DRAGIMAGE);

  if (useSystemIcons)
  {
    pPMDragImage->cptl = 0;
    pPMDragImage->sizlStretch.cx = 0;
    pPMDragImage->sizlStretch.cy = 0;
    pPMDragImage->fl = (IDMImage::ptr).asUnsignedLong();
    pPMDragImage->cxOffset = 0;
    pPMDragImage->cyOffset = 0;
  }
  else if (fadeIcon)
  {
    pPMDragImage->hImage = fadeImage;
    pPMDragImage->cptl = 0;
    pPMDragImage->sizlStretch.cx = 0;
    pPMDragImage->sizlStretch.cy = 0;
    pPMDragImage->fl = (IDMImage::ptr).asUnsignedLong();
    pPMDragImage->cxOffset = 0;
    pPMDragImage->cyOffset = 0;
  }
  else
  {
    pPMDragImage->cptl = (unsigned short)(numberOfPoints());
    pPMDragImage->sizlStretch.cx = stretchSize().width();
    pPMDragImage->sizlStretch.cy = stretchSize().height();
    pPMDragImage->fl = style().asUnsignedLong();
    pPMDragImage->cxOffset = (short)(pointerOffset().width());
    pPMDragImage->cyOffset = (short)(pointerOffset().height());
    if (pPMDragImage->fl & ((IDMImage::ptr).asUnsignedLong()))
    {
      IPointerHandle image = pointer();
      pPMDragImage->hImage = ((unsigned long)image);
    }
    else if (pPMDragImage->fl & ((IDMImage::bmp).asUnsignedLong()))
    {
      IBitmapHandle image = bitmap();
      pPMDragImage->hImage = ((unsigned long)image);
    }
    else
    //It's an array of points.
    {
      pPMDragImage->hImage = (unsigned long)pointArray();
    }
  }
}
